/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/
        
package hypercast;

import hypercast.adapters.TCP_UDP_Adapter;
import hypercast.adapters.TCP_UnicastAdapter;
import hypercast.adapters.SSL_UnicastAdapter;
import hypercast.adapters.UDP_UnicastAdapter;
import hypercast.adapters.UDP_MulticastAdapter;
import hypercast.util.XmlUtil;

/**
 * This class implements methods to create a new adapter based on the
 * specifications in the passed configuration object.
 *      
 * @author HyperCast Team
 * @author Jianping Wang 
 * @version     3.0, Mar. 15, 2004
 */
public final class AdapterFactory
{
    private boolean debugInitializationMessages;

    public AdapterFactory (final boolean debugInitializationMessages)
    {
        this.debugInitializationMessages = debugInitializationMessages;
    }

    public AdapterFactory()
    {
        this (false);
    }

    // TODO: chained factory methods? some times this method is called
    // to create a socket adapter; sometimes it is called to create a
    // node adapter, in each case some args are null, some are not,
    // also the physical address is frequently omitted by specifying
    // null. maybe there should be many chained methods (they call
    // each other, until a "unified" method is called.) Perhaps we
    // should have a createNodeAdapter method as well as a separate
    // createSocketAdapter method, that way we wouldn't have to pass
    // in string for the types.

    /**
     * This method creates an Adapter based on the attributes specified in the 
     * configuration object. 
     *
     * @param config                 an HyperCastConfig object.
     * @param adapter                an object that will be passed to the constructor of the creating adapter. TODO: Is it true that  this is non-null on creation of node adapter, in which case it is a socket adapter...
     * @param adapterPropertyString  string that specifies whether a Node Adapter or Socket Adapter will be created.
     * @param physicalAddressString  string containing physical address information.
     *
     * @throws IllegalArgumentException if an illegal node type is passed as an argument for <code>adapterPropertyString</code>
     * @throws HyperCastConfigException if values read from the configuration file are not valid or are missing
     */
    public I_UnicastAdapter createAdapter (final HyperCastConfig config,
                                           final I_UnicastAdapter       adapter,
                                           final String                 adapterPropertyString,    // TODO: maybe pick a better name 
                                           final String                 physicalAddressString)
    {
        if (! adapterPropertyString.equals ("NodeAdapter")
            &&
            ! adapterPropertyString.equals ("SocketAdapter"))
        {
            throw new IllegalArgumentException ("Property string must be \"NodeAdapter\" or \"SocketAdapter\" not: " + adapterPropertyString);
        }

		final String adapterProperty = config.getTextAttribute
				(XmlUtil.createXPath("/Public/"+adapterPropertyString));

        if (adapterProperty == null)
        {
            throw new HyperCastConfigException ("Cannot find configuration property for " + adapterPropertyString);
        }

        I_UnicastAdapter newAdapter = null;

        // TODO: Need to change name in XML schema hypercast.xsd: Sckt -> Socket, Adpt -> Adapter
        final String typePrefix = adapterProperty.startsWith ("ScktAdpt") ? "Sckt" : "Node";

        if (adapterProperty.equals (typePrefix + "AdptTriple"))
        {
            if (debugInitializationMessages)
                config.log.println ("Creating " + typePrefix + " adapter with type TCP_UDP_Adapter");

            newAdapter = new TCP_UDP_Adapter (config, adapter, physicalAddressString, adapterProperty);
        }
        else if (adapterProperty.equals (typePrefix + "AdptTCP"))
        {
            if (debugInitializationMessages)
                config.log.println ("Creating " + typePrefix + " adapter with type TCP_UnicastAdapter");

            newAdapter = new TCP_UnicastAdapter (config, adapter, physicalAddressString, adapterProperty);
        }
        else if (adapterProperty.equals (typePrefix + "AdptSSL"))
        {
            if (debugInitializationMessages)
                config.log.println ("Creating " + typePrefix + " adapter with type SSL_UnicastAdapter");

            newAdapter = new SSL_UnicastAdapter (config, adapter, physicalAddressString, adapterProperty);
        }
        else if (adapterProperty.equals (typePrefix + "AdptUDP"))
        {
            if (debugInitializationMessages)
                config.log.println ("Creating " + typePrefix + " adapter with type UDP_UnicastAdapter");

            newAdapter = new UDP_UnicastAdapter (config, adapter, physicalAddressString, adapterProperty);
        }
        else if (adapterProperty.equals (typePrefix + "AdptUDPMulticast"))
        {
            if (debugInitializationMessages)
                config.log.println ("Creating " + typePrefix + " adapter with type UDP_MulticastAdapter");

            newAdapter = new UDP_MulticastAdapter (config, adapter, physicalAddressString, adapterProperty);
        }
        else
        {
            throw new HyperCastConfigException (typePrefix + "adapter property has wrong value: " + adapterProperty + " !");
        }

        return newAdapter;
    }
}
