/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.DT;

/**
 * This class represents a point in two dimensions, where the coordinates
 * of the point are doubles.  It is used for increased accuracy over DT_Point.
 *
 * It is immutable.
 *
 * @author HyperCast Team
 * @author Michael Nahas
 * @version	2.0, Feb. 11, 2001
 */

public class DT_DoublePoint {
	/** The error used in the function <code>equals()</code> */
    public static double EPSILON = 1e-7;

    protected double x;
    protected double y;
    
    public DT_DoublePoint(DT_DoublePoint p) {
        x = p.x;
        y = p.y;
    }

    public DT_DoublePoint(DT_Point p) {
        x = p.x;
        y = p.y;
    }

    public DT_DoublePoint(double xx, double yy) {
        x = xx;
        y = yy;
    }

/** Compares this to either a DT_DoublePoint or a DT_Point.  
 * @return true if the absolute difference of both coordinates is less than <code>EPSILON</code>
 * @see DT_DoublePoint#EPSILON
*/
    public boolean equals(Object o) 
    {
        DT_DoublePoint p = null;
        

        if (o instanceof DT_DoublePoint)
            p = (DT_DoublePoint) o;     
        else if (o instanceof DT_Point)
            p = new DT_DoublePoint((DT_Point) o);
        else
            throw new IllegalArgumentException("DT_DoublePoint.equals() pass object of type not DT_DoublePoint nor DT_Point.");
        
        return Math.abs(x - p.x) < EPSILON && Math.abs(y - p.y) < EPSILON;
    }

/** Override to Object.hashCode() so that this object can be used as a key in a hashtable. */    
    public int hashCode() 
    {
        long xx = Double.doubleToLongBits(x);
        long yy = Double.doubleToLongBits(y);
        
        int hash1 = (int) (xx & 0xFFFFFFFF);
        int hash2 = (int) ((xx >> 32) & 0xFFFFFFFF);
        int hash3 = (int) (yy & 0xFFFFFFFF);
        int hash4 = (int) ((yy >> 32) & 0xFFFFFFFF);
        

        return (hash1 ^ hash2 ^ hash3 ^ hash4);
    }

    
public double getX() 
    {
        return x;
    }
    
public double getY() 
    {
        return y;
    }

/** Returns the Euclidean distance between two points. */
public static double distance(DT_DoublePoint a, DT_DoublePoint b)
    {
        double dx = a.x - b.x;
        double dy = a.y - b.y;
    
        return  Math.sqrt(dx*dx + dy*dy);
    }

/** Returns the point at the interception of two lines.
 * (f,s) and (g,t) are the x-intercept and slopes of two lines.
 * 
 * <p> <code> x = (g-f)/(s-t); </code>
 * <p> <code> y = f + s*x; </code>
 *
 * @throws ArithmeticException if s == t 
 */
public static DT_DoublePoint findInterceptOfLines(double f, double s, double g, double t) 
    {
        // Initial solution:  (caused divide by zero error)
        //double y = (f/s - g/t)/(1/s - 1/t);
        //double x = (y - f)/s;

        // solve without dividing by an individual slope.
        double x = (g-f)/(s-t);
        double y = f + s*x;

        return new DT_DoublePoint(x, y);
    }
}









