/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.DT;

import hypercast.*;

import java.awt.*;

/** This class defines a DT protocol message, which contains a type (byte), 
 * a Overlay hash(int) and 4 AddressPairs. The Src and Dst AddressPairs can 
 * not both be null. Its format in byte array is as following:
 *
 * ------------------------------------------------------------
 * | Protocol |    Length    |  Type    |   Overlay ID hash   |
 * |  8 bits  |    16 bits   |  8 bits  |      32 bits        |
 * ----------------------------------------------------------------------------
 * |       Src        |       Dst       |     ADDR1        |    ADDR2         |
 * |     112 bits     |     112 bits    |    112 bits      |   112 bits       |
 * ----------------------------------------------------------------------------
 *
 * @author HyperCast Team
 * @author Michael Nahas
 * @version 2.0, Feb. 11, 2001 
 * @author Jianping Wang
 * @version 3.0, Jan. 08, 2004 
 */

public class DT_Message implements I_Message
{
    /** A message type */
    public static final byte HelloNeighbor   = 0;
    /** A message type */
    public static final byte HelloNotNeighbor   = 1;
    /** A message type */
    public static final byte Goodbye   = 2;
    /** A message type */
    public static final byte ServerRequest = 3;
    /** A message type */
    public static final byte ServerReply = 4;
    /** A message type */
    public static final byte NewNode = 5;
    /** A message type */
    public static final byte NodePing = 6;
    /** A message type */
    public static final byte NodePong = 7;
    /** A message type used with Buddies */
    public static final byte IAmHere = 8;

	/** A message type used to broadcast leader address pair */
	public static final byte IAmLeader = 9;

    /** A message type used with Multicasting*/
	public static final byte Beacon = 10;

	/** A test type */
	public static final byte TestMessage = 11;

    
    /** Protocol Number value in the Protocol field of a DT3.0 protocol message */
    public static final byte PROTONUM_DT30 = 2;

    protected byte Proto_num;

    protected byte Proto_sub;

    /** The protocol message type. */
    private byte Type;

    /** A hash of Overlay properties.  It is used to detect either
      conflicting Overlays or detect bad packets. */
	private int overlayHash;
    

    /** Logical and Physical address of the sender. */
    private DT_AddressPair Src;

    /** Logical and Physical address of the destination. */
    private DT_AddressPair Dst;

    /** Logical and Physical address of the neighbor of the sender
        immediately clockwise (CW) of the destination. For the
        LeaderIs messages, this contains the Leader's addresses. */
        private DT_AddressPair CW; 

    /** Logical and Physical address of the neighbor of the sender
        immediately counterclockwise (CCW) of the destination. */
    private DT_AddressPair CCW; 
    
  
/** 
 * Constructor. It constructs a DT_Message instance by copying another
 * DT_Message instance.
 */
  
public DT_Message(DT_Message A) {

		//Set values to protocol number and protocol subfield.
		Proto_num = A.getProtoNum();
		Proto_sub = A.getProtoSub();

        Type    = A.getType();

        overlayHash = A.getOverlayHash();

        Src   = A.getSrc();
        Dst   = A.getDst();
        CW    = A.getCW();
        CCW   = A.getCCW();
    }


/** Constructor of DT_Message.  For LeaderIs messages, the Leader's
    AddressPair should go in the place of CW. */
public DT_Message(byte proto_sub,
		  int T, 
                  int gh,
                  DT_AddressPair Src_arg,
                  DT_AddressPair Dst_arg,
                  DT_AddressPair CW_arg,
                  DT_AddressPair CCW_arg){

		Proto_num = PROTONUM_DT30;
		Proto_sub = proto_sub;

        Type = (byte)T;

        overlayHash = gh;

        if (Src_arg == null && Dst_arg == null)
            throw new IllegalArgumentException("Both Src and Dst were null in DT_Message constructor!");

        Src = Src_arg;
        Dst = Dst_arg;
        CW = CW_arg;
        CCW = CCW_arg;
    }




/** Creates a message from a byte array.
 * @param adapter used to create new PhysicalAddress from the byte array.
 */
 public DT_Message(byte[] A, I_UnicastAdapter adapter) 
    {
		/*
		 * Get the physical address size from node adapter.
		 */
		int paSize = adapter.getPhysicalAddressSize();

		/*
		 * Get the logical address size.
		 */
        int laSize = DT_LogicalAddress.ADDRESS_ARRAY_SIZE;

        I_PhysicalAddress Addr;
        DT_LogicalAddress Coords;
        
        int index = 0;

		/*
		 * The first byte is Protocol field.
		 */
		Proto_num = (byte) ((A[index] >> 4) & 0xf);
		Proto_sub = (byte) (A[index] & 0xf);
        index += 1;
        
        if (Proto_num != PROTONUM_DT30)
            throw new IllegalArgumentException("The Protocol Number in Protocol field of the message is not PROTONUM_DT30.");

		/*
		 * Skip 2 bytes which stores the length of the message.
		 */
        index += 2;

		/*
		 * Get the message type.
		 */
        Type = A[index];
        index++;
        
		/*
		 * Get the hash value of the overlay ID carried in the message.
		 */
        overlayHash = ByteArrayUtility.toInteger(A, index);
        index += 4;

		/*
		 * Get the address pair of the source. If the fields for source
		 * address pair in the packet are all zero, do nothing.
		 */
        if (!ByteArrayUtility.isBytesZero(A, index, paSize + laSize))
        {
            Addr = adapter.createPhysicalAddress(A, index);
            Coords = new DT_LogicalAddress(ByteArrayUtility.toInteger(A, index+paSize), ByteArrayUtility.toInteger(A, index+paSize+4));
            Src = new DT_AddressPair(Addr, Coords);
        }
        index += paSize + laSize;
        
		/*
		 * Get the address pair of the destination. If the fields for 
		 * destination address pair in the packet are all zero, do nothing.
		 */
        if (!ByteArrayUtility.isBytesZero(A, index, paSize + laSize))
        {
            Addr = adapter.createPhysicalAddress(A, index);
            Coords = new DT_LogicalAddress(ByteArrayUtility.toInteger(A, index+paSize), ByteArrayUtility.toInteger(A, index+paSize+4));
            Dst = new DT_AddressPair(Addr, Coords);
        }
        index += paSize + laSize;
        
		/*
		 * Get the address pair of the CW node (with respect to the source node). 
		 * If the fields in the packet are all zero, do nothing.
		 */
       if (!ByteArrayUtility.isBytesZero(A, index, paSize + laSize))
        {
            Addr = adapter.createPhysicalAddress(A, index);
            Coords = new DT_LogicalAddress(ByteArrayUtility.toInteger(A, index+paSize), ByteArrayUtility.toInteger(A, index+paSize+4));
            CW = new DT_AddressPair(Addr, Coords);
        }
        index += paSize + laSize;
        
		/*
		 * Get the address pair of the CCW node (with respect to the source node). 
		 * If the fields in the packet are all zero, do nothing.
		 */
        if (!ByteArrayUtility.isBytesZero(A, index, paSize + laSize))
        {
            Addr = adapter.createPhysicalAddress(A, index);
            Coords = new DT_LogicalAddress(ByteArrayUtility.toInteger(A, index+paSize), ByteArrayUtility.toInteger(A, index+paSize+4));
            CCW = new DT_AddressPair(Addr, Coords);
        }
        index += paSize + laSize;

		/*
		 * In a DT protocol message, the source and destination address cannot
		 * be both 0.
		 */
        if (Src == null && Dst == null)
        {
			throw new IllegalArgumentException("Both Src and Dst were null in DT_Message constructor!");
        }
    }


/** Creates a message, if possible, from a byte array.
 * If a message is read, the value of beginValidBytes is incremented to the index
 * of the first byte not in the message.
 * @param adapter used to create new PhysicalAddress from the byte array.
 */

public static DT_Message restoreMessage(byte[] A, int[] beginValidBytes, int endValidBytes, I_UnicastAdapter adapter, int thisOverlayHash)
{
	/*
	 * Get the physical address size from node adapter.
	 */
	final int PASize = adapter.getPhysicalAddressSize();

	/*
	 * Get the logical address size and addres pair size.
	 */
	final int LASize = DT_LogicalAddress.ADDRESS_ARRAY_SIZE;
	final int pairSize = PASize + LASize;

	/* The size of a DT message is determinded based on the format 
	 * shown at the top of this file.
	 */
    if (endValidBytes - beginValidBytes[0] < 3 + 1 + 4 + 4*pairSize)
       return null;

	I_PhysicalAddress pa;
	DT_LogicalAddress la;
        
	DT_AddressPair src = null;
	DT_AddressPair dst = null;
	DT_AddressPair cw = null;
	DT_AddressPair ccw = null;

	int index = beginValidBytes[0];

	/*
	 * Get the protocol number and protocol sub field.
	 */
	byte proto_num = (byte) ((A[index] >> 4) & 0xf);
	byte proto_sub = (byte) (A[index] & 0xf);
	index++;
	
    if (proto_num != PROTONUM_DT30)
    	throw new IllegalArgumentException("The Protocol Number in Protocol field of the message is not PROTONUM_DT30.");

	/*
	 * Get the length of the parts in the packet which follows the
	 * length field.
	 */
	short length = ByteArrayUtility.toShort(A, index);
    index += 2;
    if (length != (short) (1 + 4 + 4*pairSize))
        throw new IllegalArgumentException("The length of the message is incorrect.");

	/*
	 * Get the message type.
	 */
    int type = A[index];
    index++;

	/*
	 * Get the hash value of the overlay ID carried in the message.
	 */
    int incomingOverlayHash = ByteArrayUtility.toInteger(A, index);
    index += 4;
        
    if (incomingOverlayHash != thisOverlayHash)
       throw new IllegalArgumentException("Message is not of this Overlay.");


	if (type != HelloNeighbor && type != HelloNotNeighbor && type != Goodbye && type != ServerRequest && type != ServerReply && type != NewNode
		&& type != NodePing && type != NodePong && type != TestMessage && type != IAmHere && type != IAmLeader && type != Beacon)
		throw new IllegalArgumentException("ReceiveBuffer cannot contain type DT_Message (illegal msg type)");

	/*
	 * Get the address pair of the source. If the fields for source
	 * address pair in the packet are all zero, do nothing.
	 */
    if (!ByteArrayUtility.isBytesZero(A, index, pairSize))
    {
       pa = adapter.createPhysicalAddress(A, index);
       la = new DT_LogicalAddress(ByteArrayUtility.toInteger(A, index+PASize), ByteArrayUtility.toInteger(A, index+PASize+4));
       src = new DT_AddressPair(pa, la);
    }
    index += pairSize;
        
	/*
	 * Get the address pair of the destination. If the fields for 
	 * destination address pair in the packet are all zero, do nothing.
	 */
    if (!ByteArrayUtility.isBytesZero(A, index, pairSize))
    {
       pa = adapter.createPhysicalAddress(A, index);
       la = new DT_LogicalAddress(ByteArrayUtility.toInteger(A, index+PASize), ByteArrayUtility.toInteger(A, index+PASize+4));
       dst = new DT_AddressPair(pa, la);
    }
    index += pairSize;
        
	/*
	 * Get the address pair of the CW node (with respect to the source node). 
	 * If the fields in the packet are all zero, do nothing.
	 */
    if (!ByteArrayUtility.isBytesZero(A, index, pairSize))
    {
       pa = adapter.createPhysicalAddress(A, index);
       la = new DT_LogicalAddress(ByteArrayUtility.toInteger(A, index+PASize), ByteArrayUtility.toInteger(A, index+PASize+4));
       cw = new DT_AddressPair(pa, la);
    }
    index += pairSize;
        
	/*
	 * Get the address pair of the CCW node (with respect to the source node). 
	 * If the fields in the packet are all zero, do nothing.
	 */
    if (!ByteArrayUtility.isBytesZero(A, index, pairSize))
    {
       pa = adapter.createPhysicalAddress(A, index);
       la = new DT_LogicalAddress(ByteArrayUtility.toInteger(A, index+PASize), ByteArrayUtility.toInteger(A, index+PASize+4));
       ccw = new DT_AddressPair(pa, la);
    }
    index += pairSize;

	/*
	 * Move the start position in the receive buffer by amount "index".
	 */
    beginValidBytes[0] = index;

	/*
	 * Construct a new DT message.
	 */
    DT_Message dt_m =  new DT_Message(proto_sub, type, incomingOverlayHash, src, dst, cw, ccw);

    //System.out.println("DT_Message:restoreMessage: restore message with index=" + dt_m.getIndex());

	return dt_m;
    }



/** Converts the message into a byte array for transmission. */

public byte[] toByteArray() 
    {
		/*
		 * Get the physical address size.
		 */
        int paSize;
        if (Src != null) 
            paSize = Src.getPhysicalAddress().getSize();
        else
            paSize = Dst.getPhysicalAddress().getSize();

		/*
		 * Get the logical address size.
		 */
        int laSize = DT_LogicalAddress.ADDRESS_ARRAY_SIZE;
        
        /*
         * Define the packet. Its length is calculated according to the 
         * format of DT message.
         */
        byte Result[] = new byte[3 + 1 + 4 + 4*(laSize+paSize)];
        int index = 0;
        
		/*
		 * This is protocol field.
		 */
		Result[index] = (byte) ((Proto_num << 4) | (Proto_sub & 0xf));
        index += 1;
	
		/*
		 * This is length field.
		 */
		short length = (short) (1 + 4 + 4*(laSize+paSize));
        System.arraycopy(ByteArrayUtility.toByteArray(length), 0, Result, index, 2);
        index += 2;

		/*
		 * Fill in type field.
		 */
        Result[index] = Type;
        index++;
        
        /*
         * Fill in overlay ID hash field.
         */
        System.arraycopy(ByteArrayUtility.toByteArray(overlayHash), 0, Result, index, 4);
        index += 4;

		/*
		 * If source address pair of this message is not null, convert it to byte array and 
		 * put it into the packet. 
		 */
        if (Src != null) 
        {
            System.arraycopy(Src.getPhysicalAddress().toByteArray(), 0, Result, index, paSize);
            System.arraycopy(ByteArrayUtility.toByteArray((Src.getDTLogicalAddress()!=null)?Src.getDTLogicalAddress().x:0), 0, Result, index+paSize, 4);
            System.arraycopy(ByteArrayUtility.toByteArray((Src.getDTLogicalAddress()!=null)?Src.getDTLogicalAddress().y:0), 0, Result, index+paSize+4, 4);
        }
        /*
         * If source address pair is null, set the field for soource address pair in 
         * the packet to all-zero.
         */
        else 
        {
            ByteArrayUtility.setBytesZero(Result, index, laSize+paSize);
        }
        index += laSize+paSize;
        

		/*
		 * If destination address pair of this message is not null, convert it to byte array and 
		 * put it into the packet. 
		 */
        if (Dst != null) 
        {
            System.arraycopy(Dst.getPhysicalAddress().toByteArray(), 0, Result, index, paSize);
            System.arraycopy(ByteArrayUtility.toByteArray((Dst.getDTLogicalAddress()!=null)?Dst.getDTLogicalAddress().x:0), 0, Result, index+paSize, 4);
            System.arraycopy(ByteArrayUtility.toByteArray((Dst.getDTLogicalAddress()!=null)?Dst.getDTLogicalAddress().y:0), 0, Result, index+paSize+4, 4);
        }
		/*
		 * If destination address pair is null, set the field for soource address pair in 
		 * the packet to all-zero.
		 */
        else 
        {
            ByteArrayUtility.setBytesZero(Result, index, laSize+paSize);
        }
        index += laSize+paSize;
        
		/*
		 * If CW node address pair of this message is not null, convert it to byte array and 
		 * put it into the packet. 
		 */
        if (CW != null) 
        {
            System.arraycopy(CW.getPhysicalAddress().toByteArray(), 0, Result, index, paSize);
            System.arraycopy(ByteArrayUtility.toByteArray((CW.getDTLogicalAddress()!=null)?CW.getDTLogicalAddress().x:0), 0, Result, index+paSize, 4);
            System.arraycopy(ByteArrayUtility.toByteArray((CW.getDTLogicalAddress()!=null)?CW.getDTLogicalAddress().y:0), 0, Result, index+paSize+4, 4);
        }
		/*
		 * If CW node address pair is null, set the field for soource address pair in 
		 * the packet to all-zero.
		 */
        else 
        {
            ByteArrayUtility.setBytesZero(Result, index, laSize+paSize);
        }
        index += laSize+paSize;
        
		/*
		 * If CCW node address pair of this message is not null, convert it to byte array and 
		 * put it into the packet. 
		 */
		if (CCW != null) 
        {
            System.arraycopy(CCW.getPhysicalAddress().toByteArray(), 0, Result, index, paSize);
            System.arraycopy(ByteArrayUtility.toByteArray((CCW.getDTLogicalAddress()!=null)?CCW.getDTLogicalAddress().x:0), 0, Result, index+paSize, 4);
            System.arraycopy(ByteArrayUtility.toByteArray((CCW.getDTLogicalAddress()!=null)?CCW.getDTLogicalAddress().y:0), 0, Result, index+paSize+4, 4);
        }
		/*
		 * If CCW node address pair is null, set the field for soource address pair in 
		 * the packet to all-zero.
		 */
        else 
        {
            ByteArrayUtility.setBytesZero(Result, index, laSize+paSize);
        }

        index += laSize+paSize;

        return Result;
    }
	public String toString(){
		return getLabel();
	}

/** Returns the name of the message according to its type. */
public String getLabel()
  {
      switch (Type) {
        case DT_Message.ServerRequest:
          return "ServerRequest";
        case DT_Message.ServerReply:
          return "ServerReply";
        case DT_Message.NewNode:
          return "NewNode";
        case DT_Message.Goodbye:  
          return "Goodbye ";
        case DT_Message.HelloNeighbor:
          return "HelloNeighbor";
        case DT_Message.HelloNotNeighbor:
          return "HelloNotNeighbor";
        case DT_Message.NodePing:
          return "NodePing";
        case DT_Message.NodePong:
          return "NodePong";
        case DT_Message.IAmHere:
          return "IAmHere";
        case DT_Message.IAmLeader:
          return "IAmLeader";
        case DT_Message.Beacon:
          return "Beacon";
        case DT_Message.TestMessage:
          return "TestMessage";
        default:
          return "ERROR";
      }
 }
  

    /** Returns the Proto_num field of the message. */
    byte getProtoNum() {
        return Proto_num;
    }

    /** Returns the Proto_sub field of the message. */
    byte getProtoSub() {
        return Proto_sub;
    }

    /** Returns the type of the message. */
    byte getType() {
        return Type;
    }

    /** Returns the overlayHash of this message. */
public int getOverlayHash() 
    {
        return overlayHash;
    }
    

	/** Returns the physical address of the source of the message. */
public I_PhysicalAddress getSrcPA(){
        if (Src == null)
            return null;
        else
            return Src.getPhysicalAddress();
    }
  
	/** Returns the physical address of the destination of the message. */
public I_PhysicalAddress getDstPA(){
        if (Dst == null)
            return null;
        else
            return Dst.getPhysicalAddress();
    }
  
	/** Returns the physical address of CCW neighbor with respect to the source of the message. */
public I_PhysicalAddress getCCWPA(){
        if (CCW == null)
            return null;
        else
            return CCW.getPhysicalAddress();
    }

	/** Returns the physical address of CW neighbor with respect to the source of the message. */
public I_PhysicalAddress getCWPA(){
        if (CW == null)
            return null;
        else
            return CW.getPhysicalAddress();
    }

	/** Returns the physical address of CW neighbor with respect to the source of the message. */
public I_PhysicalAddress getLeaderPA() 
    {
        if (CW == null)
            return null;
        else
            return CW.getPhysicalAddress();
    }
    

	/** Returns the logical address of the source of the message. */
public DT_LogicalAddress getSrcLA(){
        if (Src == null)
            return null;
        else
            return Src.getDTLogicalAddress();
    }
  
	/** Returns the logical address of the destination of the message. */
public DT_LogicalAddress getDstLA(){
        if (Dst == null)
            return null;
        else
            return Dst.getDTLogicalAddress();
    }
  
	/** Returns the logical address of CCW neighbor with respect to the source of the message. */
public DT_LogicalAddress getCCWLA(){
        if (CCW == null)
            return null;
        else
            return CCW.getDTLogicalAddress();
    }

	/** Returns the logical address of CW neighbor with respect to the source of the message. */
public DT_LogicalAddress getCWLA(){
        if (CW == null)
            return null;
        else
            return CW.getDTLogicalAddress();
    }

	/** Returns the logical address of CW neighbor with respect to the source of the message. */
public DT_LogicalAddress getLeaderLA() 
    {
        if (CW == null)
            return null;
        else
            return CW.getDTLogicalAddress();
    }
    

	/** Returns the address pair of the source of the message. */
public DT_AddressPair getSrc(){
        return Src;
    }
  
	/** Returns the address pair of the destination of the message. */
public DT_AddressPair getDst(){
        return Dst;
    }
  
	/** Returns the address pair of CCW neighbor with respect to the source of the message. */
public DT_AddressPair getCCW(){
        return CCW;
    }

	/** Returns the address pair of CW neighbor with respect to the source of the message. */
public DT_AddressPair getCW(){
        return CW;
    }

/** the same as getCW */
public DT_AddressPair getLeader() 
    {
        return CW;
    }
}







