/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.DT;

import hypercast.*;
import hypercast.util.XmlUtil;

import java.io.*;
import java.util.*;
import java.net.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.apache.xpath.*;

/** This class extends the basic Delaunay triangulation node to
 * provide a mechanism for joining a overlay or resolving partitions
 * through a server. 
 * Partition solving and Joins are initiated by the node sending a
 * ServerRequest message to the Server.  The server responds
 * by sending a ServerReply message back to the node. The ServerReply message
 * contains a node that is already in the overlay.
 * <p>
 * The new node then sends this node a NewNode message. The NewNode message will 
 * be routed on the topology until it reaches a node where the new node can pass
 * the neighbor test. The routing is done by compass routing.
 * That node will then attempt to add the new node as a neighbor.
 * <p>
 * @author HyperCast Team
 * @author Michael Nahas
 * @author Weisheng Si
 * @author Jianping Wang
 * @version 3.0, Jan. 08, 2004 
 */

public class DT_Node_ServerJoin extends DT_Node
{

    /** Name of property controlling the default server heartbeat time. */
	public static final String MIN_SERVER_TIME_PROPERTY_NAME = "ServerHeartbeat"; 

    /** Name of property controlling the maximum server delay before trying a backup server. */
	public static final String MAX_SERVER_TIME_PROPERTY_NAME = "TryBackupServerTime"; 

    /** Prefix for the name of the properties holding the URLs of the servers. */
	public static final String SERVER_PREFIX = "Rendezvous";

    /** Adapter timer index used for server heartbeat. */
    public static final int SERVER_TIMER_INDEX = 1;

    /** When server does not respond, ratio of backoff. */
    public static final float BACKOFF_RATIO = (float) 2.0;

    /** Times a node trys to contact DT server before exit. */
    public static final int MAX_CONTACT_SERVER_NUMBER = 20;

    /** Node state  */
    public static final int JOINING = 0;
    /** Node state  */
    public static final int STABLE = 1;
    /** Node state  */
    public static final int LEADER_STABLE = 2;
    /** Node state  */
    public static final int INCOMPLETE = 3;
    /** Node state  */
    public static final int LEADER_INCOMPLETE = 4;
    /** Node state  */
    public static final int UNCONFIRMED_LEADER_STABLE = 5;
    /** Node state  */
    public static final int UNCONFIRMED_LEADER_INCOMPLETE = 5;
    /** Node state  */
    public static final int DEAD = 6;


    /** In milliseconds, the default server heartbeat time. */
    private long MIN_SERVER_TIME; // milliseconds
    /** In milliseconds, the maximum server heartbeat time before a backup server is tried. */
    private long MAX_SERVER_TIME; // milliseconds
    
    /** Used to coordinate resetting and clearing of the server timer.*/
    private boolean serverTimerCleared;

    /** The index of the server being sent to.  
     * 0 = primary, 1 = backup, 2 = second backup, 3 = third backup, etc. */
    private int serverIndex;

    /** Number of Attempts to send message to server. */
    private int sendAttempts;

    /** Counter for exponential backoff. */
    private int backoffCounter;

    /** Time the last message was received from the server. */
    private long timeOfLastReceive;

    /** True, if at time of last receive, Server thought this was the Leader. */
    private boolean serverSaidThisWasLeader;


    /** Vector of I_PhysicalAddress objects containing addresses of servers. */
    Vector serverAddresses;

   
    /* Constructor.
	 * @param	c				an HyperCastConfig object
	 * @param	a				node adapter
	 * @param	startingCoords	initial logical address
     * @throws IllegalArgumentException if properties are incorrectly formatted.
     */
	public DT_Node_ServerJoin(HyperCastConfig c, I_UnicastAdapter a, DT_LogicalAddress startingCoords)
    {
        super(c, a, startingCoords);
        init();
    }


    /* Constructor.
	 * @param	c				an HyperCastConfig object
	 * @param	a				node adapter
	 * @param	startingCoords	initial logical address
	 * @param	prefix			property prefix
     * @throws IllegalArgumentException if properties are incorrectly formatted.
     */
	public DT_Node_ServerJoin(HyperCastConfig c, I_UnicastAdapter a, DT_LogicalAddress startingCoords, String prefix)
    {
        super(c, a, startingCoords, prefix);
        init();
    }

	/* Constructor.
	 * @param	c				an HyperCastConfig object
	 * @param	a				node adapter
	 * @throws IllegalArgumentException if properties are incorrectly formatted.
	 */
	public DT_Node_ServerJoin(HyperCastConfig c, I_UnicastAdapter a)
	{
            super(c, a);
            init();
	}


    /* Constructor.
	 * @param	c				an HyperCastConfig object
	 * @param	a				node adapter
	 * @param	prefix			property prefix
     * @throws IllegalArgumentException if properties are incorrectly formatted.
     */
	public DT_Node_ServerJoin(HyperCastConfig c, I_UnicastAdapter a, String prefix)
    {
        super(c, a, prefix);
        init();
    }


	/**
	 * This method finishes initialization task. It is called by all constructors
	 * of DT_Node_ServerJoin.
	 */
    private void init()
    {

		MIN_SERVER_TIME = config.getPositiveLongAttribute
				(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+MIN_SERVER_TIME_PROPERTY_NAME));

		MAX_SERVER_TIME = config.getPositiveLongAttribute
				(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+MAX_SERVER_TIME_PROPERTY_NAME));

        serverTimerCleared = true;

		//Set Proto_Sub variable defined in class DT_Node to PROTOSUB_SERVER (using server).
		Proto_Sub = I_Node.PROTOSUB_SERVER;
	
		//Obtain the addresses of DT servers from an HyperCastConfig object.
		obtainServerAddresses();

		/*
		 * Initialize the data structure of statistics maintained in this object.
		 */
		initStatisticsStructure();
    }

    /** Read the IP addresses of DT servers from an configuration object */
    private void obtainServerAddresses()
    {
        this.serverAddresses = new Vector();

		/*
		 * XPath index starts from 1 not 0.
		 */
        for (int i = 1; true; i++)
        {
        	String addrType;
        	try
			{
        		addrType = config.getTextAttribute(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX + SERVER_PREFIX + "[" + i + "]/UnderlayAddress"));
			}catch (HyperCastConfigException e)
			{
				/*
				 * The attribute doesn't exist.
				 */
				break;
			}
        	if(addrType == null)
        		break;
        	
			String propName = PROPERTY_PROTO_PREFIX + SERVER_PREFIX + "[" + i + "]/UnderlayAddress/" + addrType;

			String s = null;
			try
			{            
				s = config.getTextAttribute(XmlUtil.createXPath(propName));
			}
			catch (HyperCastConfigException e)
			{
				/*
				 * The attribute doesn't exist.
				 */
				break;
			}
            
            I_UnderlayAddress serverAddr = adapter.createUnderlayAddress(s);

            if (serverAddr != null)
            {
                serverAddresses.addElement(serverAddr);

                //System.out.println("The "+i+"th DT server is :"+serverAddr.toString());
            } 
            else
            {
                config.err.println("Property " + PROPERTY_PROTO_PREFIX + SERVER_PREFIX + i + " contained a bad address. ");         
            }
        }
        
        if (serverAddresses.size() == 0)
        {       
            throw new HyperCastFatalRuntimeException ("Error: Attribute " + PROPERTY_PROTO_PREFIX + SERVER_PREFIX + " did not contain correctly formated addresses!");
        }
    }




    /**
     *  Returns the number of servers available.
     */
    int numOfServers()
    {
        return serverAddresses.size();
    }


/** Joins overlay multicast overlay. */
public void joinOverlay() 
    {
        super.joinOverlay();

        sendAttempts = 0;
        backoffCounter = 0;
        serverIndex = 0;
        timeOfLastReceive = adapter.getCurrentTime() - TIMEOUT_TIME;

        // random time prevents synchronisity if large numbers are started at the same time.
        synchronized (this) 
        {
            adapter.setTimer(new Integer(SERVER_TIMER_INDEX), (long) (rand.nextFloat() * MIN_SERVER_TIME)); 
            serverTimerCleared = false;
            notifyAll();
        }
    }
    
/** Leaves overlay multicast Overlay. */
public synchronized void leaveOverlay() 
    {
        
        adapter.clearTimer(new Integer(SERVER_TIMER_INDEX));
        serverTimerCleared = true;
        notifyAll();
        sendToServer(DT_Message.Goodbye);
        
        super.leaveOverlay();
    }
    


/** I_AdapterCallback function.  It handles all messages from the
    server and passes all other messages on to
    DT_Node.messageArrivedFromAdapter. */

public synchronized void messageArrivedFromAdapter(I_Message a) 
    {
        MessageCallbackDebug = "JustCalled(DSJ)";

        if (!(a instanceof DT_Message))
            throw new IllegalArgumentException("DT_Node.messageArrivedFromAdapter() passed non-DT_Message I_Message.");
        DT_Message m = (DT_Message) a;

		//Check whether the Protocol number in a DT message is defined correctly.
        if (m.getProtoNum() != DT_Message.PROTONUM_DT30)
        {
            config.log.println("Received message with wrong Protocol Number: " + m.getProtoNum() + " != " + DT_Message.PROTONUM_DT30);
            return;  // ignore the message
        }
        if (m.getOverlayHash() != config.getOverlayHash())
            return;  // ignore the message

//        config.log.println("Received message type " + m.getLabel() + "\n");

        messagesReceivedInLastHeartbeat++;

        switch (m.getType())
        {
          case DT_Message.ServerReply: 
            handleServerReply(m);
            break;

          default:
            messagesReceivedInLastHeartbeat--;  // prevent double counting.
            super.messageArrivedFromAdapter(m);
        }

        MessageCallbackDebug = "Returned(DSJ)";
    }
    

	/** Performs actions for the ServerReply message.
	 * (i.e. attempts to add the Leader as a neighbor.) 
	 */
	private void handleServerReply(DT_Message m) 
	{
		MessageCallbackDebug = "handleServerReply(DSJ)";

		sendAttempts = 0;
		changeServerTimerToTime(MIN_SERVER_TIME);

		timeOfLastReceive = adapter.getCurrentTime();
		serverSaidThisWasLeader = (neighborhood.getPhysicalAddress()).equals(m.getLeaderPA());

		if (m.getLeaderPA() != null &&
			!serverSaidThisWasLeader &&
			neighborhood.isLeader())
		{
			MessageCallbackDebug = "SendingNewNodeToLeader(DSJ)";
			sendNewNodeMessage(m.getLeader(), neighborhood.getMyAddressPair());
		}
	}


	/** Sends a message of specified type to the server. */
	protected void sendToServer(byte packet_type)
    {
        messagesSentInLastHeartbeat++;

		I_Message NewPacket;

		NewPacket = createDTMessage(packet_type, null, null, null);

		I_UnderlayAddress ServerAddr = (I_UnderlayAddress) serverAddresses.elementAt(serverIndex);

        adapter.sendUnicastMessage(ServerAddr, NewPacket);
    }


	/** Change the frenquency of contacting DT servers */
	protected synchronized final void changeServerTimerToTime(long newTime) 
    {
        long delay = 0;

        if (!serverTimerCleared)
        {
        	adapter.setTimer(new Integer(SERVER_TIMER_INDEX), newTime);
        }
    }
    

/** Handles timer events, such as when the Leader sends a message to the server.
 * It does call super.handleTimer() to take care of other timer events. 
 */
public synchronized void timerExpired(Object TimerIndex) 
{
    TimerCallbackDebug = "JustCalled(DSJ)";

    int timer_index = ((Integer) TimerIndex).intValue();

    if (timer_index == SERVER_TIMER_INDEX) 
    {
        if (!neighborhood.isLeader()) 
        {
            TimerCallbackDebug = "Not Leader - resetting timer to min (DSJ)";

            sendAttempts = 0;
            serverIndex = 0;
            adapter.setTimer(new Integer(SERVER_TIMER_INDEX), MIN_SERVER_TIME );
            notifyAll();
        }
        else 
        {
            TimerCallbackDebug = "Leader - calculating time to wait (DSJ)";

            sendAttempts++;
            
            //It prevents a hypercast socket to run forever in the cast DT server has stopped working.
            int num_of_Servers = numOfServers();
            if (sendAttempts > num_of_Servers * MAX_CONTACT_SERVER_NUMBER) {
                throw new HyperCastFatalRuntimeException ("DT server stop working, the maximum try times " + MAX_CONTACT_SERVER_NUMBER + " is reached, exit.");
            }

            long maxTimeToWait = (long) (MIN_SERVER_TIME * Math.pow(BACKOFF_RATIO, sendAttempts));
            if (maxTimeToWait > MAX_SERVER_TIME) 
            {
                if (numOfServers() > 1)
                {
                    serverIndex = (serverIndex + 1) % numOfServers();
                    maxTimeToWait = MIN_SERVER_TIME;
                    config.log.println("DT_Node at coords "+ neighborhood.getLogicalAddress().toString() + 
                                       " had no server response - switching to server #" + serverIndex);
                }
                else 
                {
                    // bound maximum wait time
                    maxTimeToWait = MAX_SERVER_TIME;
                    config.log.println("DT_Node at coords "+ neighborhood.getLogicalAddress().toString() + 
                                       " had no server response -  will keep trying.");
                }
            }
                
            TimerCallbackDebug = "Leader - resetting timer (DSJ)";

            long waitTime = (long) (MIN_SERVER_TIME + (rand.nextFloat() * (maxTimeToWait - MIN_SERVER_TIME)));
            adapter.setTimer(new Integer(SERVER_TIMER_INDEX), waitTime );
            notifyAll();

            TimerCallbackDebug = "SendingToServer(DSJ)";
            sendToServer(DT_Message.ServerRequest);
            TimerCallbackDebug = "ReturnedFromSendToServer(DSJ)";
        
			if (!StartConsiderStable) StartConsiderStable = true;
        }
    }
    else 
    {
        super.timerExpired(TimerIndex);
    }
}

/** Returns the current state of the node.  
 *
 * State is one of: JOINING, STABLE, LEADER_STABLE, INCOMPLETE, or
 * LEADER_INCOMPLETE.  
 */
public int getState() 
{
    boolean stable = neighborhood.isStable();
    boolean leader = neighborhood.isLeader();
    boolean confirmedLeader = leader && serverSaidThisWasLeader && (adapter.getCurrentTime() - timeOfLastReceive) < TIMEOUT_TIME;
    
    if (!joined)
        return DEAD;
    else if (confirmedLeader && stable) 
        return LEADER_STABLE;
    else if (confirmedLeader && !stable)
        return LEADER_INCOMPLETE;
    else if (!leader && stable)
        return STABLE;
    else if (!leader && !stable)
        return INCOMPLETE;
    else if ( leader && stable) 
        return UNCONFIRMED_LEADER_STABLE;
    else
        return UNCONFIRMED_LEADER_INCOMPLETE;
}

/**
 * Initialize the data structure for handling the statistics defined in this object.
 * This class shares the StatsProcessor instance with its super class DT_Node.
 * In this method, the statistics, with each represented by a StatsElement instance, 
 * supported in this object are added into the StatsProcessor instance.
 */
private void initStatisticsStructure()
{
	statsPro.addStatsElement("State"/* name */, new NodeState()/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);

	statsPro.addStatsElement("SendAttempts", new SendAttemptsQuery(), 1, 1);

	statsPro.addStatsElement("ServerIndex", new CurrentServerIndex(), 1, 1);

	statsPro.addStatsElement("BackoffCounter", new BackoffTimes(), 1, 1);
}

/**
 * This class manages simple statistic "State".
 */
class NodeState extends SimpleStats
{
	/**
	 * Gets the value of a simple statistic managed by the implementation of this class.
	 * 
	 * @see SimpleStats#getStats()
	 */
	protected String getStats()
	{
		return ""+getState();
	}
	
	/**
	 * Sets the value of a simple statistic managed by the implementation of this class.
	 * 
	 * @see SimpleStats#setStats(String newvalue)
	 */
	protected String setStats(String newValue) throws HyperCastStatsException
	{
		throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
	}

	/** 
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
											null/* restriction base */, null/* pattern value */);
	}
}

/**
 * This class manages simple statistic "SendAttempts".
 */
class SendAttemptsQuery extends SimpleStats
{
	/**
	 * Gets the value of a simple statistic managed by the implementation of this class.
	 * 
	 * @see SimpleStats#getStats()
	 */
	protected String getStats()
	{
		return ""+sendAttempts;
	}
	
	/**
	 * Sets the value of a simple statistic managed by the implementation of this class.
	 * 
	 * @see SimpleStats#setStats(String newvalue)
	 */
	protected String setStats(String newValue) throws HyperCastStatsException
	{
		throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
	}

	/** 
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
											null/* restriction base */, null/* pattern value */);
	}
}

/**
 * This class manages simple statistic "ServerIndex".
 */
class CurrentServerIndex extends SimpleStats
{
	/**
	 * Gets the value of a simple statistic managed by the implementation of this class.
	 * 
	 * @see SimpleStats#getStats()
	 */
	protected String getStats()
	{
		return ""+serverIndex;
	}
	
	/**
	 * Sets the value of a simple statistic managed by the implementation of this class.
	 * 
	 * @see SimpleStats#setStats(String newvalue)
	 */
	protected String setStats(String newValue) throws HyperCastStatsException
	{
		throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
	}

	/** 
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
											null/* restriction base */, null/* pattern value */);
	}
}

/**
 * This class manages simple statistic "BackoffCounter".
 */
class BackoffTimes extends SimpleStats
{
	/**
	 * Gets the value of a simple statistic managed by the implementation of this class.
	 * 
	 * @see SimpleStats#getStats()
	 */
	protected String getStats()
	{
		return ""+backoffCounter;
	}
	
	/**
	 * Sets the value of a simple statistic managed by the implementation of this class.
	 * 
	 * @see SimpleStats#setStats(String newvalue)
	 */
	protected String setStats(String newValue) throws HyperCastStatsException
	{
		throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
	}

	/** 
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
											null/* restriction base */, null/* pattern value */);
	}
}

}









