/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.HC;

import hypercast.*;
import hypercast.events.*;
import hypercast.util.XmlUtil;

import java.net.*;
import java.io.*;
import java.util.*;


import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.apache.xpath.*;

/**	
 * HC_Node is a full implementation of the Hypercube protocol.
 * 
 * 1. One new constructor is added which contain a string "prefix"
 * in their parameters. The prefix is used for getting attributes.
 * 3. A new method createDTMessage is added which is the constructor
 * of HC_Message and is called by all sending methods. It can be 
 * overrided by the extended security node classes.
 * 4. For all timer related methods, such as setTimer, getTimer, cleanTimer
 * and timerExpired, the "timer type" parameter is redifined as a 
 * general "Object" instead of a int variable.
 * 5. The Timer in the node adapter is started/stopped explicitly.
 * 
 * @author HyperCast Team
 * @author Tyler Beam
 * @author Michael Nahas
 * @author Dongwen Wang
 * @version	2.0, Feb. 28, 2001
 * @author Jianping Wang
 * @version 3.0, Jan. 08, 2004 
 */

public final class HC_Node implements I_Node, I_AdapterCallback
{
	
    	/** The HyperCastConfig object containing the properties of this group. */
	private HyperCastConfig config;

	/** The property prefix, it can be changed by constructors. */
    protected String PROPERTY_PROTO_PREFIX = "/Public/Node/HC/";

		/** The multicast adapter that this protocol node runs on */
	private I_MulticastAdapter Adapter;
		/** The neighborhood of this node */
	private HC_Neighborhood neighbors;
		/** The logical address of this node */
	HC_LogicalAddress MyLogicalAddress;
	
		/** The physical address of this node */
	I_PhysicalAddress MyPhysicalAddress;
	
		/** The state of this node */
	private byte MyState;
	
		/** The state of this node, joinng */
	public static final byte Joining = 0;
		/** The state of this node, stable */
	public static final byte Stable = 1;
		/** The state of this node, missing neighbor */
	public static final byte MissingNeighbor = 2;
		/** The state of this node, missing neighbor action */
	public static final byte MissingNeighborAction = 3;
		/** The state of this node, hroot  */
	public static final byte HRoot = 4;
		/** The state of this node, hroot missing neighbor */
	public static final byte HRootMissingNeighbor = 5;
		/** The state of this node, hroot missing neighbor action */
	public static final byte HRootMissingNeighborAction = 6;
		/** The state of this node, leaving  */
	public static final byte Leaving = 7;
		/** The state of this node, start a hypercube  */
	public static final byte StartHypercube = 8;
		/** The state of this node, wait for joining  */
	public static final byte JoiningWait = 9;
		/** The state of this node, dead  */
	public static final byte Dead = 10;
	
	
		/** The maximum of the Hypercube dimension, pretty much must be 31 */
	final static int MaxHypercubeDim = 31;		

		/** The Heart Beat time of the node, in ms */
	private int SleepTime;

		/** The maximum time of a node's age, in heartbeats (Ttimeout / Theartbeat) */
    	int MaxAge;	

		/** The maximum missing neighbor time of a node, in heartbeats (Tmissing / Theartbeat) */
	int MaxMissingNeighbor;	

		/** The maximum supressing joining time of a node, in heartbeats (Tjoining / Theartbeat) */
	private int MaxSuppressJoinBeacon;	


    	/** True if joinGroup has not been followed by a leaveOverlay */
   	private boolean joined;
    
	/** The protocol subfield value which represents the used rendezvous and will be put 
	 * in the Proto_sub field of a HC_Message. It will be changed in the constructors. 
	 */
    byte Proto_Sub = I_Node.PROTOSUB_NOSPEC;
	
	/** Notification handler. When interesting events happen, this node 
	 * reports the events to the notification handler if it is not null.
	 */
	protected NotificationHandler n_handler = null;
	
	/** Check mode on the previous hop of OL messages.
	 */
	protected String checkmode;

	/**
	 * A StatsProcessor instance which handles the statistics in this object.
	 */
	private StatsProcessor statsPro;

	/**
	 * The statistics name for this object (as an I_Stats instance). Its default
	 * name is defined in the socket configuration schema file. It can be assigned
	 * by any other object through method setStattisticsName defined in the 
	 * I_Stats interface.
	 */
	private String statisticsName;


	/**
	 * Constructs a HC_Node object with an infrustructure I_UnicastAdapter
	 * @param	c				an HyperCastConfig object
	 * @param	madapter	 an object of I_MulticastAdapter.
	 */
	public HC_Node(HyperCastConfig c, I_MulticastAdapter madapter) {
		
		config = c;
		
		MyLogicalAddress  = new HC_LogicalAddress(MaxAddress(), true);
        
		MyState = Joining;
		
		Adapter = madapter;
		Adapter.setCallback(this);
		
        	joined = false;

		MyPhysicalAddress = Adapter.createPhysicalAddress();

		//get SleepTime (HeartBeat)
		SleepTime = config.getPositiveIntAttribute
				(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+"SleepTime"));

		//get MaxAge
		MaxAge = config.getPositiveIntAttribute
				(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+"MaxAge"));

		//get MaxMissingNeighbor
		MaxMissingNeighbor = config.getPositiveIntAttribute
				(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+"MaxMissingNeighbor"));

		//get MaxSuppressJoinBeacon
		MaxSuppressJoinBeacon = config.getPositiveIntAttribute
		(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+"MaxSuppressJoinBeacon"));

       	Proto_Sub = I_Node.PROTOSUB_MULTICAST;

		checkmode = config.getTextAttribute
				(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+"Verification"));

		initStatisticsStructure();

		// Then Initialize neighbors
		neighbors   = new HC_Neighborhood(this, statsPro);
	}

	/**
	 * Constructs a HC_Node object with an infrustructure I_UnicastAdapter and attribute prefix
	 * @param	c				an HyperCastConfig object
	 * @param	madapter	 an object of I_MulticastAdapter.
	 * @param	prefix			property prefix
	 */
	public HC_Node(HyperCastConfig c, I_MulticastAdapter madapter, String prefix) {
		
		config = c;
		
		PROPERTY_PROTO_PREFIX = prefix;

		MyLogicalAddress  = new HC_LogicalAddress(MaxAddress(), true);
        
		MyState = Joining;
		
		Adapter = madapter;
		Adapter.setCallback(this);
		
        	joined = false;

		MyPhysicalAddress = Adapter.createPhysicalAddress();

		//get SleepTime (HeartBeat)
		SleepTime = config.getPositiveIntAttribute
				(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+"SleepTime"));

		//get MaxAge
		MaxAge = config.getPositiveIntAttribute
				(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+"MaxAge"));

		//get MaxMissingNeighbor
		MaxMissingNeighbor = config.getPositiveIntAttribute
				(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+"MaxMissingNeighbor"));

		//get MaxSuppressJoinBeacon
		MaxSuppressJoinBeacon = config.getPositiveIntAttribute
		(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+"MaxSuppressJoinBeacon"));

		Proto_Sub = I_Node.PROTOSUB_MULTICAST;

		checkmode = config.getTextAttribute
				(XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+"Verification"));

		initStatisticsStructure();

		// Then Initialize neighbors
		neighbors   = new HC_Neighborhood(this, statsPro);
	}

	/**

	/**
	 * Joins a multicast group
	 */
	public void joinOverlay() {
        if (joined)
            throw new IllegalStateException("HC_Node.joinGroup() called twice without leaveOverlay()!");

        resetNeighborhood();
		MyLogicalAddress  = new HC_LogicalAddress(MaxAddress(), true);
		MyState = Joining;

		Adapter.Start();		
		Adapter.startTimer();		
		Adapter.setTimer(new Integer(0), SleepTime);

        joined = true;
	}
	
	
	/**
	 * Leaves a multicast group
	 */
	public synchronized void leaveOverlay() {
        if (!joined)
            throw new IllegalStateException("HC_Node.leaveOverlay() called without calling joinGroup!");

        // send bye messages to all neighbors
        leaveNeighborhood();
        Adapter.clearTimer(new Integer(0));

		Adapter.stopTimer();		

		Adapter.Stop();		

        joined = false;

		if (n_handler != null) {
			NODE_LEAVEOVERLAY ne = new NODE_LEAVEOVERLAY(Adapter.getCurrentTime(), null);
			n_handler.eventOccurred(ne);
		}
	}
	
	
	/**
	 * Gets the maximum logical address number of this hypercube.
	 * Basically need a 1 in each of the positions. So shift and add 1 for
	 * MaxHypercubeDim-1 times
	 * 
	 * @return	The MaxLogicalAddress
	 */

	static public int MaxAddress() {
		int Result = 0;

		for (int i = 0; i < MaxHypercubeDim; i++) {
			Result = (Result << 1) + 1;
		}

		return Result;
	}



	
	/**
	 * Creates an HC_Message if there are enough valid bytes in the buffer.
     * @see I_AdapterCallback#restoreMessage
	 */
	public I_Message restoreMessage(byte[] recvbuff, int[] validBytesStart, int validBytesEnd)  {
		
		return HC_Message.restoreMessage(recvbuff, validBytesStart, validBytesEnd, Adapter, config.getOverlayHash());
    }
	
	
	/**
	 * Creates an I_LogicalAddress object from an bytes array.
	 * @param	laddr	bytes array 
	 * @return	an object of type I_LogicalAddress
	 */
	public I_LogicalAddress createLogicalAddress(byte[] laddr, int offset) {
		if ((laddr.length - offset) < getAddressPair().getLogicalAddress().getSize())
			throw new IllegalArgumentException("The size of byte array is less than the size of HC_LogicalAddress!");

		int colex = ByteArrayUtility.toInteger(laddr, offset);
		return (I_LogicalAddress)new HC_LogicalAddress(colex, true);		
	}


  /** Creates a logical address object from a String.
   */
  public I_LogicalAddress createLogicalAddress(String laStr)
  {
	  int colex;
	  try 
	  {
		  colex = Integer.parseInt(laStr);
	  }
	  catch (NumberFormatException nfe)
	  {
		  config.err.println("HC_Node: Could not parse \""+laStr+"\" to create logical address");
		  return null;
	  }
	  return (I_LogicalAddress)new HC_LogicalAddress(colex, true);		
  }


  /** Sets the logical address of this node to the specified one.
   */
  public void setLogicalAddress(I_LogicalAddress la)
  {
	  if (!(la instanceof HC_LogicalAddress))
	  {
		 throw new IllegalArgumentException("DT_Node_Buddylist: setLogicalAddress: the logical address type of the parameter is incorrect.");
	  }

	  //TODO: check if this is correct
	  throw new HyperCastWarningRuntimeException ("The logical address of a HC node cannot be set with specified one.");

  }


	/** Verify the previous hop of the message. If checkmode is set to "neighborCheck",
	 * this method checks if the previous hop is a neighbor; otherwise this method checks
	 * if the previous hop is a valid sender of OL messages.
	 */
	public boolean previousHopCheck(I_LogicalAddress src, I_LogicalAddress dst, I_LogicalAddress prehop)
	{
		if (checkmode.equals("neighborCheck")) {
			return neighbors.contains(prehop);
		}
		else {
		   if (dst == null) //multicast or flooding
		   {
			  I_AddressPair[] Parent = getParent(src);
			  if (Parent.length != 1) return false;  //no parent found
		
			  I_LogicalAddress Parent_LA = Parent[0].getLogicalAddress();
			  return ((HC_LogicalAddress) Parent_LA).equals((HC_LogicalAddress) prehop);
		   } else //unicast
		   {
			  I_AddressPair[] Children = getChildren(dst);
			  if (Children == null) return false;
		
			  for (int i = 0; i < Children.length; i++) {
			     if ((Children[i].getLogicalAddress()).equals((HC_LogicalAddress) prehop))
				    return true;
			  }
		   }
		   return false;
		}
	}


  /**
   * Set notification handler.
   */	
  public void setNotificationHandler(NotificationHandler nh)
  {
	  n_handler = nh;
  }

	
	/** 
	 * Processes the message received from an I_UnicastAdapter.
	 * This is the kernel part of the Hypercast protocol. 
	 * The HC_Node(Hypercast protocol node) changes it's soft state based on 
	 * the Messages it receives. 
	 * 
	 * @param	msg	I_Message
	 */	
    public synchronized void messageArrivedFromAdapter(I_Message msg) {	
        if (!(msg instanceof HC_Message))
            throw new IllegalArgumentException("HC_Node.processMessage() passed I_Message not of type HC_Message");

		HC_Message X = (HC_Message) msg;

		//Check whether the Protocol number in a HC message is defined correctly.
        if (X.getProtoNum() != HC_Message.PROTONUM_HC30) {
            config.log.println("Received message with wrong Protocol Number: " + X.getProtoNum() + " != " + HC_Message.PROTONUM_HC30); 
	    return;  // ignore message from other group.
	}

        if (X.getOverlayHash() != config.getOverlayHash())
            return;  // ignore message from other group.
        

//        config.log.println(Adapter.getPhysicalAddress().toString() + " in state " + getState() + " received: " + X.toString());
//        config.log.println(MyPhysicalAddress.toString() + " in state " + getState() + " received: " + X.toString());

        byte MyOldState = MyState;

        if (X.getSrcPA().equals(MyPhysicalAddress))
            return;

SkipItReceiver:		
        switch (MyState) {

			case HC_Node.Joining:
			case HC_Node.JoiningWait:

				if ((X.getType() != HC_Message.Beacon) && !X.getDestPA().equals(MyPhysicalAddress)) {
//config.log.println("Destination physical address incorrect.\n", Debug);
					break SkipItReceiver;
				}

				switch (X.getType()) {
				case HC_Message.Ping:
//config.log.println("Taking on new logical address.\n");
					resetNeighborhood();
					MyLogicalAddress = new HC_LogicalAddress(X.getDestLA());
					
					if (n_handler != null) {
						NODE_LOGICALADDRESSCHANGED ne = new NODE_LOGICALADDRESSCHANGED(Adapter.getCurrentTime(), null);
						n_handler.eventOccurred(ne);
					}
					
					neighbors.setHRootSeqNum(-1);
					neighbors.updateHRoot(X.getHRootLA(), X.getHRootSeqNum());

					if (neighbors.getHRoot().getLogicalAddress().equals(MyLogicalAddress)) {
//config.log.println("Received ping as new root.\n");
//config.log.println("Switching to HRootMissingNeighbor.\n");
						MyState = HC_Node.HRootMissingNeighbor;
						neighbors.MNTimeOut = MaxMissingNeighbor;

						neighbors.HRootSeqNum++;
						if (neighbors.HRoot != null)
							neighbors.HRoot.setAge(MaxAge);
						multicastSend(HC_Message.Beacon);
					}
					else {
//config.log.println("Received ping as non-root.\n");
//config.log.println("Switching to MissingNeighbor.\n");
						MyState = HC_Node.MissingNeighbor;
						neighbors.MNTimeOut = MaxMissingNeighbor;
					}
					neighbors.register(new HC_AddressPair(this, X.getSrcLA(), X.getSrcPA(), false));
				break;

				case HC_Message.Beacon:
					if (X.getSrcLA().equals(new HC_LogicalAddress(MaxAddress(), true))) {
//config.log.println("Found other joining node, suppressing own beacons.\n");

						neighbors.MNTimeOut = MaxSuppressJoinBeacon;
						MyState = HC_Node.JoiningWait;
					}
					else if (X.getHRootSeqNum() != 0) {

//config.log.println("Found existing HRoot, suppressing own StartHypercube response.\n");
//						neighbors.HRoot = null;
						neighbors.setHRootSeqNum(-1);
						neighbors.updateHRoot(X.getHRootLA(), X.getHRootSeqNum());
						neighbors.HRoot.setAge(MaxAge);
					}
				break;

				default:
//config.log.println("Packet not handled.");
				break;
				}
			break;

			case HC_Node.StartHypercube:
//config.log.println("StartHypercube.\n");
			break;

			case HC_Node.Leaving:
//config.log.println("Leaving.\n");
			break;

			case HC_Node.Stable:
			case HC_Node.MissingNeighbor:
			case HC_Node.HRoot:
			case HC_Node.HRootMissingNeighbor:
			case HC_Node.MissingNeighborAction:
			case HC_Node.HRootMissingNeighborAction:

				if (X.getSrcLA().equals(MyLogicalAddress) &&
				(X.getType() != HC_Message.Leave)) {
					if (X.getSrcPA().greaterThan(MyPhysicalAddress)) {
//config.log.println("Caught an LA conflict, resetting.\n");
						leaveNeighborhood();
						resetNeighborhood();
//config.log.println("Switching to Leaving.\n");
						MyState = HC_Node.Leaving;
					} else {
//config.log.println("Caught an LA conflict, sending Kill message.\n");
						unicastSend(HC_Message.Kill, X.getSrcPA(), X.getSrcLA());
					}
					break SkipItReceiver;
				}

				if ((X.getType() != HC_Message.Beacon) && !X.getDestPA().equals(MyPhysicalAddress)) {
//config.log.println("Destination physical address incorrect.\n");
					break SkipItReceiver;
				}

				if ((X.getDestLA().getColexIndex() > MyLogicalAddress.getColexIndex()) &&
				(X.getType() != HC_Message.Beacon)) {
//config.log.println("Addressed to higher LA.\n");
					unicastSend(HC_Message.Leave, X.getSrcPA(), X.getSrcLA());
					break SkipItReceiver;
				}

				switch (X.getType()) {
				case HC_Message.Ping:
					if (X.getDestLA().equals(MyLogicalAddress)) {
						if (!neighbors.register(new HC_AddressPair(this, X.getSrcLA(), X.getSrcPA(), false))) {
//config.log.println("Could not register node in neighborhood view.\n");
							unicastSend(HC_Message.Kill,
										X.getSrcPA(), X.getSrcLA());
							break SkipItReceiver;
						}
//config.log.println("Registered node in neighborhood view.\n");

					}
					else if (X.getDestLA().getColexIndex() < MyLogicalAddress.getColexIndex()) {
//config.log.println("Pinged to a lower logical address.\n");
						leaveNeighborhood();
						resetNeighborhood();
						MyLogicalAddress = new HC_LogicalAddress(X.getDestLA());

						//Report event LACHANGED to notification handler.
						if (n_handler != null) {
							NODE_LOGICALADDRESSCHANGED ne = new NODE_LOGICALADDRESSCHANGED(Adapter.getCurrentTime(), null);
							n_handler.eventOccurred(ne);
						}
					
//config.log.println("Switching to MissingNeighbor.\n");
						MyState = HC_Node.MissingNeighbor;
						neighbors.MNTimeOut = MaxMissingNeighbor;
						break SkipItReceiver;
					}
					else {
//config.log.println("Pinged as a higher logical address.\n");
						if (X.getHRootSeqNum() > neighbors.getHRootSeqNum()) {
//config.log.println("Higher sequence number taken as meaning more recent information.\n");
							leaveNeighborhood();
							resetNeighborhood();
							MyLogicalAddress = new HC_LogicalAddress(X.getDestLA());

							//Report event LACHANGED to notification handler.
							if (n_handler != null) {
								NODE_LOGICALADDRESSCHANGED ne = new NODE_LOGICALADDRESSCHANGED(Adapter.getCurrentTime(), null);
								n_handler.eventOccurred(ne);
							}
					
//							neighbors.updateHRoot(X.getHRootLA(), X.getHRootSeqNum());
							if (neighbors.HRoot.getLogicalAddress().equals(MyLogicalAddress)) {
//config.log.println("Received ping as new root.\n");
//config.log.println("Switching to HRootMissingNeighbor.\n");
								MyState = HC_Node.HRootMissingNeighbor;
							}
							else {
//config.log.println("Received ping as non-root.\n");
//config.log.println("Switching to MissingNeighbor.\n");
								MyState = HC_Node.MissingNeighbor;
							}
							neighbors.MNTimeOut = MaxMissingNeighbor;
						}
					}


				break;

				case HC_Message.Beacon:
					if (!X.getSrcLA().equals(
							new HC_LogicalAddress(MaxAddress(), true))) {
//config.log.println("Attempting to register beacon sender in neighborhood.\n");
						neighbors.register(new HC_AddressPair(this, X.getSrcLA(), X.getSrcPA(), false));
						if (X.getSrcLA().equals(X.getHRootLA()))
							neighbors.updateHRoot(X.getHRootLA(), X.getHRootSeqNum());
					}

					if (((MyState == HC_Node.MissingNeighborAction) ||
					(MyState == HC_Node.HRootMissingNeighborAction)) &&
					(X.getSrcLA().equals(neighbors.HRoot.getLogicalAddress()) ||
					X.getSrcLA().equals(new HC_LogicalAddress(MaxAddress(), true)))) {
//config.log.println("Taking action to move HRoot or Joining node.\n");
						int tear_index = neighbors.searchTear();
						if ((tear_index != MaxAddress()) &&
						(MyLogicalAddress.getNthNeighbor(tear_index).getColexIndex() <
						 X.getSrcLA().getColexIndex())) {
//config.log.println("Executing change in neighborhood view.\n");
							unicastSend(HC_Message.Ping,
								X.getSrcPA(),
								MyLogicalAddress.getNthNeighbor(tear_index));

							break SkipItReceiver;
						}

					}
            
					if ((MyState == HC_Node.HRoot) ||
					(MyState == HC_Node.HRootMissingNeighbor) ||
					(MyState == HC_Node.HRootMissingNeighborAction)) {
						int tear_index = MyLogicalAddress.getColexIndex() + 1;
						if (X.getSrcLA().equals(new HC_LogicalAddress(MaxAddress(), true))) {
//config.log.println("Adding Joining node to top of hypercube.\n");


							neighbors.register(new HC_AddressPair(this, 
								new HC_LogicalAddress(tear_index, true), X.getSrcPA(),
								false));
							neighbors.updateHRoot(new HC_LogicalAddress(tear_index, true),
										neighbors.getHRootSeqNum() + 1);


							if (MyState == HC_Node.HRoot) {

								MyState = HC_Node.Stable;
								
								//Report event ISSTABLE to notification handler.
								if (n_handler != null) {
								   NODE_ISSTABLE ne = new NODE_ISSTABLE(Adapter.getCurrentTime(), null);
								   n_handler.eventOccurred(ne);
								}
								
							}
							if (MyState == HC_Node.HRootMissingNeighbor) {

								MyState = HC_Node.MissingNeighbor;
							}
							if (MyState == HC_Node.HRootMissingNeighborAction) {

								MyState = HC_Node.MissingNeighborAction;
							}
							break SkipItReceiver;
						}
						else if (X.getSrcLA().getColexIndex() > tear_index) {
//config.log.println("Lowering node's logical address to top of hypercube.\n");
							neighbors.register(new HC_AddressPair(this, 
								new HC_LogicalAddress(tear_index, true), X.getSrcPA(),
								false));
							neighbors.updateHRoot(new HC_LogicalAddress(tear_index, true),
										X.getHRootSeqNum() + 1);


							if (MyState == HC_Node.HRoot) {

								MyState = HC_Node.Stable;
								//Report event ISSTABLE to notification handler.
								if (n_handler != null) {
								   NODE_ISSTABLE ne = new NODE_ISSTABLE(Adapter.getCurrentTime(), null);
								   n_handler.eventOccurred(ne);
								}
								
							}
							if (MyState == HC_Node.HRootMissingNeighbor) {

								MyState = HC_Node.MissingNeighbor;
							}
							if (MyState == HC_Node.HRootMissingNeighborAction) {

								MyState = HC_Node.MissingNeighborAction;
							}
							break SkipItReceiver;
						}
					}
				break;

				case HC_Message.Kill:
					if (X.getDestLA().equals(MyLogicalAddress)) {

//config.log.println("Resetting due to Kill packet.\n");
						leaveNeighborhood();
						resetNeighborhood();
						MyState = HC_Node.Leaving;
						neighbors.MNTimeOut = 0;
						break SkipItReceiver;
					}
				break;

				case HC_Message.Leave:
					if (X.getDestLA().equals(MyLogicalAddress)) {
//config.log.println("Removing leaving node from neighborhood.\n");
						neighbors.remove(new HC_AddressPair(this, X.getSrcLA(), X.getSrcPA(), false));
					}
				break;
				}
			break;
        } // switch
    }
    




	/** 
	 * This function processes a timer event <p>
	 * This is also the kernel part of the Hypercast protocol. 
	 * The HC_Node changes it's state, pings it's neighborhod everytime there is a timer event comes in.
	 * 
	 * @param	Timer_ID		time event ID
	 */
     public synchronized void timerExpired(Object Timer_ID) {

    	int timerID = ((Integer) Timer_ID).intValue();

        if (timerID != 0)
            throw new IllegalArgumentException("Unexpected timer id "+timerID+" in HC_Node.timerExpired");

        // reset time to go off again.
        Adapter.setTimer(new Integer(0), SleepTime);

        byte MyOldState = MyState;

		int j;

SkipItPinger:	switch (MyState) {
				case HC_Node.Joining:
//config.log.println("Pinger is in Joining state.\n");
					j = neighbors.ageNeighborhood();
					// 1. Send a beacon
					multicastSend(HC_Message.Beacon);

					// 3. Check for emptiness
					if (j == 1)
						if ((neighbors.HRoot == null) || (neighbors.HRoot.getAge() <= 1)) {
//config.log.println("No neighbors found to join with.\n");
						resetNeighborhood();
//config.log.println("Switching to StartHypercube.\n");
						MyState = HC_Node.StartHypercube;
						}
				break;

				case HC_Node.JoiningWait:
//config.log.println("Pinger is in JoiningWait state.\n");
					neighbors.ageNeighborhood();

					neighbors.MNTimeOut--;
					if (neighbors.MNTimeOut <= 0) {
//config.log.println("No joining beacons detected.\n");
//config.log.println("Switching to Joining.\n");
						MyState = HC_Node.Joining;
					}
				break;

				case HC_Node.StartHypercube:
//config.log.println("Pinger is in StartHypercube state.\n");
					MyLogicalAddress = new HC_LogicalAddress(0, true);

					//Report event LACHANGED to notification handler.
					if (n_handler != null) {
						NODE_LOGICALADDRESSCHANGED ne = new NODE_LOGICALADDRESSCHANGED(Adapter.getCurrentTime(), null);
						n_handler.eventOccurred(ne);
					}
					
					neighbors.HRoot = new HC_AddressPair(this, new HC_LogicalAddress(0, true),
										MyPhysicalAddress, true);
					neighbors.setHRootSeqNum(1);
//config.log.println("Information set.\n");
//config.log.println("Switching to HRoot.\n");
					MyState = HC_Node.HRoot;
				break;

				case HC_Node.Leaving:
//config.log.println("Pinger is in Leaving state.\n");
					MyState = HC_Node.Joining;
				break;

				case HC_Node.Stable:
				case HC_Node.MissingNeighbor:
				case HC_Node.MissingNeighborAction:
				case HC_Node.HRoot:
				case HC_Node.HRootMissingNeighbor:
				case HC_Node.HRootMissingNeighborAction:

					j = neighbors.ageNeighborhood();

					// 3. Check for emptiness
					if ((j == 1) && !((MyLogicalAddress.getColexIndex() == 0) && (MyState == HC_Node.HRoot))) {
//config.log.println("No neighbors found anymore.\n");
						resetNeighborhood();
//config.log.println("Switching to StartHypercube.\n");
						MyState = HC_Node.StartHypercube;
					}

					if ((MyState == HC_Node.HRoot) ||
					(MyState == HC_Node.HRootMissingNeighbor) ||
					(MyState == HC_Node.HRootMissingNeighborAction)) {
//config.log.println("Sending Beacon as HRoot.\n");
						neighbors.HRootSeqNum++;
						if (neighbors.HRoot != null)
							neighbors.HRoot.setAge(MaxAge);
						multicastSend(HC_Message.Beacon);
					}
					if ((MyState == HC_Node.MissingNeighbor) ||
					(MyState == HC_Node.MissingNeighborAction)) {
//config.log.println("Sending Beacon as MissingNeighbor.\n");
						multicastSend(HC_Message.Beacon);
					}

					if ((MyState == HC_Node.HRootMissingNeighbor) ||
					(MyState == HC_Node.MissingNeighbor)) {
						neighbors.MNTimeOut--;
						int tear_index = neighbors.searchTear();
//						if (neighbors.MNTimeOut <= 0) {
						if (tear_index != MaxAddress()) {
//config.log.println("Switching to action mode.\n");
							if (MyState == HC_Node.MissingNeighbor)
								MyState = HC_Node.MissingNeighborAction;
							if (MyState == HC_Node.HRootMissingNeighbor)
								MyState = HC_Node.HRootMissingNeighborAction;
						}
					}

					pingNeighbors();

					int tear_index = neighbors.searchLittleTear();

					if (tear_index != MaxAddress()) {
//config.log.println("Found a hole in neighborhood list.\n");

						if (MyState == HC_Node.HRoot) {
//config.log.println("Switching to HRootMissingNeighbor.\n");
							MyState = HC_Node.HRootMissingNeighbor;
							neighbors.MNTimeOut = MaxMissingNeighbor;
							break SkipItPinger;
						}
						if (MyState == HC_Node.Stable) {
//config.log.println("Switching to MissingNeighbor.\n");
							MyState = HC_Node.MissingNeighbor;
							neighbors.MNTimeOut = MaxMissingNeighbor;
							break SkipItPinger;
						}
					}
					else {
//config.log.println("No holes in neighborhood list found.\n");
						if ((MyState == HC_Node.HRootMissingNeighbor) ||
						(MyState == HC_Node.HRootMissingNeighborAction)) {
//config.log.println("Switching to HRoot.\n");
							MyState = HC_Node.HRoot;
							neighbors.MNTimeOut = 0;
							break SkipItPinger;
						}
						if ((MyState == HC_Node.MissingNeighbor) ||
						(MyState == HC_Node.MissingNeighborAction)) {
							MyState = HC_Node.Stable;

							//Report event ISSTABLE to notification handler.
							if (n_handler != null) {
							   NODE_ISSTABLE ne = new NODE_ISSTABLE(Adapter.getCurrentTime(), null);
							   n_handler.eventOccurred(ne);
							}
								
							neighbors.MNTimeOut = 0;
							break SkipItPinger;
						}
					}

					int neighbor_index = 0;
					for (int i = 0; i < MaxHypercubeDim; i++)
						if (neighbors.getNthNeighbor(i) != null)  // WORK?
							if (!neighbors.getNthNeighbor(i).getLogicalAddress().
							equals(new HC_LogicalAddress(MaxAddress(), true)))
								neighbor_index = ((HC_LogicalAddress)neighbors.getNthNeighbor(i).
										getLogicalAddress()).getColexIndex();

					if ((neighbor_index < MyLogicalAddress.getColexIndex()) &&
					(neighbors.HRoot == null)) {
//config.log.println("In absence of known HRoot, discovered that LA is highest.\n");

						neighbors.HRoot = new HC_AddressPair(this, MyLogicalAddress,
										MyPhysicalAddress, true);
//						neighbors.HRootSeqNum++;
						neighbors.HRootSeqNum = 0;
						neighbors.HRoot.setAge(MaxAge);

						if (MyState == HC_Node.MissingNeighbor) {
//config.log.println("Switching to HRootMissingNeighbor.\n");
							MyState = HC_Node.HRootMissingNeighbor;
							break SkipItPinger;
						}
						if (MyState == HC_Node.Stable) {
//config.log.println("Switching to HRoot.\n");
							MyState = HC_Node.HRoot;
							break SkipItPinger;
						}
						if (MyState == HC_Node.MissingNeighborAction) {
//config.log.println("Switching to HRootMissingNeighborAction.\n");
							MyState = HC_Node.HRootMissingNeighborAction;
							break SkipItPinger;
						}
					}
					int temp = -1;
					if (neighbors.HRoot != null)
						temp = ((HC_LogicalAddress)neighbors.HRoot.getLogicalAddress()).getColexIndex();
					if ((neighbor_index > MyLogicalAddress.getColexIndex()) ||
					(temp > MyLogicalAddress.getColexIndex())) {
//config.log.println("A neighbor or HRoot with higher ranking has been found.\n");
						if (MyState == HC_Node.HRootMissingNeighbor) {
//config.log.println("Switching to MissingNeighbor.\n");
							MyState = HC_Node.MissingNeighbor;
							break SkipItPinger;
						}
						if (MyState == HC_Node.HRoot) {

//config.log.println("Switching to Stable.\n");
							MyState = HC_Node.Stable;

							//Report event ISSTABLE to notification handler.
							if (n_handler != null) {
							   NODE_ISSTABLE ne = new NODE_ISSTABLE(Adapter.getCurrentTime(), null);
							   n_handler.eventOccurred(ne);
							}
								
							break SkipItPinger;
						}
						if (MyState == HC_Node.HRootMissingNeighborAction) {
//config.log.println("Switching to MissingNeighborAction.\n");
							MyState = HC_Node.MissingNeighborAction;
							break SkipItPinger;
						}
					}

					if (neighbors.HRoot != null)
						if ((MyLogicalAddress.getColexIndex() >=
						((HC_LogicalAddress)neighbors.HRoot.getLogicalAddress()).getColexIndex()) &&
						(neighbor_index < MyLogicalAddress.getColexIndex())) {
//config.log.println("Discovered that LA is equal or higher than known HRoot and all neighbors.\n");

							neighbors.HRoot = new HC_AddressPair(this, MyLogicalAddress,
											MyPhysicalAddress,
											true);

							neighbors.HRootSeqNum++;
//							neighbors.HRootSeqNum = 0;
							neighbors.HRoot.setAge(MaxAge);

							if (MyState == HC_Node.MissingNeighbor) {
//config.log.println("Switching to HRootMissingNeighbor.\n");
								MyState = HC_Node.HRootMissingNeighbor;
								break SkipItPinger;
							}
							if (MyState == HC_Node.Stable) {
//config.log.println("Switching to HRoot.\n");
								MyState = HC_Node.HRoot;
								break SkipItPinger;
							}
							if (MyState == HC_Node.MissingNeighborAction) {
//config.log.println("Switching to HRootMissingNeighborAction.\n");
								MyState = HC_Node.HRootMissingNeighborAction;
								break SkipItPinger;
							}
						}
				break;
          } // switch
    }

	
	
/**
 * Sends Hypercast Messages to Unicast address through the Adapter
 * @param	type	Message type
 * @param	paddr	I_PhysicalAddress of destination
 * @param	laddr	HC_LogicalAddress of destination
 */
private void unicastSend(byte type, I_PhysicalAddress paddr, HC_LogicalAddress laddr) {

    if ((type != HC_Message.Kill) &&
        (type != HC_Message.Ping) &&
        (type != HC_Message.Leave)) {
        throw new IllegalArgumentException("HC_Node: unicastSend: Type " + type + "unrecognized. ");
    }

    byte[] data = null;

    HC_Message newpacket = null;
    if (neighbors.getHRoot() != null) {

		newpacket = createHCMessage(type, paddr, laddr, (HC_LogicalAddress)(neighbors.getHRoot().getLogicalAddress()), 
				neighbors.getHRootSeqNum(), data);
    }
    else {
		newpacket = createHCMessage(type, paddr, laddr, new HC_LogicalAddress(0, true), 0, data);
    }

    Adapter.sendUnicastMessage(paddr, newpacket);
}



/**
 * Sends Hypercast Messages to Multicast address through the Adapter
 * @param	type	Message type
 */
private void multicastSend(byte type) {
    if (type != HC_Message.Beacon) {
        throw new IllegalArgumentException("HC_Node: multicastSend: type is not Beacon!");
    }

    byte[] data = null;
    HC_Message newpacket = null;
    if ((neighbors.getHRoot() != null) && (MyState != HC_Node.Joining)) {
    	/* Multicast address might have different length than normal physical address
    	 * Thus it's dangerous to use a multicase address as an physical address
    	 */
    	/*
    	newpacket = createHCMessage(HC_Message.Beacon, Adapter.getMulticastAddress(), MyLogicalAddress, 
				(HC_LogicalAddress)(neighbors.getHRoot().getLogicalAddress()), neighbors.getHRootSeqNum(), data);
				*/
    	newpacket = createHCMessage(HC_Message.Beacon, MyPhysicalAddress, MyLogicalAddress, 
				(HC_LogicalAddress)(neighbors.getHRoot().getLogicalAddress()), neighbors.getHRootSeqNum(), data);
    }
    else {
    	/* Multicast address might have different length than normal physical address
    	 * Thus it's dangerous to use a multicase address as an physical address
    	 */
    	/*
		newpacket = createHCMessage(HC_Message.Beacon, Adapter.getMulticastAddress(), MyLogicalAddress, 
				new HC_LogicalAddress(0, true), 0, data);
				*/
    	newpacket = createHCMessage(HC_Message.Beacon, MyPhysicalAddress, MyLogicalAddress, 
				new HC_LogicalAddress(0, true), 0, data);
    }
    
    Adapter.sendMulticastMessage(newpacket);
}

/** This method creates a new DT_Message based on parameters. 
 * It is called by all sending methods and can be overrided in secure node class.
 */
public HC_Message createHCMessage (byte type, I_PhysicalAddress paddr, HC_LogicalAddress laddr, HC_LogicalAddress root, int seq, byte[] data) {
	return new HC_Message(Proto_Sub,
						type,
						config.getOverlayHash(),
						MyPhysicalAddress, MyLogicalAddress,
						paddr, laddr,
						root, seq,
						data, 
			Adapter);
}



/** 
 * Sends ping message to all neighbors
 */
private void pingNeighbors() {
	//config.log.println("Pinging all neighbors.\n");
	for (int i = 0; i < MaxHypercubeDim; i++) {
		HC_AddressPair X = neighbors.getNthNeighbor(i);

		if ((X != null) && (X.getPhysicalAddress() != null)) {		   
			unicastSend(HC_Message.Ping, 
				X.getPhysicalAddress(), 
				(HC_LogicalAddress)(X.getLogicalAddress()));
		}
	}
}



/** 
 * Resets all neighbors
 */
private void resetNeighborhood() {
	//config.log.println("Reset Neighborhood called.\n");
	neighbors.resetNeighborhood();
	MyLogicalAddress = new HC_LogicalAddress(MaxAddress(), true);

	//Report events LACHANGED and NEIGHBORHOODCHANGED to notification handler.
	if (n_handler != null) {
		NODE_LOGICALADDRESSCHANGED ne = new NODE_LOGICALADDRESSCHANGED(Adapter.getCurrentTime(), null);
		n_handler.eventOccurred(ne);
		NODE_NEIGHBORHOODCHANGED nee = new NODE_NEIGHBORHOODCHANGED(Adapter.getCurrentTime(), null);
		n_handler.eventOccurred(nee);
	}
					
}



/** 
 * Sends leave message to all neighbors
 */
private void leaveNeighborhood() {
	//config.log.println("Leave Neighborhood called.\n");
	for (int i = 0; i < MaxHypercubeDim; i++) {
		if (neighbors.getNthNeighbor(i) != null)
			if ((neighbors.getNthNeighbor(i).getAge() > 0) &&
				(neighbors.getNthNeighbor(i).getPhysicalAddress() != null)) {
				unicastSend(HC_Message.Leave,
						neighbors.getNthNeighbor(i).getPhysicalAddress(),
						(HC_LogicalAddress)(neighbors.getNthNeighbor(i).getLogicalAddress()));

			}
	}
}


  /**
   * Gets the AddressPair of this node
   * @return	an HC_AddressPair object
   */
	public I_AddressPair getAddressPair() {
        //if (!joined)
        //    throw new IllegalStateException("Must joinGroup before calling getMyAddress!");

		return new HC_AddressPair(this, MyLogicalAddress, MyPhysicalAddress, false);
	}    

	
  /**
   * Gets the AddressPair of the HRoot
   * @return	HC_AddressPair object of the HRoot
   */
	public synchronized HC_AddressPair getHRoot() {
        if (!joined)
            throw new IllegalStateException("Must joinGroup before calling getSucc!");

		return neighbors.getHRoot();
	}


	
  /**
   * Gets the AddressPair of the Sueecssor of this node in the Hypercube 
   * @return	HC_AddressPair object of the Successor
   */
	public synchronized HC_AddressPair getSucc() {
        if (!joined)
            throw new IllegalStateException("Must joinGroup before calling getSucc!");

		int myColex = MyLogicalAddress.getColexIndex();
		if ((myColex + 1) <= ((HC_LogicalAddress)neighbors.getHRoot().getLogicalAddress()).getColexIndex()) 
			return neighbors.getNeighborPair(new HC_LogicalAddress((myColex + 1), true));
		else 
			return null; 
	}

	
	
  /**
   * Gets the AddressPair of the Predecessesor of this node in the Hypercube 
   * @return	HC_AddressPair object of the Pressesor
   */
	public synchronized HC_AddressPair getPred() {
        if (!joined)
            throw new IllegalStateException("Must joinGroup before calling getPred!");

		int myColex = MyLogicalAddress.getColexIndex();
		if ((myColex-1) >= 0)
			return neighbors.getNeighborPair(new HC_LogicalAddress((myColex - 1), true));
		else 
			return null;
	}
	

	
//When parent is not found, the list of neighbors is returned as an array. 
//So return type of this method is unified to a array of I_AddressPair.
  /**
   * Gets the AddressPair[] of this node's Children with reference to a given Root.
   * If the hroot != null and children != null, return children's addresspairs. 
   * else, return null.
   * @param		root	HC_LogicalAddress object of a Root
   * @return	HC_AddressPair object of the Parent
   */
	//public synchronized I_AddressPair getParent(I_LogicalAddress root) {
	public synchronized I_AddressPair[] getParent(I_LogicalAddress root) {
        if (!joined)
            throw new IllegalStateException("Must joinGroup before calling getParent!");

		//return neighbors.getNeighborPair(MyLogicalAddress.parent((HC_LogicalAddress)root));
		I_AddressPair Pa = neighbors.getNeighborPair(MyLogicalAddress.parent((HC_LogicalAddress)root));
		if (Pa == null){
			return neighbors.getNeighbors();
		}
		else {
		  I_AddressPair[] myParent =  new I_AddressPair[1];
		  myParent[0] = Pa;
		  return myParent;
		}
	}

	
  /**
   * Gets the AddressPair[] of this node's Children with reference to a given Root
   * @param		root HC_LogicalAddress object of a Root
   * @return	HC_AddressPair array of the Children
   */
	public synchronized I_AddressPair[] getChildren(I_LogicalAddress root) {
        if (!joined)
            throw new IllegalStateException("Must joinGroup before calling getChildren!");
		
		HC_LogicalAddress hroot;
		HC_AddressPair hrootpair = null;
		
		if (neighbors != null) {	
			hrootpair = neighbors.getHRoot();
			
			if (hrootpair == null) {
				hroot = null;
			}
			else {
				hroot = (HC_LogicalAddress)(hrootpair.getLogicalAddress());
			}
		}
		else {
			hroot = null;
		}
		
		if (hroot != null) {
			HC_LogicalAddress[] childrenLAs =  MyLogicalAddress.children((HC_LogicalAddress)root, hroot);
			
			if (childrenLAs.length > 0) {
			
				HC_AddressPair[] childrenLAPAPairs = new HC_AddressPair[childrenLAs.length];
				for (int i = 0; i < childrenLAs.length; i++) {
					childrenLAPAPairs[i] = neighbors.getNeighborPair(childrenLAs[i]);
				}
				return childrenLAPAPairs;
			}
			else 
				return null;
		}	
		else
			return null;
	
	}
			

  /**
   * Gets the AddressPair[] of this node's all neighbors
   * @return	HC_AddressPair array of all the neighbors
   */	
	public synchronized I_AddressPair[] getAllNeighbors() {
        if (!joined)
            throw new IllegalStateException("Must joinGroup before calling getAllNeighbors!");
		return neighbors.getNeighbors();
	}
		
  /**
   * Gets the node's current state
   * @return	the Node's current state
   */
	public byte getState() {
		return MyState;
		
	}

	/**
	 * Initialize the data structure for handling the statistics defined in this object.
	 * In this method, a StatsProcessor instance is created which contains all statistics
	 * , with each represented by a StatsElement instance, supported in this object.
	 */
	private void initStatisticsStructure()
	{
		/*
		 * Get the statistics name defined in the configuration file for this I_Stats instance.
		 */
		statisticsName = config.getTextAttribute (XmlUtil.createXPath(PROPERTY_PROTO_PREFIX+"StatName"));

		/*
		 * Initial StatsProcessor instance.
		 */
		statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);
		
		statsPro.addStatsElement("NodeAdapter"/* name */, Adapter/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);

		statsPro.addStatsElement("State", new NodeState() , 1, 1);

		statsPro.addStatsElement("LogicalAddress", new LogicalAddressOperator(), 1, 1);

		statsPro.addStatsElement("PhysicalAddress", new PhysicalAddressOperator(), 1, 1);

		statsPro.addStatsElement("HRoot", new HRootQuery(), 1, 1);

		statsPro.addStatsElement("HRootSeqNum", new HRootSequenceNumber(), 1, 1);

		statsPro.addStatsElement("HeartBeat", new HeartBeatPeriod(), 1, 1);

		statsPro.addStatsElement("NumOfNeighbors", new NumOfCurrentNeighbors(), 1, 1);
	}


	/**
	 * This class manages simple statistic "State".
	 */
	class NodeState extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+MyState);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "LogicalAddress".
	 */
	class LogicalAddressOperator extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+MyLogicalAddress.getColexIndex());
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:simpleType"/* type */, 
												"xsd:String"/* restriction base */, "\\d+"/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "PhysicalAddress".
	 */
	class PhysicalAddressOperator extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+MyPhysicalAddress.toString());
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}


	/**
	 * This class manages simple statistic "HRoot".
	 */
	class HRootQuery extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			if (neighbors.HRoot != null)
			{
				return new String(""+neighbors.HRoot.toString());
			}
			else
			{
				return null;
			}
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}


	/**
	 * This class manages simple statistic "HRootSeqNum".
	 */
	class HRootSequenceNumber extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+neighbors.HRootSeqNum);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}


	/**
	 * This class manages simple statistic "HeartBeat".
	 */
	class HeartBeatPeriod extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+SleepTime);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			try
			{
				SleepTime = Integer.parseInt(newValue);
			}
			catch (NumberFormatException e)
			{
				throw new HyperCastStatsException(statisticsName, HyperCastStatsException.InappropriateValue);
			}

			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}


	/**
	 * This class manages simple statistic "NumOfNeighbors".
	 */
	class NumOfCurrentNeighbors extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+neighbors.getNumOfNeighbors());
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}


	/** @see I_Stats#getStatsName() */
	public String getStatsName ()
	{
		return statisticsName;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName (String name)
	{
		statisticsName = name;
	}

	/**
	 * Gets the information of a scalar statistics from an object which 
	 * implements this interface.
	 * @throws HyperCastStatsException if the statistics with given name is not
	 * supported by the object.
	 */
/*
	public String getStatsValue(String name) throws HyperCastStatsException
	{
		if(name.equals("State"))
			return new String(""+MyState);
		else if(name.equals("LogicalAddress"))
			return new String(""+MyLogicalAddress.getColexIndex());
		else if(name.equals("PhysicalAddress"))
			return new String(""+MyPhysicalAddress.toString());
		else if(name.equals("HRoot"))
		{
			if (neighbors.HRoot != null)
			{
				return new String(""+neighbors.HRoot.toString());
			}
			else
			{
				return null;
			}
		}
		else if(name.equals("HRootSeqNum"))
			return new String(""+neighbors.HRootSeqNum);
		else if(name.equals("HeartBeat"))
			return new String(""+SleepTime);
		else if(name.equals("NumOfNeighbors"))
		{
			return new String(""+neighbors.getNumOfNeighbors());
		}
		else 
			throw new HyperCastStatsException(name, HyperCastStatsException.StatisticNotFound);
	}
*/

	/**
	 * Sets the information of s writable scalar statistics to an object which 
	 * implements this interface.
	 * @throws HyperCastStatsException if the statistics with given name is not
	 * supported by the object or read-only, or if value does not conform with 
	 * expected format.
	 */
/*
	public String setStatsValue(String name, String newValue) throws HyperCastStatsException
	{
		if(name.equals("State") || name.equals("LogicalAddress") || name.equals("HRoot") ||
		   name.equals("HRootSeqNum") || 	name.startsWith("Neighbor") || 
		   name.equals("NumOfNeighbors")) 
		{
			throw new HyperCastStatsException(name, HyperCastStatsException.WriteToAReadOnlyValue);
		}
		else if(name.equals("HeartBeat"))
		{
			try
			{
				SleepTime = Integer.parseInt(newValue);
			}
			catch (NumberFormatException e)
			{
				throw new HyperCastStatsException(name, HyperCastStatsException.InappropriateValue);
			}
		}
		else
			throw new HyperCastStatsException(name, HyperCastStatsException.StatisticNotFound);

		return newValue;
	}
*/

	/** 
	 * Return the result of query for the statistics specified by the given xpath.
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance represents the statistics to be queried.
	 * @see I_Stats#getStats
	 */
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//System.out.println("HC_Node:getStats: xpath string is "+xpath.getPatternString());
		//System.out.println("HC_Node:getStats: call statsPro.getStatsResult(doc, xpath)");
		return statsPro.getStatsResult(doc, xpath);
	}

	/** 
	 * Set the statistics specified by the given xpath. The value actually set is
	 * returned.
	 * 
	 * @param doc   	Document used for create new elements or nodes.
	 * @param xpath 	XPath instance represents the statistics to be queried.
	 * @param newValue	Element representing the value or sub-tree to be set.
	 * @see I_Stats#setStats
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	{
		//System.out.println("HC_Node:setStats: xpath string is "+xpath.getPatternString());
		//System.out.println("HC_Node:setStats: call statsPro.setStatsResult(doc, xpath, newValue)");
		return statsPro.setStatsResult(doc, xpath, newValue);
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in read schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//System.out.println("HC_Node:getReadSchema: xpath string is "+xpath.getPatternString());
		//System.out.println("HC_Node:getSchema: call statsPro.getReadSchemaResult(doc, xpath)");
		return statsPro.getReadSchemaResult(doc, xpath);
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in write schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getWriteSchema(Document, XPath)
	 */
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return statsPro.getWriteSchemaResult(doc, xpath);
	}
	/**
	 * Gets neighbor's physical address or logical address
	 * @return	index		the index of the neighbor
	 * @param	attribute	PA or LA
	 * 
	 * @return	the string format of PA or LA
	 */
	public String getNeighborAttribute(int index, String attribute) {
		if(attribute.equals("PA"))
			return getPAAddress(neighbors.NView[index].toString());
		else if(attribute.equals("LA"))
			return getLAAddress(neighbors.NView[index].toString());
		else
			return "";
	}
	
	/**
	 * Gets physical address from a node's AddressPair.
	 */
	public String getPAAddress(String NodePALAPair) {
		System.out.println("NodePALAPair - "+NodePALAPair);
		int colonIndx = NodePALAPair.indexOf(":",1);
		System.out.println(":index = "+colonIndx);
		return NodePALAPair.substring(0,colonIndx);
	}

	/**
	 * Gets logical address from a node's AddressPair
	 */
	public String getLAAddress(String NodePALAPair) {
		 System.out.println("NodePALAPair - "+NodePALAPair);
		int colonIndx = NodePALAPair.indexOf(":");
		int nextColonIndx = NodePALAPair.indexOf(":",colonIndx+1);
		System.out.println("colonindex = "+colonIndx+" next - "+nextColonIndx);
		return NodePALAPair.substring(colonIndx+1, nextColonIndx);
	}
}
