/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.MonitorAndControl;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import hypercast.util.XmlUtil;

import hypercast.HyperCastFatalRuntimeException;
import hypercast.I_LogicalAddress;

/**
 * This class implements a basic Monitor/Control message exchanged
 * between Monitor and Portal. It manages message meta-information
 * such as sender and destination information, timestamp, message
 * id. Each MCRawMessage is encoded as an XML docuemnt. This XML
 * document is managed by this class, however the content of the XML
 * document, save the meta-information, is not handled at this
 * level. Classes that are derived from this class, e.g. MonMessage,
 * define the content of the XML document.
 *
 * NB: *NEVER* call setAttribute with a <code>null</code> second
 * argument. Some versions of XML libraries do not handle this very
 * well and cause a <code>NullPointerException</code> to be thrown
 * HyperCast when <code>XmlUtil.writeXml()</code> is called.
 * 
 * @author HyperCast Team
 * @author Guangyu Dong, January 2005
 * @author Greg Mattes, January 2005
 */
/* package */ abstract class MCRawMessage
                  implements Cloneable
{
    /** attribute names of the XML root */
    private static final String ATTR_SENDER_STR    = "Sender";
    private static final String ATTR_DEST_STR      = "Dest";
    private static final String ATTR_MSGID_STR     = "MsgID";
    private static final String ATTR_TIMESTAMP_STR = "TimeStamp";
    private static final String ATTR_CBINDEX_STR   = "CBIndex";

    /** Destination string used when a message is not unicast. */
    protected static final I_LogicalAddress MULTICAST_DESTINATION = null;

    /** AdvertisePortal and AdvertiseMonitor messages are not assigned unique message IDs */
    protected static final int ADVERTISE_MESSAGE_ID = 0;

    protected static final int INITIAL_UNIQUE_MESSAGE_ID = 1;

    protected static final int MESSAGE_ID_NOT_SET = -1;

    private static final long TIMESTAMP_NOT_SET = -1;

    protected long messageID = MESSAGE_ID_NOT_SET;

    /** attributes of the XML root */
    protected I_LogicalAddress sender = null;

    /**
     * The destination logical address of this message, stored as a
     * string. It is set to multicast by default, which can be
     * interpreted as "no particular unicast address has been set."
     */
    protected I_LogicalAddress destination = MULTICAST_DESTINATION;
    public final static String MULTICAST_DESTINATION_STRING = "All";

    /**
     * Set at message send time.
     *
     * @see MCSocket#sendTo
     */
    protected long timeStamp = TIMESTAMP_NOT_SET;
	
    /**
     * XML root element
     * Generated lazily
     *
     * MUST BE KEPT IN SYNC WITH doc
     *
     * @see #createDocument
     */
    protected Element messageRoot = null;

    /**
     * The XML document managed by this class. For a new message
     * (i.e. not a message that is received from the network) it is
     * generated lazily instead of at construction
     *
     * MUST BE KEPT IN SYNC WITH messageRoot
     *
     * @see #createDocument
     */
    protected Document doc = null;

    protected int type;

    /* Valid values for type */
    protected static final int ADVERTISE_MONITOR  = 0;
    protected static final int ADVERTISE_PORTAL   = 1;
    protected static final int GET_QUERY          = 2;
    protected static final int GET_REPLY          = 3;
    protected static final int READ_SCHEMA_QUERY  = 4;
    protected static final int READ_SCHEMA_REPLY  = 5;
    protected static final int SET_QUERY          = 6;
    protected static final int SET_REPLY          = 7;
    protected static final int WRITE_SCHEMA_QUERY = 8;
    protected static final int WRITE_SCHEMA_REPLY = 9;

    protected static final String ADVERTISE_MONITOR_STRING  = "AdvertiseMonitor";
    protected static final String ADVERTISE_PORTAL_STRING   = "AdvertisePortal";
    protected static final String GET_QUERY_STRING          = "GetRequest";
    protected static final String GET_REPLY_STRING          = "GetReply";
    protected static final String READ_SCHEMA_QUERY_STRING  = "ReadSchemaRequest";
    protected static final String READ_SCHEMA_REPLY_STRING  = "ReadSchemaReply";
    protected static final String SET_QUERY_STRING          = "SetRequest";
    protected static final String SET_REPLY_STRING          = "SetReply";
    protected static final String WRITE_SCHEMA_QUERY_STRING = "WriteSchemaRequest";
    protected static final String WRITE_SCHEMA_REPLY_STRING = "WriteSchemaReply";
	
    protected static String [] intToTypeString =
    {
        ADVERTISE_MONITOR_STRING,
        ADVERTISE_PORTAL_STRING,
        GET_QUERY_STRING,
        GET_REPLY_STRING,
        READ_SCHEMA_QUERY_STRING,
        READ_SCHEMA_REPLY_STRING,
        SET_QUERY_STRING,
        SET_REPLY_STRING,
        WRITE_SCHEMA_QUERY_STRING,
        WRITE_SCHEMA_REPLY_STRING,
    };

    protected static int typeStringToInt (final String typeString)
    {
        if      (typeString.equals (ADVERTISE_MONITOR_STRING))  return ADVERTISE_MONITOR;
        else if (typeString.equals (ADVERTISE_PORTAL_STRING))   return ADVERTISE_PORTAL;
        else if (typeString.equals (GET_QUERY_STRING))          return GET_QUERY;
        else if (typeString.equals (GET_REPLY_STRING))          return GET_REPLY;
        else if (typeString.equals (READ_SCHEMA_QUERY_STRING))  return READ_SCHEMA_QUERY;
        else if (typeString.equals (READ_SCHEMA_REPLY_STRING))  return READ_SCHEMA_REPLY;
        else if (typeString.equals (SET_QUERY_STRING))          return SET_QUERY;
        else if (typeString.equals (SET_REPLY_STRING))          return SET_REPLY;
        else if (typeString.equals (WRITE_SCHEMA_QUERY_STRING)) return WRITE_SCHEMA_QUERY;
        else if (typeString.equals (WRITE_SCHEMA_REPLY_STRING)) return WRITE_SCHEMA_REPLY;
        else
            throw new IllegalArgumentException ("Illegal MC message type: " + typeString);
    }

    public MCRawMessage (final int type, final I_LogicalAddress sender, final long messageID)
    {
        if (type < 0
            ||
            type >= intToTypeString.length)
            throw new IllegalArgumentException ("Illegal MC message type: " + type);

        this.type          = type;
        this.sender        = sender;
        this.messageID     = messageID;
    }

    /**
     * Construct a MonMessage from a XML document
     * @param doc
     */
    public MCRawMessage (MCSocket sock, final Document doc)
    {
        this.doc           = doc;
        this.messageRoot   = (Element) doc.getChildNodes().item(0);
        this.type          = typeStringToInt (messageRoot.getNodeName());
        this.sender        = sock.createLogicalAddress(messageRoot.getAttribute (ATTR_SENDER_STR));
        if(messageRoot.getAttribute (ATTR_DEST_STR).equals(MULTICAST_DESTINATION_STRING)){
        	this.destination = null;
        }else
        	this.destination   = sock.createLogicalAddress(messageRoot.getAttribute (ATTR_DEST_STR));
        this.messageID     = Long.parseLong (messageRoot.getAttribute (ATTR_MSGID_STR));
        this.timeStamp     = Long.parseLong (messageRoot.getAttribute (ATTR_TIMESTAMP_STR));
        
    }

    /** restore MonMessage from XML document*/
    static MCRawMessage restoreMessage (MCSocket sock, final Document document)
    {
        final Element documentRoot = (Element) document.getChildNodes().item (0);
        final String messageTypeString = documentRoot.getNodeName();
        final int messageType = typeStringToInt (messageTypeString);

        switch (messageType)
        {
            case ADVERTISE_MONITOR:
            case ADVERTISE_PORTAL:
                return new MCAdvertiseMessage (sock, document);

            case GET_QUERY:
            case GET_REPLY:
            case READ_SCHEMA_QUERY:
            case READ_SCHEMA_REPLY:
            case SET_QUERY:
            case SET_REPLY:
            case WRITE_SCHEMA_QUERY:
            case WRITE_SCHEMA_REPLY:
                return new MonMessage (sock, document);

            default:
                throw new IllegalArgumentException ("Illegal MC message type: " + messageTypeString + " (" + messageType + ")");
        }
    }

    /**
     * Create a XML representation of this message.
     * Subclass should implement this method to
     * add contents to this message.
     */
    protected abstract void createMessageContents();
	
    /**
     * Return the XML document representation of this M&C message.
     * 
     * An M&C message is transmitted on the network in XML format. So every M&C can
     * be converted to a XML document.
     * 
     * Note that this method returns a reference to an internal data
     * structure - use with care!
     *
     * @return the XML document representing this M&C message
     */
    public Document getDocument()
    {
        if (doc == null)
        {
            createDocument();
        }
		
        return doc;
    }
	
    /**
     * create an XML document
     */
    private void createDocument()
    {
        doc = XmlUtil.createDocument();
		
        messageRoot = doc.createElement (intToTypeString [type]);

        messageRoot.setAttribute (ATTR_SENDER_STR,    sender.toString());
        messageRoot.setAttribute (ATTR_DEST_STR,      getDestString());
        messageRoot.setAttribute (ATTR_MSGID_STR,     Long.toString (messageID));
        messageRoot.setAttribute (ATTR_TIMESTAMP_STR, Long.toString (timeStamp));
        
        doc.appendChild (messageRoot);

        createMessageContents();
    }

    /**
     * Get the logical address string of the destination of this message.
     * If the message is supposed to be sent to all portals in the network, this
     * field is set to "".
     * 
     * @return Returns the destination.
     */
    public I_LogicalAddress getDestination()
    {
        return destination;
    }

    /**
     * Set the logical address string of the destination of this message
     * This field is usually set by Monitor or Portal class automatically before
     * the message being sent out.
     * 
     * @param destination The destination to set.
     */
    void setDestination (final I_LogicalAddress destination)
    {
        this.destination = destination;

        if (messageRoot != null)
        {
        	messageRoot.setAttribute (ATTR_DEST_STR, getDestString());
        }
    }
    protected String getDestString(){
    	if(this.destination==null){
    		return MULTICAST_DESTINATION_STRING;
    	}else{
    		return this.destination.toString();
    	}
    }

    /**
     * Returns the message ID.
     * 
     * @return the message ID
     */
    public long getMessageID()
    {
        return messageID;
    }

    public void clearMessageID()
    {
        this.messageID = MESSAGE_ID_NOT_SET;
    }

    /**
     * Set the message ID.
     * This field is usually set by Monitor or Portal class automatically before
     * the message being sent out.
     * 
     * @param messageID The messageID to set.
     */
    void setMessageID (final long messageID)
    {
        this.messageID = messageID;

        if (messageRoot != null)
        {
            messageRoot.setAttribute (ATTR_MSGID_STR, Long.toString (messageID));
        }
    }

    /**
     * Get the logical address of the sender of this message.
     * 
     * @return The logical address of the sender.
     */
    public I_LogicalAddress getSender()
    {
        return sender;
    }

    /**
     * Set the logical address string of the sender.
     * This field is usually set by Monitor or Portal class automatically before
     * the message being sent out.
     * 
     * @param sender The sender to set.
     */
    void setSender (final I_LogicalAddress sender)
    {
        if (sender == null)
            throw new NullPointerException();

        this.sender = sender;

        if (messageRoot != null)
        {
            messageRoot.setAttribute (ATTR_SENDER_STR, sender.toString());
        }
    }

    /**
     * Get the timestamp of the message
     * 
     * @return Returns the timeStamp.
     */
    public long getTimeStamp()
    {
        return timeStamp;
    }

    /**
     * Set the timestamp of the message
     * 
     * This field is usually set by Monitor or Portal class automatically before
     * the message being sent out.
     * 
     * @param timeStamp The timeStamp to set.
     */
    void setTimeStamp (final long timeStamp)
    {
        this.timeStamp = timeStamp;

        if (messageRoot != null)
        {
            messageRoot.setAttribute (ATTR_TIMESTAMP_STR, Long.toString (timeStamp));
        }
    }
    
    /**
     * Get the String representation of the type of the message. This String will
     * be put in to the XML document representing this message as the root element
     * name.
     * 
     * @return	the String representation of the type of the message
     */
    public String getType()
    {
        return intToTypeString [type];
    }

    public int getTypeNum()
    {
    	return type;
    }

    /**
     * set the type of the message
     * 
     * This field is usually set by Monitor or Portal class automatically before
     * the message being sent out.
     * 
     */
    /* package */ void setTypeNum (final int newType) 
    {
        if (this.type != newType)
        {
            final int oldType = this.type;
            this.type = newType;

            if (messageRoot != null)
            {
                final Element oldMessageRoot = messageRoot;
                messageRoot = doc.createElement (intToTypeString [type]);
                final NodeList nodeList = oldMessageRoot.getChildNodes();

                int oldLength = nodeList.getLength();
                while (nodeList.getLength() > 0)
                {
                    messageRoot.appendChild (nodeList.item (0)); // appendChild is a move Not a copy!
                    final int newLength = nodeList.getLength();

                    if (newLength == oldLength)
                        throw new RuntimeException ("XML strangeness");

                    oldLength = newLength;
                }

                final NamedNodeMap namedNodeMap = oldMessageRoot.getAttributes();
                for (int i = 0; i < namedNodeMap.getLength(); ++i)
                {
                    final Node node = namedNodeMap.item (i);
                    messageRoot.setAttribute (node.getNodeName(), node.getNodeValue());
                }
			
                doc.removeChild (oldMessageRoot);
                doc.appendChild (messageRoot);
            }
        }
    }

    public int hashCode() { return (int) messageID; }
    
    /*
     * Equality is based solely on message id, allows responses to be matched with requests
     */
    public boolean equals (final Object object)
    {
        if (object == null)
        {
            /*
             * The Java docs for Object.equals() say that false should
             * be returned here, but I want to know about null refs so
             * I throw an exception.
             */

            //return false;
            throw new NullPointerException();
        }

        boolean result = false;
        if (object instanceof MCRawMessage)
        {
            final MCRawMessage other = (MCRawMessage) object;
            result = (other.getMessageID() == this.getMessageID());
        }
        return result;
    }

    public Object clone()
    {
        try
        {
            final MCRawMessage cloneMsg = (MCRawMessage) super.clone();
            cloneMsg.sender = (this.sender); //TODO how can we clone I_LogicalAddress??
            cloneMsg.destination = (this.destination);//TODO how can we clone I_LogicalAddress??
            cloneMsg.messageRoot = null;
            cloneMsg.doc = null;
            return cloneMsg;
        }
        catch (final CloneNotSupportedException cnse)
        {
            throw new HyperCastFatalRuntimeException(); // can never happen, by definition
        }
    }
}
