/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.MonitorAndControl;

import java.util.Vector;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import hypercast.util.XmlUtil;
import hypercast.HyperCastConfig;
import hypercast.HyperCastStatsException;
import hypercast.HyperCastFatalRuntimeException;
import hypercast.I_LogicalAddress;

/**
 * This class implements a Monitor/Control Query message 
 * exchanged between Monitor and Portal to carry the query
 * and reply for statitics and schemas.
 * 
 * The XML document represeting a MonMessage looks like:
 * <TypeOfMessage ....>
 * 		<Stats index="0" xpath="...">
 * 			<...>	//the value of the stats
 * 		</Stats> 
 * 		<Stats index="1" xpath="...">
 * 			<...>	//the value of the stats
 * 		</Stats>
 *      ... ...
 * </TypeOfMessage>
 * 
 * The type of message can be one of the following:
 * 		1. GetQuery
 *      2. GetReply
 *      3. ReadSchemaQuery
 *      4. ReadSchemaReply
 *      5. SetQuery
 *      6. SetReply
 *      7. WriteSchemaQuery
 *      8. WriteSchemaReply
 * 
 * A MonMessage contains a list of statistics ("Stats"), each of which has a XPath expression
 * in String format, and an index specifying the location of the "Stats" in the message.
 * 
 * If and only if the type is 2.,4.,5.,6.,or 8., the value Element contained in 
 * each "Stats" element is not empty.
 * 
 * A "Stats" element inside an MonMessage can be indexed either by specifying the
 * XPath string, or using the index field.
 * 
 * @author HyperCast Team
 * @author Guangyu Dong, January 2005
 * @author Greg Mattes, January 2005
 */
public class MonMessage
    extends MCRawMessage
{
    private static final boolean debug = false;
	
    private static String STATS_ITEM_NAME = "Stats";

    private static String ATTR_XPATH_STR  = "xpath";
    private static String ATTR_INDEX_STR  = "index";

    private Vector statsList = new Vector(); // StatsItem
	
    /* package */ MonMessage (final int type, final I_LogicalAddress sender, final int messageID)
    {
        super (type, sender, messageID);
    }

    /* package */ MonMessage (final int type, final I_LogicalAddress sender)
    {
        this (type, sender, MESSAGE_ID_NOT_SET);
    }

    /*
     * Creates statsList based on content of document.
     */
    /* packge */ MonMessage (MCSocket sock, final Document doc)
    {
        super (sock, doc);
		
        final NodeList nodeList = messageRoot.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); ++i)
        {
            final Element element = (Element) nodeList.item (i);
            final StatsItem statsItem = new StatsItem (element);
            statsList.addElement (statsItem);
        }
    }
	
    /**
     * Get number of statistics contained in this message.
     * 
     * @return	number of stats contained
     */
    public int getStatsCount()
    {
        return statsList.size();
    }
	
    /**
     * Return XPath string at the specified index.
     * 
     * @param index	the index of the stats
     * @return	the XPath string
     */
    public XPath getStatXPath (final int index)
        throws HyperCastStatsException
    {
        try
        {
            return XmlUtil.createXPath(((StatsItem) statsList.elementAt (index)).getXPathString());
        }
        catch (final ArrayIndexOutOfBoundsException aioobe)
        {
            throw new HyperCastStatsException ("index: " + index, HyperCastStatsException.StatisticNotFound);
        }
    }

    /**
     * Return the XML stats value at the specified index 
     * 
     * @param index the index of the stats
     * @return		the XML stats value
     */
    public Element [] getStatValue (final int index)
        throws HyperCastStatsException
    {
        try
        {
            return ((StatsItem) statsList.elementAt (index)).getValue();
        }
        catch (final ArrayIndexOutOfBoundsException aioobe)
        {
            throw new HyperCastStatsException ("index: " + index, HyperCastStatsException.StatisticNotFound);
        }
    }
	
    /**
     * Return the XML stats value which is referred by the xpath string.
     * 
     * @param xpath		the XPath string
     * @return			the XML stats value
     */
    public Element [] getStatValue (final XPath xpath)
        throws HyperCastStatsException
    {
        final StatsItem statsItem = searchStat (xpath.getPatternString());

        if (statsItem == null)
            throw new HyperCastStatsException (xpath.getPatternString(), HyperCastStatsException.StatisticNotFound);
        
        return statsItem.getValue();
    }
    
    
    /**
     * Search stats item
     */
    protected StatsItem searchStat (final String xpath)
    {
        StatsItem result = null;
        for (int i = 0; i < statsList.size(); ++i)
        {
            final StatsItem statsItem = (StatsItem) statsList.elementAt (i);
            if (statsItem.getXPathString().equals (xpath))
            {
                result = statsItem;
                break;
            }
        }
        return result;
    }

    /**
     * Adds OR updates a new stats item with value
     * Add a new stats item with <code>null</code> value.
     * NB: clobbers any existing stat with <code>null</code>
     * @param xpath
     */
    public void addStat (final XPath xpath)
    {
        addStatAndValue (xpath, null);
    }

    /**
     * Adds OR updates a new stats item with value
     * If xpath exists, update the stat with new value
     * @param Xpath		the xpath string
     * @param value		the XML stats value
     */
    public void addStat (final XPath Xpath, final Element value){
    	this.addStatAndValue(Xpath, XmlUtil.valueToList(value));
    }
    
    /* package*/ void addStatAndValue(final XPath Xpath, final Element [] value)
    {
    	
    	String xpath = Xpath.getPatternString();
        // not all stats have a value, for example in a "get" request,
        // only the stat is named, no value is set for it - the value
        // is being retrieved.

        if (value != null)
        {
            validateCanSet();
        }
		
        StatsItem statsItem = searchStat (xpath);
        if (statsItem == null)
        {
            statsItem = new StatsItem (xpath, cloneValue(value), statsList.size());
            statsList.addElement (statsItem);
            if (messageRoot != null) // if null -> doc has not been created, will be added at creation
            {
                messageRoot.appendChild (statsItem.getStatsElement());
            }
        }
        else
        {
            statsItem.update (xpath, cloneValue(value));
        }
    }
    
    /**
     * set the XML value of a stat at the specified index
     * 
     * @param index		the index of the stats
     * @param value		the XML value of the stats
     */
    public void setStat (final int index, final Element value)
        throws HyperCastStatsException{
    	this.setStatValue(index, XmlUtil.valueToList(value));
    }
	/*package*/ void setStatValue(final int index, final Element []value)
		throws HyperCastStatsException
    {
        if (value != null)
        {
            validateCanSet();
        }
		
        try
        {
            final StatsItem stats = (StatsItem) statsList.elementAt (index);

            if (stats == null)
                throw new HyperCastFatalRuntimeException ("stats should never be null");
            XPath xpath = XmlUtil.createXPath(stats.xpath);
            addStatAndValue (xpath, value);
        }
        catch (final ArrayIndexOutOfBoundsException aioobe)
        {
            throw new HyperCastStatsException ("index: " + index, HyperCastStatsException.StatisticNotFound);
        }
    }
    
    private Element[] cloneValue(Element []value){
    	return value;
    	
    	/*
    	if(value == null)
    		return null;
    	
    	Element newValue []= new Element[value.length];
    	for(int i=0;i<newValue.length;i++){
    		newValue[i] = (Element) value[i].cloneNode(true);
    	}
    	
    	return newValue;
    	*/
    }
    
    /**
     * set stats error by index
     */
    /* package */ void setStatError (final int index, final String error)
        throws HyperCastStatsException
    {
        validateCanSet();
		
        final Element value = this.getDocument().createElement ("Error");
        value.appendChild (this.getDocument().createTextNode (error));
        setStat (index, value);
    }

    /**
     * Add a error message to stats item
     */
    /* package */ void addStatError (final XPath xpath, final String error)
    {
        validateCanSet();
		
        final Element value = this.getDocument().createElement ("Error");
        value.appendChild (this.getDocument().createTextNode (error));
        addStat (xpath, value);
    }

    /**
     * Returns true if an error message exists for a stat with the specified XPath.
     * 
     * @param xpath		the xpath string
     * @return 			true if it is an error for this statistics item
     */
    public boolean isStatError (final XPath xpath)
        throws HyperCastStatsException
    {
        final StatsItem statsItem = searchStat (xpath.getPatternString());

        if (statsItem == null)
            throw new HyperCastStatsException ("cannot find: " + xpath, HyperCastStatsException.StatisticNotFound);

        return statsItem.isError();
    }

    /**
     * Returns true if an error message exists for a statistics item at the specified index.
     * 
     * @param index	the index for this stats
     * @return true if it is an error for this statistics item
     */
    public boolean isStatError (final int index)
        throws HyperCastStatsException
    {
        try
        {
            final StatsItem statsItem = (StatsItem) statsList.elementAt (index);
            return statsItem.isError();
        }
        catch (final ArrayIndexOutOfBoundsException aioobe)
        {
            throw new HyperCastStatsException ("cannot find stat at index: " + index, HyperCastStatsException.StatisticNotFound);
        }
    }

    /**
     * remove a statistics item with the specified xpath
     * 
     * @param Xpath		the xpath string
     */
    public void removeStat (final XPath Xpath)
        throws HyperCastStatsException
    {
    	String xpath = Xpath.getPatternString();
    	
        final StatsItem stats = searchStat (xpath);

        if (stats == null)
            throw new HyperCastStatsException ("cannot find stat: " + xpath, HyperCastStatsException.StatisticNotFound);

        statsList.remove (stats);
        removeFromMessageRoot (xpath);
    }

    /**
     * remove a statistics item at the specified index
     *
     * @param index		the index of the statistics item
     */
    public void removeStat (final int index)
        throws HyperCastStatsException
    {
        try
        {
            final StatsItem statsItem = (StatsItem) statsList.elementAt (index);
            statsList.remove (index);
            final String xpath = statsItem.getXPathString();
            removeFromMessageRoot (xpath);
        }
        catch (final ArrayIndexOutOfBoundsException aioobe)
        {
            throw new HyperCastStatsException ("cannot find stat at index: " + index, HyperCastStatsException.StatisticNotFound);
        }
    }
	
    /**
     * A null message root simply means that the document has not been
     * created yet, it will be created lazily.
     */
    private void removeFromMessageRoot (final String xpath)
    {
        if (messageRoot != null)
        {
            final NodeList nodeList = messageRoot.getChildNodes();
            for (int i = 0; i < nodeList.getLength(); ++i)
            {
                final Element element = (Element) nodeList.item (i);
                if (element.getAttribute (ATTR_XPATH_STR).equals (xpath))
                {
                    messageRoot.removeChild (element);
                }
            }
        }
    }

    /**
     *
     */
    protected void createMessageContents()
    {
        final Document doc = getDocument(); // ensures that messageRoot != null
        for (int i = 0; i < statsList.size(); ++i)
        {
            final StatsItem statsItem = (StatsItem) statsList.elementAt (i);
            messageRoot.appendChild (statsItem.getStatsElement());
        }
    }

    /*
     * Throws an exception if the message type is not a type that supports setting values
     */
    private void validateCanSet()
    {
        switch (type)
        {
            case GET_QUERY:
            case READ_SCHEMA_QUERY:
            case WRITE_SCHEMA_QUERY:
                //case ADVERTISE_MONITOR: ??
                throw new UnsupportedOperationException (getType() + " message cannot set statistics value for this type of message.");	
                
            default: break; /* others ok */
        }
    }

    /**
     * A three tuple: : <index, xpath, value> (of types: <int, String, Element>)
     */
    private class StatsItem
    {
        /**
         * The string that identifies the statistic - Note that this
         * *should* be a <code>String</code> and not an
         * <code>XPath</code>. An <code>XPath</code> object will be
         * created from the <code>String</code>.
         *
         * The <code>String</code> is particularly useful in the copy
         * constructor of this class because it is not known how to
         * make a copy of an <code>XPath</code> object quickly (it can
         * be serialized/deserialized).
         */
        private String xpath;

	 /** the value of some statistic in the form of a XML tree
	  *  it can never be null. 
	  **/
        private Element []value;

	/** The index of a stat */
        private int index;
	
	/**
         * The corresponding Element inside the XML MonMessage. This
         * is an "internal" variable of the StatsItem in the sense that
         * a StatsItem is thought of as a 3-tuple: <index, xpath,
         * value> but when the StatsItem appears in an XML document
         * there are two Elements for it: the stats Element which
         * encloses the value Element, for example:
         *
         *  <GetRequest ....>                  <-- XML document
         *    <Stats xpath="..." index="..">   <-- stats Element
         *      <SomeValue> ... </SomeValue>   <-- value Element
         *    </Stats>
         *  </GetRequest>
         */
        private Element statsElement;

        /**
         * @param value Value of statistic in response, <code>null</code> in request.
         */
        public StatsItem (final String xpath, final Element value[], final int index)
        {
            if (xpath == null)
                throw new NullPointerException ("XPath cannont be null");

            this.statsElement = null;

            this.xpath = xpath;
            this.value = value;
            this.index = index;
            
            if(this.value == null)
            	this.value = new Element[0];
        }

        /**
         * Copy constructor
         */
        private StatsItem (final StatsItem other)
        {
            this.statsElement = null;

            this.xpath = other.xpath;
            
            this.value =
                (other.value == null)
                ? new Element[0]
                : copyValue( other.value); // "deep" copy

            this.index = other.index;
        }
        
        private Element []copyValue(Element [] value){
        	Element [] newValue = new Element [value.length];
        	for(int i=0;i<newValue.length;i++){
        		newValue[i] = (Element) value[i].cloneNode(true);
        	}
        	
        	return newValue;
        }
        
        private Element [] getValueFromNodeList(NodeList nl){
        	Element [] vl = new Element[nl.getLength()];
        	
        	for(int i=0;i<vl.length;i++){
        		vl [i] = (Element) nl.item(i);
        	}
        	
        	return vl;
        }

        /**
         * @param statsElement
         */
        public StatsItem (final Element statsElement)
        {
            if (! statsElement.getNodeName().equals (STATS_ITEM_NAME))
                throw new IllegalArgumentException ("Invalid stats element name:" + statsElement.getNodeName());
		
            this.statsElement = statsElement;
		
            this.xpath = statsElement.getAttribute (ATTR_XPATH_STR);
		
            this.value =
                (statsElement.getChildNodes().getLength() > 0)
                ? getValueFromNodeList(statsElement.getChildNodes())
                : new Element[0];  // There is no value on a request

            this.index = Integer.parseInt (statsElement.getAttribute (ATTR_INDEX_STR));
        }

        public Element [] getValue()
        {
            return value;
        }

        public String getXPathString()
        {
            return xpath;
        }

        public Element getStatsElement()
        {
            if (statsElement == null)
            {
                final Document doc = MonMessage.this.doc;

                if (doc == null)
                    throw new IllegalStateException ("There must be some non-null document to use!.");

                statsElement = doc.createElement (STATS_ITEM_NAME);

                statsElement.setAttribute (ATTR_INDEX_STR, Integer.toString (index));
                statsElement.setAttribute (ATTR_XPATH_STR, xpath);

                updateValue();
            }
            return statsElement;
        }
        
        private Element[] changeOwner(Element []oldValue, Document newDoc){
        	Element [] newValue = new Element[oldValue.length];
        	
        	for(int i=0;i<newValue.length;i++){
        		newValue[i] = (Element) newDoc.importNode(oldValue[i], true);
        	}
        	
        	return newValue;
        }

        private void updateValue()
        {
            final Document doc = MonMessage.this.doc;

            if (doc == null)
                throw new IllegalStateException ("There must be some non-null document to use!.");
            
            /** clear the statsElement */
            while(statsElement.getChildNodes().getLength()>0){
            	Node nd = statsElement.getChildNodes().item(0);
            	statsElement.removeChild(nd);
            }

            if (value != null && value.length>0)
            {
                if (doc != value[0].getOwnerDocument())
                {
                    /*
                     * copy value Element to the current document
                     * by first creating a clone/copy and then
                     * adding that copy to this document.
                     * I.e. copy, not cut & paste.
                     */
                    value = changeOwner(value, doc);
                }
                
                /** append value array */
                for(int i=0;i<value.length;i++){
                	statsElement.appendChild (value[i]);
                }
            }
        }

        public void update (final String xpath, final Element value[])
        {
            this.xpath = xpath;
            this.value = value;

            /*
             * Could provide "dirty" bit and update only on
             * "getStatsElement" but we believe that "update" is
             * seldom if ever called.
             */
            if (statsElement != null)
            {
                statsElement.setAttribute (ATTR_XPATH_STR, xpath);
                updateValue();
            }
        }
	
        /**
         * @return true if there is an error in this message.
         */
        public boolean isError()
        {
            return (value == null && value.length>0) ? false : value[0].getNodeName().equals ("Error");
        }
    }

    private void addStatsItem (final StatsItem other)
    {
        final StatsItem newItem = new StatsItem (other);
        statsList.add (newItem);
    }

    public Object clone()
    {
        final MonMessage cloneMsg = (MonMessage) super.clone();
        synchronized (this.statsList)
        {
            cloneMsg.statsList = new Vector();

            // need "deep" copy of Vector contents so hold the lock on the Vector and do "deep" copy
            for (int i = 0; i < this.statsList.size(); ++i)
            {
                /*
                 * DANGER WILL ROBINSON! Cannot clone inner classes!
                 *
                 * See: comp.lang.java.programmer, 11 May 1999, Subject
                 * "How to clone() an inner class?" (and replies).
                 *
                 * Could not find reference to this issue in standard
                 * Java documentation, if you find it please add here.
                 */
                //final StatsItem clonedStatsItem = (StatsItem) this.statsList.elementAt(i).clone();

                cloneMsg.addStatsItem ((StatsItem) this.statsList.elementAt (i));
            }
        }
        return cloneMsg;
    }
}
