/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.MonitorAndControl;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import hypercast.MonitorAndControl.*;

import java.util.Hashtable;
import java.util.Enumeration;

import java.util.Timer;
import java.util.TimerTask;

import org.apache.xpath.XPath;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import hypercast.HyperCastStatsException;
import hypercast.HyperCastWarningRuntimeException;
import hypercast.I_LogicalAddress;
import hypercast.HyperCastConfig;
import hypercast.I_Stats;
import hypercast.SimpleStats;
import hypercast.StatsProcessor;
import hypercast.util.XmlUtil;

/**
 * For full description of the Monitor class follow the "see"
 * links. From an implementation standpoint, many methods are
 * delegated to either the <code>PortalListManager</code> or the
 * <code>MCSocket</code>. This class does implemented the send
 * methods. Also, this class implements the
 * <code>CallbackManager</code> mechanism that is responsible for
 * matching responses to requests and for issuing timeouts.
 *
 * @see <code>I_Monitor</code>
 * @see PortalListManager
 * @see MCSocket
 *
 * From the Monitor and Control design doc:
 *
 * "A monitor set to silent mode does not set entries in the portal list as inactive."
 * 
 * @author HyperCast Team
 * @author Guangyu Dong, January, 2005
 * @author Greg Mattes, January, 2005
 */
public class Monitor
    implements I_Monitor, I_Stats, I_MCRaw_Callback
{
	private static final boolean debug = false;

	private static final boolean timeoutDebug = false;

	private HyperCastConfig config;

	/**
	 * An object that keeps track of remote portals. Remote Portals are indexed
	 * by integer.
	 */
	private PortalListManager portalListManager;

	/** Object that receives reply messages and schedules timeouts. */
	private CallbackManager callbackManager = new CallbackManager();

	/** HyperCast socket used for exchanging XML messages with Portals */
	private MCSocket mcSocket;

	/**
         * Each request that is not a retransmission is given the next
         * unused messageID.
         */
	private long messageID = MCRawMessage.INITIAL_UNIQUE_MESSAGE_ID;
	
	/**
         * The timeout time for query message. This is set by
         * configuration and statistics interface only, there is no
         * API or parameter of any API to control it.
         */
	private long queryTimeout;

	/**
	 * The config properties prefix passed to the MCSocket on contruction of a
	 * socket. The config prefix allows the MCSocket to select the appropriate
	 * portion of the configuration file.
	 */
	private static final String MONITOR_CONFIG_PREFIX = "/Public/MonitorAndControl/Monitor/";

	/**
	 * @see StatsProcessor
	 */
    private StatsProcessor statsProcessor;

	/**
	 * The statistics name for this object (as an I_Stats instance). Its default
	 * name is defined in the socket configuration schema file. It can be
	 * assigned by any other object through method setStattisticsName defined in
	 * the I_Stats interface.
	 */
	private String statisticsName;

	/**
	 * Construct a Monitor with a config object
	 * 
	 * @param config
	 *            the config object to read configuration attributes from
	 */
	public Monitor (final HyperCastConfig config)
    {
		if (config == null)
			throw new NullPointerException();

		this.config = config;
		this.mcSocket = new MCSocket(config, MONITOR_CONFIG_PREFIX, this);

		this.portalListManager = new PortalListManager (config, mcSocket, MONITOR_CONFIG_PREFIX);

		statisticsName = config.getTextAttribute (XmlUtil.createXPath (MONITOR_CONFIG_PREFIX + "StatName"));
		queryTimeout = config.getPositiveLongAttribute (XmlUtil.createXPath (MONITOR_CONFIG_PREFIX + "QueryTimeout"));

		initStatisticsStructure();
	}

	/**
	 * @see I_Monitor#startTranscript(String)
	 */
	public void startTranscript(final String filename) throws FileNotFoundException {
		mcSocket.startTranscript(filename);
	}

	/**
	 * @see I_Monitor#startTranscript(String)
	 */
	public void stopTranscript() {
		mcSocket.stopTranscript();
	}

    /**
     * @see I_Monitor#getPortalAddress(int)
     */
    public I_LogicalAddress getPortalAddress (final int portalIndex)
        throws NoSuchPortalException
    {
        return portalListManager.getPortal(portalIndex).getPortalAddress();
    }

	/**
	 * @see I_Monitor#getActivePortalIndices()
	 */
	public int[] getActivePortalIndices() {
		return portalListManager.getActivePortalIndices();
	}
	
	/**
	 * @see I_Monitor#getAllPortalIndices()
	 */
	public int[] getAllPortalIndices() {
		return portalListManager.getAllPortalIndices();
	}

	/**
	 * @see I_Monitor#isPortalActive(int)
	 */
	public boolean isPortalActive(final int portalIndex) throws NoSuchPortalException {
		return portalListManager.getPortal(portalIndex).isActive();
	}

	/**
	 * @see I_Monitor#indexOfPortal(I_LogicalAddress)
	 */
	public int indexOfPortal(final I_LogicalAddress logicalAddress)
			throws NoSuchPortalException {
		return portalListManager.indexOfPortal(logicalAddress);
	}

	/**
	 * @see I_Monitor#clearPortalList()
	 */
	public void clearPortalList() {
		portalListManager.clearPortalList();
	}

	/**
	 * Generate the Message ID
	 */
    protected synchronized long getNextMessageID() {
		if (messageID <= 0) {
			messageID = MCRawMessage.INITIAL_UNIQUE_MESSAGE_ID;
		}
		return messageID++;
	}

    /**
     * @see I_Monitor#sendTo(int, MonMessage, I_ReceiveCallback, I_TimeoutCallback)
     */
    public synchronized void sendTo (final int               portalIndex,
                                     final MonMessage         message,
                                     final I_ReceiveCallback receiveCallback,
                                     final I_TimeoutCallback timeoutCallback)
        throws NoSuchPortalException
    {
        unifiedSend (portalIndex, message, receiveCallback, timeoutCallback, false);
    }

    /**
     * @see I_Monitor#sendToAll(MonMessage, I_ReceiveCallback, I_TimeoutCallback)
     */
    public synchronized void sendToAll (final MonMessage         message,
                                        final I_ReceiveCallback receiveCallback,
                                        final I_TimeoutCallback timeoutCallback)
    {
        final int dummyPortalAddress = -1;
        try
        {
            unifiedSend (dummyPortalAddress, message, receiveCallback, timeoutCallback, true);
        }
        catch (final NoSuchPortalException ignored)
        {
            // In a broadcast a specific Portal is not used.
        }
    }

    /**
     * Implements sendTo and sendToAll
     */
    private synchronized void unifiedSend (final int               portalIndex,
                                           final MonMessage         message,
                                           final I_ReceiveCallback receiveCallback,
                                           final I_TimeoutCallback timeoutCallback,
                                           final boolean           broadcast)
        throws NoSuchPortalException
    {
        if (! mcSocket.isOpen())
            throw new HyperCastWarningRuntimeException ("Monitor not in running state.");

        // Do lookup of Portal early so that other computations can be short-circuited if not found
        final I_LogicalAddress portalLogicalAddress =
            broadcast
            ? null
            : portalListManager.getPortal (portalIndex).getPortalAddress();

        final boolean isRetransmission = message.getMessageID() != MCRawMessage.MESSAGE_ID_NOT_SET;

        // Generate a new message id if this message does not have a message id set, i.e. not a retransmission
        final long messageID = isRetransmission ? message.getMessageID() : getNextMessageID();

        // This will point to either the message that was passed to
        // this method or a copy (clone), depending on whether or not
        // a timeoutCallback is given and furthermore whether or not
        // the object passed to the method is already a copy - want to
        // avoid making copies of copies.
        //
        // Speculatively set this to the message that was passed, will reset if needed to point to clone
        MonMessage messageToSend = message;

        // Message id will be cleared before return if the message is not a retransmission
        messageToSend.setMessageID (messageID);

        // The following code is purposely verbose to show various cases
 
        if (timeoutCallback == null
            &&
            receiveCallback == null)
        {
            // no need for a call back entry or a copy of the message
            // - user did not specify a receive callback or a timeout
            // callback. Simply (temporarily) set the message id (will
            // be unset before return)

        }
        else // at least one callback specified
        {
            if (timeoutCallback == null)
            {
                // on receive, user will get reference to response -> no need to save copy of request
            }
            else // on timeout, user wants the message in current state -> must make a copy, iff this is not already a copy
            {
                if (! isRetransmission)
                {
                    messageToSend = (MonMessage) message.clone();
                }
            }
            callbackManager.addCallback (messageID, messageToSend, receiveCallback, timeoutCallback, broadcast);
        }

        if (Monitor.debug
            ||
            (timeoutDebug && timeoutIDs.contains (new Long (messageID))))
        {
            System.err.println ("Monitor.sendTo" + (broadcast ? "All" : "") + " (message id -> " + messageID
                                + ", portalIndex -> " + portalIndex + ", timeout -> " + config.getPositiveLongAttribute (XmlUtil.createXPath (MONITOR_CONFIG_PREFIX + "QueryTimeout"))
//queryTimeout
                                + ")" +
                                (broadcast ? "; Portal's LA is " + portalLogicalAddress : ""));
        }

        if (broadcast)
        {
            mcSocket.sendToAll (messageToSend);
        }
        else
        {
            mcSocket.sendTo (messageToSend, portalLogicalAddress);
        }
    }

	/**
	 * @see I_Monitor#createGetRequest()
	 */
            public MonMessage createGetRequest() {
		return mcSocket.createQueryMessage(MCRawMessage.GET_QUERY);
	}

   
	/**
	 * @see I_Monitor#createSetRequest()
	 */
	public MonMessage createSetRequest() {
		return mcSocket.createQueryMessage(MCRawMessage.SET_QUERY);
	}

    
	/**
	 * @see I_Monitor#createReadSchemaRequest()
	 */
	public MonMessage createReadSchemaRequest() {
		return mcSocket.createQueryMessage(MCRawMessage.READ_SCHEMA_QUERY);
	}

   
	/**
	 * @see I_Monitor#createWriteSchemaRequest()
	 */
	public MonMessage createWriteSchemaRequest() {
		return mcSocket.createQueryMessage(MCRawMessage.WRITE_SCHEMA_QUERY);
	}

	/**
	 * @see I_Monitor#compactPortalList()
	 */
	public void compactPortalList() {
		portalListManager.compactPortalList();
	}

    /**
     * @see I_Monitor#setSilentOff
     */
    public void setSilentOn()
    {
        portalListManager.setSilentOn();
    }

    /**
     * @see I_Monitor#setSilentOff
     */
    public void setSilentOff()
    {
        portalListManager.setSilentOff();
    }

	/**
	 * @see I_Monitor#openMonitor()
	 */
	public void openMonitor()
    {
		if (!mcSocket.isOpen())
                {
                    mcSocket.open();
                    config.log.println("Opening Monitor at Logical Address: "+ mcSocket.getLogicalAddress());
                    portalListManager.signalOpenMonitor();
		}
                else
                {
                    config.log.println("Attempt to open a Monitor that is already open.");
		}
	}

    /**
     * Explicitly close monitor. NB: closes the transcript file, if there is
     * one. This file must be explicitly reopened if the monitor is reopened.
     */
    public void closeMonitor()
    {
        if (mcSocket.isOpen())
        {
            config.log.println("Closing Monitor.");

            try
            {
                stopTranscript();
            }
            catch (final IllegalStateException ignored)
            {
                // if no transcript was started there will be an
                // exception, but that is ok: we want to ensure that
                // if a transcript was opened, it will be closed.
            }

            portalListManager.signalCloseMonitor();
            mcSocket.close();

            /*
             * Do not make a new PortalListManager, that
             * would lose any changes made through stats
             * interface such as timeout values - simply
             * clear the entries of the current
             * PortalListManager object.
             */
            clearPortalList(); 
        }
        else
        {
            config.log.println("Attempt to close a Monitor that is already closed.");
        }
    }

    /**
     * This class manages simple statistic "SilentMode".
     */
    class SilentModeOperator extends SimpleStats
    {
        /**
         * Gets the value of a simple statistic managed by the implementation of this class.
         * 
         * @see SimpleStats#getStats()
         */
        protected String getStats()
        {
            return Boolean.toString (portalListManager.getSilentMode());
        }
	
        /**
         * Sets the value of a simple statistic managed by the implementation of this class.
         * 
         * @see SimpleStats#setStats(String newvalue)
         */
        protected String setStats(String newValue) throws HyperCastStatsException
        {
            final boolean mode = newValue.trim().equalsIgnoreCase ("true");
            if (mode)
            {
                setSilentOn();
            }
            else
            {
                setSilentOff();
            }
            return Boolean.toString (mode);
        }

        /** 
         * @see I_Stats#getReadSchema(Document, XPath)
         */
        public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
        {
            return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Boolean"/* type */, 
                                                 null/* restriction base */, null/* pattern value */);
        }
		
        /** 
         * @see I_Stats#getWriteSchema(Document, XPath)
         */
        public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
        {
            return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Boolean"/* type */, 
                                                 null/* restriction base */, null/* pattern value */);
        }
    }

    // debugging
    private java.util.Set timeoutIDs = timeoutDebug ? new java.util.HashSet() : null;

    /**
     * Thou shalt hold the lock for the callbackTable before thou
     * holdst the lock for any particular CallbackEntry.
     */
    private class CallbackManager
        implements I_MCRaw_Callback
    {
        // Long -> CallbackEntry (Message ID -> CallbackEntry)
        private Hashtable callbackTable = new Hashtable();

        private Timer timer = new Timer();

        /**
         * Called from Monitor application thread
         *
         * Both receiveCallback and timeoutCallback can be null for a
         * message that does not need a response. This happens
         * when the RunControl application issues
         * "kill_remote_servers" for example. The remote servers
         * die without responding.
         *
         * message can be null if there is no timeoutCallback, it must
         * be non-null if there is a timeoutCallback.
         *
         * @throws NullPointerException If message is null when
         * timeoutCallback is not null because when a timeout callback
         * is invoked there must be some message to pass to the
         * Monitor application.
         */
        public void addCallback (final long              messageID,
                                 final MonMessage         message,
                                 final I_ReceiveCallback receiveCallback,
                                 final I_TimeoutCallback timeoutCallback,
                                 final boolean           broadcast)
        {
            final Long messageIDLong = new Long (messageID);

            if (receiveCallback == null
                &&
                timeoutCallback == null)
            {
                if (Monitor.debug)
                    System.err.println ("Monitor.CallbackManager.addCallback: Both callbacks are null, message id: " + messageID);

                return; // no callbacks, nothing to do
            }

            if (timeoutCallback != null
                &&
                message == null)
            {
                throw new NullPointerException ("message must not be null when timeoutCallback is not null!");
            }
            
            if (message != null
                &&
                message.getMessageID() != messageID)
            {
                throw new IllegalArgumentException ("message.getMessageID (" + message.getMessageID() + ") and messageID (" + messageID + ") are inconsistent!");
            }

            if (Monitor.debug
                ||
                (timeoutDebug && timeoutIDs.contains (messageIDLong)))
            {
                System.err.println ("Monitor.CallbackManager.addCallback: Adding callback for message with id: " + messageID);
            }

            final CallbackEntry callbackEntry = new CallbackEntry (message, receiveCallback, timeoutCallback, broadcast);
            final TimeoutHandler timeoutHandler = new TimeoutHandler (callbackEntry);

            // Lock out timer thread, MCSocket receive thread, and any
            // other application threads.
            //
            // Note that other application threads are actually locked
            // out by serializing access (via synchronized keyword) to
            // the sendTo and sendToAll methods of the Monitor.

            synchronized (callbackTable)
            {
                final CallbackEntry preexistingCallbackEntry = (CallbackEntry) callbackTable.get (messageIDLong);
                if (preexistingCallbackEntry != null)
                {
                    // A previous send already sent a message with the
                    // same id. Replace the existing entry with the
                    // new entry, making sure to cancel the timer
                    // event for the first entry.
                    //
                    // This is considered to be strange behaviour: a
                    // retransmission should not be sent until a
                    // response or timeout occurs for the prior send.

                    config.err.println ("Monitor.CallbackManager.addCallback: Duplicate callback table entry created: retransmission before response or timeout of prior transmission, message id: " + messageID);

                    preexistingCallbackEntry.cancelTimeoutHandler();
                }

                callbackTable.put (messageIDLong, callbackEntry);
                callbackEntry.scheduleTimeoutHandler(broadcast);

                if (Monitor.debug
                    ||
                    (timeoutDebug && timeoutIDs.contains (messageIDLong)))
                {
                    displayCallbackTable ("Added new entry with key " + messageID);
                }
            }
        }

        /**
         * Called from MCSocket receive thread
         *
         * @see hypercast.MonitorAndControl.I_ReceiveCallback#receiveMessage(hypercast.MonitorAndControl.MonMessage)
         */
        public void receiveMCRawMessage (final MCRawMessage message)
        {
            final long messageID = message.getMessageID();
            final Long messageIDLong = new Long (messageID);

            if (Monitor.debug
                ||
                (timeoutDebug && timeoutIDs.contains (messageIDLong)))
            {
                System.err.println ("Monitor: Received MCRawMessage: type: " + message.getType() + " id: " + messageID);
            }

            if (! (message instanceof MonMessage))
                throw new HyperCastWarningRuntimeException ("Message received in Monitor Callback Manager that is not of type MonMessage");

            // verification of sender
            try
            {
                final I_LogicalAddress sender = message.getSender();
                final int portalIndex = indexOfPortal (sender);
            }
            catch (final NoSuchPortalException nspe)
            {
                config.err.println ("Receive: Message received from seemingly non-existent Portal, continuing; message id: "
                                    + messageID + ", \"Portal\" (sender) address: " + message.getSender());
                return;
            }

            final CallbackEntry callbackEntry;
            synchronized (callbackTable)
            {
                if (Monitor.debug
                    ||
                    (timeoutDebug && timeoutIDs.contains (messageIDLong)))
                {
                    displayCallbackTable ("Retrieving entry with message ID: " + messageID);
                }

                // Get the entry corresponding to this response. If it
                // is not in the table, maybe the reponse arrived too late.

                callbackEntry = (CallbackEntry) callbackTable.get (messageIDLong);

                if (callbackEntry == null)
                {
                    config.err.println ("No CallbackEntry for message id: " + messageID
                                        + " - possible response after timeout");
                    return;
                }

                // For non-broadcast messages: cancel the timeout
                // timer and remove the entry in the callback table
                // for we have received a response.

                if (! callbackEntry.isBroadcast())
                {
                    final boolean cancelSuccess = callbackEntry.cancelTimeoutHandler();

                    if (Monitor.debug
                        ||
                        (timeoutDebug && timeoutIDs.contains (messageIDLong)))
                    {
                        displayCallbackTable ("Attempted to cancel timeout callback for message ID (" + messageID + "): " + (cancelSuccess ? "Success" : "Failed"));
                    }

                    callbackTable.remove (messageIDLong);
                }
                else
                {
                    // Since we have received at least one response,
                    // we will never need to invoke the timeout
                    // callback - do NOT cancel the timeout handler
                    // here! It is still needed to remove the callback
                    // entry from the callback table.

                    callbackEntry.removeTimeoutCallback();
                }

                if (Monitor.debug
                    ||
                    (timeoutDebug && timeoutIDs.contains (messageIDLong)))
                {
                    displayCallbackTable ("Removed entry from callback table with message ID: " + messageID);
                }
            }

            if (Monitor.debug
                ||
                (timeoutDebug && timeoutIDs.contains (messageIDLong)))
            {
                System.err.println ("Monitor: invoking receive callback: " + messageID);
            }

            callbackEntry.invokeReceiveCallback (message);
        }

        // Used for debugging
        public void displayCallbackTable (final String text)
        {
            System.err.println("Callback Table Contents: " + text);
            for (final Enumeration enumeration = callbackTable.keys(); enumeration.hasMoreElements();)
            {
                final Long key = (Long) enumeration.nextElement();
                final CallbackEntry current = (CallbackEntry) callbackTable.get(key);
                System.err.println ("  Entry: " + key + " -> " + current);
            }
        }

        /**
         * The TimeoutHandler class serves two purposes:
         *
         * 1. To remove a CallbackEntry from the CallbackTable
         * 2. To invoke a timeoutCallback when a response is not received
         *
         * Note that for a broadcast (sendToAll) message the timeout
         * handler can do 1 or 1 & 2. For a non-broadcast message the
         * TimeoutHandler always does 1 & 2.
         *
         * To understand this, observe that a broadcast message can
         * receive responses for the entire queryTimeout time which
         * means that the callback entry should be valid for that
         * entire time - even if a response is received. This is not
         * the case for non-broadcast messages. With a non-broadcast
         * message the CallbackEntry is removed when a response is
         * received.
         */
        class TimeoutHandler
            extends java.util.TimerTask
        {
            private CallbackEntry callbackEntry;
            private long messageID;
            private Long messageIDLong;

            public TimeoutHandler (final CallbackEntry callbackEntry)
            {
                this.callbackEntry = callbackEntry;
                this.messageID = callbackEntry.getMessageID();
                this.messageIDLong = new Long (messageID);
                callbackEntry.setTimeoutHandler (this);
            }

            public void run()
            {
                if (timeoutDebug)
                    timeoutIDs.add (messageIDLong);

                if (Monitor.debug
                    ||
                    (timeoutDebug && timeoutIDs.contains (messageIDLong)))
                {
                    config.err.println ("Monitor.TimeoutHandler: Timeout for message id: " + messageID);
                }

                synchronized (callbackEntry)
                {
                    if (Monitor.debug
                        ||
                        (timeoutDebug && timeoutIDs.contains (messageIDLong)))
                    {
                        displayCallbackTable ("Monitor.TimeoutHandler: Retrieving entry with key: " + messageID);
                    }

                    final CallbackEntry removedCallbackEntry = (CallbackEntry) callbackTable.remove (messageIDLong);

                    if (removedCallbackEntry != callbackEntry)
                    {
                        config.err.println ("Monitor.TimeoutHandler: strangeness, the callback entry that I removed for this message id is not the same callback entry that I put into the table, message id: " + messageID);
                        return;
                    }

                    if (removedCallbackEntry == null)
                    {
                        config.err.println ("Monitor.TimeoutHandler: No CallbackEntry found, entry possibly removed by message receive operation, message id: " + messageID);
                        return;
                    }
                }
                
                if (Monitor.debug
                    ||
                    (timeoutDebug && timeoutIDs.contains (messageIDLong)))
                {
                    System.err.println ("Monitor: invoking timeout callback: " + messageID);
                }

                // Note: no locks are held when callback is invoked
                callbackEntry.invokeTimeoutCallback();
            }
        }

        /**
         * Stores information about a message when it is waiting for a reply or timeout.
         */
        class CallbackEntry
        {
            // multiple responses are possible with a
            // sendToAll/broadcast message. So timeout timers are not
            // cancelled when responses are received.
            private boolean broadcast;

            private TimeoutHandler timeoutHandler;

            private long messageID;

            private MonMessage message;

            private I_ReceiveCallback receiveCallback; // can be null

            private I_TimeoutCallback timeoutCallback; // can be null

            private int receiveCallbackInvocationCount = 0;

            private boolean timeoutCallbackInvoked = false;

            /**
             * It is assumed that messageID == message.getMessageID()
             *
             * @param message could be null if no timeoutCallback is specified
             */
            public CallbackEntry (final MonMessage         message,
                                  final I_ReceiveCallback receiveCallback,
                                  final I_TimeoutCallback timeoutCallback,
                                  final boolean           broadcast)
            {
                if (receiveCallback == null
                    &&
                    timeoutCallback == null)
                {
                    throw new NullPointerException ("timeoutCallback and receiveCallback are both null - No need for callback entry!");
                }

                if (timeoutCallback != null
                    &&
                    message == null)
                {
                    throw new NullPointerException ("message must not be null when timeoutCallback is not null!");
                }

                this.message         = message;
                this.messageID       = message.getMessageID();
                this.receiveCallback = receiveCallback;
                this.timeoutCallback = timeoutCallback;
                this.broadcast       = broadcast;
            }

            public boolean isBroadcast()
            {
                return broadcast;
            }

            public void scheduleTimeoutHandler(boolean broadcast)
            {
                final long queryTimeoutFOO = config.getPositiveLongAttribute (XmlUtil.createXPath (MONITOR_CONFIG_PREFIX + "QueryTimeout"));
            	if(broadcast){
            		timer.schedule (timeoutHandler, queryTimeoutFOO * 5);
            	}else{
            		timer.schedule (timeoutHandler, queryTimeoutFOO);
            	}
            }

            public boolean cancelTimeoutHandler()
            {
                return (timeoutHandler == null) ? true : timeoutHandler.cancel();
            }

            public void setTimeoutHandler (final TimeoutHandler timeoutHandler)
            {
                this.timeoutHandler = timeoutHandler;
            }

            public long getMessageID()
            {
                return messageID;
            }

            /**
             * Called when a response is received for a broadcast
             * message. Ensures that when the corresponding
             * CallbackEntry is removed from the callback table that
             * no timeout callback is invoked.
             */
            public void removeTimeoutCallback()
            {
                timeoutCallback = null;
            }

            public synchronized void invokeReceiveCallback (final MCRawMessage message)
            {
                ++receiveCallbackInvocationCount;
                receiveCallback.receiveMessage ((MonMessage) message);
            }

            public synchronized void invokeTimeoutCallback()
            {
                // null timeoutCallback means a no-op
                if (timeoutCallback != null)
                {
                    timeoutCallbackInvoked = true;
                    timeoutCallback.timeoutMessage (message);
                }
            }

            public String toString()
            {
                return "<CallbackEntry:" + " message id (" + message.getMessageID() + ")"
                    + " receiveCallback (exists: " + (receiveCallback != null) + ", invoked: "
                    + receiveCallbackInvocationCount + " times)" + " timeoutCallback ("
                    + (timeoutCallback != null) + ", invoked: " + timeoutCallbackInvoked + ")" + ">";
            }
        }
    }

    private void initStatisticsStructure()
    {
        statsProcessor = new StatsProcessor(this, true, true);

        statsProcessor.addStatsElement ("MonitorSocket", this.mcSocket.getOverlaySocket(), 1, 1);
        statsProcessor.addStatsElement ("SilentMode", new SilentModeOperator(), 1, 1);
    }

	/** @see I_Stats#getStatsName() */
	public String getStatsName() {
		return statisticsName;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName(String name) {
		statisticsName = name;
	}

	/**
	 * @see hypercast.I_Stats#getStats(org.w3c.dom.Document,
	 *      org.apache.xpath.XPath)
	 */
	public Element[] getStats(final Document doc, final XPath xpath) throws HyperCastStatsException {
		return statsProcessor.getStatsResult(doc, xpath);
	}

	/**
	 * @see hypercast.I_Stats#setStats(org.w3c.dom.Document,
	 *      org.apache.xpath.XPath, org.w3c.dom.Element)
	 */
	public Element[] setStats(final Document doc, final XPath xpath, final Element setValue)
			throws HyperCastStatsException {
		return statsProcessor.setStatsResult(doc, xpath, setValue);
	}

	/**
	 * @see hypercast.I_Stats#getReadSchema(org.w3c.dom.Document,
	 *      org.apache.xpath.XPath)
	 */
	public Element[] getReadSchema(final Document doc, final XPath xpath)
			throws HyperCastStatsException {
		return statsProcessor.getReadSchemaResult(doc, xpath);
	}

	/**
	 * @see hypercast.I_Stats#getWriteSchema(org.w3c.dom.Document,
	 *      org.apache.xpath.XPath)
	 */
	public Element[] getWriteSchema(final Document doc, final XPath xpath)
			throws HyperCastStatsException {
		return statsProcessor.getWriteSchemaResult(doc, xpath);
	}

    /* (non-Javadoc)
     * @see hypercast.MonitorAndControl.I_MCRaw_Callback#receiveMCRawMessage(hypercast.MonitorAndControl.MCRawMessage)
     */
    public void receiveMCRawMessage (final MCRawMessage message)
    {
        /* demultiplexing messages */
        switch (message.getTypeNum())
        {
            case MCRawMessage.ADVERTISE_MONITOR:
                // we hear our own advertise or from another monitor: do nothing
                break;

            case MCRawMessage.ADVERTISE_PORTAL:
                portalListManager.receiveMCRawMessage (message);
                break;

            default:
                callbackManager.receiveMCRawMessage (message);
        }
    }

	/* (non-Javadoc)
	 * @see hypercast.MonitorAndControl.I_Monitor#getInactivePortalIndices()
	 */
	public int[] getInactivePortalIndices() {
		return this.portalListManager.getInActivePortalIndices();
	}
}
