/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.MonitorAndControl;

import hypercast.I_LogicalAddress;
import hypercast.HyperCastConfig;
import hypercast.HyperCastConfigException;
import hypercast.HyperCastFatalRuntimeException;
import hypercast.util.XmlUtil;

import java.util.Random;

/**	
 * This class is responsible for two things:
 *
 * <ol>
 * <li>Periodically sending <code>AdvertisePortal</code> messages to a <code>Monitor</code>.</li>
 * <li>Determining when a <code>Monitor</code> should be considered down.</li>
 * </ol>
 *
 * From the Monitor and Control design doc:
 *
 * "Each Portal stores the logical address of at most one monitor"
 *
 * @author HyperCast Team
 * @author Guangyu Dong, January 2005
 * @author Greg Mattes, January 2005
 * @author Michael Nahas
 */
/* package */ final class PortalAdvertiser
    implements I_MCRaw_Callback, Runnable
{
    private static final boolean debug = false;

    private static final Random random = new Random();

    private static final String PROP_NAME_TIME_ADVERTISE  = "TimeAdvertise";
    private static final String PROP_NAME_MONITOR_TIMEOUT = "MonitorTimeout";
    private static final String PROP_NAME_MONITOR_ADDRESS = "MonitorAddress";
	
    /** The time interval between consecutive sends of AdvertisePortal messages. */
    private long advertisePeriod; // ms

    /**
     * The amount of time required to wait after a message is
     * received from a Monitor until it is consider to have timed out.
     */
    private long monitorTimeout; // ms
	
    /** Socket that is used by this class to send messages. */
    private MCSocket mcSocket;

    /**
     * Address of the monitor - there can be more than one monitor on
     * a control network, but each portal only stores the address of
     * the monitor from which it heard the last AdvertiseMonitor
     * message - NOT the last query message.
     */
    private I_LogicalAddress monitorAddress = null;

    /**
     * Contains the value of the configured monitor address if there
     * is one in the config, otherwise it is null.
     */
    private I_LogicalAddress configuredMonitorAddress = null;

    /** HyperCast configuration object used to read configuration information. */
    private HyperCastConfig config;
	
    /** Message that will be periodically sent to a <code>Monitor</code>. */
    private MCAdvertiseMessage advertisePortalMessage;

    /**
     * Silent mode means that no advertisement messages are being
     * sent even though the Portal is open.  In silent mode, the
     * PortalAdvertier will not send out any messages, but a Monitor
     * in silent mode will not timeout the Portal entry either.
     *
     * Note that sendAdvertisements can be true, and silentMode can
     * be true simultaneously. This corresponds to the situation
     * where the Portal Advertiser *can* sent Advertisments because
     * the Portal is open, but the user has set silentMode to true.
     *
     * Used for wireless Ad Hoc experiments - basically it makes the Monitor
     * and Portal be quiet so that they do not disturb the network.
     */
    private boolean silentMode = false;

    /**
     * Time at which the most recent <code>AdvertiseMonitor</code>
     * message was received. Always 0 for a preset Monitor.
     */
    private long mostRecentContactTime = 0L; // ms

    /**
     * The <code>PortalAdvertiser</code> will send advertisement
     * messages only when this value is <code>true</code>.
     */
    private boolean sendAdvertisements = false;

    /**
     * @throws HyperCastConfigException If relevant configuration information cannot be read.
     */
    public PortalAdvertiser (final HyperCastConfig config,
                             final MCSocket        mcSocket,
                             final String          propertiesPrefix)
        throws HyperCastConfigException
    {
        this.config = config;
        this.mcSocket = mcSocket;
		
        this.advertisePeriod = config.getLongAttribute (XmlUtil.createXPath (propertiesPrefix + PROP_NAME_TIME_ADVERTISE));
        this.monitorTimeout  = config.getLongAttribute (XmlUtil.createXPath (propertiesPrefix + PROP_NAME_MONITOR_TIMEOUT));
        
        final String monitorAddressString = config.getTextAttribute (XmlUtil.createXPath (propertiesPrefix + PROP_NAME_MONITOR_ADDRESS));

        if (PortalAdvertiser.debug)
            config.err.println ("Read monitor address: " + monitorAddressString);

        if (monitorAddressString != null && !monitorAddressString.trim().equals(""))
        {
            this.configuredMonitorAddress = mcSocket.createLogicalAddress (monitorAddressString);
            this.monitorAddress = this.configuredMonitorAddress;

            if (monitorAddress == null)
                throw new HyperCastConfigException ("ERROR parsing configured Monitor address: "+ monitorAddressString);

            config.log.println ("Portal using configured Monitor address: "+ monitorAddressString);
        }

        this.advertisePortalMessage = mcSocket.createAdvertiseMessage (MCRawMessage.ADVERTISE_PORTAL);
    }

    public synchronized void setSilentOn()
    {
        this.silentMode = true;
    }

    public synchronized void setSilentOff()
    {
        this.silentMode = false;
    }

    /* package */ synchronized boolean getSilentMode()
    {
        return silentMode;
    }

    /*
     * This is called from the Portal when the Portal.openPortal() method is
     * invoked. It signals the advertiser thread to begin advertising.
     *
     * @see Portal#openPortal()
     */
    public synchronized void signalOpenPortal()
    {
        if (PortalAdvertiser.debug)
            config.err.println ("Portal Advertiser: open portal signaled");

        sendAdvertisements = true;
        notifyAll();
    }

    /*
     * This is called from the Portal when the Portal.closePortal() method is
     * invoked. It signals the advertiser thread to stop advertising.
     *
     * @see Portal#closePortal()
     */
    public synchronized void signalClosePortal()
    {
        if (PortalAdvertiser.debug)
            config.err.println ("Portal Advertiser: close portal signaled");

        sendAdvertisements = false;
        notifyAll();
    }

    public synchronized void run()
    {
        try
        {
            if (PortalAdvertiser.debug)
                config.err.println ("Portal Advertiser started");

            // see comment below in the catch
            final int ILLEGAL_STATE_SENDS_BEFORE_THROWING_DEFAULT = 5;
            int illegalStateSendsBeforeThrowing = ILLEGAL_STATE_SENDS_BEFORE_THROWING_DEFAULT;

            while (true)
            {
                // If this thread is running but the socket has not
                // started there is nothing that this thread can do!

                while (! sendAdvertisements
                       ||
                       ! mcSocket.isOpen())
                {
                    if (PortalAdvertiser.debug)
                        config.err.println ("Portal Advertiser: waiting for portal to be opened");

                    try { wait(); } catch (final InterruptedException ignored) { }

                    // A delay for some random amount of time (uniform
                    // between (0,advertisePeriod)) to stagger
                    // advertisements of many portals that are started
                    // simultaneously.

                    final long delayForStaggering = Math.abs (random.nextLong()) % advertisePeriod;
                    //final long delayForStaggering = 0;

                    // Wait at least few seconds to make sure that the
                    // adapters have started. If the sematics of
                    // OL_Socket.joinOverlay are changed so that the
                    // caller can be assured that the adapters are
                    // prepared to receive calls to
                    // sendUnicastMessage, this extra delay can go
                    // away.

                    final long delayForAdapterStart = 2000L;

                    final long delay = Math.max (delayForAdapterStart, delayForStaggering);

                    try { wait (delay); } catch (final InterruptedException ignored) { }
                }

                /*
                 * If the Monitor is not preset and we have not heard from
                 * it for the timeout period, then forget about it.
                 *
                 * Never timeout a Monitor when in silent mode.
                 */
                if (! silentMode
                    &&
                    (monitorAddress != null
                     &&
                     (configuredMonitorAddress == null
                      ||
                      ! monitorAddress.equals (configuredMonitorAddress))))
                {
                    final long currentTime = System.currentTimeMillis();
                    final long timeSinceMostRecentContact = currentTime - mostRecentContactTime;

                    if (monitorAddress != null
                        &&
                        timeSinceMostRecentContact > monitorTimeout)
                    {
                        if (PortalAdvertiser.debug)
                            System.err.println ("PortalAdvertiser: Detected Monitor timeout after " + (timeSinceMostRecentContact / 1000L) +
                                                " seconds for monitor at: " + monitorAddress);

                        monitorAddress = configuredMonitorAddress;
                    }
                }

                /*
                 * Sending of the Advertise message
                 */
                if (! silentMode
                    &&
                    monitorAddress != null)
                {
                    if (PortalAdvertiser.debug)
                        System.err.println ("PortalAdvertiser: Sending AdvertisePortal");

                    try
                    {
                    	mcSocket.sendTo (advertisePortalMessage, monitorAddress); // non-blocking
                        illegalStateSendsBeforeThrowing = ILLEGAL_STATE_SENDS_BEFORE_THROWING_DEFAULT;
                    }
                    catch (final IllegalStateException ise)
                    {
                        // This exception can be thrown if the advertise
                        // message is sent immediately after a call to
                        // openPortal. When openPortal is called it sets a
                        // flag in the MCSocket that indicates that the
                        // Portal has been opened. The problem is that
                        // just because openPortal was called and returns
                        // does not ensure that the adapters of the
                        // olSocket of the mcSocket have been completely
                        // started. In fact, the current (11/04) HyperCast
                        // system does not provide a way for a thread to
                        // wait for this to happen.

                        // Allow this to happen a few times before deciding that it is suspicious
                        --illegalStateSendsBeforeThrowing;

                        if (illegalStateSendsBeforeThrowing == 0)
                            throw new HyperCastFatalRuntimeException (ise);
                    }
                }

                try { wait (advertisePeriod); } catch (final InterruptedException ignored) { }
            }
        }
        catch (final Exception e) // Catch all exceptions here - we do not want this thread to stop advertising.
        {
            System.err.println (Thread.currentThread().getName() + ": Caught exception, continuing to advertise: " + e);
        }
    }

    /**
     * Receives a message from the MCSocket. This message should be an
     * advertise message (either Portal or Monitor) advertise because
     * this object adds itself as a callback for those types of
     * messages.
     */
    public synchronized void receiveMCRawMessage (final MCRawMessage message)
    {
        if (MCSocket.debug)
            System.err.println ("StatsAccessor: Received MCRawMessage: " + message.getType());

        switch (message.getTypeNum())
        {
            case MCRawMessage.ADVERTISE_MONITOR:
                handleAdvertiseMonitor (message);
                break;

            case MCRawMessage.ADVERTISE_PORTAL:
                // We're hearing our own broadcast - that's ok; ignore.
                break;

            default:
                config.log.println ("PortalAdvertiser.receiveMCRawMessage(): not an Advertise message; ignoring: " + message.getType());
        }
    }

    /**
     * When a Portal receives an <code>AdvertiseMonitor</code> message
     * it sets its Monitor Address variable. Only one Monitor, the
     * most recently heard, is kept by a Portal. The time of the
     * receipt of the <code>AdvertiseMonitor</code> message is
     * recorded so that a Portal knows when to stop sending
     * <code>AdvertisePortal</code> messages to a Monitor.
     */
    private void handleAdvertiseMonitor (final MCRawMessage message)
    {
        /*
         * The arrival of a MonitorAdvertise message causes a Portal
         * to become non-silent.
         */
        if (silentMode)
        {
            config.log.println ("SilentMode has been reset.");
            setSilentOff();
        }
		
        final I_LogicalAddress monitorAddress = message.getSender();
        
        if (monitorAddress == null)
        {
            config.err.println ("WARNING: Received AdvertiseMonitor message with bad Sender Address: \"" + monitorAddress + "\"");
            return;
        }

        if (PortalAdvertiser.debug)
            System.err.println ("PortalAdvertiser: Received AdvertiseMonitor message from: " + monitorAddress);

        if (this.monitorAddress == null)
        {
            config.log.println ("Acquired Monitor at: " + monitorAddress);
        }

        this.monitorAddress = monitorAddress;
        mostRecentContactTime = System.currentTimeMillis();
    }
}
