/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import hypercast.adapters.*;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * A class for easily porting old Java Multicast programs to Hypercast.
 * Ideally, this should be constructed by calling 
 * HyperCastConfig.createJavaMulticastSocket(). 
 * <p>
 * NOTE: Some code may need to be modified if it uses functions: 
 * <code>getInterface</code>, <code>getLocalAddress</code>, <code>getLocalPort</code>, 
 * <code>getTTL</code>, <code>setTTL</code>, and <code>setInterface</code>.
 * <p>
 * If you want to skip the nice simple interface given by HyperCastConfig:
 * <p>
 * Constructor for MulticastSocket should be replaced by constructor
 * for a MulticastSocketMasquerade.
 * <p>
 * <code> MulticastSocket sock = new MulticastSocket(); </code>
 * <p>
 * Should become:
 * <p>
 *<code> MulticastSocket sock = null;
 * try { 
 *  sock = new MulticastSocketMasquerade();
 * } catch (Exception ignored) {
 * }
 * sock.setOLSocket( ... ); </code>
 * <p>
 * At which point, the variable <code> sock </code> can be used as normal.
 *
 * @author HyperCast Team
 * @author Michael Nahas
 * @version	2.0, Feb. 26, 2001
 */

public class MulticastSocketMasquerade extends MulticastSocket 
{
    /** The socket that messages should be sent to and gotten from. */
private I_OverlaySocket overlaySocket;

    /** The source address for all messages sent by this socket. */
private INETv4OnePort localAddress;

    /** Port number that should be reported. */
    final static int PORT_NUMBER = 0;
    

/**
 * Create the datagram socket.
 * @exception IOException This should be caught and ignored.
 */
    public MulticastSocketMasquerade() throws IOException
    {
        super();
    }


/** Sets the socket that messages should be sent over and gotten from. 
 * This function must be called immediately after the constructor. 
 */
    public void setOverlaySocket(I_OverlaySocket olSocket) 
    {
       overlaySocket = olSocket;

        // initialize localAddress
        InetAddress addr = null;
        try 
        {
            addr = InetAddress.getLocalHost();
        }
        catch (UnknownHostException uhe) 
        {
            throw new HyperCastFatalRuntimeException ("InetAddress.getLocalHost() threw an UnknownHostException.");
        }
        localAddress = new INETv4OnePort(addr, PORT_NUMBER);
    }

    /**
     * leave the group.
     */
    public synchronized void close() {
        overlaySocket.leaveOverlay();
    }

    /**
     *	returns local IP Address.
     */
    public InetAddress getInterface() throws SocketException {
        return localAddress.getInetAddress();
    }

    /**
     *	returns local IP Address.
     */
    public InetAddress getLocalAddress() {
        return localAddress.getInetAddress();
    }

    /**
     *	Returns designated port number (probably zero).
     */
    public int getLocalPort() {
        return localAddress.getPort();
    }

    /**
     * Returns SoTimeout of Hypercast Socket.
     */
    public int getSoTimeout() {
        return overlaySocket.getSoTimeout();
    }

    /* *
     * Returns TTL of Hypercast Socket
     * @deprecated
     */
//     public byte getTTL() throws IOException {
//         return (byte) overlaySocket.getTTL();
//     }

    /**
     * Joins overlay multicast group.
     */
    public void joinGroup(InetAddress mcastaddr) throws IOException {
        overlaySocket.joinOverlay();
    }


    /**
     * Leaves overlay multicast group.
     */
    public void leaveGroup(InetAddress mcastaddr) throws IOException {
        overlaySocket.leaveOverlay();
    }


    /**
     * Received I_OverlayMessages are translated into DatagramPackets and returned.
     * @exception IOException is never thrown.
     */
    
    public void receive(DatagramPacket p) throws IOException {
        while (true) 
        {
            I_OverlayMessage msg = overlaySocket.receive();
            byte[] payload = msg.getPayload();

            // get PA size
            int paSize = localAddress.getSize();
            
            byte[] tmp = new byte[paSize];
            System.arraycopy(payload, msg.getPayloadLength() - paSize, tmp, 0, paSize);
			INETv4OnePort srcAddr = new INETv4OnePort(tmp);

            if (p.getData() == null || (p.getData().length < msg.getPayloadLength() - paSize)) 
                p.setData(payload);
            else 
                System.arraycopy(payload, 0, p.getData(), 0, msg.getPayloadLength() - paSize);
            p.setLength(msg.getPayloadLength() - paSize);
            p.setAddress(srcAddr.getInetAddress());
            p.setPort(srcAddr.getPort());
            
            return;
        }
    }

    /**
     * Datagram Packet is translated into an I_OverlayMessage and sent.
     * A new byte array is created, so that src IP Address can be stuck at end.
     * @exception IOException is never thrown.
     */
    public void send(DatagramPacket p) throws IOException  {
        byte[] tmp = new byte[p.getLength() + localAddress.getSize()];
        System.arraycopy(p.getData(), 0, tmp, 0, p.getLength());
        System.arraycopy(localAddress.toByteArray(), 0, tmp, p.getLength(), localAddress.getSize());

        I_OverlayMessage msg = overlaySocket.createMessage(tmp);
        overlaySocket.sendToAll(msg);
    }


	/*
	 * Assume ms is an instance of MulticastSocket. Instead of calling this deprecated
	 * method, the application can call ms.setTimeToLive(int TTL) and then call
	 * ms.send(DatagramPacket p).
	 */
    /**
     * Same as Send, but with specific hop limit.
     * @param hopLimit The maximum number of hops allowed for this message in the network
     * @exception IOException is never thrown.
     *
     * There is a compiler warning here because this method has been
     * deprecated. Sun's Javadoc says to use code like this instead:
     *
     * int ttl = mcastSocket.getTimeToLive();
     * mcastSocket.setTimeToLive(newttl);
     * mcastSocket.send(p);
     * mcastSocket.setTimeToLive(ttl);
     */
/*
    public void send(DatagramPacket p, byte hopLimit) throws IOException  {
        byte[] tmp = new byte[p.getLength() + localAddress.getSize()];
        System.arraycopy(p.getData(), 0, tmp, 0, p.getLength());
        System.arraycopy(localAddress.toByteArray(), 0, tmp, p.getLength(), localAddress.getSize());

        I_OverlayMessage msg = overlaySocket.createMessage(tmp);
        msg.setHopLimit (hopLimit);
        overlaySocket.sendToAll(msg);
    }
*/


    /**
     *	ignored.
     */
    public void setInterface(InetAddress inf) throws SocketException {
    }

    /**
     * Sets Hypercast Socket's SoTimeout
     */
    public void setSoTimeout(int timeout) throws SocketException {
        overlaySocket.setSoTimeout(timeout);
    }

    /* *
     * Sets Hypercast Socket's TTL
     * @deprecated
     */
//     public void setTTL(byte ttl) {
//         overlaySocket.setTTL(ttl);
//     }


    /**
     * Ignored.
     */

    protected synchronized void finalize() {
    }
}






