/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.util.Vector;

import javax.crypto.Mac;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.Key;

//import java.io.*;

/** 
 * This class implements the I_OverlayMessage interface and defines the
 * messages exchanged between socket adapters of overlay sockets. It is 
 * used by applications to send data. It is also used to send overlay 
 * messages for monitoring and control objectives. This class is not
 * used in exchanging protocol messages via node adapters.
 * 
 * An overlay message has the following format:
 * ----------------------------     -------------
 * |    Overlay   | Extension | ... | Extension |
 * |message header|           |     |           |
 * ----------------------------     -------------
 *  
 * Here, the format of overlay message header is:
 * (Overlay Message Length and LA Size are in units of bytes)
 * 
 * -------------------------------------------------------------
 * |Version| DMode |Traffic Class|          Flow Label         |
 * | 4 bits| 4 bits|  8 bits     |           16 bits           |
 * -------------------------------------------------------------
 * |  Overlay Message Length     |          Hop Limit          |
 * |         16 bits             |           16 bits           |
 * --------------------------------------------------------------------
 * |Extension Type |   LA Size   |           Source Logical Address   /
 * |   8 bits      |    8 bits   |                 variable          /
 * ---------------------------------------------------------------------------
 * /   Previous Hop Logical Address   /   Destination Logical Address        /
 * /        variable length           / (only if DMode is Unicast, variable) /
 * ---------------------------------------------------------------------------
 * 
 * Each extension has the following format:
 *
 *                      (in bytes)
 * ----------------------------------------------------------------------
 * | Next Ext. Type | Extension Length | Extension Specific Information |
 * |     8 bits     |      16 bits     |             > 0 bytes          |
 * ----------------------------------------------------------------------
 *
 * The extension types are defined in class Extension. @see Extension
 * 
 * 1. There is only one extension with type Payload (RAW_DATA). It carries the application 
 *    payload. It does not have to be the last extension.
 *
 * 2. In an overlay message, each extension, except the Payload extension, contains
 *    plain text information. (TODO: I'm not certain that this is true)
 *
 * 3. In general, there is no assumption on the order of extensions.
 *
 * @author HyperCast Team
 * @author Dongwen Wang 
 * @author Mike Nahas
 * @author Guimin Zhang
 * @author Jianping Wang
 * @author Guangyu Dong
 * @version     3.0, Jul. 11, 2004
 */
public class OL_Message implements I_OverlayMessage, I_Message
//									, Cloneable, Serializable
{
	/**
	 * Debug flag.
	 */
	private static final boolean debug = false;
	
    /**
     * The length of all fields in the header that preceed the first variable
     * length field (source logical address).
     */
    private static final int FIXED_HEADER_FIELD_LENGTH = 10;

    /** Overlay message header version (4 bits). */
    // TODO: this should go somewhere more central perhaps? I_OverlaySocket.java? make it public?
    // TODO: accessed by OL_SEC_Message - does it really need this access?
    protected static final byte HYPERCAST_VERSION = 3 & 0x0F;

    /** Delivery mode (4 bits - low order 4 bits used) */
    protected byte deliveryMode;  

    /*
     * Available constants for deliveryMode. All constants are
     * considered 4 bit quantities.
     *
     * Note that although there is a DELIVERY_MODE_INVALID constant,
     * there are actually many settings for this value that are
     * invalid. Since the field is four bits wide there are 16
     * possible values, all of which are invalid except for the values
     * defined here. The constant DELIVERY_MODE_INVALID is useful
     * because the delivery mode of a message is set to a valid value
     * when the message is sent. After a message is constructed, but
     * before it is sent, the value DELIVERY_MODE_INVALID is used for
     * its delivery mode.
     */
    public static final byte DELIVERY_MODE_INVALID       = 0 & 0x0F;
    public static final byte DELIVERY_MODE_MULTICAST     = 1 & 0x0F;
    public static final byte DELIVERY_MODE_FLOOD         = 2 & 0x0F;
    public static final byte DELIVERY_MODE_UNICAST       = 3 & 0x0F;
			
    /** Traffic class (8 bits) for service differentiation. */
    private byte trafficClass;
	
    /** Flow Label (16 bits) for QOS (not used in current version). */
    private final short flowLabel = 0;
		
    /**
     * Overlay message length (16 bits). This length does not include
     * the length of the overlay message header. This length is the
     * sum of all of the extensions, including their type and length
     * fields.
     */
    private short messageLength; // bytes
	
    /**
     * Hop limit is interpreted as an unsigned 16 bit quantity.  Note
     * that when checking to see if a message should be allowed to go
     * to another hop do not check like this:
     *
     * if (hopLimit > 0)
     *
     * instead do this:
     *
     * if (hopLimit != 0)
     *
     * The reason for this is that negative Java short values actually
     * represent positive hop limit values since this field is
     * interpreted as unsigned by HyperCast.
     *
     * It is not advised that arithmetic be done on this value except
     * for calling the <code>decrementHopLimit</code> method of this class.
     *
     * See HyperCast design document, chapter 5 section 5.2 "Overlay Message"
     * @see #decrementHopLimit()
     */
    private short hopLimit;

    // TODO: enforce that (source|destination|previousHop)address are of the same type?

    /**
     * The address of the node that created this message.
     */
    private I_LogicalAddress sourceAddress;

    /**
     * Destination logical address of this message. It must be null if
     * delivery mode is not DELIVERY_MODE_UNICAST.
     */
    private I_LogicalAddress destinationAddress;  

    /**
     * For a message that was received from the network this value is
     * the logical address of the node that forwarded this message to
     * the local node. For a new message the previous hop address and
     * source address are set to the logical address of the local
     * node.
     */
    private I_LogicalAddress previousHopAddress;  

    /** Header Extension vector  */
    // TODO: should this structure be protected with synchroniziation?
    protected Vector extensionStore;

    /**
     * Stores the certificate and key information for the key neighbors and this
     * node.
     */
    KeyVault vault = null;

    /**
     * Security auciliary computation object.
     */
    SecurityAuxiliaryComputation sac = null;

    /**
     * Create an empty OL_Message with specified hopLimit and 
     * source logical address. 
     *
     * @param hopLimit   Maximum number of hops that message will be allowed to have in network
     * @param sourceAddress        Logical address that will be used as the source of the message
     */
    public OL_Message (final short hopLimit, final I_LogicalAddress sourceAddress)
    {
        this (hopLimit, sourceAddress, null);
    }

	/**
	 * Create an OL_Message by cloning another OL_Message instance.
	 */
	public OL_Message (OL_Message orig_msg)
	{
		this.deliveryMode = orig_msg.deliveryMode;
		//this.flowLabel = orig_msg.flowLabel; //It is final variable.
		this.hopLimit = orig_msg.hopLimit;
		this.messageLength = orig_msg.messageLength;
		this.destinationAddress = orig_msg.destinationAddress;
		this.previousHopAddress = orig_msg.previousHopAddress;
		this.sourceAddress = orig_msg.sourceAddress;
		this.trafficClass = orig_msg.trafficClass;
		this.vault = orig_msg.vault;

		if (vault != null)
		{
			this.sac = new  SecurityAuxiliaryComputation(vault, this);
		}
		
		this.extensionStore = new Vector();
		for (int i = 0; i < orig_msg.extensionStore.size(); i++)
		{
			Extension extension = (Extension) orig_msg.extensionStore.elementAt (i);
			this.extensionStore.addElement(extension.Clone());
		}
	}

	
    /**
     * Create an empty OL_Message with specified hopLimit, source logical address and
     * key vault. 
     *
     * @param hopLimit   Maximum number of hops that message will be allowed to have in network
     * @param sourceAddress        Logical address that will be used as the source of the message
     * @param kv		 an KeyVault instance for security computation 
     */
    public OL_Message (final short hopLimit, final I_LogicalAddress sourceAddress, KeyVault kv)
    {
        if (sourceAddress == null)
            throw new NullPointerException ("Cannot construct OL_Message with null source address");
            
        setHopLimit (hopLimit);

        this.deliveryMode = DELIVERY_MODE_INVALID;

        this.sourceAddress = sourceAddress;

        this.destinationAddress = null;
        this.previousHopAddress = null;

        this.messageLength = 0;
        this.extensionStore = new Vector();		

		/*
		 * Set the reference to key vault. If key vault is not null, which 
		 * means this is a secure overlay message, a auxialiary computation 
		 * object is created to handle security related computation, and 
		 * security extension is added to this message.
		 */
		vault = kv;
		if (vault != null)
		{
			sac = new  SecurityAuxiliaryComputation(vault, this);

			/*
			 * Adds security extension to the OL_SEC_Message.
			 */
			this.addExtension(new SecurityExtension(vault, sourceAddress));

			if (debug)
				System.out.println("OL_Message:OL_Message:Add security extension.");
		}
    }

    /**
     * Constructs an overlay application message from a byte array. The message
     * begins a position <code>offset</code> of the byte array.
     *
     * @param packet the byte array packet 
     * @param offset the start position in the array to create OL_Message
     * @param validBytesLen	length of valid section of the array
     * @param node 	 Overlay Node  
     * @param kv	 an KeyVault instance for security computation 
     * @throws IllegalArgumentException for invalid message version and invalid deliveryMode
     */
    protected OL_Message (final byte [] packet, final int offset, final int validBytesLen, final I_Node node, KeyVault kv) 
    {		
       	int index = offset;		
       	
		/*
		 * Set the reference to key vault.
		 */
		vault = kv;

        /*
         * The next byte contains two fields. The high-order four bits are the
         * version field. The low-order four bits are the delivery mode field.
         */
        final byte version = (byte) (packet[index] >> 4);
        if (version != HYPERCAST_VERSION) 
            throw new IllegalArgumentException ("Illegal version of application message: " + version);

        this.deliveryMode = (byte) (packet[index] & 0x0F);
        if (! isValidDeliveryMode (this.deliveryMode))
            throw new IllegalArgumentException ("Invalid delivery mode: " + this.deliveryMode);

        index += 1; // finished with version and delivery mode
		
        /*
         * The next byte is the traffic class field
         */
        this.trafficClass = packet[index];
        index += 1; // finished with traffic class

        /*
         * The next two bytes are the the flow label which is not used in
         * current version.
         */
        //this.flowLabel = ByteArrayUtility.toShort (packet, index);
        index += 2; // finished with the flow label

        /*
         * The next two bytes encode the length of the message after the
         * header. These bytes are interpretted as an unsigned quantity.
         */
        this.messageLength = ByteArrayUtility.toShort (packet, index);
        index += 2; // finished with the payload length

        /*
         * The next two bytes encode the hop limit for this message.
         */
        this.hopLimit = ByteArrayUtility.toShort (packet, index);
        index += 2; // finished with the hop limit

        /*
         * The next byte encodes the type of the extension header that follows
         * the OL_Message header.
         */
        byte nextHeaderType = packet[index];
        index += 1; // finished with the next header type

        /*
         * The next byte encodes the size of the variable length logical
         * address fields for source, destination, and previous hop addresses.
         * The logical address size field is interpretted as unsigned, so to
         * make a value that can be used for arithmetic here we must be careful
         * to interpret negative byte values as positive int values.
         */
        final int logicalAddressSize = ByteArrayUtility.toUnsignedInteger (packet, index, 1);
        index += 1; // finished with logical address size

        /*
         * The next two fields are source address and previous hop
         * address which are always present in overlay messages.
         */
        this.sourceAddress = node.createLogicalAddress (packet, index);
       	index += logicalAddressSize; // finished with the source address

        this.previousHopAddress = node.createLogicalAddress (packet, index);
        index += logicalAddressSize; // finished with the previous hop address

        /*
         * Only for an unicast overlay message, there is destination
         * address in the packet.
         */
        if (deliveryMode == DELIVERY_MODE_UNICAST)
        {
            this.destinationAddress = node.createLogicalAddress (packet, index);
            index += logicalAddressSize; // finished with the destination address
        }

        /*
         * In the following loop, all extensions are reconstructed and 
         * put into vector extensionStore.
         */
        this.extensionStore = new Vector();
		
        /*
         * In the byte array packet of an overlay message, only the last
         * extension has value HEADER_NONE in its Next Extension field.
         */
        final int extensionBeginIndex = index; // used to ensure that extensions are within message length bound
       	while (nextHeaderType != Extension.HEADER_NONE)
       	{
            final byte currentHeaderType = nextHeaderType;

            nextHeaderType = packet[index++];

            // two byte field
            final int extensionPayloadLength = ByteArrayUtility.toUnsignedInteger (packet, index, 2);
            index +=2;

            // useful for computation where 16 bits of short could overflow and become negative.
            final int messageLengthInt = (this.messageLength & 0xFF00) | (this.messageLength & 0x00FF);

            /*
            System.err.println ("OL_Message.<init>: All Extensions Length: " + messageLengthInt +
                                ", used: " + ((index-3) - extensionBeginIndex) +
                                ", remaining: " + (messageLengthInt - ((index-3) - extensionBeginIndex)) +
                                "; Extension type: " + Extension.typeToString (currentHeaderType) +
                                "; Extension length: " + (extensionPayloadLength + 3));
            */

            byte [] extensionPayload = new byte [extensionPayloadLength];
            System.arraycopy (packet, index, extensionPayload, 0, extensionPayloadLength);
            index += extensionPayloadLength;

            if (index - extensionBeginIndex > messageLengthInt) // must use int, not short to avoid signed overflow miscalculation
                throw new HyperCastWarningRuntimeException ("Could not reconstruct message because length of extensions exceeded message length given in header.");

            Extension extension = Extension.createExtension (currentHeaderType, extensionPayload, node, vault, vault != null);
            if (extension != null)
            {
                extensionStore.addElement (extension);
            }
            else
            {
                throw new HyperCastWarningRuntimeException ("Failed to reconstruct extension with type: " + currentHeaderType);
            }
       	}

		/*
		 * If key vault is not null, which means this is a secure overlay message, a auxialiary computation 
		 * object is created to handle security related computation. Andthe check for sequence number, 
		 * header Mac and payload Mac are done here.
		 */
		if (vault != null)
		{
			/*
			 * Create a byte array which only contains valid data of the received packet.
			 */
			byte[] validData = new byte[validBytesLen];
			System.arraycopy(packet, offset, validData, 0, validData.length);

			/*
			 * Create security auxiliary computation object.
			 */
			sac = new  SecurityAuxiliaryComputation(kv, this);
				

			/*
			 * Get the sender of this message and its personal key.
			 */
			SecurityExtension secExt = (SecurityExtension) this.getFirstExtensionByType(Extension.SECURITY);

			if (secExt == null)
			{
				throw new HyperCastWarningRuntimeException ("Couldn't check header MAC for a plaintext overlay message.");
			}

			/*
			 * Get the logical address of the sender of this message and its personal key.
			 */
			I_LogicalAddress prevSender = secExt.getSenderLA();

			/*
			 * If the sequence number of this message is invalid, throw exception. 
			 */
			//System.out.println("Received overlay message from "+prevSender.toString());

			if (!vault.checkSequenceNum(prevSender, secExt.getSequenceNumber())) {
				throw new IllegalArgumentException("Invalid sequence number!");
			}

			/*
			 * It is set when calling toByteArray().
			 */
			/*
			 * After the sequence number checking is passed, reassign a new sequence number
			 * for this message. The sequence number checking works in hop-by-hop mode.
			 */
			//secExt.setSequenceNumber(vault.getNextSequenceNum());

			/*
			 * Check if the header Mac field (hmac) carried in this message is correct.
			 */
			if (!sac.checkHMacComputation(validData, prevSender))
			{
				throw new IllegalArgumentException("Message authentication check failure");
			}

			/*
			 * If no Payload Extension, return; Otherwise, do payload Mac check.
			 */
			PayloadExtension payloadExtension = (PayloadExtension) this.getFirstExtensionByType (Extension.RAW_DATA);
			if (payloadExtension == null)
			{
				return;
			}

			/*
			 * Check if the payload Mac field (mac) carried in this message is correct.
			 */
			if (!sac.checkMacComputation())
			{
				throw new IllegalArgumentException("Payload authentication check failure");
			}

			/*
			 * If this message carries encrypted payload in the packet, the message key
			 * is copied from security extension and stored in the payload extension.
			 */
			if (payloadExtension instanceof EncryptedPayloadExtension)
			{
				Key msgkey = secExt.getMsgKey();
				if (msgkey == null)
					System.out.println("OL_Message: message key is null!");
				
				/*
				 * Set the message key in payload extension. The message key is needed
				 * when doing encryption or decryption on the layload.
				 */
				((EncryptedPayloadExtension) payloadExtension).setMessageKey(msgkey);
			}
		}

    }

    /**
     * Restore an OL_Message from an byte array. Actually, this
     * code validates that a complete message is inside the buffer and
     * then calls the constructor to do the actual reconstruction.
     *
     * @param	recvbuff		the byte array buffer
     *
     * @param validBytesStart start of valid section of the array,
     * this variable is essentially a pointer to an int, the caller of
     * this method expects that the value will be updated to reflect
     * the bytes of the buffer consumed by this method, i.e. on return
     * the value of this variable should be set to the next valid byte
     * after the message that has been restored.
     *
     * @param	validBytesEnd	   end of valid section of the array (actually, 1 beyond last valid byte.)
     * @param	node			an I_Node object which is used to get LogicalAddress size 
     * @param	kv			an KeyVault instance for security computation 
     * @exception	throw IllegalArgumentException for invalid version and deliveryMode
     * @return an object of OL_Message 
     */
    public static OL_Message restoreMessage (final byte [] recvbuff, final int [] validBytesStart, 
    								final int validBytesEnd, final I_Node node, KeyVault kv)
    {	
        final int validBuffLength = validBytesEnd - validBytesStart[0];

        /* 
         * For a TCP connection, the receive() method in IncomingSocketObject
         * call OL_Message.restoreMessage(...) with a piece of data. If the
         * length of the received data is greater than 0 but less than
         * FIXED_HEADER_FIELD_LENGTH, the result of this method is
         * uncertain. In that case, do nothing and return null, the TCP adapter
         * will call this method again when receive more data.
         */
        if (validBuffLength < FIXED_HEADER_FIELD_LENGTH)
        {
            return null;
        }

        /*
         * Check if the Version field contains correct version value.
         */
        byte versionPack = (byte) (recvbuff[validBytesStart[0]] >> 4);
        if (versionPack != HYPERCAST_VERSION) 
            throw new IllegalArgumentException ("OL_Message: Illegal version of application message: " + versionPack);

        /*
         * Check if the DMode field contains valid delivery mode value.
         */
        byte deliveryModePack = (byte) (recvbuff[validBytesStart[0]] & 0x0f);  //  TODO: replace number by a proper constant 

        if (! isValidDeliveryMode (deliveryModePack))
        {
            throw new IllegalArgumentException ("OL_Message: Invalid delivery mode: " + deliveryModePack);
        }
		
        /*
         * Get the payload length, i.e., the length of the parts after the fixed header.
         */
        byte[] len = new byte[4];
        System.arraycopy(recvbuff,validBytesStart[0]+ 4,len,2,2);
        int length = ByteArrayUtility.toInteger(len,0);

        /*
         * Get the logical address size that is the 10th byte in the fixed header.
         */
        final int lasize = (int) recvbuff[validBytesStart[0]+9]; // in bytes	

        /*
         * In the following, the length of the fixed header is calculated.
         */
        int mandatoryHeaderLength;
			
        /*
         * The length of source + previous hop logical address are added.
         */
        mandatoryHeaderLength = FIXED_HEADER_FIELD_LENGTH + 2*lasize; 
		
        /*
         * If it is an unicast message, the length of destination address is added.
         */
        if (deliveryModePack == OL_Message.DELIVERY_MODE_UNICAST)
        {
            mandatoryHeaderLength += lasize;
        }

        /*
         * The expected length of this packet is the sum of fixed header length 
         * (mandatoryHeaderLength) and the length of the parts following the fixed 
         * header (length).
         */
        int expectedLength = mandatoryHeaderLength + length;

        /*
         * If the actual length of this packet is less than the expected length,
         * no overlay message is reconstructed. Return null.
         */				
        if (validBuffLength < expectedLength) 
        {
            return null;
        }

        /*
         * If the actual length of this packet matchs the expected length,
         * an overlay message is reconstructed.
         */					
        OL_Message msg = new OL_Message (recvbuff, validBytesStart[0], expectedLength, node, kv);						
						
        /*
         * The start point in the receive buffer is moved with length expectedLength.
         */
        validBytesStart[0] += expectedLength;

        return msg;		
    }

	/**
	 * Increases payload length by given value.
	 * 
	 * @param delta		the amount that the payload length increases.
	 */
    private void increaseMessageLength (int delta)
    {
        this.messageLength += delta; // TODO: ensure correct arithmetic, messageLength is interpretted as unsigned short
    }

    /**
     * Convert this message into a byte array. 
     * @return a byte array representation of this message
     * @throws IllegalStateException if souce address or destination address are null
     */
    public byte [] toByteArray()
    {
        if (sourceAddress == null)
            throw new IllegalStateException ("Source Address cannot be null");

        if (previousHopAddress == null)
            throw new IllegalStateException ("Previous Hop Address cannot be null");

        /*
         * For an unicast message, its destination address shouldn't be null.
         */
        if (deliveryMode == DELIVERY_MODE_UNICAST
            &&
            destinationAddress == null)
        {
            throw new IllegalStateException ("Destination of a unicast message cannot be null");
        }
		
        /*
         * If delivery mode is not set to correct value, i.e. unicast, multicast or flood, this
         * overlay message cannot be sent out.
         */
        if (! isValidDeliveryMode (deliveryMode))
            throw new IllegalStateException ("Delivery mode is not valid");



		if ((vault != null) && (!vault.getSecPolicy().equals("plaintext")))
		{
			/*
			 * Get the security extension in this message.
			 */
			SecurityExtension secExtension = (SecurityExtension) this.getFirstExtensionByType(Extension.SECURITY);

			if (secExtension != null)
			{
	
				/*
				 * When a security extension is created, its payload MAC computation
				 * object is only initialized. The payload MAC calculation is posponed
				 * to here. If this message carries encrypted payload in the packet, the
				 * payload extension is a instance of EncryptedPayloadExtension, the call 
				 * payloadExtension.toByteArray() below returns encrypted byte array,
				 * which may first execute encryption on the payload. 
				 * See EncryptedPayloadExtension.toByteArray() for more information.
				 */
				PayloadExtension payloadExtension = (PayloadExtension) this.getFirstExtensionByType (Extension.RAW_DATA);

				if (payloadExtension != null)
				{
					if (payloadExtension instanceof EncryptedPayloadExtension)
					{
						/*
						 * Set the message key in payload extension. The message key is needed
						 * when doing encryption or decryption on the layload.
						 */
						((EncryptedPayloadExtension) payloadExtension).setMessageKey(secExtension.getMsgKey());
					}

					//System.out.println("OL_Message:toByteArray(): recalculate mac");				
					secExtension.updateMacCompute(payloadExtension.toByteArray());
				}
				else
				{
					secExtension.updateMacCompute(null);
				}
				
				/*
				 * In all sending methods, the source address of the message 
				 * is reset. So the sender address carried in the security 
				 * extension also needs to be reset.
				 */
				secExtension.setSenderLA(this.getPreviousHopAddress());
				
				/*
				 * Reset the sequential number of this message.
				 */
				secExtension.setSequenceNumber(vault.getNextSequenceNum());
			}
			else
			{
				throw new HyperCastWarningRuntimeException ("No security extension is found in a secure overlay message.");
			}
		}

		
        int length = FIXED_HEADER_FIELD_LENGTH;
        
        /*
         * Convert source logical address to byte array.
         */
        byte[] src_addr_array = sourceAddress.toByteArray();
        length += src_addr_array.length;

        /*
         * Convert previous hop logical address to byte array.
         */
        byte[] prevhop_addr_array = previousHopAddress.toByteArray();
        length += prevhop_addr_array.length;

        /*
         * If it is an unicast message, convert destination logical 
         * address to byte array.
         */
        byte[] destination_addr_array = null;
        if (deliveryMode == DELIVERY_MODE_UNICAST)
        {
            destination_addr_array = destinationAddress.toByteArray();
            length += destination_addr_array.length;
        }

        /*
         * In the following, payload length is recalculated because some 
         * extensions may be added, changed or deleted.
         */

        /*
         * Reset payload length to 0.
         */
        messageLength = 0;
		 
        /*
         * Convert extensions to byte arrays.
         */
        byte[][] extensionByteArray = new byte[extensionStore.size()][];
        byte[] extensiontype = new byte[extensionStore.size()];
        for (int i = 0; i < extensionStore.size(); i++) 
        {
            Extension extension = (Extension) extensionStore.elementAt(i);
            extensionByteArray[i] = extension.toByteArray();
            extensiontype[i] = extension.getExtensionType();
			
            /*
             * Add the length of extension byte array and 3 bytes (1 for 
             * type field and 2 for length field) to total packet length.
             */
            length += extensionByteArray[i].length + 3;    // TODO: replace number by a proper constant 

            /*
             * Add the length of extension byte array and 3 bytes (1 for 
             * type field and 2 for length field) to payload length.
             */
            increaseMessageLength (extensionByteArray[i].length + 3);
        }
 
        byte[] Result = new byte[length];
        int index = 0;

        /*
         * Fill in the first FIXED_HEADER_FIELD_LENGTH bytes of the packet. 
         */

        /*
         * Version and DMode.
         */
        Result[index+0] = (byte) ((HYPERCAST_VERSION << 4) | (deliveryMode & 0xf));

        /*
         * Traffic class.
         */
        Result[index+1] = trafficClass;
		
        /*
         * Skip two bytes for flow label.
         */
        /*
         * payload length, i.e., the length of the parts following the fixed header.
         */
        Result[index+4] = (byte) ((messageLength >>> 8) & 0xff);
        Result[index+5] = (byte) (messageLength & 0xff);

        /*
         * Hop limit.
         */
        Result[index+6] = (byte) ((hopLimit >>> 8) & 0xff);
        Result[index+7] = (byte) (hopLimit & 0xff);

        /*
         * Next extension type. It is the type of the extension following the fixed 
         * header, i.e., the first extension in extensionStore. If extensionStore
         * is empty, the Next extension type field in the fixed header is set to
         * HEADER_NONE.
         */
        byte NextHeader = ((extensionStore.size() == 0) ? Extension.HEADER_NONE : extensiontype[0]);
        Result[index+8] = NextHeader;

        /*
         * Logical address size.
         */
        Result[index+9] = (byte) sourceAddress.getSize();

        index += FIXED_HEADER_FIELD_LENGTH;

        /*
         * Copy the byte array of source logical address into packet.
         */
        System.arraycopy(src_addr_array, 0, Result, index, sourceAddress.getSize());
        index += src_addr_array.length;

        /*
         * Copy the byte array of previous hop logical address into packet.
         */
        System.arraycopy(prevhop_addr_array, 0, Result, index, sourceAddress.getSize());
        index += prevhop_addr_array.length;

        /*
         * If it is an unicast message, copy the byte array of destination 
         * logical address into packet.
         */
        if (deliveryMode == DELIVERY_MODE_UNICAST)
        {
            System.arraycopy(destination_addr_array, 0, Result, index, sourceAddress.getSize());
            index += destination_addr_array.length;
        }

        /*
         * Create extension parts of the packet.
         */
        for (int i = 0; i < extensionStore.size(); i++)
        {
            /*
             * Extension type
             */
            Result[index+0] = ((i+1 == extensionStore.size()) ? Extension.HEADER_NONE : extensiontype[i+1]);
            /*
             * Extension payload length
             */
            Result[index+1] = (byte) ((extensionByteArray[i].length >>> 8) & 0xff);
            Result[index+2] = (byte) (extensionByteArray[i].length & 0xff);
            index += 3;
			
            /*
             * Copy the byte array of this extension into the packet.
             */
            System.arraycopy(extensionByteArray[i], 0, Result, index, extensionByteArray[i].length);
            index += extensionByteArray[i].length;
        }
		
		/*
		 * If this message is a secure message, the hmac field in the packet needs 
		 * to be updated, because it is calculated on the fixed header and all 
		 * extensions (in byte array format), except the fields for hmac and payload, 
		 * which are just created above. 
		 */
		if ((vault != null) && (!vault.getSecPolicy().equals("plaintext")))
		{
			sac.updateHMac(Result);
		}		
		
        return Result;
    }

    /**
     * Validates the value of a delivery mode.
     *
     * @return true if the value of mode is one of the supported modes of delivery, false otherwise
     */
    public static boolean isValidDeliveryMode (final byte mode)
    {
        /*
         * Check to see if 4 high order bits are set. Since mode is a
         * 4 bit quantity these bits must be: 0000
         */
        if ((mode & 0xF0) != (byte) 0)
            return false;

        switch (mode)
        {
            case DELIVERY_MODE_MULTICAST:
            case DELIVERY_MODE_FLOOD:
            case DELIVERY_MODE_UNICAST:
                return true;
            default:
                return false;
        }
    }

    // TODO: "size" and "length" are used interchangably; pick one and stick to it consistently
    /* 
     * This method provides a way to get the length of this message 
     * without executing toByteArray().
     *
     * @return the length of the message when converted into byte array. 
     */	
    public int getSize()
    {
        int length = 0;
        
        /*
         * This if statement looks troubling to me. I understand that
         * it is possible using the API for an applications programmer
         * to create a message and then ask for its size -- all of
         * that happening before the message is used as a parameter to
         * one of the send methods. Before use in a send, a message
         * will have a delivery mode of invalid, so it is clearly
         * possible to call getSize correctly on such a method. That
         * said, I don't think that there should ever be a message
         * with delivery mode unicast and no destination address,
         * maybe I'm wrong about this... to sum up, basically it seems
         * to me that the first test in this if should signal an error
         * if the test fails, the third test might be ok given the
         * fact that a message can have getSize called on it before
         * that message is ever sent. I don't fully understand the
         * second test, does previousHopAddress == null imply that
         * this message was created locally, and that no locally
         * created messages can have their sizes calculated?
         */

        if ((deliveryMode == DELIVERY_MODE_UNICAST && destinationAddress == null) // TODO: isn't this an invalid state? throw exception?
            || 
            previousHopAddress == null
            ||
            ! isValidDeliveryMode (deliveryMode)) // TODO: isn't this an invalid state? (by definition) throw exception?
        {
            return length;
        }
		
        /*
         * Add the length of the fields that preceed the first variable length field
         */
        length += FIXED_HEADER_FIELD_LENGTH;
        
        /* 
         * Add the length of source logical addresses.
         */
        length += sourceAddress.getSize();

        /* 
         * Add the length of previous hop logical addresses.
         */
        length += previousHopAddress.getSize();

        /* 
         * If it is an unicast message, add the length of destination logical addresses.
         */
        if (deliveryMode == DELIVERY_MODE_UNICAST)
        {
            length += destinationAddress.getSize();
        }

        /* 
         * Add the length of all extensions.
         */
        for (int i = 0; i < extensionStore.size(); i++)
        {
            Extension extension = (Extension) extensionStore.elementAt (i);
            length += 3 + extension.getSize();    // TODO:  replace number by a proper constant 
        }
        
        return length;
    }

	
    /** 
     * @return The hop limit of the message (unsigned 16 bit quantity).
     */	
    public short getHopLimit()
    {
        return hopLimit;
    }

    /** 
     * Decrements the hop limit by one and returns the decremented
     * value. To perform this operation, Java 32 bit integers are used
     * internally when operating on the 16-bit unsigned quantity
     * because Java does not support an unsigned 16-bit quantity.
     *
     * @return The hop limit of the message after decrementing it by one (unsigned 16 bit quantity).
     *         Note that any value other than zero is a <em>positive</em> value because this value is <em>not</em> interpreted as a signed quantity.
     *         Do compare this value to zero, e.g. <code>if (msg.decrementHopLimit() != 0)</code>. Do <em>not</em>
     *         use greater than or less than in comparisons, e.g. <code>if (msg.decrementHopLimit() &gt; 0)</code>.
     *         The latter comparison will fail when the hop limit is large because a large hop limit is represented as a negative Java short.
     *
     * @throws HyperCastWarningRuntimeException if the current hop limit is zero
     */	
    public short decrementHopLimit()
    {
        if (this.hopLimit == 0)
            throw new HyperCastWarningRuntimeException ("Cannot decrement zero hop limit value");

        /*
         * Map from 16 bit space to 32 bit space, interpreting all bits as unsigned.
         * Simple casting won't work because highest-order bit will be sign-extended.
         */
        int largerScratchSpace = (hopLimit & 0xFF00) | (hopLimit & 0x00FF);

        /*
         * Do arithmetic in 32 bit space. Java has 32 bit signed
         * integers, but since we are only using the lowest 16 bits,
         * the sign bit never comes into play, i.e. all results are
         * guaranteed to be non-negative (we have already checked for
         * zero above).
         */
        --largerScratchSpace;

        /*
         * Map from 32 bit space to 16 bit space; Java will interpret
         * the high bit as a sign bit, but all arithmetic on this
         * quantity is done in this method so it will always be
         * interpreted by the HyperCast system as unsigned.
         */
        short result = (short) largerScratchSpace;
        this.hopLimit = result;
        return result;
    }

    /**
     * Sets the hop limit of this message.
     * Application can not call this function to set hop limit of message directly.
     * This function can ONLY be called by a OL_Socket.
     *
     * All values of a short are valid. The short is interpreted as
     * an unsigned 16 bit quantity, e.g. a value of -1 (0xFFFF) means a hop
     * limit of (2^16)-1.
     */
    public void setHopLimit (final short hopLimit)
    {
        this.hopLimit = hopLimit;
    }

    /**
     * Returns the length of application payload. It is the length of the application
     * data (byte array) in the extension with type RAW_DATA.
     */				
    public int getPayloadLength() 
    {
        PayloadExtension payloadExtension = (PayloadExtension) this.getFirstExtensionByType (Extension.RAW_DATA);
        return (payloadExtension == null) ? 0 : payloadExtension.getPayload().length;
    }
	
    /**
     * Returns the plaintext application payload.
     */
    public byte [] getPayload()
    {
        PayloadExtension payloadExtension = (PayloadExtension) this.getFirstExtensionByType (Extension.RAW_DATA);
        return (payloadExtension == null) ? null : payloadExtension.getPayload();
    }
	
    /**
     * Sets the payload.
     */
    public void setPayload (final byte [] newPayload)
    {
        for (int i = 0; i < extensionStore.size(); i++)
        {
            Extension extension = (Extension) extensionStore.elementAt(i);
            if (extension.getExtensionType() == Extension.RAW_DATA)
            {
                PayloadExtension payloadExtension = (PayloadExtension) extension;
                payloadExtension.setPayload (newPayload);
                extensionStore.setElementAt (payloadExtension, i);

                return;
            }
        }
        throw new HyperCastWarningRuntimeException ("OL_Message: no extension with type RAW_DATA is found.");
    }	
		
    /** 
     * Returns the version of the message.
     */	
    public byte getVersion()
    {
        return HYPERCAST_VERSION;
    }
	
    /** 
     * Returns the delivery mode of the message
     */
    public byte getDeliveryMode()
    {
        return deliveryMode;
    }

    // TODO: enforce that if deliveryMode == UNICAST then destination logical address must not be null?
    /**
     * Sets the delivery mode in this message
     * @param		deliveryMode	delivery mode 
     * @exception	throw IllegalArgumentException for invalid delivery mode
     */
    public void setDeliveryMode (final byte deliveryMode)
    {
        if (! isValidDeliveryMode (deliveryMode))
            throw new IllegalArgumentException ("OL_Message: Invald delivery mode: " + deliveryMode);

        this.deliveryMode = deliveryMode;
    }

    /** 
     * Returns the traffice class of the message
     */
    public byte getTrafficClass()
    {
        return trafficClass;
    }

    /** 
     * Sets the traffic class of the message. 
     */	
    public void setTrafficClass (final byte trafficClass)
    {
        this.trafficClass = trafficClass;
    }

    /**
     * Returns the unique identifier of this message.
     * 
     * @deprecated use getMessageIdentifier
     */
    public byte [] getMessageID()
    {
        return getMessageIdentifier();
    }
 	
    /**
     * Returns the service number carried by this message.
     * 
     * @deprecated use getServiceIdentifier
     */
    public short getServiceNumber()
    {
        return this.getServiceIdentifier();
    }

	/**
	 * Returns the service number carried by this message.
	 * 
     * @throws HyperCastWarningRuntimeException if the message is not an enhanced service message.
	 */
	public short getServiceIdentifier()
	{
		FSM_Extension fsm_extension = (FSM_Extension) getFirstExtensionByType(Extension.FSM);
		
		if (fsm_extension == null)
		{
			throw new HyperCastWarningRuntimeException ("Cannot get service identifier from a non-enhanced service message!");
		}
		else
		{
			return fsm_extension.getFSMID();
		}
	}

	/**
	 * Sets the service number in the enhanced service message.
	 * 
	 * @param serviceID	service identifier
	 * 
     * @throws HyperCastWarningRuntimeException if the message is not an enhanced service message ,
     * 	or setting new service identifier will change the message type.
	 */
	public void setServiceIdentifier(short serviceID)
	{
		FSM_Extension fsm_extension = (FSM_Extension) getFirstExtensionByType(Extension.FSM);
		
		short newMsgType = MessageStore.getMsgType(serviceID);
		short oldMsgType = MessageStore.getMsgType(fsm_extension.getFSMID());
		
		if ((fsm_extension == null) || (oldMsgType == MessageStore.OTHER_MESSAGE))
		{
			throw new HyperCastWarningRuntimeException ("Cannot set service ID to a non-enhanced service message!");
		}
		/*
		 * If setting service identifier changes the message type, i.e., from STREAM_MESSAGE to MESSAGESTORE_MESSAGE
		 * or from MESSAGESTORE_MESSAGE to STREAM_MESSAGE, it is inhibited, because different type messages have different
		 * formats.
		 */
		else if (newMsgType != oldMsgType)
		{
			throw new HyperCastWarningRuntimeException ("Setting new service ID changes the message type. It is not allowed!");
		}
		else
		{
			fsm_extension.setFSMID(serviceID);
		}
	}

 	
    /**
     * @return the stream identifier carried by this message.
     * 
     * @deprecated use getStreamIdentifier
     */
    public byte [] getStreamID()
    {
        return ByteArrayUtility.toByteArray(this.getStreamIdentifier());
    }

 	
	/**
	 * Returns the stream identifier carried by this message. It returns -1
	 * if the message is not a enhanced service message.
	 * 
     * @throws HyperCastWarningRuntimeException if the message is not a stream oriented message.
	 * 
	 */
	//public byte [] getStreamIdentifier()
	public int getStreamIdentifier()
	{
		final FSM_Extension fsmExtension = (FSM_Extension) getFirstExtensionByType (Extension.FSM);
		
		if ((fsmExtension == null) || (MessageStore.getMsgType(fsmExtension.getFSMID()) != MessageStore.STREAM_MESSAGE))
		{
			throw new HyperCastWarningRuntimeException ("Non-stream oriented message does not contain stream id information!");
		}
		else
		{
			//return fsmExtension.getUniqueID();
			return ByteArrayUtility.toInteger(fsmExtension.getUniqueID(),0);
		}
	}

	/**
	 * Sets the stream identifier in this message if the message is an enhanced service message.
	 * 
	 * @param streamID	stream identifier in byte array format
	 * 
     * @throws HyperCastWarningRuntimeException if the message is not a stream oriented message.
	 */
	//public void setStreamIdentifier(byte[] streamID)
	public void setStreamIdentifier(int streamID)
	{
		final FSM_Extension fsmExtension = (FSM_Extension) getFirstExtensionByType (Extension.FSM);
		
		if ((fsmExtension == null) || (MessageStore.getMsgType(fsmExtension.getFSMID()) != MessageStore.STREAM_MESSAGE))
		{
			throw new HyperCastWarningRuntimeException ("Cannot set stream ID to a non-stream oriented message!");
		}
		else
		{
			//fsmExtension.setUniqueID(streamID);
			fsmExtension.setUniqueID(ByteArrayUtility.toByteArray(streamID));
		}
	}

 	
	/**
	 * Returns the sequence number carried by this message.
	 * 
     * @throws HyperCastWarningRuntimeException if the message is not a stream oriented message.
	 */
	public long getSequenceNumber()
	{
		final FSM_Extension fsmExtension = (FSM_Extension) getFirstExtensionByType (Extension.FSM);
		
		if ((fsmExtension == null) || (MessageStore.getMsgType(fsmExtension.getFSMID()) != MessageStore.STREAM_MESSAGE))
		{
			throw new HyperCastWarningRuntimeException ("Non-stream oriented message does not contain sequence number information!");
		}
		else
		{
			return fsmExtension.getSequenceNumber();
		}
	}

	/**
	 * Sets the sequence number in this message if the message is an stream-oriented message.
	 * 
	 * @param sequenceNum	the sequence number
	 * 
     * @throws HyperCastWarningRuntimeException if the message is not a stream oriented message.
	 */
	public void setSequenceNumber(long sequenceNum)
	{
		final FSM_Extension fsmExtension = (FSM_Extension) getFirstExtensionByType (Extension.FSM);
		
		if ((fsmExtension == null) || (MessageStore.getMsgType(fsmExtension.getFSMID()) != MessageStore.STREAM_MESSAGE))
		{
			throw new HyperCastWarningRuntimeException ("Cannot set sequence number to a non-stream oriented message!");
		}
		else
		{
			fsmExtension.setSequenceNumber(sequenceNum);
		}
	}

    /**
     * Returns the message identifier carried by this message. Message
     * identifier is defined only when the message has a message store
     * (fsm/enhanced service) extension.
     * 
     * @throws HyperCastWarningRuntimeException if the message is not a messagestroe oriented message.
     */
    public byte [] getMessageIdentifier()
    {
        final FSM_Extension fsmExtension = (FSM_Extension) getFirstExtensionByType (Extension.FSM);
		
        if (fsmExtension == null
            ||
            MessageStore.getMsgType (fsmExtension.getFSMID()) != MessageStore.MESSAGESTORE_MESSAGE)
        {
            throw new HyperCastWarningRuntimeException ("Non-messagestore oriented message does not contain message id information!");
        }
        else
        {
            final byte [] uniqueId = fsmExtension.getUniqueID();
            return uniqueId;
        }
    }

	/**
	 * Sets the message identifier in this message if the message is an enhanced service message.
	 * 
	 * @param msgID	the message identifier
	 * 
     * @throws HyperCastWarningRuntimeException if the message is not a messagestroe oriented message.
	 */
	//public void setMsgIdentifier(byte[] msgID)
	public void setMsgIdentifier(int msgID)
	{
		final FSM_Extension fsmExtension = (FSM_Extension) getFirstExtensionByType (Extension.FSM);
		
		if ((fsmExtension == null) || (MessageStore.getMsgType(fsmExtension.getFSMID()) != MessageStore.MESSAGESTORE_MESSAGE))
		{
			throw new HyperCastWarningRuntimeException ("Cannot set message ID to a non-messagestore oriented message!");
		}
		else
		{
			//fsmExtension.setUniqueID(msgID);
			fsmExtension.setUniqueID(ByteArrayUtility.toByteArray(msgID));
		}
	}


    /**
     * Add a extension to the Overlay application message. <p>
     * 
     * @param	extension	the extension to be added into extension vector.
     * @exception	throw IllegalArgumentException if the type of the extension is not supported.
     */
    public void addExtension (final Extension extension)
    {
        if (! Extension.isValidHeader (extension.getExtensionType()))
        {
            throw new IllegalArgumentException ("OL_Message: Extension type is not supported!" + extension.getExtensionType());
        }
        extensionStore.addElement(extension);
    }

    /**
     * Returns the first extension with specified extension type.
     */
    public Extension getFirstExtensionByType (final byte type)
    {
        Extension result = null;
        for (int i = 0; i < extensionStore.size(); i++)
        {
            Extension extension = (Extension) extensionStore.elementAt (i);
            if (extension.getExtensionType() == type)
            {
                result = extension;
                break;
            }
        }
        return result;
    }

    // TODO: provide better interface for removing extensions, see comment in MessageStoreFSM_Naming

    /**
     * Sets the first extension with specified extension type.
     */
    public void setFirstExtensionByType (final int type, final Extension ee)
    {
        /*
         * If an extension is found which has the specified type, 
         * replace it with parameter extension.
         */
        for (int i = 0; i < extensionStore.size(); ++i)
        {
            Extension e = (Extension) extensionStore.elementAt (i);
            if(e.getExtensionType() == type)
            {
                extensionStore.setElementAt (ee, i);
                return;
            }
        }
		
        /*
         * If no extension is replaced, add the parameter extension
         * to extensionStore.
         */
        addExtension (ee);
    }

    /**
     * Gets the extension store (vector) in this message
     */
    // TODO: if this method gets the entire extension store, not just one extension, why isn't it called getExtensionStore?
    public Vector getExtension()
    {
        return extensionStore;
    }

    /**
     * Gets the extension at specified position in the extension store (vector) of this message.
     */
    public Extension getExtension (final int index)
    {
        return (extensionStore == null) ? null : (Extension) extensionStore.elementAt (index);
    }


    /**
     * Sets the extension at specified position in the extension store (vector) of this message.
     */
    public void setExtension (final int index, final Extension e)
    {
        if (! Extension.isValidHeader (e.getExtensionType()))
        {
            throw new IllegalArgumentException("OL_Message: Extension header not supported!!!"+e.getExtensionType());
        }
	
        if (index < 0
            ||
            index > extensionStore.size ())
        {
            addExtension (e);
        }
        else
        {
            extensionStore.setElementAt (e, index);
        }
    }


    /** 
     * Accesses the flow label. NB The flow label is not an unused
     * field in this version of HyperCast.
     *
     * @return the flow label of the message.
     */
    public byte getFlowLabel()
    {
        return flowLabel;
    }

    /*
     * Source Address get/set
     */
	
    /**
     * Gets the source logical address in this message
     *
     * @deprecated name has changed from Src Address to Source Address, use getSourceAddress
     */
    public I_LogicalAddress getSrcAddress()
    {
        return getSourceAddress();
    }

    /**
     * @return the source logical address of this message.
     */
    public I_LogicalAddress getSourceAddress()
    {
        return sourceAddress;
    }

    /**
     * Sets the logical address in this message
     *
     * @deprecated name has changed from Src Address to Source Address, use setSourceAddress
     */
    public void setSrcAddress (final I_LogicalAddress rtaddress)
    {
        if (rtaddress == null)
            throw new IllegalArgumentException ("Tried to setSrcAddress of OL_Message to null!");

        this.sourceAddress = rtaddress;
    }

    /**
     * Sets the logical address in this message
     */
    public void setSourceAddress (final I_LogicalAddress sourceAddress)
    {
        if (sourceAddress == null)
            throw new NullPointerException ("Source address cannot be null");

        this.sourceAddress = sourceAddress;
    }

    /*
     * Destination Address get/set
     */

    /**
     * Gets the destination logical address in this message 
     *
     * @deprecated name has changed: Routing Address to Destination Address, use getDestinationAddress
     */
    public I_LogicalAddress getRoutingAddress()
    {
        return destinationAddress;
    }
	
    /**
     * Gets the destination logical address in this message 
     */
    public I_LogicalAddress getDestinationAddress()
    {
        return destinationAddress;
    }
	
    /**
     * Sets the destination address in this message
     *
     * @deprecated name has changed: Routing Address to Destination Address, use setDestinationAddress
     */
    public void setRoutingAddress (final I_LogicalAddress destinationAddress)
    {
        this.destinationAddress = destinationAddress;
    }

    /**
     * Sets the destination address in this message
     */
    public void setDestinationAddress (final I_LogicalAddress destinationAddress)
    {
        this.destinationAddress = destinationAddress;
    }

    /*
     * Previous Hop Address get/set
     */

    /**
     * Gets the previous hop logical address in this message 
     *
     * @deprecated name has changed from Prevhop Address to Previous Hop Address, use getPreviousHopAddress
     */
    public I_LogicalAddress getPrevhopAddress()
    {
        return previousHopAddress;
    }

    /**
     * Gets the previous hop logical address in this message 
     */
    public I_LogicalAddress getPreviousHopAddress()
    {
        return previousHopAddress;
    }

    /**
     * Sets the previous hop address in this message
     *
     * @deprecated name has changed from Prevhop Address to Previous Hop Address, use setPreviousHopAddress
     */
    public void setPrevhopAddress (final I_LogicalAddress previousHopAddress)
    {
        this.previousHopAddress = previousHopAddress;
    }

    /**
     * Sets the previous hop address in this message
     */
    public void setPreviousHopAddress (final I_LogicalAddress previousHopAddress)
    {
        this.previousHopAddress = previousHopAddress;
    }


	/**
	 * Sets key vault instance in this message. It also creates the
	 * security auxialiary computation object in this message.
	 */
	public void setKeyVault (KeyVault kv)
	{
		if (kv == null)
		{
			return;
		}
		
		vault = kv;
		sac = new  SecurityAuxiliaryComputation(vault, this);
	}

    /** 
     * Converts the overaly message object to String representation.
     *
     * TODO: is this needed? if so, is it only for debugging? statistics?
     *
     * @return a string representation of the OL_Message object
     */
    public String toString()
    {
        final StringBuffer result = new StringBuffer();

        result.append ("\n");
        result.append ("\tVersion:\t" + HYPERCAST_VERSION + "\n");
        switch (deliveryMode)
        {
            case DELIVERY_MODE_INVALID:
                result.append ("\tDelivery Mode:\t" + "DELIVERY_MODE_INVALID (explicit)" + "\n");
                break;

            case DELIVERY_MODE_MULTICAST:
                result.append ("\tDelivery Mode:\t" + "DELIVERY_MODE_MULTICAST" + "\n");
                break;

            case DELIVERY_MODE_FLOOD:
                result.append ("\tDelivery Mode:\t" + "DELIVERY_MODE_FLOOD" + "\n");
                break;

            case DELIVERY_MODE_UNICAST:
                result.append ("\tDelivery Mode:\t" + "DELIVERY_MODE_UNICAST" + "\n");
                break;

            default:
                result.append ("\tDelivery Mode:\t" + "DELIVERY_MODE_INVALID (implicit)" + "\n");
                break;
        }		

        // ensure that only positive values are displayed for traffic class.
        final int trafficClass = this.trafficClass & 0x00FF;
        result.append ("\tTraffic Class:\t" + trafficClass + "\n");

        // ensure that only positive values are displayed for flow label.
        final int flowLabel = (this.flowLabel & 0xFF00) | (this.flowLabel & 0x00FF);
        result.append ("\tFlow Label:\t" + flowLabel + "\n");

        // ensure that only positive values are displayed for message length.
        final int messageLength = (this.messageLength & 0xFF00) | (this.messageLength & 0x00FF);
        result.append ("\tMessage Length:\t" + messageLength + "\n");

        // ensure that only positive values are displayed for hop limit.
        final int hopLimit = (this.hopLimit & 0xFF00) | (this.hopLimit & 0x00FF);
        result.append ("\tHop Limit:\t" + hopLimit + "\n");

        result.append ("\tLA Size:\t" + sourceAddress.getSize() + "\n");
        result.append ("\tSource Address:\t" + sourceAddress + "\n");
        if (deliveryMode == DELIVERY_MODE_UNICAST)
        {
            result.append ("\tDestination Address:\t" + destinationAddress + "\n");
        }
	
        for (int i = 0; i < extensionStore.size(); i++)
        {
            final Extension extension = (Extension) extensionStore.elementAt (i);
            result.append (extension);
        }

        final String resultString = result.toString();
        return resultString;
    }
    
    
    /**
     * This class defines a set of auxiliary functions which execute security related 
     * computation for an secure OL_Message. For a plaintext OL_Message, they are not 
     * used.
     */
    class SecurityAuxiliaryComputation
    {
		/**
		 * Key vault. It maintains the certificates and keys of local node and other
		 * nodes which exchange certificate and key with local node.
		 */
		KeyVault keyVault;
  	
    	/**
    	 * The OL_Message instance to which this object is bound.
    	 */
    	OL_Message OLMessage;

    	
    	/**
    	 * Constructor.
    	 */
		public SecurityAuxiliaryComputation(KeyVault kv, OL_Message olm)
		{
			keyVault = kv;
			
			OLMessage = olm;
		}
		

		/**
		 * Update the HMac field in the byte array format of an overlay message.
		 * 
		 * @param packet	byte array converted from an instance of OL_Message.
		 */
		private void updateHMac(byte[] packet)
		{
			/*
			 * Get the security extension in the OL_message.
			 */
			SecurityExtension secExt = (SecurityExtension) OLMessage.getFirstExtensionByType(Extension.SECURITY);

			if (secExt == null)
			{
				throw new HyperCastWarningRuntimeException ("Couldn't calculate header MAC for a plaintext overlay message.");
			}

			/*
			 * Get the byte array which is constructed from the packet
			 * and used for hmac calculation.
			 */
			byte[] fieldBytes = getByteArrayForHMacComputation(packet, secExt);

			/*
			 * Update hmac computation object with the created byte array.
			 */
			secExt.updateHMacCompute(fieldBytes);

			/*
			 * Get hmac computation object from the security extension
			 * and convert it to byte array.
			 */
			Mac hMacCompute = secExt.getHMacCompute();
			byte[] newHMac = hMacCompute.doFinal();

			/*
			 * Get the localtion of hmac in the packet (byte array).
			 */
			int hmacLocInPacket = getHMacLoc(packet, secExt);

			System.arraycopy(newHMac, 0, packet, hmacLocInPacket, newHMac.length);
		}
	
		/**
		 * Create a byte array from the packet which is used to calculate hmac.
		 * 
		 * @param packet	byte array converted from OL_Mesage instance
		 * @param secExt	security extension stored in the OL_Message
		 */
		private byte[] getByteArrayForHMacComputation(byte[] packet, SecurityExtension secExt)
		{
			/*
			 * The hmac is calculated on all fields in a byte array packet, except
			 * hmac field and payload field. A byte array packet has the following format:
			 * 
			 *   fixed header      security extension         payload extension
			 * ---------------      ------------------      -----------------------
			 * |             | ... |      |hmac|     | ...  |type|  Len   |payload| ...
			 * ---------------     -------------------      -----------------------
			 * 
			 * There is no assumption on the order of security extension and payload
			 * extension.
			 */


			int hmacLoc = this.getHMacLoc(packet, secExt);
			int hmacLen = secExt.getHMacLen();
		
			PayloadExtension payloadExtension = (PayloadExtension) OLMessage.getFirstExtensionByType (Extension.RAW_DATA);
				
			/*
			 * Change: An OL_Message may have no Payload Extension, such as MessageStore control messages.
			 */
/*
			if (payloadExtension == null)
			{
				throw new HyperCastWarningRuntimeException("This message contains no payload extension.");
			}
*/
			int payloadLoc;
			int payloadLen;
			if (payloadExtension == null)
			{
				payloadLoc = 0;
				payloadLen = 0;
			}
			else
			{

				payloadLoc = this.getPayloadLoc(packet);
				payloadLen = payloadExtension.toByteArray().length;
			}

			/*
			 * Figure out two pieces ordered by their location in the packet 
			 * that aren't used for calculating hmac.
			 */
			int locOfFirstPartToSkip;
			int lenOfFirstPartToSkip;
			int locOfSecondPartToSkip;
			int lenOfSecondPartToSkip;
			if (hmacLoc < payloadLoc)
			{
				locOfFirstPartToSkip = hmacLoc;
				lenOfFirstPartToSkip = hmacLen;
				locOfSecondPartToSkip = payloadLoc;
				lenOfSecondPartToSkip = payloadLen;
			}
			else
			{
				locOfFirstPartToSkip = payloadLoc;
				lenOfFirstPartToSkip = payloadLen;
				locOfSecondPartToSkip = hmacLoc;
				lenOfSecondPartToSkip = hmacLen;
			}

			/*
			 * Construct the byte array for hmac update.
			 */		
			byte[] fieldBytes = new byte[packet.length - lenOfFirstPartToSkip - lenOfSecondPartToSkip];
		
			System.arraycopy(packet, 0, fieldBytes, 0, locOfFirstPartToSkip);
			System.arraycopy(packet, locOfFirstPartToSkip+lenOfFirstPartToSkip, 
					fieldBytes, locOfFirstPartToSkip, 
					locOfSecondPartToSkip-(locOfFirstPartToSkip+lenOfFirstPartToSkip));
			System.arraycopy(packet, locOfSecondPartToSkip+lenOfSecondPartToSkip, 
					fieldBytes, locOfSecondPartToSkip-lenOfFirstPartToSkip, 
					packet.length - (locOfSecondPartToSkip+lenOfSecondPartToSkip));
				
			return fieldBytes;
		}

		/**
		 * Get the location of hmac field in the packet.
		 * 
		 * @param packet	byte array converted from an instance of OL_Message.
		 * @param secExt	security extension stored in the OL_Message
		 */
		private int getHMacLoc(byte[] packet, SecurityExtension secExt)
		{
		
			/*
			 * Get the extension type field (The 10th byte) in the fixed header.
			 */
			byte extType = packet[9];

			/*
			 * headerIndex is initialized to the first byte after fixed header.
			 */
			int headerIndex = packet.length - OLMessage.messageLength;

			/*
			 * Search for the security extension in the packet.
			 */
			while (extType != Extension.SECURITY)
			{
				if (extType == Extension.HEADER_NONE) 
				{
					throw new IllegalArgumentException("This message contains no security extension when doing header MAC calculation.");
				}

				extType = packet[headerIndex];

				headerIndex += 1;

				byte[] LengthBytes = new byte[4];
				System.arraycopy(packet, headerIndex, LengthBytes, 2, 2);
				headerIndex += 2;

				int len = ByteArrayUtility.toInteger(LengthBytes, 0);
				headerIndex += len;

				if (headerIndex > packet.length) 
				{
					throw new IllegalArgumentException("This message is corrupted.");
				}
			}
		
			/*
			 * The location of hmac field is the location of security extension plus the location
			 * of hmac in the security extension (in byte array format).
			 */
			return headerIndex + 3 + secExt.getHMacLoc();
		}


		/**
		 * Get the location of payload field in the packet.
		 * 
		 * @param packet	byte array converted from an instance of OL_Message.
		 */
		private int getPayloadLoc(byte[] packet)
		{
			/*
			 * Get the extension type field (The 10th byte) in the fixed header.
			 */
			byte extType = packet[9];

			/*
			 * headerIndex is initialized to the first byte after fixed header.
			 */
			int headerIndex = packet.length - OLMessage.messageLength;

			/*
			 * Search for the payload extension in the packet.
			 */
			while (extType != Extension.RAW_DATA)
			{
				if (extType == Extension.HEADER_NONE) {
					throw new IllegalArgumentException("This message contains no payload extension.");
				}

				extType = packet[headerIndex];

				headerIndex += 1;

				byte[] Length = new byte[4];
				System.arraycopy(packet, headerIndex, Length, 2, 2);
				headerIndex += 2;

				int len = ByteArrayUtility.toInteger(Length, 0);
				headerIndex += len;

				if (headerIndex > packet.length) 
				{
					throw new IllegalArgumentException("This message is corrupted.");
				}
			}
		
			/*
			 * The location of payload field is the location of Payload extension plus the location
			 * of payload inside the Payload extension.
			 */
			return headerIndex + 3;
		}


		/**
		 * Verify fixed header mac field.
		 * 
		 * @param packet		byte array converted from an instance of OL_Message.
		 * @param prevSender	logical address of message sender
		 */
		private boolean checkHMacComputation(byte[] packet, I_LogicalAddress prevSender)
		{
			/*
			 * Get the key of message sender.
			 */
			Key prevSenderKey = keyVault.getKey(prevSender);
			if(prevSenderKey == null)
				return false;
		
			/*
			 * Get the MAC calculation algorithm for this message.
			 */
			String integrityAlgorithm = keyVault.getIntegrityAlgorithm();

			/*
			 * Get the security extension in this message.
			 */
			SecurityExtension secExtension = (SecurityExtension) OLMessage.getFirstExtensionByType(Extension.SECURITY);

			if (secExtension == null)
			{
				throw new HyperCastWarningRuntimeException ("Couldn't check header MAC for a plaintext overlay message.");
			}

			/*
			 * Get the hmac (stored in the security extension) carried in the received packet.
			 */
			byte[] hmac = secExtension.getHMac();

			/*
			 * Get the byte array which is constructed from the received packet
			 * and is used for hmac calculation.
			 */
			byte[] fieldBytes = getByteArrayForHMacComputation(packet, secExtension);
		
			/*
			 * Calculate hmac from the byte array constructed above and compare it with
			 * the hmac carried in the packet.
			 */
			Mac hmacCompute;
		
			try
			{
				hmacCompute = Mac.getInstance(integrityAlgorithm);
				hmacCompute.init(prevSenderKey);

				hmacCompute.update(fieldBytes, 0, fieldBytes.length);

				byte[] hhmac = hmacCompute.doFinal();

				//System.out.println("SecurityAuxiliaryComputation:checkHMacComputation:calculated header mac="+HTTP_ServerUtility.toHexString(hhmac)+"***");
				//System.out.println("SecurityAuxiliaryComputation:checkHMacComputation:carried header mac="+HTTP_ServerUtility.toHexString(hmac)+"***");

				if (!HTTP_ServerUtility.toHexString(hhmac).equals(HTTP_ServerUtility.toHexString(hmac))) 
				{
					//throw new IllegalArgumentException("Message authentication check failure");
					keyVault.setKeyInvalid(prevSender, true);
					keyVault.sp.requestKeyFor(prevSender);
					return false;
				}

				/*
				 * Message authentication passed, so reset sender's watchdog timer.
				 */ 
				if (keyVault.getKeyMode().equals("NeighborhoodKeys")) 
				{
					keyVault.setKeyInvalid(prevSender, false);
				}
			}
			catch (InvalidKeyException e)
			{
				e.printStackTrace();
				throw new IllegalArgumentException("Message authentication check failure");
			}
			catch (NoSuchAlgorithmException e)
			{
				e.printStackTrace();
				throw new IllegalArgumentException("Message authentication check failure");
			}
		
			return true;
		}

		/**
		 * Verify the payload mac field.
		 */
		private boolean checkMacComputation()
		{
			/*
			 * Get the MAC calculation algorithm for this message.
			 */
			String integrityAlgorithm = keyVault.getIntegrityAlgorithm();

			/*
			 * Get the security extension in this message.
			 */
			SecurityExtension secExtension = (SecurityExtension) OLMessage.getFirstExtensionByType(Extension.SECURITY);

			if (secExtension == null)
			{
				throw new HyperCastWarningRuntimeException ("Couldn't check payload MAC for a plaintext overlay message.");
			}

			/*
			 * Get the mac (stored in the security extension) carried in the received packet.
			 */
			byte[] mac = secExtension.getMac();

			/*
			 * Get the message key stored in the security extension.
			 */
			Key msgKey = secExtension.getMsgKey();


			Mac macCompute;

			PayloadExtension payloadExtension = (PayloadExtension) OLMessage.getFirstExtensionByType (Extension.RAW_DATA);
				
			if (payloadExtension == null)
			{
				throw new HyperCastWarningRuntimeException("This message contains no payload extension.");
			}

			byte[] payload = payloadExtension.toByteArray();

			/*
			 * Calculate mac from the payload and compare it with
			 * the mac carried in the packet.
			 */
			try
			{
				macCompute = Mac.getInstance(integrityAlgorithm);
				macCompute.init(msgKey);

				//System.out.println("SecurityAuxiliaryComputation:checkHMacComputation:payload used to cal mac ="+HTTP_ServerUtility.toHexString(payload)+"***");

				macCompute.update(payload, 0, payload.length);

				byte[] mmac = macCompute.doFinal();

				//System.out.println("SecurityAuxiliaryComputation:checkHMacComputation:calculated payload mac="+HTTP_ServerUtility.toHexString(mmac)+"***");
				//System.out.println("SecurityAuxiliaryComputation:checkHMacComputation:carried payload mac="+HTTP_ServerUtility.toHexString(mac)+"***");

				if (!HTTP_ServerUtility.toHexString(mmac).equals(HTTP_ServerUtility
																				 .toHexString(mac)))
				{
					//throw new IllegalArgumentException("Payload authentication check failure");
					return false;
				}
			}
			catch (InvalidKeyException e1)
			{
				throw new IllegalArgumentException("Payload authentication check failure");
			}
			catch (NoSuchAlgorithmException e)
			{
				throw new IllegalArgumentException("Payload authentication check failure");
			}
		
			return true;
		}
	
    }//end of class SecurityAuxiliaryComputation
}//end of class OL_Message
