/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.Pastry;

import hypercast.Pastry.rice.pastry.ScheduledMessage;
import hypercast.Pastry.rice.pastry.PastryNode;
import hypercast.Pastry.rice.pastry.messaging.Message;
import hypercast.Pastry.rice.pastry.join.InitiateJoin;
import hypercast.Pastry.rice.pastry.leafset.InitiateLeafSetMaintenance;
import hypercast.Pastry.rice.pastry.routing.InitiateRouteSetMaintenance;

/**
 *  This class implements an interface between the native Pastry (Rice
 *  University) timer mechanism and the HyperCast timer mechanism.
 *
 *  @author HyperCast Team
 *  @author Greg Mattes
 *  @version May 2004
 */
public class HyperCastScheduledMessage
    extends ScheduledMessage
{
    /**
     *  The delay from when this event is scheduled until the first
     *  occurrence of this event.
     */
    private long delay;

    /**
     *  The delay between periodic occurrences of an event. If the
     *  event should not be rescheduled for repeated occurrences, this
     *  is 0.
     */
    private long period = 0;

    /** 
     *  True if this event has been scheduled to occur via a
     *  HyperCastTimer.
     */
    private boolean scheduled = false;

    /**
     *  True if cancel() has been called on this event.
     */
    private boolean cancelled = false;

    /**
     *  @see java.util.Timer
     */
    private boolean fixedRate = false;

    /**
     *  Timer that scheduled this event. We need to know this to be
     *  able to cancel this event.
     */
    private HyperCastTimer timer = null;

    /**
     *  Constructor for event that is scheduled to occur once.
     */
    public HyperCastScheduledMessage (final PastryNode pn, final Message msg, final long delay)
    {
        this (pn, msg, delay, 0);
    }

    /**
     *  Constructor for event that is scheduled to occur periodically.
     */
    public HyperCastScheduledMessage (final PastryNode pn, final Message msg, final long delay, final long period)
    {
        super (pn, msg);
        this.delay = delay;
        this.period = period;
    }

    /**
     * Delay until message run() is called.
     */
    public long getDelay ()
    {
        return delay;
    }

    /**
     * The amount of time (in ms) that elapses between message runs. 0 if only one run occurs.
     */
    public long getPeriod ()
    {
        return period;
    }

    /**
     * True is this message should be rescheduled. The HyperCastTimer
     * object will call this method for every message that runs to see
     * if it should be scheduled again.
     */
    public boolean shouldReschedule()
    {
        return period > 0;
    }

    /**
     * For a discussion of "fixed-rate" see:
     * @see java.util.Timer
     */
    public boolean isFixedRate()
    {
        return fixedRate;
    }

    /**
     * True if this event has been cancelled.
     */
    public synchronized boolean isCancelled()
    {
        return cancelled;
    }

    /**
     * Schedules this message for delayed delivery.
     *
     * For a discussion of "fixed-rate" see:
     * @see java.util.Timer
     */
    public synchronized void setScheduledFixedRate (final HyperCastTimer timer)
    {
        this.timer = timer;
        scheduled = true;
        fixedRate = true;
    }

    /**
     * Schedules this message for delayed delivery.
     */
    public synchronized void setScheduled (final HyperCastTimer timer)
    {
        this.timer = timer;
        scheduled = true;
    }

    /*
     *  Override the cancel event so that calling it does not cause
     *  nasty java.util.Timer thread stuff (NB ScheduleMessage is
     *  derived from java.util.TimerTask).
     *
     *  See Javadoc for the semantics of the return value.
     */
    public synchronized boolean cancel ()
    {
        cancelled = true;
        timer.clearTimer (this);

        boolean returnValue = scheduled || (period > 0);

        period = 0;
        scheduled = false;

        return returnValue;
    }

    /*
     *  Not clear that rice.pastry.messaging.Message can be compared
     *  with equals. Equality based solely on type.
     */
    public boolean equals (Object obj)
    {
        boolean result = false;
        if (obj instanceof HyperCastScheduledMessage)
        {
            HyperCastScheduledMessage other = (HyperCastScheduledMessage) obj;
            Message otherMessage = other.getMessage ();
            Message myMessage = this.getMessage ();

            result =
                ((otherMessage instanceof InitiateJoin
                  &&
                  myMessage instanceof InitiateJoin)
                 ||
                 (otherMessage instanceof InitiateLeafSetMaintenance
                  &&
                  myMessage instanceof InitiateLeafSetMaintenance)
                 ||
                 (otherMessage instanceof InitiateRouteSetMaintenance
                  &&
                  myMessage instanceof InitiateRouteSetMaintenance));
        }
        return result;
    }

    /*
     *  Sometimes useful when debugging
     */
    public String getType ()
    {
        Message myMessage = this.getMessage ();
        if (myMessage instanceof InitiateJoin)
        {
            return "InitiateJoin";
        }
        else if (myMessage instanceof InitiateLeafSetMaintenance)
        {
            return "InitiateLeafSetMaintenance";
        }
        else if (myMessage instanceof InitiateRouteSetMaintenance)
        {
            return "InitiateRouteSetMaintenance";
        }
        else
        {
            return "Unknown Message Type";
        }
    }
}
