/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.Pastry;

import hypercast.I_AddressPair;
import hypercast.I_Message;
import hypercast.I_PhysicalAddress;
import hypercast.I_UnicastAdapter;
import hypercast.I_LogicalAddress;
import hypercast.I_AdapterCallback;
import hypercast.I_Node;
import hypercast.OL_Socket;
import hypercast.HyperCastConfig;
import hypercast.ByteArrayUtility;
import hypercast.SimpleStats;
import hypercast.StatsProcessor;
import hypercast.StatsElement;
import hypercast.NeighborhoodStats;

import hypercast.util.XmlUtil;

import hypercast.HyperCastStatsException;
import hypercast.HyperCastWarningRuntimeException;
import hypercast.HyperCastFatalRuntimeException;
import hypercast.HyperCastConfigException;

import hypercast.NotificationHandler;
import hypercast.events.NODE_LEAVEOVERLAY;
import hypercast.events.NODE_ISSTABLE;

import hypercast.Pastry.HyperCastTimer;
import hypercast.Pastry.HyperCastNodeAdapter;

import hypercast.Pastry.rice.pastry.NodeHandle;
import hypercast.Pastry.rice.pastry.NodeId;
import hypercast.Pastry.rice.pastry.routing.RoutingTable;
import hypercast.Pastry.rice.pastry.hc.HCPastryNode;
import hypercast.Pastry.rice.pastry.hc.HCNodeHandle;
import hypercast.Pastry.rice.pastry.hc.HCPastryNodeFactory;
import hypercast.Pastry.rice.pastry.standard.RandomNodeIdFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import java.lang.UnsupportedOperationException;

import org.apache.xpath.XPath;

import java.util.Vector;
import java.util.Collections;

/**
 *  This class implements a Pastry node and also
 *  provides a mechanism for joining an overlay
 *  using a buddy list rendezvous mechanism. 
 *
 *  @author HyperCast Team
 *  @author Greg Mattes
 *  @version May 2004
 */
public class Pastry_Node_BuddyList
    implements I_AdapterCallback, I_Node
{
    /*
     *  Debugging constants
     */
    private static final boolean debug                                  = false;
    private static final boolean debug_config                           = false;
    private static final boolean debug_state_change                     = false;
    private static final boolean debug_buddy                            = false;
    private static final boolean debug_join_phase_2                     = false;
    private static final boolean debug_pendingProximityResponseMessages = false;

    /*
     *  There are 3 distinct Phases of the Join Sequence: rendezvous
     *  (using a buddy list), find the physically closest node in
     *  the existing ring, and issue the Pastry Join operation.
     *
     *  First Phase of the Join Sequence: Rendezvous by finding a live
     *  "buddy," doing this allows the new node to make initial
     *  contact with an existing Pastry ring. The code for this phase
     *  is based on the Delaunay Triangulation Buddy mechanism. Note
     *  that this phase does not exist in the FreePastry
     *  implementation, rather rendezvous in FreePastry is done by
     *  providing hostnames and port numbers on the command line when
     *  starting a new FreePastry node.
     *
     *  Second Phase of the Join Sequence: Run the FreePastry
     *  "getNearest" algorithm originally implemented in
     *  PastryNodeFactory.java. The FreePastry implemenation of this
     *  algorithm cannot be called from within HyperCast due to the
     *  unidirectional nature of HyperCast channels so it is
     *  reimplemented in Pastry_Node_BuddyList.java. The variable
     *  joinSubState tracks the progress of this algorithm.
     *
     *  Third Phase of the Join Sequence: The Pastry Join algorithm
     *  proper is run after the doneNode() method is invoked on the
     *  node. The join algorithm proper exchanges Pastry protocol
     *  messages beginning with InitiateJoin, see
     *  HCPastryNode.java. NB these messages are passed back and
     *  forth between embedded FreePastry "black-boxes" within
     *  Hypercast. They are designated as type
     *  Pastry_Message.PASTRY_ENCAPSULATED to differentiate them from
     *  messages that are interpretted and processed in this file.
     */

    /**
     *  If true, do not search for a buddy; immediately start a new
     *  Pastry ring, i.e. skip Phases 1 & 2 and go to Phase 3.
     */
    private static boolean disableJoinSequencePhase1 = false;

    /**
     *  If true, do not search for the physically closest node; issue
     *  InitiateJoin message immediately to the rendezvous node.
     */
    private static boolean disableJoinSequencePhase2 = false;

    /*
     *  Join Sequence Phase 3, i.e. calling doneNode(), is obligatory.
     */

    /**
     *  If true, do not request the leaf set of the rendezvous node.
     */
    private static boolean disableJoinSequencePhase2Stage1 = false;

    /**
     *  If true, do not "recursively examine the top level routing row
     *  of the nodes until no more progress can be made."
     */
    private static boolean disableJoinSequencePhase2Stage2 = false;

    /**
     *  If true, do not "recursively examine the top level routing row
     *  of the nodes until no more progress can be made;" issue
     *  InitiateJoin message immediately using current closely
     *  physical node.
     */
    private static boolean disableJoinSequencePhase2Stage3 = false;

    /*
     *  Object used to notify application of node events
     */
    private NotificationHandler notificationHandler = null;

    /*
     *  Pastry <-> HyperCast adapter objects.
     *  These object implement the object adapter pattern.
     */
    private HyperCastTimer hyperCastTimer;
    private HyperCastNodeAdapter hyperCastNodeAdapter;

    /*
     *  Pastry objects designed for use with HyperCast
     */
    private static HCPastryNodeFactory factory = (HCPastryNodeFactory) HCPastryNodeFactory.getFactory (new RandomNodeIdFactory());
    private HCPastryNode hcPastryNode;

    /*
     *  Node adapter, this is the object that allows this node to
     *  access the network. It has methods that the node can call to
     *  send messages, it also notifies the node of message arrival
     *  by calling the messageArrived() method of the node. In
     *  addition, it supports setting timer events and will call
     *  timerExpired() when a scheduled event occurs.
     *
     *  Access is protected so that Buddy_TimedStopper can see it.
     */
    protected I_UnicastAdapter adapter;

    /*
     *  Used to read the configuration file.
     */
    private HyperCastConfig config;
    private static String PROPERTY_PROTO_PREFIX = "/Public/Node/Pastry/";

    /*
     *  Constants that keep track of the state of the node.
     */

    private int state = STATE_STOPPED;

    public static final int STATE_STOPPED        = 0; // after init or call to leaveOverlay()
    public static final int STATE_BUDDY_SEARCH   = 1; // after call to join group
    public static final int STATE_BUDDY_FOUND    = 2; // after NODE_ID_RESPONSE received from buddy
    public static final int STATE_JOIN_INITIATED = 3; // after call to startNewPastryRing or joinEstablishedPastryRing
    public static final int STATE_JOINED         = 4; // after call to nodeIsReady (from the innards of the Pastry code)

    public static String getStringOfState (int x)
    {
        switch (x)
        {
            case STATE_STOPPED:        return "STATE_STOPPED";
            case STATE_BUDDY_SEARCH:   return "STATE_BUDDY_SEARCH";
            case STATE_BUDDY_FOUND:    return "STATE_BUDDY_FOUND";
            case STATE_JOIN_INITIATED: return "STATE_JOIN_INITIATED";
            case STATE_JOINED:         return "STATE_JOINED";
            default: return "UNKNOWN STATE " + x;
        }
    }

    /*
     *  The "join substates" happen between the node states
     *  STATE_JOIN_INITIATED and STATE_JOINED.
     *
     *  After the physically closest node has been determined (Phase
     *  2), JOIN_SUBSTATE_PASTRY_JOIN is entered (Phase 3) to indicate
     *  that doneNode() has been called and thus the Pastry Join
     *  protocol is taking place.
     */

    private int joinSubState = JOIN_SUBSTATE_NOT_IN_JOIN;

    public static final int JOIN_SUBSTATE_INVALID                = -1;
    public static final int JOIN_SUBSTATE_NOT_IN_JOIN            = 0; // Phase 1 or not doing join
    public static final int JOIN_SUBSTATE_LEAFSET                = 1; // Phase 2
    public static final int JOIN_SUBSTATE_LEAFSET_CLOSEST        = 2; // Phase 2
    public static final int JOIN_SUBSTATE_ROUTEROW_PASS1         = 3; // Phase 2
    public static final int JOIN_SUBSTATE_ROUTEROW_PASS1_CLOSEST = 4; // Phase 2
    public static final int JOIN_SUBSTATE_ROUTEROW_PASS2         = 5; // Phase 2
    public static final int JOIN_SUBSTATE_ROUTEROW_PASS2_CLOSEST = 6; // Phase 2
    public static final int JOIN_SUBSTATE_PASTRY_JOIN            = 7; // Phase 3

    public static String getStringOfSubState (int x)
    {
        switch (x)
        {
            case JOIN_SUBSTATE_INVALID:                return "JOIN_SUBSTATE_INVALID";
            case JOIN_SUBSTATE_NOT_IN_JOIN:            return "JOIN_SUBSTATE_NOT_IN_JOIN";
            case JOIN_SUBSTATE_LEAFSET:                return "JOIN_SUBSTATE_LEAFSET";
            case JOIN_SUBSTATE_LEAFSET_CLOSEST:        return "JOIN_SUBSTATE_LEAFSET_CLOSEST";
            case JOIN_SUBSTATE_ROUTEROW_PASS1:         return "JOIN_SUBSTATE_ROUTEROW_PASS1";
            case JOIN_SUBSTATE_ROUTEROW_PASS1_CLOSEST: return "JOIN_SUBSTATE_ROUTEROW_PASS1_CLOSEST";
            case JOIN_SUBSTATE_ROUTEROW_PASS2:         return "JOIN_SUBSTATE_ROUTEROW_PASS2";
            case JOIN_SUBSTATE_ROUTEROW_PASS2_CLOSEST: return "JOIN_SUBSTATE_ROUTEROW_PASS2_CLOSEST";
            case JOIN_SUBSTATE_PASTRY_JOIN:            return "JOIN_SUBSTATE_PASTRY_JOIN";
            default: return "UNKNOWN SUBSTATE " + x;
        }
    }

    /*
     *  Used by Phase 2 join algorithm to determine the physically closest node.
     */
    private HCNodeHandle currentClosest;
    private long nearestdist = Long.MAX_VALUE; // distance to current closest

    /*
     *  Used by Stage 2 of Phase 2 of the Join Sequence.  These
     *  variables set the maximum number of Stage 2 iterations as well
     *  as keep track of the current Stage 2 iteration.
     */
    private static final int JOIN_ROUTING_TABLE_DEPTH = NodeId.nodeIdBitLength / RoutingTable.idBaseBitLength;
    private int routeRowPass1CurrentDepth = 0;

    /*
     *  Used by Stage 3 of Phase 2 of the Join Sequence. This variable
     *  records the value found in the previous iteration of the third
     *  stage so that it can be compared to the current value.
     */
    private HCNodeHandle stage3PreviousClosest;

    /*
     *  This variable is set to false when a route row request is
     *  made. When a route row response is received it is set to
     *  true. This information is useful because when a route row
     *  request timeout occurs we need to know whether or not a route
     *  row has been received. If a route row was received the timeout
     *  can be ignored, alternatively, if a route row was not
     *  received, the timeout should cause the next step of Phase 2 of
     *  the Join Sequence to occur.
     */
    private boolean routeRowReceived = false;

    /**
     * Amount of time (in ms) to wait for a response to a join message to be received.
     */
    private long joinMessageTimeout;

    /**
     *  Time at which a substate of the Second Phase of the Join
     *  Sequence begins. Knowing this time allows us to determine
     *  which messages and timeouts to recognize.
     *
     *  This value begins at zero (the epoch)
     */
    private long subStateStartTime = 0; // milliseconds

    /**
     *  Keeps track of how messages have been sent for a particular
     *  round of the join protocol.
     */
    private int pendingProximityResponseMessages = 0;

    /**
     * List of buddies read from configuration file.
     */
    private Vector staticBuddyList;

    /**
     * Keeps track of the index into staticBuddyList of the next buddy
     * that we will attempt to contact.
     */
    private int buddySearchIndex;

    /**
     * Keeps track of the remaining number of retries for the
     * staticBuddyList before we decide to start our own network.
     */
    private int buddySearchRetries;

    /** Time in ms between leaf set maintenance operations. */
    private long leafSetMaintFreq;
    
    /** Time in ms between route set maintenance operations. */
    private long routeSetMaintFreq;

    /** The size of the leaf set. */
    private int leafSetSize;

    /** The number of handles to have for each entry of the routing table. */
    private int routeSetSize;

    /** true after joinOverlay and before leaveOverlay, false after leaveOverlay but before joinOverlay, initially false */
    private boolean inOverlay = false;

    /** Time to wait before declaring a remote node to be dead. */
    private long timeoutTime;

    /** The logical address read from the configuration file or null if random node id should be generated. */
    private NodeId configuredNodeId;

    /**
     *  Set to true when a buddy replies, this variable allows
     *  handleNodeIdRequestTimeout to ignore subsequent buddy replies
     */
    private boolean buddyFound = false;

    /** amount of time in milliseconds to wait for a response */
    private long nodeIdRequestTimeout;

    /** Object used to prevent problems while starting and stopping the node. */
    protected Object startStopLock = new Object();

    /** Object used to stop all threads TIMEOUT_TIME seconds after the call to leaveOverlay(). */
    protected Buddy_TimedStopper stopper;

    /** time to wait after leave group until thread shutdown */
    private long LEAVE_OVERLAY_TIMEOUT_TIME = 10000; // milliseconds

    /** Object to which statistics calls are delegated */
    private StatsProcessor statsProcessor = new StatsProcessor (this, true, false);

    /** The name by which this particular Pastry node will be known */
    private String statisticsName;

    // Useful for logging/debugging
    private final java.text.SimpleDateFormat sdf;

    /**
     * The configuration is read in this method, here is an example:
     *
     *  <Node>
     *    <Pastry>
     *      <LeafSetMaintenanceFrequency>60000</LeafSetMaintenanceFrequency>
     *      <RouteSetMaintenanceFrequency>900000</RouteSetMaintenanceFrequency>
     *      <LeafSetSize>24</LeafSetSize>
     *      <RouteSetSize>1</RouteSetSize>
     *      <PingTimeout>500</PingTimeout>
     *      <JoinMessageTimeout>5000</JoinMessageTimeout>
     *      <TimeoutTime>1800000</TimeoutTime>
     *      <UsingStaticBuddyList>true</UsingStaticBuddyList>
     *      <BuddyNum>1</BuddyNum>
     *      <Buddy>
     *        <UnderlayAddress>
     *          <INETV4AndOnePort>192.168.0.30:41405</INETV4AndOnePort>
     *        </UnderlayAddress>
     *      </Buddy>
     *      <BuddyRetries>10</BuddyRetries>
     *      <LogicalAddress>
     *        <RANDOM/>
     *        <!--
     *        <FIXED>
     *          <NodeId>0x12345678901020304050607080ABCDEF</NodeId>
     *        </FIXED>
     *        -->
     *      </LogicalAddress>
     *    </Pastry>
     */
    public Pastry_Node_BuddyList (HyperCastConfig config, I_UnicastAdapter adapter, String statisticsPrefix){
        this(config, adapter, statisticsPrefix, null);
    }
    public Pastry_Node_BuddyList (HyperCastConfig config, I_UnicastAdapter adapter, String statisticsPrefix, Pastry_LogicalAddress la)
    {
        sdf = new java.text.SimpleDateFormat();
        sdf.applyPattern ("hh:mm:ss.SSS");

        PROPERTY_PROTO_PREFIX = statisticsPrefix;
        this.config = config;
        this.adapter = adapter;

        adapter.setCallback (this);

        // Read configuration for timers of periodic pastry operations.
        this.leafSetMaintFreq = config.getPositiveLongAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "LeafSetMaintenanceFrequency"));
        if (debug_config)
            config.err.println ("leafSetMaintFreq: " + leafSetMaintFreq);

        this.routeSetMaintFreq = config.getPositiveLongAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "RouteSetMaintenanceFrequency"));
        if (debug_config)
            config.err.println ("routeSetMaintFreq: " + routeSetMaintFreq);

        // Read configuration for route and leaf set sizes.
        this.leafSetSize = config.getPositiveIntAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "LeafSetSize"));
        if (debug_config)
            config.err.println ("leafSetSize: " + leafSetSize);

        this.routeSetSize = config.getPositiveIntAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "RouteSetSize"));
        if (debug_config)
            config.err.println ("routeSetSize: " + routeSetSize);

        // Read configuration that tells us how many times to try each buddy before starting our own network
        this.buddySearchRetries = config.getNonNegativeIntAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "BuddyRetries"));
        if (debug_config)
            config.err.println ("buddySearchRetries: " + buddySearchRetries);

        // Read configuration that tells us how often to try buddies.
        this.nodeIdRequestTimeout = config.getPositiveLongAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "PingTimeout"));
        if (debug_config)
            config.err.println ("nodeIdRequestTimeout: " + nodeIdRequestTimeout);

        // Read configuration that tells us how long to wait for join
        this.joinMessageTimeout = config.getPositiveLongAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "JoinPhase2MessageTimeout"));
        if (debug_config)
            config.err.println ("joinMessageTimeout: " + joinMessageTimeout);

        // Read configuration that tells us how long to wait before declaring a node dead (twice route set maintenance time?)
        this.timeoutTime = config.getPositiveLongAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "TimeoutTime"));
        if (debug_config)
            config.err.println ("timeoutTime: " + timeoutTime);

        final String laString = config.getTextAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "LogicalAddress"));

        if (debug_config)
            config.err.println ("logical address choice: " + laString);
        
        if(la!=null){
            this.configuredNodeId = la.getNodeId();
        }
        else if (laString.equals ("RANDOM"))
        {
            if (debug_config)
                config.err.println ("RANDOM logical address");

            this.configuredNodeId = null;
        }
        else if (laString.equals ("FIXED"))
        {
            final String logicalAddressString = config.getTextAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "LogicalAddress/FIXED/NodeId"));

            if (debug_config)
                config.err.println ("FIXED logical address: " + logicalAddressString);

            try
            {
                this.configuredNodeId = new Pastry_LogicalAddress (logicalAddressString).getNodeId();
            }
            catch (final NumberFormatException nfe)
            {
                throw new HyperCastConfigException ("Invalid fixed node id for pastry node: \"" + configuredNodeId + "\"", nfe);
            }
        }

        if (debug_config)
            config.err.println ("Configured NodeId: " + configuredNodeId);

        /*
         *  The adapter is the object that controls the timer
         *  thread. The hyperCastTimer is the "timer bridge" between
         *  Pastry and HyperCast. It intercepts scheduling events in
         *  Pastry and schedules them on the timer thread controlled
         *  by the adapter. When the Pastry timer events occur, the
         *  object returned by the adapter's timer callback is sent to
         *  Pastry by calling hyperCastTimer.timerExpired().
         */
    hyperCastTimer = new HyperCastTimer (adapter);
    hyperCastNodeAdapter = new HyperCastNodeAdapter (this);
        hcPastryNode = (HCPastryNode) factory.newNode (configuredNodeId,
                                                       hyperCastTimer, hyperCastNodeAdapter,
                                                       (int) (leafSetMaintFreq / 1000), (int) (routeSetMaintFreq / 1000),
                                                       leafSetSize, routeSetSize,
                                                       timeoutTime);
        
        /*
         *  Initialize buddy data structures. Static in this context
         *  does not refer to the notion of static in the Java
         *  programming language, rather, it means buddy address data
         *  that is statically declared in the hypercast configuration
         *  file instead of buddy address data that is collected
         *  dynamically at run time.
         */

        this.buddySearchIndex = 0;
        staticBuddyList = new Vector();

        // Read buddy information from config file
        if ("true".equals (config.getTextAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "UsingStaticBuddyList"))))
        {
            final int buddyNum = config.getIntAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "BuddyNum"));

            if (debug_config)
                config.err.println ("buddyNum: " + buddyNum);
            
            String underlayAddress = null;
            for (int i = 0; i < buddyNum; i++)
            {
                /*
                 * Read the address of a buddy from configuration file.
                 */
                try
                {
                    final String addrType = config.getTextAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "Buddy[" + (i + 1) + "]/UnderlayAddress"));
                    underlayAddress = config.getTextAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "Buddy[" + (i + 1) + "]/UnderlayAddress/" + addrType));
                }
                catch (final HyperCastConfigException ignored)
                {
                    // The attribute doesn't exist.
                    break;
                }

                /*
                 * If the address is null, stop adding BuddyRecord to the buddy list and continue
                 * to finish other work.
                 */
                if ((underlayAddress == null) || (underlayAddress.startsWith("0.0.0.0")))
                {
                    break;
                }

                /*
                 * If the address is not null, add a BuddyRecord in the buddy list.
                 */
                try
                {
                    final BuddyRecord buddyRecord = new BuddyRecord (adapter, underlayAddress);

                    /*
                     * A node doesn't put itself into static buddy list. Do not compare socket adapter numbers.
                     */
                    if (! (getAddressPair().getPhysicalAddress()).addressCheck (buddyRecord.getAddressPair().getPhysicalAddress()))
                    {
                        staticBuddyList.addElement (buddyRecord);

                        if (debug_config)
                            System.out.println ("Add buddy " + buddyRecord + ".");
                    }
                    else
                    {
                        if (debug_config)
                            System.out.println ("Buddy " + buddyRecord + " is this node.");
                    }
                }
                catch (final IllegalArgumentException iae)
                {
                    config.err.print ("Buddy address is inconsistent with local physical address: " + iae);
                }
            }
        }

        /*
         *  Randomly permute staticBuddyList so that when lots of new
         *  nodes are simultaneously instantiated using the same
         *  configuration file, all of them do not attempt to contact
         *  the same buddy.
         */
        Collections.shuffle (staticBuddyList);

        initStatisticsStructure();

        changeState (STATE_STOPPED);
    }

    private String stateString()
    {
        return "state (" + getStringOfState (state) + "), substate (" + getStringOfSubState (joinSubState) + ")";
    }

    private void changeState (int newState)
    {
        changeState (newState, joinSubState);
    }

    private void changeSubState (int newSubState)
    {
        if (state != STATE_JOIN_INITIATED)
        {
            config.log.println ("Ignoring substate change in: " + stateString());
            //throw new AssertionError ("state != JOIN_INITIATED: " + stateString());}
        }
        subStateStartTime = adapter.getCurrentTime();
        changeState (state, newSubState);
    }

    private void changeState (int newState, int newJoinSubState)
    {
        state = newState;
        joinSubState = newJoinSubState;
        if (debug_state_change) System.err.println (lastr() + ": STATE CHANGE: " + stateString());
    }

    /*
     *  Called by the embedded Pastry "black-box" via HyperCastNodeAdapter
     */
    /* package */ void nodeIsReady()
    {
        changeState (STATE_JOINED, JOIN_SUBSTATE_NOT_IN_JOIN);

        if (debug_state_change)
            System.err.println (lastr() + ": join operation complete");

        if (notificationHandler != null)
        {
            final NODE_ISSTABLE nodeIsStableEvent = new NODE_ISSTABLE (adapter.getCurrentTime(), null);
            notificationHandler.eventOccurred (nodeIsStableEvent);
        }
    }

    private void joinEstablishedPastryRing (NodeHandle rendezvous)
    {
        if (debug_state_change)
            System.err.println (lastr() + ": found a buddy node at " + rendezvous + ", finding physically nearest node in existing Pastry ring...");

        changeState (STATE_JOIN_INITIATED);

        if (! disableJoinSequencePhase2)
        {
            /*
             *  Now begins Phase 2 of the Join Sequence: finding the
             *  physically closest node. This algorithm is not trivial
             *  to implement, existing FreePastry code cannot be used
             *  because it assumes that bidirectional channels can be
             *  used. No such channels are possible with HyperCast so
             *  the algorithm that is implemented in FreePastry must
             *  be reimplemented to use the HyperCast adapters with
             *  their uni-directional channels and retransimission
             *  timers.
             *
             *  Initially only the rendezvous node is the only other
             *  known node, so it is the best guess of the closest
             *  node.
             */
            currentClosest = (HCNodeHandle) rendezvous;

            if (! disableJoinSequencePhase2Stage1)
            {
                beginLeafSetRequestStage();
            }
            else
            {
                if (! disableJoinSequencePhase2Stage2)
                {
                    beginRouteRowPass1();
                }
                else
                {
                    if (! disableJoinSequencePhase2Stage3)
                    {
                        beginRouteRowPass2();
                    }
                    else
                    {
                        beginPhase3 (rendezvous);
                    }
                }
            }
        }
        else
        {
            /*
             *  Phase 2 of the Join Sequence can be bypassed by
             *  calling doneNode() here. Doing so results in a Join
             *  Sequence that is different than Pastry. The option to
             *  disable is included here for use in debugging: perhaps
             *  the source of some problems can found by disabling
             *  Phase 2.
             */
            beginPhase3 (rendezvous);
        }
    }

    private void startNewPastryRing()
    {
        if (debug_state_change)
            System.err.println (lastr() + ": could not find a buddy node, starting a new Pastry ring...");

        beginPhase3 (null);
    }

    /*
     *  This method invokes the Pastry join mechanism. The parameter
     *  "joinNode" can be null which will cause a new ring to be
     *  formed, i.e. an InitiateJoin message will not be sent to
     *  another node.
     */
    private void beginPhase3 (final NodeHandle joinNode)
    {
        changeState (STATE_JOIN_INITIATED, JOIN_SUBSTATE_PASTRY_JOIN);
        hcPastryNode.doneNode (joinNode);
    }

    /**
     *  joinOverlay()
     *
     *  (1) Start the adapter
     *  (2) Contact buddy:
     *      if no buddies specified
     *        start new ring (i.e. call donenode(null))
     *      else
     *        ping first buddy
     *        set buddy timer
     *
     *  There are three separate issues to deal with when creating a new node:
     *
     *  (A) Creating the data-struture(s) that represent the node.
     *  (B) Discovering (Rendezvousing) with an existing node.
     *  (C) Performing a "join" operating once an existing node is contacted.
     *
     *  In FreePastry, the order of these operations is (B), (A), (C);
     *  in HyperCast the order is (A), (B), (C).
     *
     *  In FreePastry, the discovery of an existing node happens
     *  before node creation. See the "getBootStrap" code in
     *  DistHelloWorld.java of the FreePastry distribution to
     *  understand the discovery mechanism. See the
     *  "HcPastryNodeFactory.newNode" method to understand the
     *  creation mechanism. The initiation of "join" takes place after
     *  the call to "Node.doneNode."
     *
     *  HyperCast performs data-structure creation first, then
     *  discovery/rendezvous, and finally "join." The newNode method
     *  of the HcPastyNodeFactory object was altered so that it
     *  constructs a node, but does not spawn a thread and attempt to
     *  join a ring. The ring joining logic is here. No thread is
     *  created for the node, we rely on HyperCast adapter threads.
     */

    public synchronized void joinOverlay()
    {
        inOverlay = true;

        startAdapter();

        if (disableJoinSequencePhase1 || staticBuddyList.size() == 0)
        {
            startNewPastryRing();
        }
        else
        {
            changeState (STATE_BUDDY_SEARCH);
            final I_PhysicalAddress buddyAddress = getNextBuddyAddress();

            if (buddyAddress == null)
                throw new AssertionError ("Could not get the address of a single buddy even though staticBuddyList.size = " + staticBuddyList.size());

            requestNodeIdOfBuddy (buddyAddress);
        }
    }

    private I_PhysicalAddress getNextBuddyAddress()
    {
        if (buddySearchRetries < 0)
        {
            return null;
        }

        BuddyRecord buddy;
        try
        {
            buddy = (BuddyRecord) staticBuddyList.elementAt (buddySearchIndex);
        }
        catch (ArrayIndexOutOfBoundsException aioobe)
        {
            throw new RuntimeException ("Error in buddy list logic: buddySearchRetries(" + buddySearchRetries + ") buddySearchIndex(" + buddySearchIndex + ") staticBuddyList.size(" + staticBuddyList.size() +")");
        }

        buddySearchIndex = (buddySearchIndex + 1) % staticBuddyList.size();
        if (buddySearchIndex == 0)
        {
            --buddySearchRetries;
        }

        final Pastry_AddressPair addressPair = buddy.getAddressPair();
        final I_PhysicalAddress nextBuddyPhysicalAddress = addressPair.getPhysicalAddress();

        return nextBuddyPhysicalAddress;
    }

    /*
     *  The burden of starting the node adapter is on the
     *  implementation of the joinOverlay method of the Node interface.
     */

    public void startAdapter()
    {
        synchronized (startStopLock) 
        {
            if (stopper != null) 
            {
                stopper.die();
                stopper = null;
            }
            else 
            {
                adapter.Start();
                adapter.startTimer();
            }
        }
    }

    public void leaveOverlay() {
        leaveOverlay (null);
    }

    public synchronized void leaveOverlay (final Pastry_LogicalAddress newLA)
    {
        // reset buddy search retries so that the next join overlay will start with the original number of attempts
        this.buddySearchRetries = config.getNonNegativeIntAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "BuddyRetries"));

        synchronized (startStopLock) 
        {
            stopper = new Buddy_TimedStopper (this, LEAVE_OVERLAY_TIMEOUT_TIME);
        }

        changeState (STATE_STOPPED);

        if (notificationHandler != null)
        {
            final NODE_LEAVEOVERLAY nodeLeaveOverlayEvent = new NODE_LEAVEOVERLAY (adapter.getCurrentTime(), null);
            notificationHandler.eventOccurred (nodeLeaveOverlayEvent);
        }

        clearNeighborhoodTable();

        //currentClosest = null;
        nearestdist = Long.MAX_VALUE;
        routeRowPass1CurrentDepth = 0;
        //stage3PreviousClosest = null;
        routeRowReceived = false;
        subStateStartTime = 0;
        pendingProximityResponseMessages = 0;
        buddySearchIndex = 0;
        buddyFound = false;

        final NodeId nodeId = (newLA == null) ? hcPastryNode.getNodeId() : newLA.getNodeId();
        hcPastryNode.leaveOverlay();
        hcPastryNode = (HCPastryNode) factory.newNode (nodeId,
                                                       hyperCastTimer, hyperCastNodeAdapter,
                                                       (int) (leafSetMaintFreq / 1000), (int) (routeSetMaintFreq / 1000),
                                                       leafSetSize, routeSetSize,
                                                       timeoutTime);

        inOverlay = false;
    }

    public I_AddressPair[] getParent (I_LogicalAddress root)
    {
        return hyperCastNodeAdapter.getParent (root);
    }

    /**
     *  In the Pastry protocol it is not possible to known which
     *  neighbors are children in a multicast tree. All neighbors must
     *  be considered potential children. Each potential child must
     *  send a message to a confused parent saying that it is not the
     *  child of that node.
     *
     *  This schem is called "reverse path forwarding" and it is used
     *  to implement a multicast operation.
     */
    public I_AddressPair[] getChildren (I_LogicalAddress root)
    {
        return getAllNeighbors();
    }

    /**
     *  Return everything in the routing table and leaf set
     */
    public synchronized I_AddressPair[] getAllNeighbors()
    {
        return inOverlay ? hyperCastNodeAdapter.getAllNeighbors() : new I_AddressPair[0];
    }

    public I_AddressPair getAddressPair()
    {
        I_PhysicalAddress p = createPhysicalAddress();
        Pastry_LogicalAddress l = (Pastry_LogicalAddress) createLogicalAddress();
        I_AddressPair pair = new Pastry_AddressPair (p, l);
        return pair;
    }

    public I_LogicalAddress createLogicalAddress()
    {
        return new Pastry_LogicalAddress (hcPastryNode.getNodeId());
    }

    public I_LogicalAddress createLogicalAddress (byte[] laddr, int offset)
    {
        int size = Pastry_LogicalAddress.getSizeStatic();
        if (laddr.length - offset < size)
        {
            return null;
        }
        else
        {
            byte [] bytes = new byte [size];
            System.arraycopy (laddr, offset, bytes, 0, size);
            return new Pastry_LogicalAddress (bytes);
        }
    }

    public I_LogicalAddress createLogicalAddress (String logicalAddressString)
    {
        return new Pastry_LogicalAddress (logicalAddressString);
    }

    public I_PhysicalAddress createPhysicalAddress (String physicalAddressString)
    {        
        return adapter.createPhysicalAddress (physicalAddressString);
    }

    public I_PhysicalAddress createPhysicalAddress()
    {
        return adapter.createPhysicalAddress();
    }

    public long getCurrentTime() {
        return adapter.getCurrentTime();
    }

    public void setLogicalAddress (I_LogicalAddress address)
    {
        if (! (address instanceof Pastry_LogicalAddress))
        {
            IllegalArgumentException iae = new IllegalArgumentException ("Not a Pastry logical address: " + address);
            throw new HyperCastFatalRuntimeException (iae);
        }
        else
        {
            leaveOverlay ((Pastry_LogicalAddress) address);
            joinOverlay();
        }
    }

    public void setNotificationHandler (NotificationHandler nh)
    {
        this.notificationHandler = nh;
    }

    /**
     * Unlike the DT protocol, a Pastry node cannot make a local
     * decision to compute is children, instead it must send to all
     * neighbors and each neighbor decides if it should have received any particular message.
     *
     * basically: if the previous hop is my parent on the path to the
     * source: ok, else not ok
     *
     * "reverse path forwarding"
     */
    public boolean previousHopCheck (final I_LogicalAddress src, final I_LogicalAddress dst, final I_LogicalAddress prehop)
    {
        final I_AddressPair [] parentArray = getParent (src); // always array of length one in Pastry
        final I_LogicalAddress parent = parentArray [0].getLogicalAddress();

        final I_LogicalAddress me = createLogicalAddress();

        if (debug)
            System.err.println ("previousHopCheck: prehop (" + prehop + ") parent (" + parent + ") src (" + src + ") dst (" + dst + ") me (" + me + ")");

        return (dst != null) || (parent.equals (prehop) && ! me.equals (prehop));
    }

    /* package */ void sendUnicastMessage (final Pastry_AddressPair dstPair, final Pastry_AddressPair srcPair, final HyperCastMessage hcMessage)
    {
        Pastry_Message msg = new Pastry_Message ((byte) '0', Pastry_Message.PASTRY_ENCAPSULATED, config.getOverlayHash(), srcPair, dstPair, hcMessage);
        I_PhysicalAddress dst = dstPair.getPhysicalAddress();

        if (debug)
            System.err.println ("PASTRY DEBUG: sendUnicastMessage dst (from bytes): " + dst);

        adapter.sendUnicastMessage (dst, msg);
    }

    /* package */ void sendUnicastMessage (final Pastry_AddressPair dstPair, final Pastry_AddressPair srcPair, final I_Message msg)
    {
        I_PhysicalAddress dst = dstPair.getPhysicalAddress();

        if (debug)
            System.err.println ("PASTRY DEBUG: sendUnicastMessage dst (from pastry message): " + dst);

        adapter.sendUnicastMessage (dst, msg);
    }

    /*
     *  The computeClosest call is designed to set the value of the
     *  variable currentClosest. It can do this if all of the
     *  proximity values needed for the computation exist in the
     *  proximity cache.
     *
     *  If the necessary proximity cache values exist, the function
     *  returns a value of true, meaning that currentClosest is set
     *  properly.
     *
     *  If the proximity cache is missing any needed value,
     *  computeClosest acts like an asynchronous method invocation,
     *  returning before the computation for which it was called has
     *  completed; this result is indicated by returning a value of
     *  false to the caller. In this case, it will initiate send
     *  operations for node id request messages that will determine
     *  any missing proximity values. Only when the responses to those
     *  message arrive (or timeout) can the computation be deemed
     *  complete.
     */

    private boolean computeClosest (HCNodeHandle [] handles)
    {
        /*
         *  Optimistically set this to false; on a proximity cache miss
         *  it will be set to true.
         */
        boolean cacheMiss = false;

        Long delta = factory.proximityCacheGet (currentClosest.getNodeId());
        if (delta == null)
        {
            if (debug_join_phase_2)
                System.err.println (lastr() + ": proximityCache.get (currentClosest) == null... BUG: current closest should always have an entry in the proximity cache");
        }
        else
        {
            nearestdist = delta.longValue();
        }

        if (debug_join_phase_2) System.err.println (lastr() + ": Computing closest node from list of " + handles.length + " nodes");
        for (int i=0; i < handles.length; i++)
        {
            HCNodeHandle tempNode = handles [i];

            delta = factory.proximityCacheGet (tempNode.getNodeId());
            if (delta != null)
            {
                long dist = delta.longValue();
                if (debug_join_phase_2) System.err.println (lastr() + ": Using cached value for proximity: " + tempNode + " -> " + dist);
                if (dist < nearestdist)
                {
                    updateCurrentClosest (tempNode, dist);
                }
            }
            else
            {
                cacheMiss = true;
                sendNodeIdRequestMessageForCloseness (tempNode);
            }
        }
        return (! cacheMiss);
    }

    private void updateCurrentClosest (HCNodeHandle newClosest, long distance)
    {
        nearestdist = distance;
        currentClosest = newClosest;
        if (debug_join_phase_2) System.err.println (lastr() + ": New current closest is: " + currentClosest + " at a distance of " + nearestdist);
    }

    /*
     *  Called from the adapter's message processor thread whenever a
     *  message is read from the input queue. Processes the message.
     */
    public synchronized void messageArrivedFromAdapter (I_Message typeless_msg)
    {
        if (! (typeless_msg instanceof Pastry_Message))
        {
            throw new IllegalArgumentException ("Pastry_Node_BuddyList.messageArrivedFromAdapter() passed non-Pastry_Message I_Message.");
        }
        Pastry_Message msg = (Pastry_Message) typeless_msg;

        I_PhysicalAddress otherPhysicalAddress = msg.getSrc().getPhysicalAddress();
        I_LogicalAddress otherLogicalAddress = msg.getSrc().getLogicalAddress();
        Pastry_LogicalAddress otherPastryLogicalAddress = (Pastry_LogicalAddress) otherLogicalAddress;
        byte [] payload = msg.getPayload();

        hyperCastNodeAdapter.markAlive (otherPastryLogicalAddress.getNodeId());

        switch (msg.getType())
        {
            case Pastry_Message.NODE_ID_REQUEST:
                respondToNodeIdRequestMessage (otherPhysicalAddress, payload);
                break;

            case Pastry_Message.NODE_ID_RESPONSE:
                handleNodeIdResponse (otherPhysicalAddress, otherPastryLogicalAddress, payload);
                break;

            case Pastry_Message.LEAFSET_REQUEST:
                respondToLeafSetRequest (otherPhysicalAddress);
                break;

            case Pastry_Message.LEAFSET_RESPONSE:
                handleLeafSetResponse (otherPhysicalAddress, payload);
                break;

            case Pastry_Message.ROUTEROW_REQUEST:
                respondToRouteRowRequest (otherPhysicalAddress, payload);
                break;

            case Pastry_Message.ROUTEROW_RESPONSE:
                handleRouteRowResponse (otherPhysicalAddress, payload);
                break;

            case Pastry_Message.PASTRY_ENCAPSULATED:
            {
                if (debug) System.err.println (lastr() + ": PASTRY DEBUG: messageArrivedFromAdapter: Encapsulated Pastry Message");
                HyperCastMessage hcm = msg.getHyperCastMessage();
                if (hcm != null)
                    hyperCastNodeAdapter.receiveMessage (hcm);
            }
            break;

            default:
                System.err.println (lastr() + ": Unknown message type received \"" + msg.getType() + "\", ignoring...");
        }
    }

    private void requestNodeIdOfBuddy (final I_PhysicalAddress buddyAddress)
    {
        if (debug_buddy) System.err.println (lastr() + ": Sending NODE_ID_REQUEST (buddy search) to " + buddyAddress);
        long now = adapter.getCurrentTime();
        byte [] payload = ByteArrayUtility.toByteArray (now);
        sendUnicastCommandMessage (buddyAddress, Pastry_Message.NODE_ID_REQUEST, payload);
        adapter.setTimer (new NodeIdRequestTimeout(), nodeIdRequestTimeout);
    }

    /* package */ void ping (final HyperCastPhysicalAddress otherPhysicalAddress, final NodeId otherLogicalAddress)
    {
        final HCNodeHandle pingTarget = (HCNodeHandle) factory.generateNodeHandle (otherPhysicalAddress, otherLogicalAddress, timeoutTime);
        sendNodeIdRequestMessageForCloseness (pingTarget);
    }

    private void sendNodeIdRequestMessageForCloseness (final HCNodeHandle otherNode)
    {
        if (state == STATE_BUDDY_SEARCH)
        {
            // ok
        }
        else if (state != STATE_JOINED) // ok for ping
        {
            switch (joinSubState)
            {
                case JOIN_SUBSTATE_PASTRY_JOIN:            /* fall-through */
                case JOIN_SUBSTATE_LEAFSET_CLOSEST:        /* fall-through */
                case JOIN_SUBSTATE_ROUTEROW_PASS1_CLOSEST: /* fall-through */
                case JOIN_SUBSTATE_ROUTEROW_PASS2_CLOSEST:
                    /* these states are acceptable */
                    break;

                default:
                    config.log.println ("Ignoring sendNodeIdRequestMessageForCloseness in: " + stateString());
                    //throw new AssertionError (stateString());
            }
        }

        if (debug) System.err.println (lastr() + ": Sending NODE_ID_REQUEST (proximity) to " + otherNode);
        long now = adapter.getCurrentTime();
        byte [] payload = ByteArrayUtility.toByteArray (now);
        ++pendingProximityResponseMessages;
        if (debug_pendingProximityResponseMessages) System.err.println (lastr() + ": Incrementing pendingProximityResponseMessages in send request for closeness: " + pendingProximityResponseMessages);
        sendUnicastCommandMessage (otherNode, Pastry_Message.NODE_ID_REQUEST, payload);
        adapter.setTimer (new NodeIdRequestClosenessTimeout (now, otherNode), nodeIdRequestTimeout);
    }

    private void respondToNodeIdRequestMessage (final I_PhysicalAddress otherPhysicalAddress, byte [] payload)
    {
        /*
         *  We can respond with our node id only if we have joined a ring
         */
        if (state == STATE_JOINED)
        {
            if (debug || debug_buddy) System.err.println (lastr() + ": Received NODE_ID_REQUEST from " + otherPhysicalAddress);
            sendUnicastCommandMessage (otherPhysicalAddress, Pastry_Message.NODE_ID_RESPONSE, payload);
        }
        else
        {
            if (debug || debug_buddy) System.err.println (lastr() + ": Received NODE_ID_REQUEST from " + otherPhysicalAddress + ": " + stateString() + ", ignoring...");
        }
    }

    private void handleNodeIdResponse (final I_PhysicalAddress otherPhysicalAddress, final Pastry_LogicalAddress otherPastryLogicalAddress, final byte [] payload)
    {
        long then = ByteArrayUtility.toLong (payload, 0);
        long now = adapter.getCurrentTime();
        long delta = now - then;
        long oneWayDelayApprox = delta / 2;

        switch (state)
        {
            case STATE_JOINED:
            {
                if (debug) System.err.println (lastr() + " RECEIVED PING RESPONSE from " + otherPhysicalAddress + " " + otherPastryLogicalAddress);
                adapter.clearTimer (new NodeIdRequestClosenessTimeout (then, otherPhysicalAddress));
                hyperCastNodeAdapter.markAliveWithProximity (otherPastryLogicalAddress.getNodeId(), (int) oneWayDelayApprox);
            }
            break;

            case STATE_JOIN_INITIATED:
            {
                switch (joinSubState)
                {
                    case JOIN_SUBSTATE_LEAFSET_CLOSEST:
                    case JOIN_SUBSTATE_ROUTEROW_PASS1_CLOSEST:
                    case JOIN_SUBSTATE_ROUTEROW_PASS2_CLOSEST:
                    {
                        if (debug_join_phase_2) System.err.println (lastr() + ": Received NODE_ID_RESPONSE (proximity) from " + otherPastryLogicalAddress + "@" + otherPhysicalAddress + ", while in state: " + stateString());
                                
                        adapter.clearTimer (new NodeIdRequestClosenessTimeout (then, otherPhysicalAddress));
                        factory.proximityCachePut (otherPastryLogicalAddress.getNodeId(), oneWayDelayApprox);

                        if (then >= subStateStartTime
                            &&
                            delta < nearestdist)
                        {
                            final HCNodeHandle newClosest = (HCNodeHandle) factory.generateNodeHandle (new HyperCastPhysicalAddress (otherPhysicalAddress), otherPastryLogicalAddress.getNodeId(), timeoutTime);
                            newClosest.setLocalNode (hcPastryNode);
                            updateCurrentClosest (newClosest, delta);
                        }
                        --pendingProximityResponseMessages;
                        if (debug_pendingProximityResponseMessages) System.err.println (lastr() + ": PASTRY DEBUG: decrementing pendingProximityResponseMessages in message receive: " + pendingProximityResponseMessages);
                        if (pendingProximityResponseMessages < 0) throw new AssertionError ("Underflow pendingProximityResponseMessages");

                        if (pendingProximityResponseMessages == 0)
                        {
                            joinPhase2DoNextStep();
                        }
                    }
                    break;
                              
                    default:
                    {
                        if (debug) System.err.println (lastr() + ": Received NODE_ID_RESPONSE from " + otherPastryLogicalAddress + "@" + otherPhysicalAddress + ", while in state: " + stateString() + ", ignoring...");
                    }
                }
            }
            break;
                        
            case STATE_BUDDY_SEARCH:
            {
                /*
                 *  handle buddy reply
                 *  
                 *  cancel buddy timer
                 *  create node handle with (address, id) tuple
                 *  call donenode
                 */

                if (debug_buddy) System.err.println (lastr() + ": Received NODE_ID_RESPONSE (buddy search) from " + otherPastryLogicalAddress + "@" + otherPhysicalAddress + ", while in state: " + stateString());

                factory.proximityCachePut (otherPastryLogicalAddress.getNodeId(), oneWayDelayApprox);

                buddyFound = true;
                changeState (STATE_BUDDY_FOUND);
                adapter.clearTimer (new NodeIdRequestTimeout());

                NodeId buddyNodeId = otherPastryLogicalAddress.getNodeId();

                NodeHandle buddyHandle = factory.generateNodeHandle (new HyperCastPhysicalAddress (otherPhysicalAddress), buddyNodeId, timeoutTime);
                buddyHandle.setLocalNode (hcPastryNode);
                joinEstablishedPastryRing (buddyHandle);
            }
            break;

            default:
            {
                if (debug) System.err.println (lastr() + ": Received NODE_ID_RESPONSE from " + otherPastryLogicalAddress + "@" + otherPhysicalAddress + ", while in state: " + stateString() + ", ignoring...");
            }
        }
    }

    /*
     *  This is Stage 1 of Phase 2 of the Join Sequence
     */
    private void sendLeafSetRequest (HCNodeHandle otherNode)
    {
        if (debug_join_phase_2) System.err.println (lastr() + ": Sending Leaf Set request to " + otherNode);
        sendUnicastCommandMessage (otherNode, Pastry_Message.LEAFSET_REQUEST, null);
        adapter.setTimer (new JoinPhase2MessageTimeout (otherNode), joinMessageTimeout);
    }

    private void respondToLeafSetRequest (final I_PhysicalAddress otherNode)
    {
        if (debug_join_phase_2) System.err.println (lastr() + ": Received LEAFSET_REQUEST from: " + otherNode);
        byte [] payload = hyperCastNodeAdapter.getSerializedLeafSet();
        sendUnicastCommandMessage (otherNode, Pastry_Message.LEAFSET_RESPONSE, payload);
    }

    private void handleLeafSetResponse (final I_PhysicalAddress otherPhysicalAddress, final byte [] payload)
    {
        adapter.clearTimer (new JoinPhase2MessageTimeout (otherPhysicalAddress));

        if (joinSubState == JOIN_SUBSTATE_LEAFSET)
        {
            if (debug_join_phase_2) System.err.println (lastr() + ": Received LEAFSET_RESPONSE from: " + otherPhysicalAddress);
            byte [] serializedLeafSet = payload;
            HCNodeHandle [] handles = hyperCastNodeAdapter.constructNodeHandleArrayFromSerializedLeafSet (serializedLeafSet);

            if (debug)
            {
                if (handles.length == 0)
                {
                    if (debug)
                        System.err.println (lastr() + ": LEAFSET_RESPONSE contained an empty leaf set");
                }
                else
                {
                    if (debug_join_phase_2) System.err.println (lastr() + ": LEAFSET_RESPONSE contained this leaf set:");
                    for (int i = 0; i < handles.length; ++i)
                    {
                        System.err.println (lastr() + ":     handles[" + i + "]: " + handles[i]);
                    }
                }
            }

            /*
             *  We asked for the leaf set of the bootstrap so that we
             *  can compute the leaf that is closest to us. We do this
             *  by "pinging," i.e. sending NODE_ID_REQUESTs, to each
             *  leaf. Before doing this we must set the substate to
             *  "leafset closest" so that when NODE_ID_RESPONSEs come
             *  back, or when timeouts occur, we can interpret them
             *  correctly. The computeClosest method is the one that
             *  sends all of the NodeId requests (and sets timers),
             *  that method consults a cache of previous ping values,
             *  so it is possible that it doesn't have to send any
             *  messages over the wire. In that case it sets the value
             *  of newClosest - this should be rare if not impossible
             *  for the leaf set response stage (Stage 1) of the
             *  algorithm, but for consistency, the code here is
             *  exactly the same as the code for the route row stages
             *  (Stages 2 & 3).
             */

            doComputeClosest (JOIN_SUBSTATE_LEAFSET_CLOSEST, handles);
        }
        else
        {
            if (debug_join_phase_2) System.err.println (lastr() + ": Received LEAFSET_RESPONSE: not expecting this message: " + stateString() + ", ignoring...");
        }
    }

    private void sendRouteRowRequest (HCNodeHandle otherNode, int row)
    {
        if (debug_join_phase_2) System.err.println (lastr() + ": Requesting Route Row " + row + " from " + otherNode);
        long now = adapter.getCurrentTime();
        byte [] payload = new byte [8 + 4];
        System.arraycopy (ByteArrayUtility.toByteArray (now), 0, payload, 0, 8);
        System.arraycopy (ByteArrayUtility.toByteArray (row), 0, payload, 8, 4);
        routeRowReceived = false;
        sendUnicastCommandMessage (otherNode, Pastry_Message.ROUTEROW_REQUEST, payload);
        adapter.setTimer (new JoinPhase2MessageTimeout (now, otherNode), joinMessageTimeout);
    }

    private void respondToRouteRowRequest (final I_PhysicalAddress otherAddress, final byte [] payload)
    {
        long then = ByteArrayUtility.toLong (payload, 0);
        int row = ByteArrayUtility.toInteger (payload, 8);

        if (debug_join_phase_2) System.err.println (lastr() + ": Received ROUTEROW_REQUEST from " + otherAddress + " for row " + row + " (time " + then + ")");

        int returnPayloadLength = 8; // timestamp
        byte [] routerow = hyperCastNodeAdapter.getSerializedRouteRow (row);
        if (routerow != null)
        {
            returnPayloadLength += routerow.length;
        }

        byte [] returnPayload = new byte [returnPayloadLength];
        System.arraycopy (ByteArrayUtility.toByteArray (then), 0, returnPayload, 0, 8);
        if (routerow != null)
        {
            System.arraycopy (routerow, 0, returnPayload, 8, routerow.length);
        }

        sendUnicastCommandMessage (otherAddress, Pastry_Message.ROUTEROW_RESPONSE, returnPayload);
    }

    private void handleRouteRowResponse (final I_PhysicalAddress otherPhysicalAddress, final byte [] payload)
    {
        /*
         *  It only makes sense to receive a
         *  ROUTEROW_RESPONSE when in the route row
         *  request states (either pass 1 or pass 2)
         */
        int nextSubState = JOIN_SUBSTATE_INVALID;
        switch (joinSubState)
        {
            case JOIN_SUBSTATE_ROUTEROW_PASS1:
                nextSubState = JOIN_SUBSTATE_ROUTEROW_PASS1_CLOSEST;
                break;

            case JOIN_SUBSTATE_ROUTEROW_PASS2:
                nextSubState = JOIN_SUBSTATE_ROUTEROW_PASS2_CLOSEST;
                break;
                        
            default:
            {
                if (debug_join_phase_2) System.err.println (lastr() + ": Received ROUTEROW_RESPONSE, not expecting this type of message " + stateString() + ", ignoring...");
            }
        }
        if (nextSubState == JOIN_SUBSTATE_INVALID) return;

        long then = ByteArrayUtility.toLong (payload, 0);

        adapter.clearTimer (new JoinPhase2MessageTimeout (then, otherPhysicalAddress));

        /*
         *  If a properly formatted ROUTEROW_RESPONSE
         *  arrives in timely fashion, use its contents to
         *  compute the closest node. Otherwise move to the
         *  next step of the Phase 2 algorithm.
         */
        if (payload.length < 8)
        {
            System.err.println (lastr() + ": Ill-formed ROUTEROW_RESPONSE");
            joinPhase2DoNextStep(); // give up on this row and move on
        }
        else if (payload.length == 8)
        {
            //long then = ByteArrayUtility.toLong (payload, 0);
            System.err.println (lastr() + ": Received ROUTEROW_RESPONSE from time " + then + " with no route row information, ignoring...");
            joinPhase2DoNextStep(); // give up on this row and move on
        }
        else // payload.length > 8
        {
            //long then = ByteArrayUtility.toLong (payload, 0);

            if (then >= subStateStartTime)
            {
                if (debug_join_phase_2) System.err.println (lastr() + ": Received ROUTEROW_RESPONSE, (time " + then + ") arrived in time");
                routeRowReceived = true;
                byte [] serializedRouteRow = new byte [payload.length - 8];
                System.arraycopy (payload, 8, serializedRouteRow, 0, payload.length - 8);
                HCNodeHandle [] handles = hyperCastNodeAdapter.constructNodeHandleArrayFromSerializedRouteRow (serializedRouteRow);
                doComputeClosest (nextSubState, handles);
            }
            else
            {
                if (debug_join_phase_2) System.err.println (lastr() + ": Received ROUTEROW_RESPONSE, (time " + then + ", subStateStartTime " + subStateStartTime + ", diff (subStateStartTime - then) " + (subStateStartTime - then) + ", timeout " + joinMessageTimeout + ") arrived too late! ignoring...");
                joinPhase2DoNextStep();
            }
        }
    }

    private void sendUnicastCommandMessage (final HCNodeHandle dst, final int type, final byte [] payload)
    {
        I_PhysicalAddress dstAddress = dst.getAddress().getPhysicalAddress (hyperCastNodeAdapter);
        sendUnicastCommandMessage (dstAddress, type, payload);
    }

    private void sendUnicastCommandMessage (final I_PhysicalAddress dstPhysical, final int type, final byte [] payload)
    {
        final I_PhysicalAddress srcPhysical = createPhysicalAddress();
        final Pastry_LogicalAddress srcLogical = (Pastry_LogicalAddress) createLogicalAddress();

        final Pastry_AddressPair dstPair = new Pastry_AddressPair (dstPhysical, null);
        final Pastry_AddressPair srcPair = new Pastry_AddressPair (srcPhysical, srcLogical);

        final Pastry_Message pastryMsg = new Pastry_Message ((byte) '0', type, config.getOverlayHash(), srcPair, dstPair, payload);
        sendUnicastMessage (dstPair, srcPair, pastryMsg);
    }

    /*
     *  This is Stage 1 of Phase 2 of the Join Sequence
     */
    private void beginLeafSetRequestStage()
    {
        changeSubState (JOIN_SUBSTATE_LEAFSET);
        sendLeafSetRequest (currentClosest);
    }

    /*
     *  This is Stage 2 of Phase 2 of the Join Sequence
     */
    private void beginRouteRowPass1()
    {
        if (! disableJoinSequencePhase2Stage2)
        {
            if (debug_join_phase_2) System.err.println (lastr() + ": Beginning Route Row exchange #1");

            routeRowPass1CurrentDepth = 0;
            requestNextRouteRowPass1();
        }
        else
        {
            beginRouteRowPass2();
        }
    }

    private void requestNextRouteRowPass1()
    {
        if (routeRowPass1CurrentDepth >= JOIN_ROUTING_TABLE_DEPTH)
        {
            if (debug_join_phase_2) System.err.println (lastr() + ": All route row exchanges complete for pass #1, initiating exchanges of route rows for pass #2");
            beginRouteRowPass2();
        }
        else
        {
            changeSubState (JOIN_SUBSTATE_ROUTEROW_PASS1);
            sendRouteRowRequest (currentClosest, routeRowPass1CurrentDepth);
            ++routeRowPass1CurrentDepth;
        }
    }

    /*
     *  This unconditional step implements the first iteration of the
     *  do loop from PastryNodeFactory.getNearest().
     *
     *  This is Stage 3 of Phase 2 of the Join Sequence
     */
    private void beginRouteRowPass2()
    {
        if (! disableJoinSequencePhase2Stage3)
        {
            if (debug_join_phase_2) System.err.println (lastr() + ": Beginning Route Row exchange #2");
            routeRowPass2Iteration();
        }
        else
        {
            completePhase2();
        }
    }
  
    private void routeRowPass2Iteration()
    {
        changeSubState (JOIN_SUBSTATE_ROUTEROW_PASS2);
        stage3PreviousClosest = currentClosest;
        sendRouteRowRequest (currentClosest, JOIN_ROUTING_TABLE_DEPTH - 1);
    }

    private void nextIterationRouteRowPass2()
    {
        if (stage3PreviousClosest.equals (currentClosest))
        {
            completePhase2();
        }
        else
        {
            routeRowPass2Iteration();
        }
    }

    private void completePhase2()
    {
        if (debug_state_change)
            System.err.println (lastr() + ": nearest physical node identified as: " + currentClosest + ", initiating Pastry Join operation...");

        currentClosest.setLocalNode (hcPastryNode);
        beginPhase3 (currentClosest);
    }

    private void doComputeClosest (int nextSubState, HCNodeHandle [] handles)
    {
        changeSubState (nextSubState);
        boolean completedSynchronously = computeClosest (handles);
        if (completedSynchronously)
        {
            joinPhase2DoNextStep();
        }
    }

    /*
     *  This method replaces PastryNodeFactory.getNearest()
     */
    private void joinPhase2DoNextStep()
    {
        //if (pendingProximityResponseMessages != 0) throw new AssertionError ("pendingProximityResponseMessages != 0: " + pendingProximityResponseMessages);

        switch (joinSubState)
        {
            case JOIN_SUBSTATE_LEAFSET:
            {
                if (debug_join_phase_2) System.err.println (lastr() + ": joinPhase2DoNextStep, sending leaf set request");
                sendLeafSetRequest (currentClosest);
            }
            break;

            case JOIN_SUBSTATE_LEAFSET_CLOSEST:
            {
                if (debug_join_phase_2) System.err.println (lastr() + ": joinPhase2DoNextStep, initiating route row pass 1");
                beginRouteRowPass1();
            }
            break;

            case JOIN_SUBSTATE_ROUTEROW_PASS1:
            case JOIN_SUBSTATE_ROUTEROW_PASS1_CLOSEST:
            {
                if (debug_join_phase_2) System.err.println (lastr() + ": joinPhase2DoNextStep, continuing with next route row pass 1 iteration");
                requestNextRouteRowPass1();
            }
            break;

            case JOIN_SUBSTATE_ROUTEROW_PASS2:
            case JOIN_SUBSTATE_ROUTEROW_PASS2_CLOSEST:
            {
                if (debug_join_phase_2) System.err.println (lastr() + ": joinPhase2DoNextStep, continuing with next route row pass 2 iteration");
                nextIterationRouteRowPass2();
            }
            break;
                      
            default:
            {
                throw new AssertionError ("joinPhase2DoNextStep called but not in Phase 2: " + stateString());
            }
        }
    }

    /*
     *  Base class of all timer objects that are created in this
     *  file. The handler method is similar to the run method of
     *  Java's TimerTask, see the JavaDoc.
     */
    private abstract class TimerObject
    {
        public abstract void handler();
    }

    private abstract class TimestampAddressTimeout extends TimerObject
    {
        private long timestamp;
        private I_PhysicalAddress address;

        private TimestampAddressTimeout() {}

        protected TimestampAddressTimeout (final HCNodeHandle otherNode) {
            this (0, otherNode);
        }

        protected TimestampAddressTimeout (final I_PhysicalAddress otherPhysicalAddress) {
            this (0, otherPhysicalAddress);
        }

        protected TimestampAddressTimeout (final long timestamp, final HCNodeHandle otherNode) {
            this (timestamp, otherNode.getAddress().getPhysicalAddress (hyperCastNodeAdapter));
        }

        protected TimestampAddressTimeout (final long timestamp, final I_PhysicalAddress otherPhysicalAddress)
        {
            this.timestamp = timestamp;
            this.address = otherPhysicalAddress;
        }

        public long getTimestamp() { return timestamp; }
        public I_PhysicalAddress getAddress() { return address; }

        public boolean equals (Object obj)
        {
            boolean result = false;
            if (obj instanceof TimestampAddressTimeout)
            {
                TimestampAddressTimeout other = (TimestampAddressTimeout) obj;
                result =
                    other.timestamp == this.timestamp
                    &&
                    other.address.equals (this.address);
                /*
                System.out.println ("TimestampAddressTimeout equals comparing (" +
                                    this.timestamp + ", " + this.address + ") to (" +
                                    other.timestamp + ", " + other.address + "). Returning: " + result);
                */
            }
            return result;
        }

        public String toString() {
            return address + "@" + timestamp;
        }
    }

    private class NodeIdRequestClosenessTimeout extends TimestampAddressTimeout
    {
        private NodeId otherLogicalAddress;

        public NodeIdRequestClosenessTimeout (final long timestamp, final HCNodeHandle otherNode) {
            super (timestamp, otherNode);
            this.otherLogicalAddress = otherNode.getNodeId();
        }
        
        public NodeIdRequestClosenessTimeout (final long timestamp, final I_PhysicalAddress otherPhysicalAddress) {
            super (timestamp, otherPhysicalAddress);
        }

        public boolean equals (Object obj)
        {
            boolean result = false;
            if (obj instanceof NodeIdRequestClosenessTimeout)
            {
                result = super.equals (obj);
            }
            //System.out.println ("equals for NodeIdRequestClosenessTimeout returning " + result);
            return result;
        }

        public void handler()
        {
            if (debug)
                System.err.println (lastr() + ": Timeout: NODE_ID_REQUEST (proximity): " + this.toString());
            
            if (state != STATE_JOINED && joinSubState != JOIN_SUBSTATE_PASTRY_JOIN) // not ping
            {
                if (getTimestamp() >= subStateStartTime)
                {
                    --pendingProximityResponseMessages;
                    if (debug_pendingProximityResponseMessages) System.err.println (lastr() + ": decrementing pendingProximityResponseMessages in timeout: " + pendingProximityResponseMessages);
                    if (pendingProximityResponseMessages == 0)
                    {
                        joinPhase2DoNextStep();
                    }
                }
            }
        }
    }

    private class JoinPhase2MessageTimeout extends TimestampAddressTimeout
    {
        public JoinPhase2MessageTimeout (HCNodeHandle otherNode) {
            super (otherNode);
        }

        public JoinPhase2MessageTimeout (final I_PhysicalAddress otherPhysicalAddress) {
            super (otherPhysicalAddress);
        }

        public JoinPhase2MessageTimeout (final long timestamp, final HCNodeHandle otherNode) {
            super (timestamp, otherNode);
        }

        public JoinPhase2MessageTimeout (final long timestamp, final I_PhysicalAddress otherPhysicalAddress) {
            super (timestamp, otherPhysicalAddress);
        }

        public boolean equals (Object obj)
        {
            boolean result = false;
            if (obj instanceof JoinPhase2MessageTimeout)
            {
                result = super.equals (obj);
            }
            return result;
        }

        public void handler()
        {
            switch (joinSubState)
            {
                case JOIN_SUBSTATE_LEAFSET:
                {
                    /*
                     *  We asked for a Leafset from the current physically
                     *  closest node, but it did not respond fast enough
                     *  for us. We will go to the next step of the
                     *  algorithm which is route row exchange, pass 1.
                     */
                    if (debug_join_phase_2) System.err.println (lastr() + ": Timeout: Join Phase 2: leaf set was not returned promptly: " + this);
                    joinPhase2DoNextStep();
                }
                break;

                case JOIN_SUBSTATE_ROUTEROW_PASS1:
                {
                    long msgTimeStamp = getTimestamp();

                    String debugString =
                        lastr() + ": Timeout: Join Phase 2: " +
                        "msgTimeStamp (" + msgTimeStamp + ") " +
                        "subStateStartTime (" + subStateStartTime + ") " +
                        "diff subStateStartTime - msgTimeStamp (" + (subStateStartTime - msgTimeStamp) + ") " +
                        "routeRowReceived (" + routeRowReceived + ")";

                    if ((msgTimeStamp >= subStateStartTime) && (! routeRowReceived))
                    {
                        if (debug_join_phase_2) System.err.println (debugString + ": route row for pass #1 was not returned promptly: " + this);
                        joinPhase2DoNextStep();
                    }
                }
                break;

                case JOIN_SUBSTATE_ROUTEROW_PASS2:
                {
                    long msgTimeStamp = getTimestamp();

                    String debugString =
                        lastr() + ": Timeout: Join Phase 2: " +
                        "msgTimeStamp (" + msgTimeStamp + ") " +
                        "subStateStartTime (" + subStateStartTime + ") " +
                        "diff subStateStartTime - msgTimeStamp (" + (subStateStartTime - msgTimeStamp) + ") " +
                        "routeRowReceived (" + routeRowReceived + ")";

                    if ((msgTimeStamp >= subStateStartTime) && (! routeRowReceived))
                    {
                        if (debug_join_phase_2) System.err.println (debugString + ": route row for pass #2 was not returned promptly: " + this.toString());
                        completePhase2();
                    }
                }
                break;

                default:
                {
                    if (debug_join_phase_2) System.err.println (lastr() + ": " + stateString() + ", ignoring...");
                }
            }
        }
    }

    private class NodeIdRequestTimeout extends TimerObject
    {
        public String toString()
        {
            return "NodeIdRequestTimeout";
        }

        public boolean equals (Object obj)
        {
            boolean result = false;
            if (obj instanceof NodeIdRequestTimeout)
            {
                result = true;
            }
            return result;
        }

        public void handler()
        {
            if (debug || (buddyFound == false && debug_buddy)) System.err.println (lastr() + ": Timeout: NODE_ID_REQUEST (buddy search): " + this.toString());

            if (buddyFound == true)
            {
                if (debug_buddy) System.err.println (lastr() + ": I already have a buddy");
                return;
            }

            I_PhysicalAddress buddyAddress = getNextBuddyAddress();
            if (buddyAddress != null)
            {
                requestNodeIdOfBuddy (buddyAddress);
            }
            else
            {
                buddyFound = true; // true in the sense that I am my own buddy
                startNewPastryRing();
            }
        }
    }

    /*
     *  Called from the adapter's timer thread whenever a timer event
     *  occurs.
     */
    public synchronized void timerExpired (Object timerID)
    {
        if (timerID instanceof TimerObject) // Timer event set by code in this file
        {
            TimerObject timerObject = (TimerObject) timerID;
            timerObject.handler();
        }
        else if (timerID instanceof HyperCastScheduledMessage) // Timer event set by Pastry embedded "black-box"
        {
            if (debug) System.err.println (lastr() + ": Timeout: HyperCastScheduledMessage");
            HyperCastScheduledMessage hcsm = (HyperCastScheduledMessage) timerID;
            hyperCastTimer.timerExpired (hcsm);            
        }
        else
        {
            throw new AssertionError (lastr() + ": Timeout: Unknown timer event type");
        }
    }

    /*
     *  Called from the node adapter's message receiver thread
     *  whenever a message is read from the network. Parses the message. 
     */
    public synchronized I_Message restoreMessage (byte [] receiveBuffer, int [] validBytesStart, int validBytesEnd)
    {
        return Pastry_Message.restoreMessage (receiveBuffer, validBytesStart, validBytesEnd, adapter, config.getOverlayHash());
    }

    private void initStatisticsStructure()
    {
        statisticsName = config.getTextAttribute (XmlUtil.createXPath (PROPERTY_PROTO_PREFIX + "StatName"));

        statsProcessor.addStatsElement ("NodeAdapter",                  adapter,                                1, 1);
        statsProcessor.addStatsElement ("State",                        new StateStat(),                        1, 1);
        statsProcessor.addStatsElement ("JoinSubState",                 new JoinSubStateStat(),                 1, 1);
        statsProcessor.addStatsElement ("LogicalAddress",               new LogicalAddressStat(),               1, 1);
        statsProcessor.addStatsElement ("PhysicalAddress",              new PhysicalAddressStat(),              1, 1);
        statsProcessor.addStatsElement ("NumOfNeighbors",               new NumOfNeighborsStat(),               1, 1);
        statsProcessor.addStatsElement ("LeafSetMaintenanceFrequency",  new LeafSetMaintenanceFrequencyStat(),  1, 1);
        statsProcessor.addStatsElement ("RouteSetMaintenanceFrequency", new RouteSetMaintenanceFrequencyStat(), 1, 1);
        statsProcessor.addStatsElement ("PingTimeout",                  new NodeIdRequestTimeoutStat(),         1, 1);
        statsProcessor.addStatsElement ("JoinMessageTimeout",           new JoinPhase2MessageTimeoutStat(),     1, 1);
        statsProcessor.addStatsElement ("BuddyRetries",                 new BuddyRetriesStat(),                 1, 1);
        statsProcessor.addStatsElement ("LeafSetSize",                  new LeafSetSizeStat(),                  1, 1);
        statsProcessor.addStatsElement ("RouteSetSize",                 new RouteSetSizeStat(),                 1, 1);
    }

    abstract class PastryStat extends SimpleStats
    {
        protected final String statType;

        public PastryStat (final String statType) { this.statType = statType; }

        protected String setStats (String newValue) throws HyperCastStatsException {
            throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
        }

        public final Element [] getReadSchema (Document doc, XPath xpath) throws HyperCastStatsException {
            return XmlUtil.createSchemaElement (doc, statisticsName, statType, null, null);
        }

        public final Element [] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
            return XmlUtil.createSchemaElement (doc, statisticsName, statType, null, null);
        }
    }

    final class StateStat extends PastryStat {
        public StateStat() { super ("xsd:Integer"); }
        protected String getStats() { return Integer.toString (state); }
    }

    final class JoinSubStateStat extends PastryStat {
        public JoinSubStateStat() { super ("xsd:Integer"); }
        protected String getStats() { return Integer.toString (joinSubState); }
    }

    final class LogicalAddressStat extends PastryStat {
        public LogicalAddressStat() { super ("xsd:String"); }
        protected String getStats() { return createLogicalAddress().toString(); }

        protected String setStats (final String newValue)
            throws HyperCastStatsException
        {
            try {
                setLogicalAddress (new Pastry_LogicalAddress (newValue));
            }
            catch (final NumberFormatException nfe) {
                throw new HyperCastStatsException (statisticsName, HyperCastStatsException.InappropriateValue);
            }
            return newValue;
        }
    }

    final class PhysicalAddressStat extends PastryStat {
        public PhysicalAddressStat() { super ("xsd:String"); }
        protected String getStats() { return createPhysicalAddress().toString(); }
    }

    final class NumOfNeighborsStat extends PastryStat {
        public NumOfNeighborsStat() { super ("xsd:Integer"); }
        protected String getStats() { return Integer.toString (hyperCastNodeAdapter.getNeighborCount()); }
    }

    final class LeafSetMaintenanceFrequencyStat extends PastryStat {
        public LeafSetMaintenanceFrequencyStat() { super ("xsd:Integer"); }
        protected String getStats() { return Long.toString (leafSetMaintFreq); }
    }

    final class RouteSetMaintenanceFrequencyStat extends PastryStat {
        public RouteSetMaintenanceFrequencyStat() { super ("xsd:Integer"); }
        protected String getStats() { return Long.toString (routeSetMaintFreq); }
    }

    final class NodeIdRequestTimeoutStat extends PastryStat {
        public NodeIdRequestTimeoutStat() { super ("xsd:Integer"); }
        protected String getStats() { return Long.toString (nodeIdRequestTimeout); }
    }

    final class JoinPhase2MessageTimeoutStat extends PastryStat {
        public JoinPhase2MessageTimeoutStat() { super ("xsd:Integer"); }
        protected String getStats() { return Long.toString (joinMessageTimeout); }
    }

    final class BuddyRetriesStat extends PastryStat {
        public BuddyRetriesStat() { super ("xsd:Integer"); }
        protected String getStats() { return Integer.toString (buddySearchRetries); }
    }

    final class LeafSetSizeStat extends PastryStat {
        public LeafSetSizeStat() { super ("xsd:Integer"); }
        protected String getStats() { return Integer.toString (leafSetSize); }
    }

    final class RouteSetSizeStat extends PastryStat {
        public RouteSetSizeStat() { super ("xsd:Integer"); }
        protected String getStats() { return Integer.toString (routeSetSize); }
    }

    /** @see I_Stats#getStatsName */
    public String getStatsName()
    {
        return statisticsName;
    }

    /** @see I_Stats#setStatsName */
    public void setStatsName (String name)
    {
        statisticsName = name;
    }

    /** @see I_Stats#getStats */
    public Element [] getStats (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        createNeighborhoodTable (xpath);
        return statsProcessor.getStatsResult (doc, xpath);
    }

    /** @see I_Stats#setStats */
    public Element[] setStats(final Document doc, final XPath xpath, final Element newValue)
        throws HyperCastStatsException
    {
        return statsProcessor.setStatsResult (doc, xpath, newValue);

    }

    /** I_Stats#getReadSchema */
    public Element[] getReadSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        createNeighborhoodTable (xpath);
        return statsProcessor.getReadSchemaResult (doc, xpath);
    }

    /** I_Stats#getWriteSchema */
    public Element[] getWriteSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getWriteSchemaResult (doc, xpath);
    }

    /**
     * There is no "neighbor hood table" in Pastry. Rather, there is a
     * leaf set and routing table (which is composed of route
     * sets). These structures are maintained in the core of the
     * FreePastry implementation so the points at which the
     * "neighbors" are "added" and "removed" are not easily
     * accessible.
     *
     * This method creates neighborhood table statistics "on-demand."
     * It senses when the NeighborTable is requested (or the entire
     * Node is requested) and creates neighbor entries at that time.
     */
    public synchronized void createNeighborhoodTable (final XPath xpath)
    {
        if (inOverlay)
        {
            final String xpathString = xpath.getPatternString();
            if (xpathString.equals ("/" + statisticsName) // whole Node
                ||
                xpathString.startsWith ("/" + statisticsName + "/NeighborTable"))
            {
                clearNeighborhoodTable();
                final I_AddressPair [] neighbors = getAllNeighbors();
                for (int i = 0; i < neighbors.length; ++i)
                    statsProcessor.addStatsElement ("NeighborTable", new NeighborhoodStats (neighbors [i]), 0, -1);
            }
        }else{
            clearNeighborhoodTable();
        }
    }

    public void clearNeighborhoodTable() {
        statsProcessor.removeAllStatsElement ("NeighborTable");
    }

    /*
     *  Used for debugging/logging output.
     */
    private String lastr()
    {
        String la = createLogicalAddress().toString();
        if (la == null)
        {
            la = "<Logical Address not set>";
        }
        long timestamp = adapter.getCurrentTime();
        return la + " " + sdf.format (new java.util.Date (timestamp));
    }
}

class Buddy_TimedStopper implements Runnable
{
    Pastry_Node_BuddyList node;
    long time;
    boolean dying;
    
    Buddy_TimedStopper(Pastry_Node_BuddyList n, long t)
    {
        node = n;
        time = t;
        dying = false;

        Thread th = new Thread(this);
        th.start();
    }
    
    void die() 
    {
        synchronized (node.startStopLock) 
        {
            dying = true;
        }
    }
    
    public void run() 
    {
        try 
        {
            Thread.sleep(time);
        }
        catch (InterruptedException ie) 
        {
        }

        synchronized (node.startStopLock) 
        {
            if (!dying) 
            {

                node.adapter.stopTimer();
                node.adapter.Stop();
                node.stopper = null;
            }
        }
    }
}
