/*************************************************************************

"FreePastry" Peer-to-Peer Application Development Substrate

Copyright 2002, Rice University. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

- Neither  the name  of Rice  University (RICE) nor  the names  of its
contributors may be  used to endorse or promote  products derived from
this software without specific prior written permission.

This software is provided by RICE and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall RICE or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

********************************************************************************/

package hypercast.Pastry.rice.pastry.hc;

import java.util.HashMap;

import hypercast.Pastry.rice.pastry.*;
import hypercast.Pastry.rice.pastry.messaging.*;
import hypercast.Pastry.rice.pastry.security.*;
import hypercast.Pastry.rice.pastry.standard.*;
import hypercast.Pastry.rice.pastry.routing.*;
import hypercast.Pastry.rice.pastry.leafset.*;

import hypercast.Pastry.HyperCastTimer;
import hypercast.Pastry.HyperCastNodeAdapter;
import hypercast.Pastry.HyperCastMessage;
import hypercast.Pastry.HyperCastPhysicalAddress;

/**
 * Pastry node factory for HyperCast-linked nodes.
 *
 * @author Greg Mattes
 * @version May 2004
 */
public class HCPastryNodeFactory
    extends PastryNodeFactory
{
    private static final boolean debug = false;

    private NodeIdFactory nidFactory;

    public HCPastryNodeFactory (final NodeIdFactory nf) {
        nidFactory = nf;
    }

    private HashMap proximityCache = new HashMap(); // NodeId -> Long, smaller values mean closer node

    public Long proximityCacheGet (final NodeId key)
    {
        synchronized (proximityCache)
        {
            return (Long) proximityCache.get (key);
        }
    }

    public void proximityCachePut (final NodeId key, final long delta)
    {
        synchronized (proximityCache)
        {
            if (proximityCache.get (key) == null)
            {
                if (debug)
                    System.err.println ("*** PROXIMITY CACHE *** adding: " + key + " -> " + delta);

                proximityCache.put (key, new Long (delta));
            }
        }
    }

    /*
     *  In regular Pastry, node handles may be generated by specifying
     *  an internet address. The node is contacted using a
     *  bi-directional channel, which hypercast does not allow.  This
     *  operation is not supported, instead, another method that has
     *  two parameters, including a nodeId is provided. The nodeId is
     *  obtained in the Pastry_Node_Buddylist code by exchanging ping
     *  and pong messages. When the pong is received, the node handle
     *  can be generated.
     */
    public synchronized NodeHandle generateNodeHandle (final HyperCastPhysicalAddress address, final NodeId nodeId, final long timeoutTime) {
        return new HCNodeHandle (nodeId, address, timeoutTime);
    }

    private HyperCastPhysicalAddress getAddress (final HyperCastNodeAdapter hyperCastNodeAdapter) {
        return new HyperCastPhysicalAddress (hyperCastNodeAdapter.getPhysicalAddress());
    }

    /*
     * With hypercast, a HyperCastTimer and HyperCastNodeAdapter must
     * be given in order for a new node to be created.
     */
    public PastryNode newNode (final NodeHandle bootstrap) {
        throw new UnsupportedOperationException();
    }

    public PastryNode newNode (final NodeHandle bootstrap, final NodeId nodeId) {
        throw new UnsupportedOperationException();
    }

    public synchronized PastryNode newNode (final NodeId               nodeIdParam,
                                            final HyperCastTimer       hyperCastTimer,
                                            final HyperCastNodeAdapter hyperCastNodeAdapter,
                                            final int                  leafSetMaintFreq,
                                            final int                  routeSetMaintFreq,
                                            final int                  lSetSize,
                                            final int                  rtMax,
                                            final long                 timeoutTime)
    {
        final NodeId nodeId = (nodeIdParam == null) ? nidFactory.generateNodeId() : nodeIdParam;
        final HCPastryNode pn = new HCPastryNode (nodeId, hyperCastTimer, hyperCastNodeAdapter);
        final HyperCastPhysicalAddress address = getAddress (hyperCastNodeAdapter);
        final HCNodeHandle localhandle = new HCNodeHandle (nodeId, address, timeoutTime);

	hyperCastNodeAdapter.setMessageReceiver (localhandle);

        HCNodeHandlePool pool = new HCNodeHandlePool();

        HCPastrySecurityManager secureMan = new HCPastrySecurityManager (localhandle, pool);

        MessageDispatch msgDisp = new MessageDispatch();

        RoutingTable routeTable = new RoutingTable (localhandle, rtMax);
        LeafSet leafSet = new LeafSet (localhandle, lSetSize);

        StandardRouter router = new StandardRouter (localhandle, routeTable, leafSet, secureMan);
        StandardLeafSetProtocol lsProtocol = new StandardLeafSetProtocol (pn, localhandle, secureMan, leafSet, routeTable);
        StandardRouteSetProtocol rsProtocol = new StandardRouteSetProtocol (localhandle, secureMan, routeTable);
        StandardJoinProtocol jProtocol = new StandardJoinProtocol (pn, localhandle, secureMan, routeTable, leafSet);

        msgDisp.registerReceiver (router.getAddress(), router);
        msgDisp.registerReceiver (lsProtocol.getAddress(), lsProtocol);
        msgDisp.registerReceiver (rsProtocol.getAddress(), rsProtocol);
        msgDisp.registerReceiver (jProtocol.getAddress(), jProtocol);

        pn.setElements (localhandle, secureMan, msgDisp, leafSet, routeTable);
        pn.setHCElements (pool, leafSetMaintFreq, routeSetMaintFreq);
        secureMan.setLocalPastryNode (pn);

        pool.coalesce (localhandle);
        localhandle.setLocalNode (pn);

        return pn;
    }

    /*
     * These operations cannot be supported here due to the hypercast
     * architecture's uni-directional communication channels. The
     * functionality provided by these methods is implemented in
     * Pastry_Node_Buddylist.java
     */

    public LeafSet getLeafSet (final NodeHandle handle) {
        throw new UnsupportedOperationException();
    }

    public RouteSet[] getRouteRow (final NodeHandle handle, final int row) {
        throw new UnsupportedOperationException();
    }

    public int getProximity (final NodeHandle local, final NodeHandle handle) {
        throw new UnsupportedOperationException();
    }

    /**
     * Static method which is designed to be used by clients needing a
     * HyperCast pastry node factory.
     *
     * @return An HCPastryNodeFactory
     */
    public static HCPastryNodeFactory getFactory (final NodeIdFactory nf) {
        return new HCPastryNodeFactory (nf);
    }
}
