/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.SPT;

import hypercast.*;

import java.util.*;

/** 
 * This class encapsulates a SPT protocol message.
 * The message contains a type (byte), a Overlay hash(int) and 4 AddressPairs. 
 * Other fields are related to SPT protocol
 *<p>
 * 
 * @author HyperCast Team
 * @author Guangyu Dong
 * @version 3.0, May. 23, 2003
 */

public class SPT_Message implements I_Message {
	/** A message type */
	/** Beacon message */
	public static final byte Hello = 0;
	/** Goodbye message. Send when leaving the overlay */
	public static final byte Goodbye = 1;
	/** Sent when the routing information for some destination is unknown */
	public static final byte RouteRequest = 2;
	/** Sent to reply a route request message */
	public static final byte RouteReply = 3;
	
	/** The protocol message type. */
	private byte Type;

	/** A hash of Overlay properties.  It is used to detect either
	  conflicting Overlays or detect bad packets. */
	private int overlayHash;

	/** The address of the node itself */
	private SPT_AddressPair selfAddr;

	/** The address of the root node */
	private SPT_LogicalAddress rootAddr;

	/** The address of the parent node leading to root */
	private SPT_LogicalAddress parentAddr;

	/** Number of hops to root */
	private int cost;

	/** Time the packet is sent from the root*/
	private long time_stamp;

	/** Sender List. A list of nodes which are in contact with the beacon sender. */
	private Hashtable sender_list;

	/** for RouteReply message */
	private SPT_LogicalAddress nextHop;
	
	/** for RouteRequest/RouteReply message */
	private SPT_LogicalAddress pathDst;

	/** Protocol Number value in the Protocol field of a SPT3.0 protocol message */
	public static final byte PROTONUM_SPT30 = 3;

	protected byte Proto_num;

	protected byte Proto_sub;
	
	
	/** The aggregated reliability of this path leading to root */
	short pathMetric=0;

	/** Constructor a new SPT_Message A from an existing message
	 * @param A		the existing message
	 */

	public SPT_Message(SPT_Message A) {
		/* Set values to protocol number and protocol subfield */
		Proto_num = A.getProtoNum();
		Proto_sub = A.getProtoSub();

		Type = A.getType();

		overlayHash = A.getOverlayHash();

		rootAddr = A.getRoot();
		selfAddr = A.getSelf();
		parentAddr = A.getParent();
		cost = A.getCost();
		time_stamp = A.getTimeStamp();

		sender_list = A.sender_list;

		nextHop = A.nextHop;
		pathDst = A.pathDst;
		
		pathMetric=A.pathMetric;
	}

	/** Constructor for Hello message.
	 */
	public SPT_Message(
		byte proto_sub,
		byte T,
		int gh,
		SPT_AddressPair ma,
		SPT_LogicalAddress ra,
		SPT_LogicalAddress pa,
		int ct,
		long ts,
		Hashtable sl,
		short reli
		) {
			
		if(T!=SPT_Message.Hello)throw new IllegalArgumentException("Constructor only for SPT Hello message");

		/* Set values to protocol number and protocol subfield. */
		Proto_num = PROTONUM_SPT30;
		Proto_sub = proto_sub;

		Type = (byte) T;

		overlayHash = gh;

		rootAddr = ra;
		selfAddr = ma;
		parentAddr = pa;
		cost = ct;
		time_stamp = ts;

		sender_list = sl;
		
		pathMetric=reli;
		
		//System.out.println("Sender "+selfAddr.getLogicalAddress()+" PathMetric="+pathMetric);
	}

	/**
	 * Constructor for Goodbye message
	 */
	public SPT_Message(byte proto_sub, byte T, int gh, SPT_AddressPair ma) {
		
		if(T!=SPT_Message.Goodbye)throw new IllegalArgumentException("Constructor only for SPT Goodbye message");
		
		/* Set values to protocol number and protocol subfield. */
		Proto_num = PROTONUM_SPT30;
		Proto_sub = proto_sub;

		Type = (byte) T;
		overlayHash = gh;
		selfAddr = ma;
	}
	/**
	 * Constructor for RouteRequest message
	 */
	public SPT_Message(byte proto_sub, byte T, int gh, SPT_AddressPair ma, SPT_LogicalAddress pd) {

		if(T!=SPT_Message.RouteRequest)throw new IllegalArgumentException("Constructor only for SPT RouteRequest message");

		Proto_num = PROTONUM_SPT30;
		Proto_sub = proto_sub;

		Type = (byte) T;
		overlayHash = gh;
		selfAddr = ma;

		pathDst = pd;
	}

	/**
		 * Constructor for RouteReply message
		 */
	public SPT_Message(
		byte proto_sub,
		byte T,
		int gh,
		SPT_AddressPair ma,
		SPT_LogicalAddress nh,
		SPT_LogicalAddress pd) {
			
		if(T!=SPT_Message.RouteReply)throw new IllegalArgumentException("Constructor only for SPT RouteReply message");
		
		Proto_num = PROTONUM_SPT30;
		Proto_sub = proto_sub;

		Type = (byte) T;
		overlayHash = gh;
		selfAddr = ma;

		nextHop = nh;
		pathDst = pd;
	}

	/** Creates a message from a byte array.
	 * @param adapter used to create new PhysicalAddress from the byte array.
	 */
	public SPT_Message(byte[] A, I_UnicastAdapter adapter) {
	  /*
	   * This is a newly introduced change. The adapter knows the type of physical 
	   * but doesn't store physical address. Check.
	   * 
	   */
		//int paSize = adapter.getPhysicalAddress().getSize();
		int paSize = adapter.getPhysicalAddressSize();

		int laSize = SPT_LogicalAddress.ADDRESS_ARRAY_SIZE;

		I_PhysicalAddress pa = null;
		SPT_LogicalAddress la;

		int index = 0;

		//It is for the new protocol message defined in HCAST3.0
		//The first byte is Protocol field
		Proto_num = (byte) ((A[index] >> 4) & 0xf);
		Proto_sub = (byte) (A[index] & 0xf);

		index += 1;
		if (Proto_num != PROTONUM_SPT30)
			throw new IllegalArgumentException("The Protocol Number in Protocol field of the message is not PROTONUM_SPT30.");

		//It is for the new protocol message defined in HCAST3.0
		//Skip 2 bytes which stores the length of the message
		index += 2;

		Type = A[index];
		index++;
		overlayHash = ByteArrayUtility.toInteger(A, index);
		index += 4;

		switch (Type) {
			case Hello :
				if (!ByteArrayUtility.isBytesZero(A, index, paSize + laSize)) {
					pa = adapter.createPhysicalAddress(A, index);
					la = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index + paSize));
					selfAddr = new SPT_AddressPair(pa, la);
				}
				index += paSize + laSize;

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					rootAddr = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));
					
				}
				index += laSize;

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					parentAddr = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));
					
				}
				index += laSize;

				cost = ByteArrayUtility.toInteger(A, index);
				index += 4;

				time_stamp = ByteArrayUtility.toLong(A, index);
				index += 8;

				int sender_count = ByteArrayUtility.toInteger(A, index);
				index += 4;
				sender_list = new Hashtable();

				for (int i = 0; i < sender_count; i++) {
					SPT_LogicalAddress sender_la =
						new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));
					index += laSize;
					Byte sender_byte=new Byte(A[index]);
					sender_list.put("" + sender_la, sender_byte);
					index +=1;
					
				}

				break;
			case Goodbye :
				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					pa = null;
					la = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));
					selfAddr = new SPT_AddressPair(pa, la);
				}
				index += laSize;
				break;
			case RouteRequest :
				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					pa = null;
					la = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));
					selfAddr = new SPT_AddressPair(pa, la);
				}
				index += laSize;

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					pathDst = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));

				}
				index += laSize;

			case RouteReply :
				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					pa = null;
					la = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));
					selfAddr = new SPT_AddressPair(pa, la);
				}
				index += laSize;

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					nextHop = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));

				}
				index += laSize;

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					pathDst = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));

				}
				index += laSize;

		}
	}

	/** Creates a message, if possible, from a byte array.
	 * If a message is read, the value of beginValidBytes is incremented to the index
	 * of the first byte not in the message.
	 * @param adapter used to create new PhysicalAddress from the byte array.
	 */

	public static SPT_Message restoreMessage(
		byte[] A,
		int[] beginValidBytes,
		int endValidBytes,
		I_UnicastAdapter adapter,
		int thisOverlayHash) {


		  /*
		   * This is a newly introduced change. The adapter knows the type of physical 
		   * but doesn't store physical address. Check.
		   * 
		   */
		//		TODO: Follow up on above comment
		//final int paSize = adapter.getPhysicalAddress().getSize();
		final int paSize = adapter.getPhysicalAddressSize();

		final int laSize = SPT_LogicalAddress.ADDRESS_ARRAY_SIZE;
		final int Reliability_Size=2;
		
		
		SPT_Message spt_msg;

		I_PhysicalAddress pa = null;
		SPT_LogicalAddress la;

		int message_size = endValidBytes - beginValidBytes[0];
		if (message_size < 1 + 2 + 1 + 3 + 4)
			throw new IllegalArgumentException("The size of the message is incorrect.");

		int index = beginValidBytes[0];

		//It is for the new protocol message defined in HCAST3.0
		byte proto_num = (byte) ((A[index] >> 4) & 0xf);
		byte proto_sub = (byte) (A[index] & 0xf);

		index++;
		if (proto_num != PROTONUM_SPT30)
			throw new IllegalArgumentException("The Protocol Number in Protocol field of the message is not PROTONUM_SPT30.");

		//It is for the new protocol message defined in HCAST3.0 protocol
		short length = ByteArrayUtility.toShort(A, index);
		index += 2;
		
		if ((length + 1 + 2) > message_size || length < 3 + 4)
			throw new IllegalArgumentException("The length of the message is incorrect.");

		message_size = (length + 1 + 2);

		byte type = A[index];
		index++;

		int incomingOverlayHash = ByteArrayUtility.toInteger(A, index);
		index += 4;

		if (incomingOverlayHash != thisOverlayHash) {

			throw new IllegalArgumentException("Message is not of this Overlay.");

		}
		SPT_LogicalAddress root_addr = null;
		SPT_AddressPair self_addr = null;
		SPT_LogicalAddress parent_addr = null;
		SPT_LogicalAddress next_hop = null, path_dst = null;
		int h_cost;
		long t_stamp;

		switch (type) {
			case Hello :

				if (message_size < index + 3 * laSize + paSize + 4 + 8 + 4)
					throw new IllegalArgumentException("The length of the message is incorrect.");

				if (!ByteArrayUtility.isBytesZero(A, index, paSize + laSize)) {
					pa = adapter.createPhysicalAddress(A, index);
					la = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index + paSize));
					self_addr = new SPT_AddressPair(pa, la);
				}
				index += paSize + laSize;

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					root_addr = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));
					
				}
				index += laSize;

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					parent_addr = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));
					
				}
				index += laSize;

				h_cost = ByteArrayUtility.toInteger(A, index);
				index += 4;

				t_stamp = ByteArrayUtility.toLong(A, index);
				index += 8;

				int sender_count = ByteArrayUtility.toInteger(A, index);

				index += 4;

				Hashtable s_list = new Hashtable();

				if (message_size < 4 + 3 * laSize + paSize + 4 + 8 + 4 + (laSize+1) * sender_count +Reliability_Size)
					return null;

				for (int i = 0; i < sender_count; i++) {

					SPT_LogicalAddress sender_la =
						new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));
					index += laSize;
					Byte sender_byte=new Byte(A[index]);
					index +=1;
					s_list.put("" + sender_la, sender_byte);
					
				}
				//parse reliability
				short reli=ByteArrayUtility.toShort(A, index);
				index+=Reliability_Size;
				
				beginValidBytes[0] = index;

				spt_msg= new SPT_Message(
					I_Node.PROTOSUB_MULTICAST,
					type,
					incomingOverlayHash,
					self_addr,
					root_addr,
					parent_addr,
					h_cost,
					t_stamp,
					s_list,
					reli);
				
				break;

			case Goodbye :
				if (message_size < index + laSize)
					throw new IllegalArgumentException("The length of the message is incorrect.");

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					pa = null;
					la = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));
					self_addr = new SPT_AddressPair(pa, la);
				}
				index += laSize;

				beginValidBytes[0] = index;

				spt_msg= new SPT_Message(
					I_Node.PROTOSUB_MULTICAST,
					type,
					incomingOverlayHash,
					self_addr);
					
				break;

			case RouteRequest :

				if (message_size < index + 2 * laSize)
					throw new IllegalArgumentException("The length of the message is incorrect.");

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					pa = null;
					la = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));
					self_addr = new SPT_AddressPair(pa, la);
				}
				index += laSize;

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					path_dst = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));

				}
				index += laSize;

				beginValidBytes[0] = index;

				spt_msg= new SPT_Message(
					I_Node.PROTOSUB_MULTICAST,
					type,
					incomingOverlayHash,
					self_addr,
					path_dst);
				
				break;

			case RouteReply :

				if (message_size < index + 3 * laSize)
					throw new IllegalArgumentException("The length of the message is incorrect.");

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					pa = null;
					la = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));
					self_addr = new SPT_AddressPair(pa, la);
				}
				index += laSize;

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					next_hop = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));

				}
				index += laSize;

				if (!ByteArrayUtility.isBytesZero(A, index, laSize)) {
					path_dst = new SPT_LogicalAddress(ByteArrayUtility.toInteger(A, index));

				}
				index += laSize;

				beginValidBytes[0] = index;

				spt_msg= new SPT_Message(
					I_Node.PROTOSUB_MULTICAST,
					type,
					incomingOverlayHash,
					self_addr,
					next_hop,
					path_dst);
					
				break;

			default :

				throw new IllegalArgumentException("ReceiveBuffer cannot contain type SPT_Message (illegal msg type)");

		}

		
		return spt_msg;

	}

	public int getSenderSize() {
		if (sender_list == null)
			return 0;
		else
			return sender_list.size();
	}
	
	/** Converts the message into a byte array for transmission.
	 * @return 	the byte array of the message
	 */
	public byte[] toByteArray() {
		byte Result[] = null;
		int index;

		int paSize = selfAddr.getPhysicalAddress().getSize();
		int laSize = SPT_LogicalAddress.ADDRESS_ARRAY_SIZE;
		
		final int Reliability_Size=2;
		
		switch (Type) {
			case Hello :
				Result =
					new byte[
						+ 1
						+ 4
						+ paSize
						+ 3 * laSize
						+ 4
						+ 8
						+ 4
						+ (laSize+1) * getSenderSize()
						+ Reliability_Size
						];
				index = 0;

				Result[index] = Type;
				index++;

				System.arraycopy(ByteArrayUtility.toByteArray(overlayHash), 0, Result, index, 4);
				index += 4;

				if (selfAddr != null) {
					System.arraycopy(
						selfAddr.getPhysicalAddress().toByteArray(),
						0,
						Result,
						index,
						paSize);
					System.arraycopy(
						ByteArrayUtility.toByteArray(selfAddr.getSPTLogicalAddress().getId()),
						0,
						Result,
						index + paSize,
						laSize);
				} else
					ByteArrayUtility.setBytesZero(Result, index, laSize + paSize);
				index += laSize + paSize;

				if (rootAddr != null) {
					System.arraycopy(
						ByteArrayUtility.toByteArray(rootAddr.getId()),
						0,
						Result,
						index,
						laSize);
				} else
					ByteArrayUtility.setBytesZero(Result, index, laSize);
				index += laSize;

				if (parentAddr != null) {
					System.arraycopy(
						ByteArrayUtility.toByteArray(parentAddr.getId()),
						0,
						Result,
						index,
						laSize);
				} else
					ByteArrayUtility.setBytesZero(Result, index, laSize);
				index += laSize;

				System.arraycopy(ByteArrayUtility.toByteArray(cost), 0, Result, index, 4);
				index += 4;

				System.arraycopy(ByteArrayUtility.toByteArray(time_stamp), 0, Result, index, 8);
				index += 8;

				/** sender list */
				if (sender_list != null) {
					System.arraycopy(
						ByteArrayUtility.toByteArray(sender_list.size()),
						0,
						Result,
						index,
						4);
					index += 4;

					for (Enumeration e = sender_list.keys(); e.hasMoreElements();) {
						String sender_la=(String) e.nextElement();
						int s_la = Integer.parseInt(sender_la);
						Byte sender_byte=(Byte)sender_list.get(sender_la);
						
						System.arraycopy(
							ByteArrayUtility.toByteArray(s_la),
							0,
							Result,
							index,
							laSize);
						index += laSize;
						
						Result[index]=sender_byte.byteValue();
						index +=1;
					}
				} else {
					System.arraycopy(ByteArrayUtility.toByteArray(0), 0, Result, index, 4);
					index += 4;
				}
				
				//put path reliability
				System.arraycopy(ByteArrayUtility.toByteArray(pathMetric), 0, Result, index, Reliability_Size);
				index+=Reliability_Size;
				
				break;

			case Goodbye :
				Result = new byte[1 + 4 + laSize];
				index = 0;
				
				Result[index] = Type;
				index += 1;

				System.arraycopy(ByteArrayUtility.toByteArray(overlayHash), 0, Result, index, 4);
				index += 4;

				if (selfAddr != null) {
					System.arraycopy(
						ByteArrayUtility.toByteArray(selfAddr.getSPTLogicalAddress().getId()),
						0,
						Result,
						index,
						laSize);
				} else
					ByteArrayUtility.setBytesZero(Result, index, laSize);
				index += laSize;

				break;
			case RouteRequest :

				Result = new byte[1 + 4 + 2 * laSize];
				index = 0;

				Result[index] = Type;
				index++;

				System.arraycopy(ByteArrayUtility.toByteArray(overlayHash), 0, Result, index, 4);
				index += 4;

				if (selfAddr != null) {
					System.arraycopy(
						ByteArrayUtility.toByteArray(selfAddr.getSPTLogicalAddress().getId()),
						0,
						Result,
						index,
						laSize);
				} else
					ByteArrayUtility.setBytesZero(Result, index, laSize);
				index += laSize;

				if (pathDst != null) {
					System.arraycopy(
						ByteArrayUtility.toByteArray(pathDst.getId()),
						0,
						Result,
						index,
						laSize);
				} else
					ByteArrayUtility.setBytesZero(Result, index, laSize);
				index += laSize;

				break;

			case RouteReply :
				Result = new byte[1 + 4 + 3 * laSize];
				index = 0;
				

				Result[index] = Type;
				index++;

				System.arraycopy(ByteArrayUtility.toByteArray(overlayHash), 0, Result, index, 4);
				index += 4;

				if (selfAddr != null) {
					System.arraycopy(
						ByteArrayUtility.toByteArray(selfAddr.getSPTLogicalAddress().getId()),
						0,
						Result,
						index,
						laSize);
				} else
					ByteArrayUtility.setBytesZero(Result, index, laSize);
				index += laSize;

				if (nextHop != null) {
					System.arraycopy(
						ByteArrayUtility.toByteArray(nextHop.getId()),
						0,
						Result,
						index,
						laSize);
				} else
					ByteArrayUtility.setBytesZero(Result, index, laSize);
				index += laSize;

				if (pathDst != null) {
					System.arraycopy(
						ByteArrayUtility.toByteArray(pathDst.getId()),
						0,
						Result,
						index,
						laSize);
				} else
					ByteArrayUtility.setBytesZero(Result, index, laSize);
				index += laSize;

				break;

		}
		
		//It is for the new protocol message defined in HCAST3.0
		// The added 3 bytes is for protocol and length fields. 
		byte newResult[] = new byte[3 + Result.length];
		int newindex = 0;

		//This is protocol field
		newResult[newindex] = (byte) ((Proto_num << 4) | (Proto_sub & 0xf));
		newindex += 1;

		//This is length field.
		short length = (short) Result.length;
		System.arraycopy(ByteArrayUtility.toByteArray(length), 0, newResult, newindex, 2);
		newindex += 2;

		System.arraycopy(Result, 0, newResult, newindex, Result.length);

		return newResult;

	}
	public String toString(){
		return getLabel();
	}
	/** Returns the name of the message according to its type.
	 *  It's for displaying purpose
	 * @return	the label of the message type, or "ERROR" if current type is invalid
	 */
	public String getLabel() {
		switch (Type) {
			case SPT_Message.Hello :
				return "Beacon";
			case SPT_Message.Goodbye :
				return "Goodbye";
			case SPT_Message.RouteRequest :
				return "RouteRequest";
			case SPT_Message.RouteReply :
				return "RouteReply";
			default :
				return "Unknown";
		}
	}

	/** @return the Proto_num field of the message. */
	byte getProtoNum() {
		return Proto_num;
	}

	/** @return the Proto_sub field of the message. */
	byte getProtoSub() {
		return Proto_sub;
	}

	/** Returns the type of the message.
	 * @return		message type
	 */
	byte getType() {
		return Type;
	}

	/**
	 * @return		the message's source's physical address
	 */
	public I_PhysicalAddress getSrcPA() {
		return selfAddr.getPhysicalAddress();
	}
	/**
	 * @return	the message's source's logical address	
	 */
	public SPT_LogicalAddress getSrcLA() {
		return selfAddr.getSPTLogicalAddress();
	}

	/** Returns the overlayHash of this message. 
	 * @return		overlayhash
	 */
	public int getOverlayHash() {
		return overlayHash;
	}

	/**
	 * @return		the root's logical address
	 */
	public SPT_LogicalAddress getRoot() {
		return rootAddr;
	}
	/**
	 * @return		the root's ID
	 */
	public int getRootId() {
		return rootAddr.getId();
	}

	/**
	 * @return		the address pair of message's source
	 */
	public SPT_AddressPair getSelf() {
		return selfAddr;
	}
	/**
	 * @return		the address pair of message's source
	 */
	public SPT_AddressPair getSrc() {
		return selfAddr;
	}
	/**
	 * @return		the ID of the message's source
	 */
	public int getSrcId() {
		return selfAddr.getSPTLogicalAddress().getId();
	}

	/**
	 * @return		the parent's logical address
	 */
	public SPT_LogicalAddress getParent() {
		return parentAddr;
	}

	/**
	 * @return		the path cost to the root
	 */
	public int getCost() {
		return cost;
	}

	/**
	 * @return		the time stamp
	 */
	public long getTimeStamp() {
		return time_stamp;
	}
	/**
	 * for RouteReply message
	 * @return	the unicast path source address
	 */
	public SPT_LogicalAddress getNextHop() {
		return nextHop;
	}
	/**
	 * for RouteRequest/RouteReply message
	 * @return	the unicast path destination address
	 */
	public SPT_LogicalAddress getPathDst() {
		return pathDst;
	}

	/**
	 * To do adjacency test
	 * @param la	a spt logical address
	 * @return		whether this address is contained in the sender list of this message
	 */
	public boolean containsSender(SPT_LogicalAddress la) {
		return sender_list != null && (sender_list.get("" + la) != null );
	}
	public boolean isNeighbor(SPT_LogicalAddress la){
		if(sender_list==null)
			return false;
		
		Byte sender_byte=(Byte)sender_list.get(""+la);
		if(sender_byte==null)return false;

		return (sender_byte.byteValue() & 0x80)!=0;
		
		//return sender_list != null && (sender_list.get(""+(-la.getId()))!=null );
	}
	public int getPathMetric(){
		return pathMetric;
	}
	/** get one hop reversed reliability */
	public int getReversedOneHopReliability(SPT_LogicalAddress la){
		if(sender_list==null)
			return 0;
		
		Byte sender_byte=(Byte)sender_list.get(""+la);
		if(sender_byte==null)return 0;
		
		return (sender_byte.byteValue() & 0x7F);
	}
}
