/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.security.*;

import javax.crypto.*;


	/**
	* This class defines the OL_Message extension with type SECURITY. In the byte
	* array format of an overlay message, the security extension has the following
	* format: 
 	* 
 	* (security policy is set to integrity)
 	* -------------------------------------------------------------------------------
 	* |Next Header |     Length       |   Sequence number  |        EP(Mk)          /
 	* |  8 bits    |     16 bits      |       32 bits      |     128 - 256 bits     /
 	* -------------------------------------------------------------------------------
 	* /      SPI         /Length of LA/      LA of sender     /
 	* /      16 bits     /   8 bits   /       variable        /
 	* -------------------------------------------------------------------------------
 	* / OL_H MAC Len /        OL_H MAC       /Payload MAC Len/      Payload MAC     /
 	* /    8 bits    /  variable(>=128 bits) /     8 bits    / variable(>=128 bits  /
 	* -------------------------------------------------------------------------------
 	* 
  	* (security policy is set to privacy)
 	* -------------------------------------------------------------------------------
 	* |Next Header |     Length       |   Sequence number  |        EP(Mk)          /
 	* |  8 bits    |     16 bits      |       32 bits      |     128 - 256 bits     /
 	* -------------------------------------------------------------------------------
 	* /      SPI         /Length of LA/     LA of sender      /
 	* /      16 bits     /   8 bits   /       variable        /
 	* -------------------------------------------------------------------------------
 	* / OL_H MAC Len /        OL_H MAC       /Payload MAC Len/      Payload MAC     /
 	* /    8 bits    /  variable(>=128 bits) /     8 bits    / variable(>=128 bits  /
 	* -------------------------------------------------------------------------------
	* 
	* This extension doesn't handle Next Header and Length fields, which are processed 
	* in OL_Message. 
	*
	* @author HyperCast Team
	*/

	public class SecurityExtension extends Extension 
	{
		/**
		 * Security processor which handles the sending/receiving of secure protocol
		 * and overlay messages.
		 */
		//SecurityProcessor sp = null;

		/**
		 * Key vault. It maintains the certificates and keys of local node and other
		 * nodes which exchange certificate and key with local node.
		 */
		KeyVault keyVault;


		/**
		 * Sequence number of the message.
		 */
		int sequenceNum = -1;

		/**
		 * Security Parameter Index (not used in current version).
		 */
		int SPI;

		/**
		 * Logical address size.
		 */
		int LASize;

		/**
		 * The logical address of the sender of the message.
		 */
		private I_LogicalAddress prevSender = null;

		/**
		 * The length of hmac.
		 */
		int hmacLength = 0;

		/**
		 * The MAC applied to the fixed header and security extension.
		 */
		Mac hmacCompute = null;

		/**
		 * The length of mac.
		 */
		int macLength = 0;

		/**
		 * The MAC applied to the payload/encrypted payload.
		 */
		Mac macCompute = null;

		/**
		 * Message key.
		 */
		Key msgKey = null;


		/**
		 * Encrypted message key.
		 */
		byte[] encryptedMsgKey = null;

		/**
		 * Encrypted key size.
		 */
		int encryptedMsgKeySize;

		/**
		 * Header MAC in byte array format.
		 */
		byte[] hmac = null;

		/**
		 * Payload MAC in byte array format.
		 */
		byte[] mac = null;


		/**
		 * Constructs an SecurityExtension Extension from an existing one.
		 */
		public SecurityExtension(SecurityExtension se)
		{
			this.encryptedMsgKey = se.encryptedMsgKey;
			this.encryptedMsgKeySize = se.encryptedMsgKeySize;
			this.hmac = se.hmac;
			this.mac = se.mac;
			this.hmacLength = se.hmacLength;
			this.hmacCompute = se.hmacCompute;
			this.macLength = se.macLength;
			this.macCompute = se.macCompute;
			this.keyVault = se.keyVault;
			this.LASize = se.LASize;
			this.msgKey = se.msgKey;
			this.prevSender = se.prevSender;
			this.sequenceNum = se.sequenceNum;
			this.SPI = se.SPI;
		}
		
		/**
		 * Constructor.
		 * 
	 	 * @param _kv		key vault.
	 	 * @param senderla	the logical address of message sender.
		 */
		public SecurityExtension(KeyVault _kv, I_LogicalAddress senderla)
		{
			/* 
			 * Get the key vault for security computation.
			 */
			keyVault = _kv;

			/*
			 * Get a new message key from key vault. If the key distribution mode is
			 * end-to-end, the group key is used as message key.
			 */
			msgKey = keyVault.getNewMsgKey();


			//sequenceNum may not be set here. It is set when calling toByteArray().
			/*
			 * Gets the sequence number for this message.
			 */
			//sequenceNum = keyVault.getNextSequenceNum();
			
			/*
			 * Get the source logical address.
			 */
			prevSender = senderla;
			
			/*
			 * Initialize hmac and mac computation objects.
			 */
			//initializeHMACAndMACComputation();

		}

		
		/**
		 * Constructor.
		 */
		public SecurityExtension(byte[] contents, KeyVault _kv)
		{
			/*
			 * Get the key vault kept in the security processor.
			 */
			keyVault = _kv;

			/*
			 * Get the overlay node ketp in the security processor.
			 */
			I_Node node = keyVault.getNode();

			if (node == null)
			{
			}

			int index = 0;

			/* 
			 * Get sequence number.
			 */
			byte[] sequenceNumBytes = new byte[4];
			System.arraycopy(contents, index, sequenceNumBytes, 0, 4);
			sequenceNum = ByteArrayUtility.toInteger(sequenceNumBytes, 0);

			index += 4;


			/* 
			 * Get the encrypted message key. Only when this message is
			 * being forwarded to other neighbors, the original message 
			 * key is reconstructed from this encrypted message key. The
			 * reconstruction of original key is done in method toByteArray()
			 * of this class.
			 */
			encryptedMsgKeySize = keyVault.getWrappedKeySize();
			encryptedMsgKey = new byte[encryptedMsgKeySize];
			System.arraycopy(contents, index, encryptedMsgKey, 0,
							 encryptedMsgKeySize);

			index += encryptedMsgKeySize;

			//System.out.println("SecurityExtension:restoreMessage: received encrypted message key is: *****"
			//		+ HTTP_ServerUtility.toHexString(encryptedMsgKey)+"*****");

			/* 
			 * Get SPI (currently not used)
			 */
/*
			byte[] SPIbytes = new byte[4];
			ByteArrayUtility.setBytesZero(SPIbytes, 0, 4);
			System.arraycopy(secHeader, index, SPIbytes, 2, 2);
			SPI = ByteArrayUtility.toInteger(SPIbytes, 0);
*/
			index += 2;


			/* 
			 * Get Length of LA.
			 */
			byte[] LALenBytes = new byte[4];
			ByteArrayUtility.setBytesZero(LALenBytes, 0, 4);
			System.arraycopy(contents, index, LALenBytes, 3, 1);
			LASize = ByteArrayUtility.toInteger(LALenBytes, 0);

			index++;

			/* 
			 * Get LA of sender
			 */
			byte[] prevSenderBytes = new byte[LASize];
			System.arraycopy(contents, index, prevSenderBytes, 0, LASize);
			prevSender = node.createLogicalAddress(prevSenderBytes, 0);

			if (keyVault.getKeyMode().equals("NeighborhoodKeys")) {
				/* 
				 * Set the time of receiving last secure overlay message
				 * from prevSender.
				 */
				keyVault.setLastSecureMessageTime(prevSender);
				/*
				 * make the key as valid
				 */
				keyVault.setKeyInvalid(prevSender, false);
			}

			index += LASize;

			/*
			 * Get hmac length.
			 */
			byte[] hmacLenBytes = new byte[4];
			ByteArrayUtility.setBytesZero(hmacLenBytes, 0, 4);
			System.arraycopy(contents, index, hmacLenBytes, 3, 1);
			hmacLength = ByteArrayUtility.toInteger(hmacLenBytes, 0);

			index++;

			/* 
			 * Get hmac.
			 */
			hmac = new byte[hmacLength];
			System.arraycopy(contents, index, hmac, 0, hmacLength);

			index += hmacLength;

			/* 
			 * Get mac length.
			 */
			byte[] macLenBytes = new byte[4];
			ByteArrayUtility.setBytesZero(macLenBytes, 0, 4);
			System.arraycopy(contents, index, macLenBytes, 3, 1);
			macLength = ByteArrayUtility.toInteger(macLenBytes, 0);

			index++;

			/* 
			 * Get mac.
			 */
			mac = new byte[macLength];
			System.arraycopy(contents, index, mac, 0, macLength);
		}


		/**
		 * Construct message key from received encrypted message key.
		 */
		private Key constructMsgKeyFromEncryptedMsgKey()
		{
			
			Key newMsgKey = null;

			/*
			 * If encrypted message is not provided, throw exception.
			 */
			if (encryptedMsgKey == null)
			{
				throw new HyperCastWarningRuntimeException ("Message key cannot be constructed without encrypted message key.");
			}

			 /*
			  * Get the personal key of the message sender.
			  */
			Key prevSenderKey = keyVault.getKey(prevSender);


			//System.out.println(
			//		"SecurityExtension:toByteArray: personal key used to decrypt message key is: "
			//				+ HTTP_ServerUtility.toHexString(prevSenderKey.getEncoded()));

			/*
			 * Get the encryption algorithm for this message.
			 */
			String confidentialityAlgorithm = keyVault.getConfidentialityAlgorithm();

			Cipher wrapper;

			/*
			 * If the key distribution mode is set to hop-by-hop, the message key
			 * is reconstructed from the encrypted message key carried in the packet.
			 */
			if (keyVault.getKeyMode().equals("NeighborhoodKeys")) {
				try {
					wrapper = Cipher.getInstance(confidentialityAlgorithm);
					wrapper.init(Cipher.UNWRAP_MODE, prevSenderKey);

					newMsgKey =
						wrapper.unwrap(encryptedMsgKey, confidentialityAlgorithm,
									   Cipher.SECRET_KEY);
				} catch (InvalidKeyException e) {
					e.printStackTrace();
					throw new IllegalArgumentException("Message key failure");
				} catch (NoSuchAlgorithmException e) {
					e.printStackTrace();
					throw new IllegalArgumentException("Message key failure");
				} catch (NoSuchPaddingException e) {
					e.printStackTrace();
					throw new IllegalArgumentException("Message key failure");
				} catch (IllegalStateException e) {
					e.printStackTrace();
					throw new IllegalArgumentException("Message key failure");
				}
			}
			/*
			 * If the key distribution mode is set to end-to-end, the message key 
			 * is the group key. See method KeyVault.getMyPersonalKey().
			 */
			else 
			{
				newMsgKey = keyVault.getMyPersonalKey();
			}

			//System.out.println(
			//		"EncryptedPayloadExtension:encryptePayload: received message key is: "
			//			  + HTTP_ServerUtility.toHexString(newMsgKey.getEncoded()));


			
			return newMsgKey;
		}

		/**
		 * Initialize hmac and mac computation objects.
		 */
		private void initializeHMACAndMACComputation()
		{
			if (msgKey == null)
			{
				throw new HyperCastWarningRuntimeException ("Message key is not constructed when doing mac initialization.");
			}

			/*
			 * Get the MAC calculation algorithm for this message.
			 */
			String integrityAlgorithm = keyVault.getIntegrityAlgorithm();

			/*
			 * Initialize the MAC for headers and payload. In the method toByteArray() 
			 * of OL_Message, they will be updated with the packet (byte array) 
			 * converted from the entire OL_Message message.
			 */
			try {
				hmacCompute = Mac.getInstance(integrityAlgorithm);
				hmacCompute.init(keyVault.getMyPersonalKey());

				//System.out.println("SecurityExtension:innitializeHMACAndMACComputation: personal key for calculat hmac is: *****"
				//		+ HTTP_ServerUtility.toHexString(keyVault.getMyPersonalKey().getEncoded())+"*****");

				macCompute = Mac.getInstance(integrityAlgorithm);
				macCompute.init(msgKey);
			} catch (NoSuchAlgorithmException e) {
				e.printStackTrace();
				throw new IllegalArgumentException("No such integrity algorithm.");
			} catch (InvalidKeyException e) {
				System.out.println("Invalid personal key.");
				e.printStackTrace();
				throw new IllegalArgumentException("Invalid personal key.");
			}

			/* 
			 * Gets hmac length.
			 * 
			 */
			hmacLength = hmacCompute.getMacLength();

			/* 
			 * Gets mac length.
			 * 
			 */
			macLength = macCompute.getMacLength();

		}


		/**
		 * Return a new SecurityExtension extension instance by cloning this extension.
		 */
		public Extension Clone()
		{
			return new SecurityExtension(this);
		}

		/**
		 * Returns the type of extension.
		 */
		public byte getExtensionType()
		{
			return Extension.SECURITY;
		}


		/**
		 * Converts this object to byte array.
		 */
		public byte[] toByteArray()
		{
			/*
			 * Get the MAC calculation algorithm for this message.
			 */
			String integrityAlgorithm = keyVault.getIntegrityAlgorithm();

			/*
			 * Get the encryption algorithm for this message.
			 */
			String confidentialityAlgorithm = keyVault.getConfidentialityAlgorithm();

			/*
			 * Re-construct message key from encrypted message key. 
			 * Only when this message is being forwarded to next hop,
			 * the message key is recnstructed from the received 
			 * encrypted message key. 
			 * If this node is the source of the message, the message 
			 * key is assigned by call method KeyVault.getNewMsgKey() 
			 * when the message instance is created.
			 */
			if (msgKey == null)
			{
				msgKey = constructMsgKeyFromEncryptedMsgKey();
			} 

			/*
			 * Initialize hmac and mac computation objects. It needs message
			 * key to finish initialization.
			 */
			//initializeHMACAndMACComputation();


			int extensionLen = 0;
			
			/* 
			 * Convert the sequence number into byte array.
			 * 
			 */
			byte[] sequenceNumBytes =
				ByteArrayUtility.toByteArray(sequenceNum);
		
			extensionLen += 4;



			Cipher wrapper;
			int wrappedKeySize = keyVault.getWrappedKeySize();

			/*
			 * Encrypted message key. In either of the following two cases,
			 * The length of the field reserved for message key is wrappedKeySize.
			 */
			byte[] encryptedMsgKey;


			//System.out.println(
			//	"SecurityExtension:toByteArray: message key to be encrypted is: "
			//		+ HTTP_ServerUtility.toHexString(msgKey.getEncoded()));



			/*
			 * If key distribution mode is set to Dynamic, create encrypted 
			 * message key from the original mesage key.
			 */
			if (keyVault.getKeyMode().equals("NeighborhoodKeys"))
			{
				try {
					wrapper = Cipher.getInstance(confidentialityAlgorithm);
					wrapper.init(Cipher.WRAP_MODE, keyVault.getMyPersonalKey());

				//System.out.println(
				//			  "SecurityExtension:toByteArray: personal key used to encrypt message key is: "
				//				  + HTTP_ServerUtility.toHexString(keyVault.getMyPersonalKey().getEncoded()));


					encryptedMsgKey = wrapper.wrap(msgKey);

				} 
				catch (InvalidKeyException e1) 
				{
					throw new IllegalArgumentException("Message authentication check failure");
				} 
				catch (NoSuchAlgorithmException e) 
				{
					throw new IllegalArgumentException("Message authentication check failure");
				} 
				catch (NoSuchPaddingException e) 
				{
					throw new IllegalArgumentException("Message authentication check failure");
				} 
				catch (IllegalStateException e) 
				{
					throw new IllegalArgumentException("Message authentication check failure");
				} 
				catch (IllegalBlockSizeException e) 
				{
					throw new IllegalArgumentException("Message authentication check failure");
				}
			}
			/*
			 * If key distribution mode is end-to-end, do nothing. In this case, the field
			 * for encrypted message key is defined but no contents filled in.
			 */
			else
			{ 
				encryptedMsgKey = new byte[wrappedKeySize];
			}

			extensionLen += encryptedMsgKey.length;

			/*
			 * Store the size of encrypted message key for later use.
			 */
			encryptedMsgKeySize = encryptedMsgKey.length;

			/* 
			 * Add two bytes for SPI (2 bytes)
			 */
			extensionLen += 2;

			/*
			 * Gets the LA size.
			 */
			LASize = prevSender.getSize();

			/*
			 * Add length for the fields of LA length (1 byte) and LA of the sender (LAsize).
			 */
			extensionLen += 1 + LASize;

			/*
			 * Add the length of hmac length field and hmac field.
			 */
			extensionLen += 1 + hmacLength;

			/*
			 * Add the length of mac length field and mac field.
			 */
			extensionLen += 1 + macLength;

			/*
			 * In the following steps, the byte array of security extension is constructed.
			 */
			byte[] secHeader = new byte[extensionLen];
			int index = 0;
			
			/*
			 * Copy the byte array of sequence number into packet.
			 */
			System.arraycopy(sequenceNumBytes, 0, secHeader, 0, 4);
			index += 4;


			//System.out.println("SecurityExtension:toByteArray: encrypted message key is: *****"
			//		+ HTTP_ServerUtility.toHexString(encryptedMsgKey)+"*****");

			/*
			 * If the key distribution is set to hop-by-hop, copy encrypted 
			 * message key into packet. Otherwise don't do memory copy.
			 */
			if (keyVault.getKeyMode().equals("NeighborhoodKeys"))
			{
				//System.arraycopy(encryptedMsgKey, 0, secHeader, index, wrappedKeySize);
				System.arraycopy(encryptedMsgKey, 0, secHeader, index, encryptedMsgKey.length);
			}
			index += encryptedMsgKey.length;

			/* 
			 * Skip two bytes for SPI (2 bytes)
			 */
			index += 2;

			/*
			 * Convert LA size to byte array and copy it into packet (1 byte) .
			 */
			System.arraycopy(ByteArrayUtility.toByteArray(LASize), 3,
							 secHeader, index, 1);

			/*
			 * Convert source LA to byte array and copy it into packet.
			 */
			System.arraycopy(prevSender.toByteArray(), 0, secHeader, index+1, LASize);

			index += 1 + LASize;

			/*
			 * Convert hmac length to byte array and copy it into packet (1 byte).
			 */
			System.arraycopy(ByteArrayUtility.toByteArray(hmacLength), 3,
							 secHeader, index, 1);
			index += 1;

			/*
			 * Skip length hmacLength for hmac field. This field will be reset 
			 * in OL_Message.toByteArray().
			 */
			//System.arraycopy(hmacCompute.doFinal(), 0, secHeader, index, hmacLength);
			index += hmacLength;

			/*
			 * Convert mac length to byte array and copy it into packet (1 byte).
			 */
			System.arraycopy(ByteArrayUtility.toByteArray(macLength), 3,
							 secHeader, index, 1);
			index += 1;

			//System.out.println("SecurityExtension: toByteArray: macLength is "+macLength);

			/*
			 * Convert mac to byte array and copy into pcket.
			 */
			//System.arraycopy(macCompute.doFinal(), 0, secHeader, index, macLength);

			if (mac != null)
			{
				System.arraycopy(mac, 0, secHeader, index, macLength);
			}
			index += macLength;

			return secHeader;
		}


		/**
		 * Get the source logical address.
		 */
		public I_LogicalAddress getSenderLA()
		{
			return prevSender;
		}

		/**
		 * Update the source logical address.
		 */
		public void setSenderLA(I_LogicalAddress senderla)
		{
			prevSender = senderla;
		}

		/**
		 * Get the sequence number.
		 */
		public int getSequenceNumber()
		{
			return sequenceNum;
		}

		/**
		 * Set the sequence number.
		 */
		public void setSequenceNumber(int seqNum)
		{
			sequenceNum = seqNum;
		}

		/**
		 * Update hmac computation object with a byte array.
		 */
		public void updateHMacCompute(byte[] contents)
		{
			if (hmacCompute != null)
			{
		//System.out.println("SecurityExtension:updateHMacCompute: payload for hmac calculation is: *****"
		//	 + HTTP_ServerUtility.toHexString(contents)+"*****");

				hmacCompute.update(contents, 0, contents.length);
			}
		}

		/**
		 * Get hmac computation object in this extension.
		 */
		public Mac getHMacCompute()
		{
			return hmacCompute;
		}

		/**
		 * Get hmac in this extension.
		 */
		public byte[] getHMac()
		{
			return hmac;
		}

		/**
		 * Get hmac length in this extension.
		 */
		public int getHMacLen()
		{
			return hmacLength;
		}

		/**
		 * Get the location of hmac in packet.
		 */
		public int getHMacLoc()
		{
			/*
			 * See the format of security extension defined in the top of this file.
			 * The Next Header and Length fields are not included.
			 */
			return 4 + encryptedMsgKeySize + 2 + 1 + LASize + 1; 
		}

		/**
		 * Update mac computation object with a byte array.
		 */
		public void updateMacCompute(byte[] contents)
		{
			initializeHMACAndMACComputation();

			if ((macCompute != null) && (contents != null))
			{
				//System.out.println("SecurityExtension:updateMacCompute:payload used to cal mac ="+HTTP_ServerUtility.toHexString(contents)+"***");

				macCompute.update(contents, 0, contents.length);
				
				mac = macCompute.doFinal();

				//System.out.println("SecurityExtension:updateMacCompute:payload mac="+HTTP_ServerUtility.toHexString(mac)+"***");
			}
		}

		/**
		 * Get mac in this extension.
		 */
		public byte[] getMac()
		{
			return mac;
		}

		/**
		 * Get the message key in this extension.
		 */
		public Key getMsgKey()
		{
			/*
			 * If message key has not been reconstructed, create from the received 
			 * encrypted message key.
			 */
			if (msgKey == null)
			{
				msgKey = constructMsgKeyFromEncryptedMsgKey();
			}
			//if (msgKey == null)
			//	System.out.println("SecurityExtension: getMsgKey is still null after reconstruction!");
			
			return msgKey;
		}


		/*
		 * This method provides a way for all extensions to get the length of 
		 * the byte array of an extension without actually executing toByteArray().
		 */
		/**
		 * Returns the length of the byte array of this extension.
		 */
		public int getSize()
		{
			int extensionLen = 0;
			
			/*
			 * Add length for sequence number (4 bytes) and encrypted message key.
			 */
			extensionLen += 4 + keyVault.getWrappedKeySize();
			
			/*
			 * Add length for SPI (2 bytes), LA size (1 byte), LA of sender.
			 */
			extensionLen += 2 + 1 + prevSender.getSize();
			
			/*
			 * Add the length of hmac length, hmac, mac length and mac.
			 */
			extensionLen += 1 + hmacLength + 1 + macLength;

			return extensionLen;
		}

		
	}
