/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;


import hypercast.events.NODE_NEIGHBORHOODCHANGED;
import hypercast.util.XmlUtil;

import java.security.PrivateKey;
import java.security.cert.*;

import java.util.*;

import javax.crypto.spec.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.xpath.*;

/**
 * This class defines the functions for handling the sending/receiving of secure
 * messages. Once the instance of this class is created, it works as the node
 * adapter for an overlay node and builds a bridge between the overlay node and
 * the normal node adapter. The relationship between the overlay node, security
 * processor and normal node adapter are represented by the following two
 * chains:
 * 
 * (I_AdapterCallBack object) node <- ((I_AdapterCallBack object) this object
 * <- (I_UnicastAdapter) adapter
 * 
 * node -> (I_UnicastAdapter) this object -> (I_AdapterCallBack object) adapter
 * 
 * @author HyperCast Team
 * @author Guangyu Dong
 * @author Jianping Wang
 * @version 3.0, Jan. 27, 2004 
 */
public class SecurityProcessor extends NotificationHandler
	implements I_Stats, I_AdapterCallback, I_MulticastAdapter
{

	/**
	 * Debug flag.
	 */
	protected boolean debug = false;

	/** protocol node */
	protected I_Node node;

	/*
	 * It points to the overlay node.
	 */
	/** Object that implements I_AdapterCallback interface */
	protected I_AdapterCallback callback;

	/*
	 * It points to the normal node adapter.
	 */
	/**  node adapter */
	protected I_UnicastAdapter adapter;
	

	/** Overlay socket configuration object */
	protected HyperCastConfig config;
	
	/**
	 * In milliseconds, the timeout time of waiting for a certificate reply after
	 * sending out a certificate request.
	 */
	protected long CERT_REQUEST_TIMEOUT;

	/**
	 * In milliseconds, the timeout time of waiting for a key update after
	 * sending out a key request.
	 */
	protected long KEY_REQUEST_TIMEOUT;

	/**
	 * In milliseconds, the timeout time for an entry in the key neighbor table.
	 */
	protected long MAX_AGE_OF_AUTH_NODE;
	
	/**
	 * In millisecond, the maximum age of an neighbor key
	 */
	protected long MAX_AGE_OF_KEY;

	
	/** Prefix for all public security property names read by the Security Processor. */
	protected static final String PUBLIC_SECURITY_PROPERTY_NAME_PREFIX = "/Public/Security/";
	
	/**
	 * Stores the certificate, key and authentication information for this
	 * node and its neighbors.
	 */
	KeyVault vault;

	/**
	 * Constructor.
	 * 
	 * @param c			overlay socket configuration object.
	 * @param keyvault	a KeyVault instance.
	 * @param _adapter	node adapter.
	 */
	public SecurityProcessor(HyperCastConfig c, KeyVault keyvault,	I_UnicastAdapter _adapter) 
	{
		/*
		 * Store the configuration
		 */
		config = c;
		config.log.println("Start Security Processor (" + getVersion() + ")");
		
		/*
		 * Get the timeout time for waiting for the reply of a certificate request message. 
		 */
		CERT_REQUEST_TIMEOUT = 
			config.getLongAttribute (XmlUtil.createXPath(PUBLIC_SECURITY_PROPERTY_NAME_PREFIX + "CertificateRequestTimeout"));

		/*
		 * Get the timeout time for waiting for the reply of a key request message. 
		 */
		KEY_REQUEST_TIMEOUT = 
			config.getLongAttribute (XmlUtil.createXPath(PUBLIC_SECURITY_PROPERTY_NAME_PREFIX + "KeyRequestTimeout"));

		/*
		 * Get the timeout time for the personal key of a remote overlay socket. 
		 */
		MAX_AGE_OF_AUTH_NODE = 
			config.getLongAttribute (XmlUtil.createXPath(PUBLIC_SECURITY_PROPERTY_NAME_PREFIX + "MaxAgeOfAuthNode"));
		
		/*
		 * Get the maximum age of key
		 */
		MAX_AGE_OF_KEY = 
			config.getLongAttribute (XmlUtil.createXPath(PUBLIC_SECURITY_PROPERTY_NAME_PREFIX + "MaxAgeOfKey"));
		

		/*
		 * Sets the node adapter stored in this object.
		 */
		setAdapter(_adapter);

		/*
		 * Pass this object to the node adpater as the I_AdapterCallback object. 
		 * So that the adapter will call timerExpired(), messageArrivedFromAdapter()
		 * methods in this class.
		 */
		adapter.setCallback(this);
		
		/*
		 * Get the reference to key vault.
		 */
		vault = keyvault;

		/*
		 * Set the creation time to be associated with my key.
		 */
		vault.setMyKeyTimestamp(adapter.getCurrentTime());

		/*
		 * Set the reference to SecurityProcessor in the key vault.
		 */
		vault.setSecurityProcessor(this);
	}



	/**
	 * Gets the key vault object kept by this object.
	 * 
	 * @return the key vault object.
	 */
	public KeyVault getKeyVault() {
		return vault;
	}

	/**
	 * @return the current version information.
	 */
	public String getVersion() {
		return "General SecurityProcessor: Ver 1.0";
	}

	/** 
	 * Create an address pair from a physical address and a logical address.
	 * 
	 * @return an address pair.
	 */
	public I_AddressPair createAddressPair(I_PhysicalAddress pa, I_LogicalAddress la) {
		return new G_AddressPair(pa, la);
	}
	
	
	/**
	 * Gets all the nodes in the authenticated neighbor table.
	 * 
	 * @return	the address pairs of all the nodes in the key neighbor table.
	 */
	protected I_AddressPair[] getKeyExchangeNeighbors() {
		try
		{
			return vault.getAllKeyNeighbors();
		} catch (Exception e) {
			return new I_AddressPair[0];
		}
	}

	/**
	 * Checks all incoming protocol messages for security extensions and
	 * performs appropriate integrity checks and/or decryption before handing
	 * the message to the node to process.  This function will also
	 * handle all security-related protocol messages.
	 *
	 * @param a	a message to be processed.
	 *
	 * @throws IllegalArgumentException if the given message is not an instance of 
	 * 	SecInfoExchange_Message.
	 */
	public synchronized void messageArrivedFromAdapter(I_Message a)
	{

		if (!(a instanceof SecInfoExchange_Message))
		{
			throw new IllegalArgumentException("SecurityProcessor.messageArrivedFromAdapter() passed non-SecInfoExchange_Message I_Message.");
		}

		SecInfoExchange_Message m = (SecInfoExchange_Message) a;

		/*
		 * If the local overlay ID hash value is not the same as the overlay
		 * ID hash value carried in the packet, drop the message.
		 */
		if (m.getOverlayHash() != config.getOverlayHash())
		{
			return;
		}
		
		/*
		 * Get the source address pair in the message.
		 */
		I_AddressPair src = m.getSrc();

		/*
		 *  If the message comes from myself, drop the message.
		 */
		if (((src.getLogicalAddress()).equals(node.getAddressPair().getLogicalAddress())) &&
			((src.getPhysicalAddress()).equals(node.getAddressPair().getPhysicalAddress())))
		{
			return;
		}
		
		/*
		 * Update the time when last message (protocol oe secure) is received for the 
		 * source of the message. If there is no entry for the source node in the 
		 * key neighbor table, do nothing.
		 */
		vault.updateEntry(src, adapter.getCurrentTime()); //update time_stamp

		/*
		 * If this is a certificate request, certificate reply, key request
		 * or key update message, Update the time when last secure message is 
		 * received for the  source of the message. If there is no entry for 
		 * the source node in the key neighbor table, it does nothing.
		 * 
		 */
		if (m.getType() != SecInfoExchange_Message.ProtoMsg)
		{
			vault.secureMessageReceived(src, adapter.getCurrentTime()); //update lastSecureMessgeReceived
		}

		/*
		 * Handle the security messages.
		 */
		switch (m.getType())
		{
			case SecInfoExchange_Message.CertRequest :
				
				handleCertRequest(m);

				break;

			case SecInfoExchange_Message.CertReply :
				
				handleCertReply(m);

				break;

			case SecInfoExchange_Message.KeyRequest :

				handleKeyRequest(m);

				break;

			case SecInfoExchange_Message.KeyUpdate :

				handleKeyUpdate(m);

				break;
			/*
			 * Otherwise, it is a wrap (with type ProtoMsg) mesage.
			 */
			default :

				int[] begin = new int[1];
				begin[0] = 0;

				I_LogicalAddress srcLA = src.getLogicalAddress();
				I_PhysicalAddress srcPA = src.getPhysicalAddress();
				
				/*
				 * Get the protocol message bytes
				 */
				byte[] protoMsgBytes = m.getProtoMsgBytes();
				
				/*
				 * Restore the protocol message
				 */
				I_Message protoMsg;
				if(protoMsgBytes!=null){
				
					int end = protoMsgBytes.length;

					/*
					 * Reconstruct the plaintext protocol message.
					 */
					protoMsg = callback.restoreMessage(m.getProtoMsgBytes(), begin, end);
				}else{
					protoMsg = m.getProtoMsg();
				}
				
				if (debug)
					System.out.println("SecurityProcessor:messageArrivedFromAdapter: received protocol msg from "+src);

				/*
				 * In current implementation, all protocol and secure messages need to be
				 * authenticated
				 */
				boolean requireAuth = true;
			
				/*
				 * If this message needs to be authenticated, update the time when 
				 * last secure message is received for the  source of the message. 
				 * If there is no entry for the source node in the  key neighbor 
				 * table, do nothing. 
				 */
				if (requireAuth)
				{
					vault.secureMessageReceived(src, adapter.getCurrentTime());//update lastSecureMessgeReceived
				}

				/*
				 * If this message comes from a node whose certificate is not known by
				 * local node, initialize certificate authentication process. In this case, 
				 * the message is droped. No matter the key mode method is set to NeighborhoodKeys or
				 * GroupKeys, the certificate authentication is excuted between two nodes who
				 * exchange messages.
				 */
				if((requireAuth && !vault.haveCert(srcLA)))
				{
					
					/*
					 * Initiate certificate exchange.
					 */
					//System.err.println("start auth with "+src);
					initiateAuthentication(src);
					
				}
				
			    /*
			     * If key distribution mode method is set to GroupKeys or NeighborhoodKey3, 
			     * once this node has remote node's certificate, send the message to 
			     * the overlay node to process.
			     */
			    else if (requireAuth && (vault.getKeyMode().equals("GroupKeys") || vault.getNeighborKeyMode().equals(KeyVault.NKM_SHARED_PROTOCOL_KEY)))
			    {
			    	/*
					 * Check sequence number
					 */
					if (!vault.checkSequenceNum(srcLA, m.getSequenceNum())) {
						//System.err.println("Invalid sequence number! "+m.getSequenceNum());
						return;
					}
					//messagesReceivedInLastHeartbeat--; // prevent double counting.
	
					/*
					 * Pass the plaintext protocol message to overlay node to handle.
					 */
					callback.messageArrivedFromAdapter(protoMsg);
	
					break;
			    }
				/*
				 * Otherwise the local node has the certificate of the message sender.
				 */
				else
				{
					/*
					 * Check sequence number
					 */
					if (!vault.checkSequenceNum(srcLA, m.getSequenceNum())) {
						//System.err.println("Invalid sequence number! "+m.getSequenceNum());
						return;
					}
					/*
					 * If the message is a wrap message and the local node doesn't has the 
					 * key of the message sender, drop the message and try to get the new key 
					 * from the message sender.
					 */
					if (requireAuth
						&& (!vault.haveKey(srcLA))
						&& (m.getType() == SecInfoExchange_Message.ProtoMsg))
					{
						//if (debug)
							System.out.println("There's no key for " + src);
						
						this.requestKeyFor(srcLA);
						break;
					}
					//messagesReceivedInLastHeartbeat--; // prevent double counting.

					/*
					 * Pass the plaintext protocol message to overlay node to handle.
					 */
					callback.messageArrivedFromAdapter(protoMsg);
				}
				//System.out.println("Finishing handling message");
		}

		//MessageCallBackDebug = "Returned(DSJ)";
	}

	/**
	 * Initialize certificate exchange process with given node.
	 *
	 * @param src the logical address of the node with witch certificate
	 * 	is exchanged.
	 */
	public void initiateAuthentication(I_AddressPair src)
	{

		I_LogicalAddress srcLA = src.getLogicalAddress();
		I_PhysicalAddress srcPA = src.getPhysicalAddress();

		/*
		 * If this node is not waiting for the certificate reply from the given 
		 * node, send certificate request to this node. Otherwise do nothing.
		 */
		if (!vault
			.isWaitingCertReply(src, CERT_REQUEST_TIMEOUT, adapter.getCurrentTime()))
		{
			try
			{
				//config.log.println("Initiate authentication with " + src);

				/*
				 * Send certificate request to the given node.
				 */
				sendCertRequest(src);

				/*
				 * Add the given node in the key neighbor table. If this node exists,
				 * update the time when last message is received from this node.
				 */
				vault.addEntry(src, adapter.getCurrentTime());

				/*
				 * Record the time when the last certificate request is sent to this node.
				 */ 
				vault.certRequestSent(src, adapter.getCurrentTime());

			} catch (IllegalStateException e) {
				return;
			}
		}
	}

	/**
	 * Verifies certificate of requesting party and, if valid, send own
	 * certificate.
	 *
	 * @param m		a certificate request message.
	 */
	protected void handleCertRequest(SecInfoExchange_Message m)
	{
		
		//System.err.println("cert request from"+m.getSrc());
		
		X509Certificate cert;

		I_AddressPair src = m.getSrc();
		I_LogicalAddress srcLA = src.getLogicalAddress();
		I_PhysicalAddress srcPA = src.getPhysicalAddress();

		/*
		 * Add the given node in the key neighbor table. If the given node exists,
		 * update the time when last message is received from that node.
		 */
		vault.addEntry(src, adapter.getCurrentTime());

		/*
		 * If this node is waiting for the certificate reply from the given node, the tie
		 * is broken by the following rule: if the logical address of message sender is 
		 * larger than the local node, drop the certificate request message.
		 */
		if (vault
			.isWaitingCertReply(src, this.CERT_REQUEST_TIMEOUT, adapter.getCurrentTime()))
		{
			if (srcLA.greaterThan(node.getAddressPair().getLogicalAddress()))
			{
				return;
			}
		}

		
		/*
		 * Get certificate from the message.
		 */
		cert = m.getCertificate();
		
		/*
		 * Add the certificate into the certificate hash table in the key vault.
		 */
		vault.addCert(src.getLogicalAddress(), cert);
		
		/*
		 * Set the lastTimeCertReplyReceived time for the given node to current time.
		 * A CertRequest also contains the sender's Certificate. So this time value
		 * just work as a record as the last time a certificate was received from the
		 * given node.
		 */
		vault.certReplyReceived(src);

		/*
		 * If this node is not waiting for the certificate reply from given node, set
		 * the time when certificate request is sent to the source node of the message.
		 * 
		 */
		if (!vault
			.isWaitingCertReply(src, this.CERT_REQUEST_TIMEOUT, adapter.getCurrentTime()))
		{

			vault.certRequestSent(src, adapter.getCurrentTime());
		}

		/*
		 * Send back a certificate reply to the sender of the certificate request message.
		 */
		sendCertReply(src);
		
		/*
		 * Send KeyUpdates if not in "GroupKeys" mode
		 */
		if(!vault.getKeyMode().equals("GroupKeys")){
			if(vault.getNeighborKeyMode().equals(KeyVault.NKM_UNIFORM_DYNAMIC_KEY)){
				/*
				 * If it's in NeighborhoodKey1, update the neighborhood key, and send
				 * KeyUpdates to all key neighbors.
				 */
				this.updatePersonalKey(null);
				sendKeyUpdateToAllNeighbors(null);
			}else if(vault.getNeighborKeyMode().equals(KeyVault.NKM_STATIC_NONNEIGHBOR_KEY)){
				/*
				 * If it's in NeighborhoodKey3 mode, then generate a new key for the sender
				 * and send KeyUpdate to it.
				 */
				vault.generateMyNoNeighborKeyFor(src.getLogicalAddress(), adapter.getCurrentTime());
				sendKeyUpdate(src);
			}
			
		}
	}

	/**
	 * Verifies incoming certificate and, if valid, adds to list of key neighbors.
	 *
	 * @param m		a certificate reply message.
	 */
	protected void handleCertReply(SecInfoExchange_Message m)
	{
		//System.err.println("cert reply from"+m.getSrc());
		
		X509Certificate cert;

		I_AddressPair src = m.getSrc();
		I_LogicalAddress srcLA = src.getLogicalAddress();
		I_PhysicalAddress srcPA = src.getPhysicalAddress();

		/*
		 * Add the given node in the key neighbor table. If the given node exists,
		 * update the time when last message is received from that node.
		 */
		vault.addEntry(src, adapter.getCurrentTime());

		/*
		 * Get certificate from the message.
		 */
		cert = ((SecInfoExchange_Message) m).getCertificate();

		/*
		 * Add the certificate into the certificate hash table in the key vault.
		 */
		vault.addCert(srcLA, cert);
		
		/*
		 * Update the time when certificate reply is received from the source node of the message.
		 * 
		 */
		vault.certReplyReceived(src);

		/*
		 * Send KeyUpdates if not in "GroupKeys" mode
		 */
		if(!vault.getKeyMode().equals("GroupKeys")){
			if(vault.getNeighborKeyMode().equals(KeyVault.NKM_UNIFORM_DYNAMIC_KEY)){
				/*
				 * If it's in NeighborhoodKey1, update the neighborhood key, and send
				 * KeyUpdates to all key neighbors.
				 */
				this.updatePersonalKey(null);
				sendKeyUpdateToAllNeighbors(null);
			}else if(vault.getNeighborKeyMode().equals(KeyVault.NKM_STATIC_NONNEIGHBOR_KEY)){
				/*
				 * If it's in NeighborhoodKey3 mode, then generate a new key for the sender
				 * and send KeyUpdate to it.
				 */
				vault.generateMyNoNeighborKeyFor(src.getLogicalAddress(), adapter.getCurrentTime());
				sendKeyUpdate(src);
			}
			
		}
		
	}

	/**
	 * Handle a KeyRequest message.
	 *
	 * @param m		a key request message.
	 */
	protected void handleKeyRequest(SecInfoExchange_Message m)
	{
		/*
		 * If key distribution mode is set to shared group key, do nothing.
		 */
		//if (vault.getKeyMode().equals("EndToEnd"))
		if (vault.getKeyMode().equals("GroupKeys"))
		{
			return;
		}
		//System.err.println("key request from"+m.getSrc());
		
		SecretKeySpec keySpec;

		I_AddressPair src = m.getSrc();
		I_LogicalAddress srcLA = src.getLogicalAddress();
		I_PhysicalAddress srcPA = src.getPhysicalAddress();

		/*
		 * If this node has the certificate of the message sender, 
		 * update the time of receiving certificate and send out key update.
		 */
		if (vault.haveCert(srcLA))
		{
			
			/*
			 * Update the time of receiving certificate from the sender of the message.
			 */
			vault.certReplyReceived(src);

			/*
			 * Add the given node in the key neighbor table. If the given node exists,
			 * update the time when last message is received from that node.
			 */
			vault.addEntry(src, adapter.getCurrentTime());

			/*
			 * If this node has the key of the message sender, send my key to it.
			 */
			sendKeyUpdate(m.getSrc());
			
		}
		/*
		 * Otherwise, if the local node is not waiting for the certificate reply from the
		 * given node, send certificate request to it.
		 */
		else
		{
			if (!vault
				.isWaitingCertReply(src, this.CERT_REQUEST_TIMEOUT, adapter.getCurrentTime()))
			{
				sendCertRequest(src);
				
				vault.addEntry(src, adapter.getCurrentTime());
				vault.certRequestSent(src, adapter.getCurrentTime());
			}
		}
	}

	/**
	 * If this message comes from a valid neighbor, updates that neighbor's key entry
	 * in key table.
	 *
	 * @param m		a key update message.
	 */
	protected void handleKeyUpdate(SecInfoExchange_Message m)
	{
		//System.err.println("key update from"+m.getSrc());
		
		/*
		 * If key distribution mode is set to shared group key, do nothing.
		 */
		//if (vault.getKeyMode().equals("EndToEnd")) {
		if (vault.getKeyMode().equals("GroupKeys")) {
			return;
		}

		/*
		 * Get the key from the message.
		 */
		SecretKeySpec keySpec = m.getKey();
		
		I_AddressPair src = m.getSrc();
		I_LogicalAddress srcLA = src.getLogicalAddress();
		I_PhysicalAddress srcPA = src.getPhysicalAddress();

		/*
		 * If this node has the certificate of the message sender, 
		 * update the time of receiving certificate and the time of receiving
		 * key update, and put the key in the key hashtable, update 
		 * sequence number hashtable.
		 */
		if (vault.haveCert(srcLA))
		{
			vault.addEntry(src, adapter.getCurrentTime());

			vault.certReplyReceived(src);

			vault.keyUpdateReceived(src);

			vault.addKey(srcLA, keySpec, m.getKeyTimestamp());

			vault.setSequenceNum(srcLA, 0);
		}
		/*
		 * Otherwise, send certificate request to the sender of the message.
		 * Add it to the key neighbor table and set the time when certificate
		 * is sent out. In this case, the received key is droped.
		 */
		else
		{
			if (!vault
					.isWaitingCertReply(src, this.CERT_REQUEST_TIMEOUT, adapter.getCurrentTime()))
			{
				sendCertRequest(src);
				
				vault.addEntry(src, adapter.getCurrentTime());
				
				vault.certRequestSent(src, adapter.getCurrentTime());
			}
		}
	}

	/**
	 * Sends a request for a node's certificate.  Includes a copy of my
	 * certificate.
	 *
	 * @param dst	the address pair of the node the certificate request
	 * is sent to.
	 */
	protected final void sendCertRequest(I_AddressPair dst)
	{
		//if (debug)
			//System.out.println("Sending CertRequest to " + dst);

		//messagesSentInLastHeartbeat++;

		SecInfoExchange_Message NewPacket =
				new SecInfoExchange_Message(
				SecInfoExchange_Message.CertRequest,
				config.getOverlayHash(),
				node.getAddressPair(),
				dst,
				this);

		adapter.sendUnicastMessage(dst.getPhysicalAddress(), NewPacket);
	}

	/**
	 * Sends a reply to the sender of certificate request message. A copy of 
	 * my certificate to a node is included in the message.
	 *
	 * @param dst	the address pair of the node the certificate reply
	 * is sent to.
	 */
	protected final void sendCertReply(I_AddressPair dst)
	{
		//messagesSentInLastHeartbeat++;

		if (debug)
			System.out.println("Sending CertificateReply to " + dst);

		SecInfoExchange_Message NewPacket =
				new SecInfoExchange_Message(
				SecInfoExchange_Message.CertReply,
				config.getOverlayHash(),
				node.getAddressPair(),
				dst,
				this);

		adapter.sendUnicastMessage(dst.getPhysicalAddress(), NewPacket);
	}

	/**
	 * Sends a request for a node's neighborhood key.
	 *
	 * @param dst	the address pair of the node the key request is sent to.
	 */
	protected final void sendKeyRequest(I_AddressPair dst)
	{

		/*
		 * If key distribution mode is shared group key, do nothing.
		 */
		if (vault.getKeyMode().equals("GroupKeys"))
		{
			return;
		}
		
		/*
		 * record the time
		 */
		
		vault.keyRequestSent(dst, adapter.getCurrentTime());	

		SecInfoExchange_Message NewPacket =
			new SecInfoExchange_Message(
				SecInfoExchange_Message.KeyRequest,
				config.getOverlayHash(),
				node.getAddressPair(),
				dst,
				this);

		adapter.sendUnicastMessage(dst.getPhysicalAddress(), NewPacket);
	}

	/**
	 * Sends a copy of my neighborhood key to the given node.
	 *
	 * @param dst node that needs to receive key but hasn't yet.
	 */
	protected final void sendKeyUpdate(I_AddressPair dst)
	{
		/*
		 * If key distribution mode is set to shared group key, do nothing.
		 */
		if (vault.getKeyMode().equals("GroupKeys"))
		{
			return;
		}

		if (debug)
			System.err.println("SecurityProcessor:sendKeyUpdate: Sending KeyUpdate to " + dst);

		SecInfoExchange_Message NewPacket;

		if (dst == null)
		{
			return;
		}

		NewPacket =
			new SecInfoExchange_Message(
				SecInfoExchange_Message.KeyUpdate,
				config.getOverlayHash(),
				node.getAddressPair(),
				dst,
				this);
				
		/*
		 * Set my neighborhood key in the message.
		 */
		NewPacket.setKey(this.getMyKeyFor(dst, node.getAddressPair()));

		adapter.sendUnicastMessage(dst.getPhysicalAddress(), NewPacket);
	}

	/**
	 * Update my private or personal key. If the given private
	 * key is not null, use it to replace my private key; 
	 * Otherwise, create a new personal key.
	 * @param pKey	a private key.
	 */
	public void updatePersonalKey(PrivateKey pKey)
	{
		if (pKey == null)
		{
			vault.updateMyPersonalKey();
		}
		else
		{
			vault.setMyPrivateKey(pKey);
		}

		vault.setMyKeyTimestamp(adapter.getCurrentTime());
		if(MAX_AGE_OF_KEY<1000000)
			this.setSecurityTimer(new RekeyTimer(), MAX_AGE_OF_KEY);
	}

	/**
	 * Set timer for updating key to given node after predefined delay.
	 */
	protected void setKeyVaultCleanupTimer()
	{
		if(MAX_AGE_OF_AUTH_NODE>0){
			KeyVaultCleanupTimer timer = new KeyVaultCleanupTimer();
			this.setSecurityTimer(timer, this.MAX_AGE_OF_AUTH_NODE);
		}
	}

	/**
	 * Sends a copy of my personal key to all of my neighbors.
	 *
	 * @param dst node that needs to receive key but hasn't already.
	 */
	protected final void sendKeyUpdateToAllNeighbors(I_AddressPair dst)
	{
		/*
		 * If key distribution mode is set to shared group key, do nothing.
		 */
		if (vault.getKeyMode().equals("GroupKeys")) {
			return;
		}
		
		vault.setSequenceNum(1);

		SecInfoExchange_Message NewPacket;
		
		I_AddressPair myAddressPair = node.getAddressPair();
		
		I_AddressPair[] neighbors;
		
		/*
		 * Decides the actual destination nodes according to different key modes
		 */
		if(vault.getNeighborKeyMode().equals(KeyVault.NKM_UNIFORM_DYNAMIC_KEY)){
			neighbors = vault.getAllKeyNeighbors();
		}else{
			neighbors = node.getAllNeighbors();
		}

		int numOfNeighbors = neighbors.length;

		boolean sentToDst;

		/*
		 * This flag is used to make sure that a message is sent to the given node.
		 * If it is a key neighbor, a message is sent to it in the following loop;
		 * otherwise a message is sent to it at the end of this method.
		 */
		if (dst != null)
		{
			sentToDst = false;
		}
		else
		{
			sentToDst = true;
		}

		/*
		 * Get my personal key. If this method is called to send a new personal key
		 * to all key neighbors, the new key is created before calling this method.
		 * See this.timerExpired().
		 * 
		 */
		SecretKeySpec personalKey = vault.getMyPersonalKey();

		/*
		 * Send key update message to all neighbors for which the certicifate is known.
		 */
		for (int i = 0; i < numOfNeighbors; i++)
		{
			//messagesSentInLastHeartbeat++;
			//System.err.println("Sending Neighborhood KeyUpdate to " + neighbors[i]);

			NewPacket =
				new SecInfoExchange_Message(
					SecInfoExchange_Message.KeyUpdate,
					config.getOverlayHash(),
					myAddressPair,
					neighbors[i],
					this);

			/*
			 * Set my personal key in the message.
			 */
			NewPacket.setKey(personalKey);

			/*
			 * If the certificate of the neighbor is known, send message to it;
			 * Otherwies don't send, actually thisneighbor is exchanging certificate
			 * with local node.
			 */
			if (vault.getCertificate(neighbors[i].getLogicalAddress()) != null)
			{
				adapter.sendUnicastMessage(neighbors[i].getPhysicalAddress(), NewPacket);
			}

			/*
			 * If the given nide is one of the key neighbors, set the sentToDst flag to true.
			 */
			if (sentToDst != true)
			{
				if (neighbors[i].toString().equals(dst.toString())) {
					sentToDst = true;
				}
			}
		}

		/*
		 * Makes sure that the message gets sent to the joining node that isn't yet
		 * in the key neighborhood table.
		 */ 
		if (!sentToDst)
		{

			//System.out.println("Sending to " + dst);

			NewPacket =
				new SecInfoExchange_Message(
					SecInfoExchange_Message.KeyUpdate,
					config.getOverlayHash(),
					node.getAddressPair(),
					dst,
					this);

			/*
			 * Set my personal key in the message.
			 */
			NewPacket.setKey(vault.getMyPersonalKey());

			adapter.sendUnicastMessage(dst.getPhysicalAddress(), NewPacket);
		}
	}


	/**
     * Restore an SecInfoExchange_Message from byte array.
     *
     * @param	receiveBuffer			the byte array buffer which contains the packet
     * @param	validBytesStart	start position of valid section of the array
     * @param	validBytesEnd	  	end of valid section of the array (actually, 1 beyond last valid byte.)
     * 
	 * @see I_AdapterCallback#restoreMessage(byte[], int[], int)
	 */
	public I_Message restoreMessage(
		byte[] receiveBuffer,
		int[] validBytesStart,
		int validBytesEnd)
	{
			
		try
		{
			return SecInfoExchange_Message.restoreMessage(
				receiveBuffer,
				validBytesStart,
				validBytesEnd,
				adapter,
				config.getOverlayHash(),
				this);
		}
		catch(Exception e)
		{
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * This timer update the neighborhood key if the age of key exceed
	 * a maximum limit.
	 */
	class RekeyTimer
	{
		public boolean equals(Object o)
		{
			if (o == null)
			{
				return false;
			}
			else if (!(o instanceof RekeyTimer))
			{
				return false;
			}
			else return true;
		}
	}
	
	/**
	 * This class defines a timer to retransmit the KeyRequest
	 * if the KeyReply has not been received in time
	 */
	class KeyRequestRetransmitTimer
	{
		/*
		 * The logical address of the node to be watched.
		 */
		I_LogicalAddress LA;
		public KeyRequestRetransmitTimer(I_LogicalAddress la){
			LA = la;
		}
		/*
		 * Check if two WatchDogTimer instances are equal.
		 */
		public boolean equals(Object o)
		{
			if (o == null)
			{
				return false;
			}
			else if (!(o instanceof KeyRequestRetransmitTimer))
			{
				return false;
			}
			else if (((KeyRequestRetransmitTimer) o).LA == null)
			{
				if (LA == null)
				{
					return true;
				}
				else
				{
					return false;
				}
			}
			return LA.equals(((KeyRequestRetransmitTimer) o).LA);
		}
	}
	
	
	/**
	 * This class defines a timer clean up the expired key vault entry
	 */
	class KeyVaultCleanupTimer
	{
		
		public boolean equals(Object o)
		{
			if (o == null)
			{
				return false;
			}
			else if (!(o instanceof KeyVaultCleanupTimer))
			{
				return false;
			}
			else return true;
		}
	}
	
	
	

	/**
	 * Initiate a key exchange with a node 
	 *
	 * @param nodeLA the logical address of the node with which 
	 * 	the key is exchanged.
	 */
	public void requestKeyFor(I_LogicalAddress nodeLA)
	{
		/*
		 * If key distribution mode is shared group key, do nothing.
		 */
		if (vault.getKeyMode().equals("GroupKeys")) {
			return;
		}

		KeyVaultEntry ke = vault.getEntry(nodeLA);
		
		/*
		 * Try to get the address pair of the node
		 */
		I_AddressPair nodeAP;
		if (ke != null)
		{
			nodeAP = ke.getAddressPair();
		}
		else
		{
			nodeAP = null;
		}

		/*
		 * If the node doesn't exist in key vault table, do nothing.
		 */
		if (nodeAP == null || ke.getCert()==null)
		{
			return;
		}

		long currentTime = adapter.getCurrentTime();
		
		/*
		 * If it's currently waiting the KeyUpdate, to avoid sending out KeyRequest
		 * too frequently, don't send KeyRequest this time.
		 */
		if (vault.isWaitingKeyUpdate(nodeAP, KEY_REQUEST_TIMEOUT, currentTime))
		{
			return;
		}
		
		/*
		 * send out key request
		 */
		sendKeyRequest(nodeAP);
		/*
		 * Set a retransmission timer
		 */
		this.setSecurityTimer(new KeyRequestRetransmitTimer(nodeLA), this.KEY_REQUEST_TIMEOUT);

	}


	/**
	 * Set the time when the last secure message was received from the given
	 * node.
	 *
	 * @param nodeLA 	the logical address of the node from which the secure message
	 * 	is received.
	 */
	public void setLastSecureMessageTime(I_LogicalAddress nodeLA)
	{
		KeyVaultEntry kn = vault.getEntry(nodeLA);
		if (kn != null)
		{
			kn.secureMessageReceived(adapter.getCurrentTime());
		}
	}

	/**
	 * @return	the adapter stored in the instance of this class.
	 */
	public I_UnicastAdapter getAdapter()
	{
		return adapter;
	}

	/**
	 * @return	the overlay node object stored in the instance of this class.
	 */
	public I_Node getNode()
	{
		return node;
	}

	/**
	 * Sets the adapter stored in the instance of this class.
	 * @param adapter	an I_UnicastAdapter object.
	 */
	public void setAdapter(I_UnicastAdapter adapter)
	{
		this.adapter = adapter;
	}

	/**
	 * Sets the overlay node stored in the instance of this class.
	 * @param node	an I_Node object.
	 */
	public void setNode(I_Node node)
	{
		this.node = node;
	}


	/**
	 * @return	the adapter callback object stored in ths instance of this class.
	 */
	public I_AdapterCallback getCallBack()
	{
		return callback;
	}

	/**
	 * This function, a member of I_AdapterCallBack, is called for the
	 * WatchDogTimer timer and DelayNewKeyTimer. When a WatchDogTimer timer 
	 * occurs, the node initiates a new key exchange with the problematic node.  
	 * When a DelayNewKeyTimer timer occurs, the node check if key neighbor
	 * table is changed in the past NEW_KEY_DELAY mss. If there are changes,
	 * the node create new personal key and send it to all key neighbors.
	 *
	 * @param Timer_Object index of the timer that has expired
	 */

	public synchronized void timerExpired(Object Timer_Object)
	{
		/*
		 * If it is instance of timer for security, it is processed here.
		 */
		if (Timer_Object instanceof SecurityTimer)
		{
			Object Timer_Index = ((SecurityTimer) Timer_Object).getTimerObject();
			
			if (Timer_Index instanceof KeyRequestRetransmitTimer)
			{
				I_LogicalAddress nodeLA = ((KeyRequestRetransmitTimer) Timer_Index).LA;
				
				/*
				 * Retransmit the KeyRequest if the key is still invalid.
				 * A key is invalid means that there's no KeyUpdate received
				 * since the last KeyRequest is sent.
				 */
				if(vault.isKeyInvalid(nodeLA))
				{
					requestKeyFor(nodeLA);
				}
			}
			else if (Timer_Index instanceof KeyVaultCleanupTimer)
			{
				/*
				 * clean up expired key neighbors
				 */
				vault.clearExpiredNeighbors(
					this.MAX_AGE_OF_AUTH_NODE,
					adapter.getCurrentTime());
				
				/*
				 * Set a new KeyVaultCleanupTimer timer.
				 */
				this.setKeyVaultCleanupTimer();
			}else if(Timer_Index instanceof RekeyTimer){
				
				/*
				 * update the personal key
				 * and send it to all neighbors
				 */
				this.updatePersonalKey(null);
				this.sendKeyUpdateToAllNeighbors(null);
				
			}
			else
			{
				config.err.println("Unknow security timer: " + Timer_Index.getClass().getName());
			}

		}
		/*
		 * Otherwise pass it to the overlay node to process.
		 */
		else
		{
			callback.timerExpired(Timer_Object);
		}

	}


	/*
	 * Below are the functions defined in the I_MulticastAdapter interface.
	 */

	/**
	 * Starts the adapter. It in turn starts the node adapter.
	 * 
	 * @see hypercast.I_UnicastAdapter#Start()
	 */
	public void Start()
	{
		adapter.Start();
	}

	/**
	 * Starts the timer thread. It in turn starts the timer thread in the 
	 * node adapter.
	 * 
	 * @see hypercast.I_UnicastAdapter#startTimer()
	 */
	public void startTimer()
	{
		adapter.startTimer();
		
		/*
		 * set the key vault cleanup timer
		 */
		this.setKeyVaultCleanupTimer();
	}

	/**
	 * Stops the adapter. It in turn stops the node adapter.
	 * 
	 * @see hypercast.I_UnicastAdapter#Stop()
	 */
	public void Stop()
	{
		adapter.Stop();

	}

	 /**
	  * Stops the timer thread. It in turn stops the timer thread in the 
	  * node adapter.
	  * 
	  * @see hypercast.I_UnicastAdapter#stopTimer()
	  */
	public void stopTimer()
	{
		adapter.stopTimer();
	}

	/** 
	 * Sends out an unicast message. It creates a secure message that wraps the plaintext protocol
	 * message and calls the same method in the node adapter to send the wraped message out..
	 * 
	 * @param dst	the physical or underlay address of the destination of this message.
	 * @param msg	plaintext protocol message to be sent out.
	 * 
	 * @see hypercast.I_UnicastAdapter#sendUnicastMessage
	 * 	(hypercast.I_PhysicalAddress, hypercast.I_Message)
	 */
	public void sendUnicastMessage(I_NetworkAddress dst, I_Message msg)
	{
		//adapter.sendUnicastMessage(dst, getWrappedMessage((I_PhysicalAddress) dst, msg));
		adapter.sendUnicastMessage(dst, getWrappedMessage(dst, msg));
	}

	/**
	 * Sends out a multicast message. It in turn calls the same method in the
	 * node adapter.
	 * 
	 * @see hypercast.I_MulticastAdapter#sendMulticastMessage(hypercast.I_Message)
	 */
	public void sendMulticastMessage(I_Message msg) {
		((I_MulticastAdapter) adapter).sendMulticastMessage(getWrappedMessage(node.getAddressPair().getPhysicalAddress(), msg));
	}

	/**
	 * Creates a secure message that wraps the plaintext message.
	 * 
	 * @param dstpa	the destination of this message.
	 * @param msg	a plaintext message.
	 * 
	 * @return	secure message.
	 */
	private I_Message getWrappedMessage(I_NetworkAddress dstpa, I_Message msg)
	{
		/*
		 * For a wrap (with type ProtoMsg) message, the destination (variable dst
		 * in SecInfoExchange_Message) is not used. So below it is  set to null.
		 */
		SecInfoExchange_Message secMsg =
			new SecInfoExchange_Message(
				SecInfoExchange_Message.ProtoMsg,
				config.getOverlayHash(),
				node.getAddressPair(),
				vault.getAddressPairByPA(dstpa),
				this);

		/*
		 * Store the plaintext protocol message in the wrap message.
		 */
		secMsg.setProtoMsg(msg);

		return secMsg;
	}

	/**
	 * Creates physical address from a byte array.
	 * 
	 * @param	byteaddr  an byte array containing a physical address.
	 * @param	offset  the position in byteAddr from where availabe
	 * 			data is read.
	 * @return	a new instance of <code>I_PhysicalAddress</code>.
	 * 
	 * @see hypercast.I_UnicastAdapter#createPhysicalAddress(byte[], int)
	 */
	public I_PhysicalAddress createPhysicalAddress(byte[] byteaddr, int offset) {
		return adapter.createPhysicalAddress(byteaddr, offset);
	}

	/**
	 * Creates physical address from a string.
	 * 
	 * @param	addrString  an string representing a physical address.
	 * @return	a new instance of <code>I_PhysicalAddress</code>.
	 * 
	 * @see hypercast.I_UnicastAdapter#createPhysicalAddress(java.lang.String)
	 */
	public I_PhysicalAddress createPhysicalAddress(String addrString) {
		return adapter.createPhysicalAddress(addrString);
	}

	/**
	 * Creates a new physical address.
	 *
	 * @return	a new instance of <code>I_PhysicalAddress</code>.
	 * 
	 * @see hypercast.I_UnicastAdapter#createPhysicalAddress()
	 */
	public I_PhysicalAddress createPhysicalAddress() {
		return adapter.createPhysicalAddress();
	}

	
	/** Returns the size of physical address, when carried in a packet,
	 *  recognized by this adapter. 
	 */
	public int getPhysicalAddressSize()
	{
		return adapter.getPhysicalAddressSize();
	}

	/** 
	 * Creates underlay address from a byte array of address - the inverse of 
	 * <code>toByteArray</ode>. The first 4 bytes reprsent the IP address, 
	 * and the last 4 bytes represent the port number.
	 *
	 * @param	byteAddressAndPort  an byte array containing a physical address
	 * @param	offset  the position in byteAddressAndPort from where availabe
	 * 			data is read.
	 * @return	a new instance of <code>I_UnderlayAddress</code>
	 * @see		INETv4OnePort#toByteArray()
	 */
	 public I_UnderlayAddress createUnderlayAddress(byte[] byteAddressAndPort, int offset) {
		return adapter.createUnderlayAddress(byteAddressAndPort, offset);
	 }
	
		
	 /** 
	  * Creates underlay address object from a string.
	  * @param	addrString  an string representing a physical address.
	  * @return	a new instance of <code>I_UnderlayAddress</code>.
	  */
	 public I_UnderlayAddress createUnderlayAddress(String addrString) {
		return adapter.createUnderlayAddress(addrString);
	 }
	
	
	 /** 
	  * Creates a new underlay address.
	  * @return	a new instance of <code>I_UnderlayAddress</code>.
	  */
	 public I_UnderlayAddress createUnderlayAddress() {
		return adapter.createUnderlayAddress();
	 }
	
	
	/** Returns the size of underlay address, when carried in a packet,
	 *  recognized by this adapter. 
	 */
	public int getUnderlayAddressSize()
	{
		return adapter.getUnderlayAddressSize();
	}

	/**
	 * Reset the sockets in the adapter. It in turn call the same method of the 
	 * node adapter.
	 * 
	 * @see hypercast.I_UnicastAdapter#Reset(java.lang.Object)
	 */
	public boolean Reset(Object socketObj) {
		return adapter.Reset(socketObj);
	}


	/** Get information of the adapter. It in turn calls the same method
	 * in the node adapter.
	 *
	 * @see hypercast.I_UnicastAdapter#getInfo()
	 */
	public Object getInfo() {
		return adapter.getInfo();
	}


	/**
	 * Set the adapter callback object.
	 * 
	 * @see I_UnicastAdapter#setCallback(I_AdapterCallback)
	 */
	public void setCallback(I_AdapterCallback adaptercallbackObj)
	{
		/*
		 * Sets the adapter callback object stored in this object.
		 */
		callback = adaptercallbackObj;
		
		/*
		 * Sets the overlay node object stored in this object.
		 */
		this.node = (I_Node) adaptercallbackObj;
		
	}
	
	/**
	 * The function return the key that the local node is exchanging with a remote
	 * destination node. When in NeighborhoodKey2 mode, the node might use different
	 * keys with different remote nodes.
	 * 
	 * @param dst_ap	the address of the remote node
	 * @param src_ap	the address of the current node
	 * @return	the key
	 */
	public SecretKeySpec getMyKeyFor(I_AddressPair dst_ap, I_AddressPair src_ap){
		if(vault.getKeyMode().equals("GroupKeys")){
			return vault.getGroupKey();
		}else if(vault.getNeighborKeyMode().equals(KeyVault.NKM_UNIFORM_DYNAMIC_KEY)){
			return vault.getMyPersonalKey();
		}else if(vault.getNeighborKeyMode().equals(KeyVault.NKM_SHARED_PROTOCOL_KEY)){
			return vault.getGroupKey();//shared key
		}else {
			SecretKeySpec key = vault.getMyKeyFor(dst_ap, src_ap);
			//System.err.println("key="+HTTP_ServerUtility.toHexString(key.getEncoded()));
			return key;
		}
		
	}
	/**
	 * @see #getMyKeyFor(I_AddressPair, I_AddressPair)
	 * return the key timestamp for the key that is return by getMyKeyFor() function
	 */
	public long getMyKeyTimestampFor(I_AddressPair dst_ap, I_AddressPair src_ap){
		
		if(vault.getKeyMode().equals("GroupKeys")){
			return 0;
		}else if(vault.getNeighborKeyMode().equals(KeyVault.NKM_UNIFORM_DYNAMIC_KEY)){
			return vault.getMyKeyTimestamp();
		}else if(vault.getNeighborKeyMode().equals(KeyVault.NKM_SHARED_PROTOCOL_KEY)){
			return 0; //shared key
		}else{
			return vault.getMyKeyTimestampFor(dst_ap, src_ap);
		}
	}


	/**
	 * This class defines a wrap timer for a secure event.
	 * A timer used by the SecurityProcessor is wrapped in this
	 * object in order for the SecurityProcessor to recognize
	 * later.
	 * 
	 */
	class SecurityTimer
	{
		static final int PROTOCOL = 0;
		static final int SECURITY = 1;
		
		/**
		 * a secure eventtimer that is wraped by this object.
		 */
		Object timer_id;
		
		/**
		 * The type of this SecurityTimer object.
		 */
		int type;

		/**
		 * Constructor.
		 * 
		 * @param ti	a secure timer event.
		 * @param tp	the type of this SecurityTimer object.
		 */
		SecurityTimer(Object ti, int tp)
		{
			timer_id = ti;
			type = tp;
		}

		/**
		 * Get the secure timer event wraped by this this object.
		 * @return	a secure timer event object.
		 */
		public Object getTimerObject()
		{
			return timer_id;
		}

		/**
		 * Checks if two SecurityTimer object are equal.
		 * 
		 * @return	true if they are eaual; false otherwise.
		 */
		public boolean equals(Object o)
		{
			if (o == null)
			{
				return false;
			}
			else if (!(o instanceof SecurityTimer))
			{
				return false;
			}

			SecurityTimer o2 = (SecurityTimer) o;
			if (o2.getType() != getType())
			{
				return false;
			}

			return o2.getTimerObject().equals(getTimerObject());
		}

		/**
		 * Gets the type of this SecurityTimer object.
		 * 
		 * @return the type of this timer event.
		 */
		public int getType()
		{
			return type;
		}
	}


	/** 
	 * Set a <code>Timer</code> event. <p>. It calls the same method in the node
	 * adapter.
	 * 
	 * @param	timer_id	a time event ID
	 * @param	delay_ms		a time in ms
	 */		
	public void setTimer(Object timer_id, long delay_ms)
	{
		adapter.setTimer(timer_id, delay_ms);
	}

	/** 
	 * Set a <code>SecurityTimer</code> event <p>. It calls the same method in the node
	 * adapter.
	 * @param	timer_id	a time event ID
	 * @param	delay_ms	a time in ms
	 */		
	public void setSecurityTimer(Object timer_id, long delay_ms)
	{
		adapter.setTimer(new SecurityTimer(timer_id, SecurityTimer.SECURITY), delay_ms);
	}

	/** 
	 * Clear a <code>Timer</code> event <p>. It calls the same method in the node
	 * adapter.
	 * @param	timer_id	a time event ID
	 */				
	public void clearTimer(Object timer_id)
	{
		adapter.clearTimer(timer_id);
	}

	/** 
	 * Clear a <code>SecurityTimer</code> event<p>. It calls the same method in the node
	 * adapter.
	 * @param	timer_id	a time event ID
	 */				
	public void clearSecurityTimer(Object timer_id)
	{
		adapter.clearTimer(new SecurityTimer(timer_id, SecurityTimer.SECURITY));
	}

	/** 
	 * Get a <code>Timer</code> event <p>. It calls the same method in the node
	 * adapter.
	 * @param	timer_id	a time event ID
	 * @return	time in ms
	 */			
	public long getTimer(Object timer_id)
	{
		return adapter.getTimer(timer_id);
	}

	/** 
	 * Get a <code>SecurityTimer</code> event <p>. It calls the same method in the node
	 * adapter.
	 * @param	timer_id	a security time event ID
	 * @return	time in ms
	 * 
	 * @see hypercast.I_UnicastAdapter#getCurrentTime()
	 */			
	public long getSecurityTimer(Object timer_id)
	{
		return adapter.getTimer(new SecurityTimer(timer_id, SecurityTimer.SECURITY));
	}
	
	/**
	 * Gets the current time.
	 * 
	 * @return	current time in ms.
	 * 
	 * @see hypercast.I_UnicastAdapter#getCurrentTime()
	 */
	public long getCurrentTime()
	{
		return adapter.getCurrentTime();
	}


	/**
	 * Gets the multicast address of the adapter. It in turn the same method in the 
	 * node adapter.
	 * 
	 * @see hypercast.I_MulticastAdapter#getMulticastAddress()
	 */
	public I_PhysicalAddress getMulticastAddress() {
		return ((I_MulticastAdapter) adapter).getMulticastAddress();
	}



	/** 
	 * Return the result of query for the statistics specified by the given xpath.
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance represents the statistics to be queried.
	 * @see I_Stats#getStats
	 */
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return adapter.getStats(doc, xpath);
	}

	/** 
	 * Set the statistics specified by the given xpath. The value actually set is
	 * returned.
	 * 
	 * @param doc   	Document used for create new elements or nodes.
	 * @param xpath 	XPath instance represents the statistics to be queried.
	 * @param newValue	Element representing the value or sub-tree to be set.
	 * @see I_Stats#setStats
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	{
		return adapter.setStats(doc, xpath, newValue);
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in read schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return adapter.getReadSchema(doc, xpath);
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in write schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getWriteSchema(Document, XPath)
	 */
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return adapter.getWriteSchema(doc, xpath);
	}
	

	/**
	 * Gets the element name of the adapter. It in turn call the same method in the 
	 * node adapter.
	 *
	 * @see I_Stats#getStatsName()
	 */
	public String getStatsName ()
	{
		return adapter.getStatsName ();
	}

	/**
	 * Sets the element name of the adapter. It in turn call the same method in the 
	 * node adapter.
	 *
	 * @param name the statistics name to be set.
	 *  
	 * @see I_Stats#setStatsName(String)
	 */
	public void setStatsName (String name)
	{
		adapter.setStatsName(name);
	}
	
	/**
	 * Handle neighborhood change event.
	 */
	public void handle_NODE_NEIGHBORHOODCHANGED (final NODE_NEIGHBORHOODCHANGED event) {
		
		/*
		 * If key distribution mode is set to shared group key, do nothing.
		 */
		if (vault.getKeyMode().equals("GroupKeys")) {
			return;
		}
		
		if(!vault.getNeighborKeyMode().equals(KeyVault.NKM_UNIFORM_DYNAMIC_KEY)){
			//debug
			//System.err.println("Neighborhood Change Event");
			
			/* synchronize neighbors and key_neighbors */
			boolean changed = false;
			I_AddressPair neighbors[] = node.getAllNeighbors();
			
			I_AddressPair key_neighbors[] = vault.getAllKeyNeighbors();
			
			for(int i=0;i<key_neighbors.length;i++){
				KeyVaultEntry entry = vault.getEntry(key_neighbors[i]);
				boolean isNeighbor = false;
				for(int j=0;j<neighbors.length;j++){
					if(key_neighbors[i].equals(neighbors[j])){
						if(!entry.isOverlayNeighbor()){
							changed = true;
							entry.setIsOverlayNeighbor(true);
						}
						isNeighbor = true;
						break;
					}
				}
				if(!isNeighbor){
					if(entry.isOverlayNeighbor()){
						changed = true;
						entry.setIsOverlayNeighbor(false);
						/* set the current neighbor key as the no neighbor key for this node*/
						vault.setMyNoNeighborKeyFor(key_neighbors[i].getLogicalAddress(), vault.getMyPersonalKey(), vault.getMyKeyTimestamp());
					}
				}
			}
			
			if(changed){
				//rekeying
				//debug
				//System.err.println("Rekeying due to neighborhood change");
				this.updatePersonalKey(null);
				this.sendKeyUpdateToAllNeighbors(null);
			}
		}
		
	}

}


/**
 * This is a class of a general address pair. It's used in 
 * SecurityProcessor to generate an address pair for
 * any logical address.
 * 
 * @author Guangyu Dong 2004-7-22
 * @version 3.0
 *
 */
class G_AddressPair implements I_AddressPair {

	/** the physical address */
	private I_PhysicalAddress PA;

	/** the logical address */
	private I_LogicalAddress LA;

	/**Constructor
	 * @param pa	physical address
	 * @param la	logical address
	 */
	public G_AddressPair(I_PhysicalAddress pa, I_LogicalAddress la) {
		PA = pa;
		LA = la;
	}

	/** @return the physical address*/
	public I_PhysicalAddress getPhysicalAddress() {
		return PA;
	}

	/** @return the logical address*/
	public I_LogicalAddress getLogicalAddress() {
		return LA;
	}
	public String toString() {
		return "PA:" + PA.toString() + " LA:" + LA.toString();
	}
	public boolean equals(Object o){
		if(o==null){
			return false;
		}
		if(! (o instanceof I_AddressPair)){
			return false;
		}
		I_AddressPair ap = (I_AddressPair)o;
		return ap.getLogicalAddress().equals(LA) && ap.getPhysicalAddress().equals(PA);
	}

}
