/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.adapters;

import hypercast.*;
import hypercast.util.XmlUtil;

import java.io.*;
import java.util.*;
import java.net.*;


import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.apache.xpath.*;

/** 
 * This class implements an adapter which supports the folowing 
 * three kinds of data transmission:
 * 1. UDP unicast;
 * 2. TCP unicast;
 * 3. UDP multicast.
 * This class employs a UDP_MulticastAdapter object and 
 * TCP_UnicastAdapter object to fulfill data transmission
 * task, instead of implementing all required functions 
 * by itself.
 * This class has its own Timer object and Timer processor.
 *
 * This class defines a INETv4TwoPorts physical address in which the 
 * "port-plus-one" is applied.
 * 
 * @author HyperCast Team
 * @author Jianping Wang
 * @version 3.0, Feb 02, 2004
 */

public class TCP_UDP_Adapter implements I_MulticastAdapter
{

	/** Maximum queue length of incomming connection */
final static int MAX_BACKLOG = 50;

    /** Prefix for all property names read by this adapter.*/
protected static String PROPERTY_PREFIX;

    /** property object for configration */
protected HyperCastConfig config;

	/** UDP unicast socket in this adapter */
DatagramSocket UDPSock = null;

/** TCP unicast  socket in this adapter */
ServerSocket TCPSock = null;


    /** The Callback object */
I_AdapterCallback callback;

    /** The timer that works with this adapter */
Timer Uni_Timer;
    /** Runnable object that hosts the thread that checks the timers */
private Uni_TimeEventProcessor uniTimeProcessor;

	/** The timer status string for debuging */
String uniTimerProcDebug;

    /** The UDP_MulticastAdapter object */
protected I_MulticastAdapter multicast_adapter = null;

    /** The TCP_UnicastAdapter object */
protected I_UnicastAdapter tcp_adapter = null;

    /** Current state of the node.  It goes from STOPPED to STARTING
         * to STARTED to STOPPING and then back to STOPPED. */
    protected int state;
        
    /** value for <code>state</code>. */
    final static int STOPPED = 0;
    /** value for <code>state</code>. */
    final static int STARTING = 1;
    /** value for <code>state</code>. */
    final static int STARTED = 2;
    /** value for <code>state</code>. */
    final static int STOPPING = 3;
                
    /** Used to protect <code>state</code> when its value is changed. */
	protected Object stateGuard;
	
     /** The object used to wait() on when waiting for other threads to die. */
        Object deathGuard;

	/** The string which indicate this adapter is Node Adapter or
	 * Socket Adapter: 
	 * "nodeadpt" - Node Adapter, "sockadpt" - Socket Adapter.
	 */
	String AdptIndicator;   

	/** Determine if Node adapter should try random port number 
	 * if physical addresss assignment fails.
	 */
	String AllowRandomPhyAddrPort;   

	/** The pointer to socket adapter */
	I_UnicastAdapter socketAdpter;

	/** The number of sub-adapters contained in this adapter. */
	final static int NUM_OF_SUBADAPTERS = 2;

	/**
	 * A StatsProcessor instance which handles the statistics in this object.
	 */
	private StatsProcessor statsPro;

	/**
	 * The statistics name for this object (as an I_Stats instance). Its default
	 * name is defined in the socket configuration schema file. It can be assigned
	 * by any other object through method setStattisticsName defined in the 
	 * I_Stats interface.
	 */
	private String statisticsName;


/** 
 * Constructs a TCP_UDP_Adapter object with an HyperCastConfig
 * object and an I_UnicastAdapter object.<p>
 * 
 * @param	c			a HyperCastConfig object.
 * @param	scktAdpt	object representing the Socket Adapter. It is set to null when 
 * 						creating the Socket Adapter.
 * @param	AddrString	string containing physical address information.
 * @param	prefix		property prefix.
 */
public TCP_UDP_Adapter(HyperCastConfig c, I_UnicastAdapter scktAdpt, String AddrString, String prefix){
    config = c;

	/*
	 * Assign the adapter indicator.
	 */
	if (prefix.startsWith("NodeAdpt"))
	{
		AdptIndicator = "nodeadpt";
		PROPERTY_PREFIX = "/Public/NodeAdapter/"+prefix+"/";
	} 
	else
	{
		AdptIndicator = "scktadpt";
		PROPERTY_PREFIX = "/Public/SocketAdapter/"+prefix+"/";
	} 

	socketAdpter = scktAdpt;
    
	if (prefix.startsWith("NodeAdpt"))
	{
		AllowRandomPhyAddrPort = config.getTextAttribute
				(XmlUtil.createXPath(PROPERTY_PREFIX+"AllowRandomPhyAddrPort"));
	} 

	/* 
	 * Initialize the sockets running on Node Adapter and Socket Adapter this adapter. 
	 * If the initialization fails, a runtime exception is thrown. Only Node Adapter starts
	 * sockets initialization. It not only initializes the socket in Node Adapter, 
	 * but also initialize the socket in Socket Adapter.
	 */
	if ((!AdptIndicator.equals("scktadpt")) && !InitializeDatagramServerSocket(scktAdpt, config, AddrString))
	{
		throw new HyperCastFatalRuntimeException ("Could not initialize local unicast socket!");
	}


	/*
	 * Create the UDP_MulticastAdapter object contained in this adapter.
	 */
	multicast_adapter = new UDP_MulticastAdapter(config, UDPSock, PROPERTY_PREFIX + ".UDPMulticast");

	/*
	 * Create the TCP_UniastAdapter object contained in this adapter.
	 */
    tcp_adapter = new TCP_UnicastAdapter(config, TCPSock, PROPERTY_PREFIX + ".TCP");


	Uni_Timer = null;
	uniTimeProcessor = null;

    state = STOPPED;
    stateGuard = new Object();
	deathGuard = new Object();

	InitStatisticsStructure();
}


/**
 * Initialize DatagramSocket and ServerSocket used by this adapter.
 * 
 * @param	ScktAdpter	 A I_UnicastAdapter object representing Socket Adapter.
 * 			ScktAdpter is null when this method is called in Socket Adapter.
 * @param	config	 a HyperCastConfig object.
 * @param	AddrString	 A string carring physical address information.
 */
public boolean InitializeDatagramServerSocket(I_UnicastAdapter ScktAdpter, HyperCastConfig config, String AddrString)
{

	/*
	 * Set initial values for two ports and InetAddress.
	 */
	int portNodeAdpt = 0;
	int portScktAdpt = 0;
	InetAddress ipAddr = null;

	/*
	 * If this is Socket Adapter, return. The initialization of sockets in Socket Adapter
	 * is trigged from Node Adapter.
	 */
	if (AdptIndicator.equals("scktadpt"))
	{
		return true;
	}
		
	/*
	 * If this is Node Adapter and physical address is specified (in string format) with papameter,
	 * the specified physical address string is used to create the physical address of this adapter.
	 */
	String addrStr;
	if (AddrString != null) {
		addrStr = AddrString;
	}
	/*
	 * If this is Node Adapter and no physical address is specified (in string format) with papameter,
	 * the adapter reads configuration file to get the information of physical address.
	 */
	else {
		String addrType = config.getTextAttribute
				(XmlUtil.createXPath(PROPERTY_PREFIX + "PhyAddr"));
		addrStr = config.getTextAttribute
				(XmlUtil.createXPath(PROPERTY_PREFIX + "PhyAddr/" + addrType));

	}
		
	String[] paStr = INETv4TwoPorts.getPartsFromAddrAndPortsString(addrStr);

	/*
	 * Create InetAddress.
	 */			
	if (!paStr[0].equals("")) {
		try
		{
		   ipAddr = InetAddress.getByName(paStr[0]);
		}
		catch (UnknownHostException e)
		{
		   //throw new HyperCastFatalRuntimeException (""+addrStr + " is a bad address string.**" + e);
		   config.err.println("Cannot create InetAddress from " + ipAddr + " : " + e);
		   ipAddr = null;
		}
	}

	/*
	 * Obtain the first port number.
	 */
	if (!paStr[1].equals("")) {
		try 
		{
		  portNodeAdpt = Integer.parseInt(paStr[1]);
		  if ((portNodeAdpt < 0) && (portNodeAdpt != -1))
		  {
			throw new IllegalArgumentException("" + paStr[1] + " is not a available value for port number.");
		  }
		}
		catch (NumberFormatException e) 
		{
		  throw new IllegalArgumentException("Unable to parse  \"" + paStr[1] + "\" as a port number.");
		}
	}
		
	/*
	 * Obtain the second port number.
	 */
	if (!paStr[2].equals("")) {
		try 
		{
			portScktAdpt = Integer.parseInt(paStr[2]);
			if ((portScktAdpt < 0) && (portScktAdpt != -1))
			{
			  throw new IllegalArgumentException("" + paStr[2] + " is not a available value for port number.");
			}
		}
		catch (NumberFormatException e) 
		{
		  throw new IllegalArgumentException("Unable to parse  \"" + paStr[2] + "\" as a port number.");
		}
	}


	/*
	 * Initialize the DatagramSocket with obtained IP address and port information.
	 */
	if (!ProcessingSocketInitialization(ScktAdpter, portNodeAdpt, portScktAdpt, ipAddr)) {
		/*
		 * When the initialization of the DatagramSocket fails, if AllowRandomPhyAddrPort
		 * is specified to "Yes", the adapter will try to initialize the DatagramSocket
		 * without specifying port number, i.e., asking system to pick port number.
		 */
		if ((AllowRandomPhyAddrPort.equalsIgnoreCase("Yes")) && ((portNodeAdpt > 0) || (portScktAdpt > 0))) {
			if (!ProcessingSocketInitialization(ScktAdpter, 0, 0, ipAddr)) {
				return false;
			}
			else {
				return true;
			}
		}
		else {
			return false;
		}
	}
	else {
		return true;
	}

}


/**
 * Process the initialization of underlay network sockets.
 * 
 * @param	ScktAdpter	 	A I_UnicastAdapter object representing Socket Adapter.
 * @param	portNodeAdpt	port number of the socket in Node Adapter.
 * @param	portScktAdpt	port number assigned to the socket in Socket Adapter.
 * @param	ipAddr			IP address to which the socket is bounded.
 */
private boolean ProcessingSocketInitialization(I_UnicastAdapter ScktAdpter, int portNodeAdpt, int portScktAdpt, InetAddress ipAddr)
{
	/*
	 * If the port number of Node Adapter is specified as a positive number,
	 * first initialize the socket in Node Adapter and then the socket in Socket
	 * Adapter. 
	 */
	if (portNodeAdpt > 0)
	{
		/*
		 * Initialize the socket in Node Adapter. If it fails, 
		 * stop the initialization process and return.
		 */
		if (!createDatagramServerSocket(portNodeAdpt, ipAddr))
		{
			return false;
		}
			
		/*
		 * If Socket Adapter is not specified, return.
		 */
		if (ScktAdpter == null)
		{
		   return true;
		}

		/*
		 * Initialize the socket in Socket Adapter by calling Reset method. If it fails, 
		 * stop the initialization process and return.
		 */
		String addressStr;
		if (ipAddr != null) 
		{
			addressStr = new String(getIPString(ipAddr)+":"+portScktAdpt);
		}
		else
		{
			addressStr = new String(":"+portScktAdpt);
		}
		if (!ScktAdpter.Reset(addressStr))
		{
			return false;
		}
	}
	/*
	 * If the port number of Node Adapter is not specified (-1 or 0), first the 
	 * socket in Socket Adapter and then the socket in Node Adapter.
	 */
	else
	{

		/*
		 * If Socket Adapter is not specified, skip the process of initializing the 
		 * socket for Socket Adapter. Otherwise initialize the socket in Socket Adapter.
		 */
		if (ScktAdpter != null)
		{
			/*
			 * Initialize the socket in Socket Adapter by calling Reset method. If it fails, 
			 * stop the initialization process and return.
			 */
			String addressStr;
			if (ipAddr != null) 
			{
				addressStr = new String(getIPString(ipAddr)+":"+portScktAdpt);
			}
			else
			{
				addressStr = new String(":"+portScktAdpt);
			}
			if (!ScktAdpter.Reset(addressStr))
			{
				return false;
			}
		}
		
		/*
		 * Initialize the socket in Node Adapter. If it fails, 
		 * stop the initialization process and return.
		 */
		if (!createDatagramServerSocket(portNodeAdpt, ipAddr))
		{
			return false;
		}
			
	}

	/*
	 * The initialization is successful. Return.
	 */
	return true;
}


private static Random rand = new Random();

/**
 * Creates a DatagramSocket object and a ServerSocket object.
 * 
 * @param	portNum	 the port number for the TCP and UDP unicast sockets. 
 * 					 If portN is set to 0, the port will be the minimal
 * 					 available port shosen by system; If portN is -1,
 * 					 the port number is chosen randomly.
 * @param	ipAddr 	 IP address to which the created socket is binded.
 */
public boolean createDatagramServerSocket(int portNum, InetAddress ipAddr)
{

	/*
	 * If IP address is not specified, the local sockets are bound to 
	 * the default IP address chosen by system.
	 */
	if (ipAddr == null) {

		for (int retries = 0; retries < MAX_RETRIES_TO_OPEN_SOCKET; retries++)
		{
		   try 
		   {
				/*
				 * If port number parameter is set to 0, system picks the minimum available port number to use.
				 */
			  if (portNum == 0)
			  {
				UDPSock = new DatagramSocket();
			  }
				/*
				 * If port number parameter is set to -1, a random number is used as the port number.
				 */
			  else if (portNum == -1)
			  {
				UDPSock = new DatagramSocket(Math.abs(rand.nextInt()) % (256*256));
			  }
				/*
				 * If port number parameter is set to a positive number, it is used as the port number.
				 */
			  else {
				UDPSock = new DatagramSocket(portNum);
			  }
		   }
		   catch (SocketException se) 
		   {
				/*
				 * If portNum is -1, which means the port number is randomly chosen,
				 * it may conflict with the port number used in another adapter (Node
				 * Adapter or Socket Adapter). In this case, retry to get another port 
				 * number. Otherwise, the creation of DatagramSocket fails. Return.
				 */
				//System.out.println("TCP_UDP_Adapter(1): IOException:"+se); 
				if (portNum == -1)
				{
				   continue;
				}
				else
				{
				   return false;
				}
		   }
	

			/*
			 * Use the port number of DatagramSocket to create ServerSocket.
			 */
		   try 
		   {
			  TCPSock = new ServerSocket(UDPSock.getLocalPort());
		   }
		   catch (IOException se) 
		   {
				/*
				 * If portNum is -1, which means the port number is randomly chosen,
				 * it may conflict with the port number used in another adapter (Node
				 * Adapter or Socket Adapter). In this case, retry to get another port 
				 * number. Otherwise, the creation of DatagramSocket fails. Return.
				 */
				//System.out.println("TCP_UDP_Adapter(1): IOException:"+se); 
				UDPSock.close();
				if (portNum == -1)
				{
				   continue;
				}
				else
				{
				   return false;
				}
		   }
	
		   return true;
		}
	}
	else {

		for (int retries = 0; retries < MAX_RETRIES_TO_OPEN_SOCKET; retries++)
		{
		   // designed to throw an exception if no socket can be created.
		   try 
		   {
				/*
				 * If port number parameter is set to 0, system picks the minimum available port number to use.
				 */
			  if (portNum == 0)
			  {
				UDPSock = new DatagramSocket(portNum, ipAddr);
			  }
				/*
				 * If port number parameter is set to -1, a random number is used as the port number.
				 */
			  else if (portNum == -1)
			  {
				UDPSock = new DatagramSocket(Math.abs(rand.nextInt()) % (256*256), ipAddr);
			  }
				/*
				 * If port number parameter is set to a positive number, it is used as the port number.
				 */
			  else {
				UDPSock = new DatagramSocket(portNum, ipAddr);
			  }
		   }
		   catch (SocketException se) 
		   {
				/*
				 * If portNum is -1, which means the port number is randomly chosen,
				 * it may conflict with the port number used in another adapter (Node
				 * Adapter or Socket Adapter). In this case, retry to get another port 
				 * number. Otherwise, the creation of DatagramSocket fails. Return.
				 */
				if (portNum == -1)
				{
				   continue;
				}
				else
				{
				   return false;
				}
		   }
	
		   /*
			* Use the port number of DatagramSocket to create ServerSocket.
			*/
		   try 
		   {
			  TCPSock = new ServerSocket(UDPSock.getLocalPort(), MAX_BACKLOG, ipAddr);
		   }
		   catch (IOException se) 
		   {
				/*
				 * If portNum is -1, which means the port number is randomly chosen,
				 * it may conflict with the port number used in another adapter (Node
				 * Adapter or Socket Adapter). In this case, retry to get another port 
				 * number. Otherwise, the creation of DatagramSocket fails. Return.
				 */
				//System.out.println("TCP_UDP_Adapter(1): IOException:"+se); 
				UDPSock.close();
				if (portNum == -1)
				{
				   continue;
				}
				else
				{
				   return false;
				}
		   }
	
		   return true;
		}
	}
	return false;
}


/**
 * Get string representation of IP address in a InetAddress object.
 * @param	addr	An InetAddress object. 
 */
public String getIPString(InetAddress addr) {
	if (addr == null) return new String("");
		
	String addrstr = addr.toString();
	int slash_loc = addrstr.indexOf('/');
	return addrstr.substring(slash_loc+1, addrstr.length());
}


/** Get information of this adapter.
 */ 
public Object getInfo() {
	/*
	 * It returns a String object that carries the port number 
	 * used by the socket in this adapter.
	 */
   int port = UDPSock.getLocalPort();
   return new String(""+port);
}


/**
 * Resets the local unicast socket with a object which
 * implemnents an unicast socket.
 */
public boolean Reset(Object SocketObj) {
	/*
	 * If parameter infoObj is a string, this string will bring IP address and port number
	 * information that are used to reset the DatagramSocket in this adapter. The passed
	 * string has format "IPaddrString:port".
	 */
	if (SocketObj instanceof String) {
		String addrstring = (String) SocketObj;

		/*
		 * If addString is "", close the DatagramSocket and ServerSocket and return.
		 */
		if (addrstring.equals(""))
		{
			if (UDPSock != null) UDPSock.close();
		
			try {
			   if (TCPSock != null) TCPSock.close();
			}
			catch (IOException  se) 
			{
				return false;
			}
			return true;
		}


		int cl_loc = addrstring.indexOf(':');
		String ipString = addrstring.substring(0, cl_loc);
		int portnumber = Integer.parseInt(addrstring.substring(cl_loc+1, addrstring.length()));
			
		/*
		 * If the local DatagramSocket and ServerSocket satisfy the requirement of reset parameters, do nothing and return.
		 */			
		if ((TCPSock != null) && 
			((portnumber <= 0) ||
			((portnumber > 0) && (TCPSock.getLocalPort() == portnumber) && (ipString.equals(getIPString(TCPSock.getInetAddress()))))))
		{
			return true;
		}

		/*
		 * Otherwise the DatagramSocket is reset.
		 */
		if (UDPSock != null) 
		{
			UDPSock.close();
		} 
		
		try {
		   if (TCPSock != null)
		   {
		   		TCPSock.close();
		   } 
		}
		catch (IOException  se) 
		{
			return false;
		}
	
		InetAddress ipaddress = null;
		/*
		 * If ipString is specified (not ""), it is used to create the local DatagramSocket and
		 * ServerSocket, i.e. the local DatagramSocket and ServerSocket are bound to it. Another 
		 * case is that the ipString is obtained from calling USocket.getLocalAddress(), where 
		 * USocket is a DatagramSocket instance. In this case, ipAddress is always "0.0.0.0".
		 */
		if ((!ipString.equals("")) && (!ipString.equals("0.0.0.0")))
		{
			try
			{
				ipaddress = InetAddress.getByName(ipString);
			}
			catch (UnknownHostException e)
			{
			   System.err.println(ipString + " is a bad address string.**" + e);
			   ipaddress = null;
			}
		}

		/*
		 * If ipString is "" or "0.0.0.0", the default IP address picked by system will be used 
		 * to create the local Datagram and Server Socket, i.e. the local DatagramSocket and
		 * ServerSocket are bound to this default IP address.
		 */
		return createDatagramServerSocket(portnumber, ipaddress);
	}
	
	return true;
}



	/**
     * Set <code>I_AdapterCallback</code> object. <p>
     * @param   cb  an object of type <code>I_AdapterCallback</code>
     */
public void setCallback(I_AdapterCallback cb) {
    if (state != STOPPED)
       throw new IllegalStateException("Tried to change callback while TCP_UDP_Adapter was running!");
    callback = cb;

	if (multicast_adapter != null) multicast_adapter.setCallback(cb);
	if (tcp_adapter != null) tcp_adapter.setCallback(cb);
}


     /**
      * Set a <code>Timer</code> event. <p>
      * @param       timer_id        a time event ID
      * @param       time_ms         a time in ms
      */
public synchronized void setTimer(Object timer_id, long time_ms) {
        if (state == STOPPED || state == STARTING)
            throw new IllegalStateException("TCP_UDP_Adapter: tried to set Timer when adapter was not running.");
                Uni_Timer.setTimer(timer_id, time_ms);
}

     /**
      * Get a <code>Timer</code> event. <p>
      * @param       timerid        a time event ID
      * @return      time in ms
      */
public long getTimer(Object timerid) {
        if (state == STOPPED || state == STARTING)
            throw new IllegalStateException("TCP_UDP_Adapter: tried to get Timer when adapter was not running.");
                return Uni_Timer.getTimer(timerid);

}

     /**
      * Clear a <code>Timer</code> event. <p>
      * @param       timerid        a time event ID
      */
public void clearTimer(Object timerid) {
        if (state == STOPPED || state == STARTING)
            throw new IllegalStateException("TCP_UDP_Adapter: tried to clear Timer when adapter was not running.");
                Uni_Timer.clearTimer(timerid);

}

    /**
     * Get current system time.
     */
public long getCurrentTime() {
        return System.currentTimeMillis();
}

    /**
     * Starts all the threads in this Adapter
     */
public void Start() {
    synchronized (stateGuard)
    {
        while (true)
        {
           if (state == STOPPED)
           {
              state = STARTING;
              break;
           }
           else if (state == STARTED)
           {
              throw new IllegalStateException("Tried to start TCP_UDP_Adapter twice!");
           }

           try
           {
              stateGuard.wait();
           }
           catch (InterruptedException ie)
           {
           }
       }
    }

    if (callback == null)
       throw new IllegalStateException("Tried to start TCP_UDP_Adapter without setting callback!");

	/*
	 * starts the threads in the UDP_MulticastAdapter and TCP_UnicastAdapter objects 
	 * contained in this adapter.
	 */
	if (multicast_adapter != null) multicast_adapter.Start();
	if (tcp_adapter != null) tcp_adapter.Start();

    synchronized (stateGuard)
    {
        state = STARTED;
        stateGuard.notify();
    }
}

/** This function starts the Timer Processor thread.
  */        
public  void startTimer() {            
	if (Uni_Timer != null) {
	   if (!uniTimeProcessor.isDead()) {
		 System.out.println("Attempt to start timer thread twice!");
		 return;
	   }
	   else {
		 uniTimeProcessor = new Uni_TimeEventProcessor(this);
		 Thread uniTimerThread = new Thread(uniTimeProcessor);
		 uniTimerThread.start();
		 return;     
	   }
   }
   Uni_Timer = new Timer();
   uniTimeProcessor = new Uni_TimeEventProcessor(this);
   Thread uniTimerThread = new Thread(uniTimeProcessor);
   uniTimerThread.start();
}


   /**
     * Stops all these threads on this Adapter.  Callbacks may still occur while this function
     * is being executed; it will only return when it can guarantee no more Callbacks will be called.
     */
public  void Stop() {
        synchronized (stateGuard)
        {
            while (true)
            {
                if (state == STARTED)
                {
                    state = STOPPING;
                    break;
                }
                else if (state == STOPPED)
                {   
                    throw new IllegalStateException("Tried to stop UDP_UnicastAdapter twice!");
                }
                
                try {
                    stateGuard.wait();
                }
                catch (InterruptedException ie)
                {
                }
            }
        }

		/*
		 * stops the threads in the UDP_MulticastAdapter and TCP_UnicastAdapter objects 
		 * contained in this adapter.
		 */
		if (multicast_adapter != null) multicast_adapter.Stop();    
		if (tcp_adapter != null) tcp_adapter.Stop();    

		/*
		 * stops the threads in this adapter.
		 */
        myStop();

        
        synchronized (stateGuard)
        {   
            state = STOPPED;
            stateGuard.notify();
        }
}

    /** This function actually stops the timer processor threads. 
     *  It is called by myStop().
     */
protected void myStop()
{

    if (uniTimeProcessor != null)
       uniTimeProcessor.die();
           
    synchronized (deathGuard)
    {
       if (!(uniTimeProcessor == null)) {
          while (!uniTimeProcessor.isDead())
          {
             try
             {
                 deathGuard.wait();
             }   
             catch (InterruptedException ie)
             {
             }
          }   
       }   
    }   
}

    /** This function stops the TimerProc thread.
      */
public  void stopTimer() {
   if (uniTimeProcessor != null) uniTimeProcessor.die();
}

    		
/** 
 * Creates physical address from a byte array of address - the inverse of 
 * <code>toByteArray</ode>. The first 4 bytes reprsent the IP address, 
 * and the last 8 bytes represent the Node Adapter port number and Socket
 * Adapter port number.
 *
 * @param	byteAddressAndPort  an byte array containing a physical address
 * @param	offset  the position in byteAddressAndPort from where availabe
 * 			data is read.
 * @return	a new instance of <code>INETv4TwoPorts</code>
 * @see		INETv4TwoPorts#toByteArray
 */
 public I_PhysicalAddress createPhysicalAddress(byte[] byteAddressAndPort, int offset) {
	 byte[] withoutShift = new byte[INETv4TwoPorts.ADDRESS_SIZE];
	 System.arraycopy(byteAddressAndPort, offset, withoutShift, 0, INETv4TwoPorts.ADDRESS_SIZE);
	 return new INETv4TwoPorts(withoutShift);
 }

	
 /** 
  * Creates physical address object from a string.
  * @param	addrString  an string representing a physical address.
  * @return	a new instance of <code>INETv4TwoPorts</code>
  */
 public I_PhysicalAddress createPhysicalAddress(String addrString) {
	 INETv4TwoPorts newINETv4 = new INETv4TwoPorts(addrString);
	 if (newINETv4.getInetAddress() == null) return null;
	 else return newINETv4;
 }



 /** 
  * Creates a new physical address.
  * @return	a new instance of <code>INETv4TwoPorts</code>
  */
 public I_PhysicalAddress createPhysicalAddress() {
	 /*
	  * Get the port number used in Socket Adapter.
	  */
	 int portScktAdpt = Integer.parseInt((String) socketAdpter.getInfo());

	 /*
	  * Get the port number used in NodeAd Adapter.
	  */
	 int portNodeAdpt = UDPSock.getLocalPort();

	 InetAddress localAddr = UDPSock.getLocalAddress();

	 /*
	  * Create a INETv4TwoPorts object with two ports. Ths IP address is determined by
	  * the system.
	  */
	 //INETv4TwoPorts newINETv4 = new INETv4TwoPorts(portNodeAdpt, portScktAdpt);
	 INETv4TwoPorts newINETv4 = new INETv4TwoPorts(localAddr, portNodeAdpt, portScktAdpt);
	 if (newINETv4.getInetAddress() == null)
	 {
		 throw new HyperCastFatalRuntimeException ("Could not create physical address.");
	 }
	 return newINETv4;
 }


 /** Returns the size of physical address, when carried in a packet,
  *  of an I_UnicastAdapter object. 
  */
 public int getPhysicalAddressSize()
 {
	 return INETv4TwoPorts.ADDRESS_SIZE;
 }



 /** 
  * Creates underlay address from a byte array of address - the inverse of 
  * <code>toByteArray</ode>. The first 4 bytes reprsent the IP address, 
  * and the last 4 bytes represent the port number.
  *
  * @param	byteAddressAndPort  an byte array containing a physical address
  * @param	offset  the position in byteAddressAndPort from where availabe
  * 			data is read.
  * @return	a new instance of <code>INETv4AndOnePort</code>
  * @see		INETv4OnePort#toByteArray()
  */
  public I_UnderlayAddress createUnderlayAddress(byte[] byteAddressAndPort, int offset) {
	  byte[] withoutShift = new byte[INETv4OnePort.ADDRESS_SIZE];
	  System.arraycopy(byteAddressAndPort, offset, withoutShift, 0, INETv4OnePort.ADDRESS_SIZE);
	  return new INETv4OnePort(withoutShift);
  }

	
  /** 
   * Creates underlay address object from a string.
   * @param	addrString  an string representing a physical address.
   * @return	a new instance of <code>INETv4AndOnePort</code>
   */
  public I_UnderlayAddress createUnderlayAddress(String addrString) {
	  INETv4OnePort newINETv4 = new INETv4OnePort(addrString);
	  if (newINETv4.getInetAddress() == null) return null;
	  else return newINETv4;
  }


  /** 
   * Creates a new underlay address.
   * @return	a new instance of <code>INETv4AndOnePort</code>
   */
  public I_UnderlayAddress createUnderlayAddress() {
	  /*
	   * Get the port number used in NodeAd Adapter.
	   */
	  int port = UDPSock.getLocalPort();

	  InetAddress localAddr = UDPSock.getLocalAddress();

	  /*
	   * Create a INETv4OnePort object with one port. Ths IP address is determined by
	   * the system.
	   */
	  //INETv4OnePort newINETv4 = new INETv4OnePort(port);
	  INETv4OnePort newINETv4 = new INETv4OnePort(localAddr, port);
	  if (newINETv4.getInetAddress() == null)
	  {
		  throw new HyperCastFatalRuntimeException ("Could not create physical address.");
	  }
	  return newINETv4;
  }


 /** Returns the size of underlay address, when carried in a packet,
  *  recognized by this adapter. 
  */
 public int getUnderlayAddressSize()
 {
	 return INETv4OnePort.ADDRESS_SIZE;
 }


 /**
  * Initialize the data structure for handling the statistics defined in this object.
  * In this method, a StatsProcessor instance is created which contains all statistics
  * , with each represented by a StatsElement instance, supported in this object.
  */
 private void InitStatisticsStructure()
 {
	/*
	 * Get the statistics name defined in the configuration file for this I_Stats instance.
	 */
	statisticsName = config.getTextAttribute (XmlUtil.createXPath(PROPERTY_PREFIX+"StatName"));

	/*
	 * Initial StatsProcessor instance.
	 */
	statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, false/* writable flag */);
		
	statsPro.addStatsElement("udpMulticastAdapter"/* name */, multicast_adapter/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);

	statsPro.addStatsElement("tcpUulticastAdapter", tcp_adapter, 1, 1);
 }

 /** @see I_Stats#getStatsName() */
 public String getStatsName ()
 {
	 return statisticsName;
 }

 /** @see I_Stats#setStatsName(String) */
 public void setStatsName (String name)
 {
	 statisticsName = name;
 }


 /** 
  * Return the result of query for the statistics specified by the given xpath.
  * 
  * @param doc   Document used for create new elements or nodes.
  * @param xpath XPath instance represents the statistics to be queried.
  * @see I_Stats#getStats
  */
 public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException
 {
	 //System.out.println("TCP_UDP_Adapter:getStats: xpath string is "+xpath.getPatternString());
	 //System.out.println("TCP_UDP_Adapter:getStats: call statsPro.getStatsResult(doc, xpath)");
	 return statsPro.getStatsResult(doc, xpath);
 }

 /** 
  * Set the statistics specified by the given xpath. The value actually set is
  * returned.
  * 
  * @param doc   	Document used for create new elements or nodes.
  * @param xpath 	XPath instance represents the statistics to be queried.
  * @param newValue	Element representing the value or sub-tree to be set.
  * @see I_Stats#setStats
  */
 public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
 {
	 //System.out.println("TCP_UDP_Adapter:setStats: xpath string is "+xpath.getPatternString());
	 //System.out.println("TCP_UDP_Adapter:setStats: call statsPro.setStatsResult(doc, xpath, newValue)");
	 return statsPro.setStatsResult(doc, xpath, newValue);
 }

 /** 
  * Return the schema element which represents the root of the sub-tree, specified
  * by the given xpath, in read schema tree. 
  * 
  * @param doc   Document used for create new elements or nodes.
  * @param xpath XPath instance representing the statistics which is the root
  * 	of the sub-tree to be queried.
  * @see I_Stats#getReadSchema(Document, XPath)
  */
 public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
 {
	 //System.out.println("TCP_UDP_Adapter:getReadSchema: xpath string is "+xpath.getPatternString());
	 //System.out.println("TCP_UDP_Adapter:getSchema: call statsPro.getReadSchemaResult(doc, xpath)");
	 return statsPro.getReadSchemaResult(doc, xpath);
 }

 /** 
  * Return the schema element which represents the root of the sub-tree, specified
  * by the given xpath, in write schema tree. 
  * 
  * @param doc   Document used for create new elements or nodes.
  * @param xpath XPath instance representing the statistics which is the root
  * 	of the sub-tree to be queried.
  * @see I_Stats#getWriteSchema(Document, XPath)
  */
 public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
 {
	 return statsPro.getWriteSchemaResult(doc, xpath);
 }

    /**
     * Sends a message to a unicast address. There is no protpcol specified
     * in the parameters, by default the message is sent out using 
     * UDP unicast protocol.
     */
//public void sendUnicastMessage(I_PhysicalAddress dstaddr, I_Message msg ) {
public void sendUnicastMessage(I_NetworkAddress dstaddr, I_Message msg ) {

	multicast_adapter.sendUnicastMessage(dstaddr, msg);

	//System.out.println("TCP_UDP_Adapter: send UDP Unicast message.");
}


    /**
     * Sends a message to a multicast address. It is sent out through
     * the UDP_MulticastAdapter object contained in this adapter. 
     */
public void sendMulticastMessage(I_Message msg) {
	multicast_adapter.sendMulticastMessage(msg );
}

    /**
     * Get the multicast address used by this adpater.
     */
public I_PhysicalAddress getMulticastAddress() {
	return multicast_adapter.getMulticastAddress();
}

}//end TCP_UDP_Adapter


/**
 * This class starts a thread to process time event. <p>
 * <br>         
 * The processor takes the time event with the smallest timestamp from
 * the time_event queue. <br>
 * If this timestamp > currenttime,
 * the processor sleep for (timestamp - currenttime), then processes the time event. <br>
 * If this timestamp <= currenttime,
 * the processor processes the time event at once.
 */             
                    
class Uni_TimeEventProcessor implements Runnable {
    /** The adapter. */ 
        TCP_UDP_Adapter Parent;
                            
    /** The receiving thread. */
        private Thread myThread;

    /** True if the function die() has been called. */
        private boolean dying;

    /** True if <code>myThread</code> has returned from run(). */
    	private boolean died;
                        
    /** True while <code>myThread</code> is waiting to receive. */
    	boolean waiting;    

    /** Guard used to protect <code>waiting</code> when it is changed. */
        private Object waitingGuard;
                    
                    
        /**             
         * Contructs the timer processor object with a TCP_UDP_Adapter object.
         */                 
     Uni_TimeEventProcessor(TCP_UDP_Adapter adapter) {
            Parent = adapter;
                   
            dying = false;
            died = false;
                
            waiting = false;
        	waitingGuard = new Object();
     }               
                       

     /** 
      * Starts running the thread to take a time event from the TimeEventQueue and
      * send it to the protocol node (I_AdapterCallback) for processing.
      */
     public void run()
     {


        	Parent.uniTimerProcDebug = "JustStarted";

                //System.out.println("TCP_UDP_TimeEventProcessor started.....");
                myThread = Thread.currentThread();

                Object id;

            while (true) {
              try
              {
                try
                {
                    try
                    {
                        synchronized (waitingGuard)
                        {
                            waiting = true;
                            Parent.uniTimerProcDebug = "Waiting";
                        }

                        if (dying)
                            break;

                        //may block here
                        id = Parent.Uni_Timer.sleepUntilNextEvent();

                    }
                    finally
                    {
                        synchronized (waitingGuard)
                        {
                            waiting = false;
                            Parent.uniTimerProcDebug = "DoneWaiting";
                        }
                    }
                }
                finally
                {
                    if (Thread.interrupted())
                    {
                        Parent.uniTimerProcDebug = "ClearingInterruptBit";
                        Thread.sleep(1000);
                    }
                }

                if (dying)
                {
                    Parent.uniTimerProcDebug = "ResettingTimer";
                    // stick timer back in queue
                    Parent.Uni_Timer.setTimer(id, 0);
                    break;
                }

                try
                {
                    Parent.uniTimerProcDebug = "InCallback";
                    Parent.callback.timerExpired(id);
                }
                catch (Exception e)
                {
                    Parent.config.err.println("Encountered exception \""+e.getMessage() + "\" in timerExpired()");
                    e.printStackTrace(Parent.config.err);
                }
            }
            catch (InterruptedException e)
            {
            }
            
        	Parent.uniTimerProcDebug = "EndOfLoop";
        }

        synchronized (Parent.deathGuard)
        {
            died = true;
            Parent.uniTimerProcDebug = "Dead";
            Parent.deathGuard.notify();
        }
     }


     /** 
      * This function finally stops the timer processor (myThread) thread.
      */
     public void die() {
         synchronized (waitingGuard) {
             dying = true;
             if (waiting)
                 myThread.interrupt();  // incase it is blocking on the wait call.
         }
     }

    /** Returns true if <code>myThread</code> will never call an I_AdapterCallback method again. */
	public boolean isDead()
    {
        return (died || waiting);
    }

}
