/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.adapters;

import hypercast.*;
import hypercast.util.XmlUtil;

import java.io.*;
import java.util.*;
import java.net.*;


import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.xpath.*;

/** 
 * This class implements the I_UnicastAdapter interface while using
 * TCP links.  TCP links are created when necessary to send data.
 * Links are shutdown when not used for TIMEOUT_TIME millseconds.
 * Each link is unidirectional, so upto 2 TCP sockets can exist for
 * each neighbor.  To prevent deadlock, when the thread assigned to
 * receive from a link has not read from it in DROP_TIME seconds, a
 * message is dropped from the link.  
 *<p>
 *
 * In implementation, there are 3 threads plus a thread for each
 * incoming link plus a thread for each dropped message.  The 3
 * threads are responsible for (1) accept()ing connections on the
 * server socket, (2) performing the timer functions and (3) detecting
 * deadlocks and timeouts.  When a connection is accepted, a new
 * receiving thread is created for the socket.  When the deadlock
 * thread computes that data isn't being read from an incoming socket,
 * it creates a thread to drop a single message from the socket.  (The
 * thread dies once the one message has been dropped.)
 * <p>
 *
 * For each incoming socket and outgoing socket, an object is created.
 * For the incoming socket the IncomingSocketObject object is used to
 * synchronize multiple threads trying to read and manages the buffers
 * for reading messages from the socket.  For outgoing sockets, the
 * OutgoingSocketObject is used to synchronize multiple threads trying
 * to write. This class supports INETv4TwoPorts type physical address.
 *
 * @author HyperCast Team
 * @author Dongwen Wang
 * @author Michael Nahas
 * @version 2.0, February 13, 2001.
 * @author Jianping Wang
 * @version 3.0, Jul. 24, 2004 
 */

public class TCP_UnicastAdapter implements I_IPv4_UnicastAdapter
{	

	    /** Prefix for all property names read by this socket.*/
	protected static String PROPERTY_NAME_PREFIX;
	
	    /** Size of receive buffer for each incoming TCP connection. */
	int RECEIVE_BUFFER_SIZE;
	
	    /** Name suffix of property containing size of receive buffer for each incoming TCP connection. */
	private static final String RECEIVE_BUFFER_SIZE_PROPERTY_SUFFIX = "MaximumPacketLength";
	
	    /** In milliseconds, the time before an idle TCP connection is closed. */
	private int TIMEOUT_TIME;
	
	    /** Name suffix of property containing the time before an idle TCP connection is closed. */
	private static final String TIMEOUT_TIME_PROPERTY_SUFFIX = "Timeout";
	
	    /** In milliseconds, the time a receiving thread can ignore a socket
	      before a message is dropped. */
	private int DROP_TIME;
	
	    /** Name suffix of property containing the time a receiving thread
	      can ignore a socket before a message is dropped. */
	private static final String DROP_TIME_PROPERTY_SUFFIX = "MaxIdleTime";

	/** Maximum queue length of incomming connection */
	int MAX_BACKLOG;

	/** Name suffix of property containing the maximum queue length of incoming connection. */
	private static final String MAX_BACKLOG_PROPERTY_SUFFIX = "MaxBacklog";
	
	
	    /** property object for configration */	
	HyperCastConfig config;			
	
	    /** The Callback object */
	I_AdapterCallback callback;	
	
	
	    /** The TCP ServerSocket in the OL_SocketAdapter */
	ServerSocket serverSocket = null;	
	
		/** The string which indicate this adapter is Node Adapter or
		 * Socket Adapter: 
		 * "nodeadpt" - Node Adapter, "sockadpt" - Socket Adapter.
		 */
	String AdptIndicator;   
	
		/** Determine if Node adapter should try random port number 
		 * if physical addresss assignment fails.
		 */
	String AllowRandomPhyAddrPort;   
	
		/** The pointer to socket adapter */
	I_UnicastAdapter socketAdpter;
	
	    /** Runnable object that hosts the thread for checking ServerSocket */
	private TcpServer tcpServer;
	
	    /** The timer that works with this adapter */
	    Timer IP_Timer = null;
	
	    /** Runnable object that hosts the thread that checks the timers */
	private Tcp_TimeEventProcessor tcpTimeProcessor;
	
	    /** Runnable object that hosts the thread that checks for deadlock & timeouts. */
	private DeadlockMonitor DeadlockMonitor;

    /** Hashtable of Node's child's IP_Address -> OutgoingSocketObject */
    Hashtable outgoings;
    /** Vector of IncomingSocketObjects */
    Vector incomings;
    /** Vector of TcpReceiver */
    Vector receivers;

    /** Total number of packets received by this TcpAdapter */
	long UReceivedPackets;
    /** Total number of packets sent from this TcpAdapter */
	long USentPackets;
    /** Total number of bytes received by this TcpAdapter */	
	long UReceivedBytes;
    /** Total number of bytes sent from this TcpAdapter */
	long USentBytes;

    /** Used for wait/notify while adapter is shutting down. */
    protected Object deathGuard;

    /** Current state of the node.  It goes from STOPPED to STARTING
         * to STARTED to STOPPING and then back to STOPPED. */
    protected int state;

    /** value for <code>state</code>. */
    protected final static int STOPPED = 0;
    /** value for <code>state</code>. */
    protected final static int STARTING = 1;
    /** value for <code>state</code>. */
    protected final static int STARTED = 2;
    /** value for <code>state</code>. */
    protected final static int STOPPING = 3;

    /** Used to protect <code>state</code> when its value is changed. */
    protected Object stateGuard;

	/**
	 * A StatsProcessor instance which handles the statistics in this object.
	 */
	private StatsProcessor statsPro;

	/**
	 * The statistics name for this object (as an I_Stats instance). Its default
	 * name is defined in the socket configuration schema file. It can be assigned
	 * by any other object through method setStattisticsName defined in the 
	 * I_Stats interface.
	 */
	private String statisticsName;


    /** 
	 * Constructs a new <code>TCP_UnicastAdapter</code> object from an <code>HyperCastConfig</code> 	 
     * object. 
     *  
	 * @param	c		a HyperCastConfig object.
	 * @param	scktAdpt	object representing the Socket Adapter. It is set to null when 
	 * 						creating the Socket Adapter.
	 * @param	AddrString	string containing physical address information.
	 * @param	prefix		property prefix.
     */ 
public TCP_UnicastAdapter(HyperCastConfig c,  I_UnicastAdapter scktAdpt, String AddrString, String prefix) 
    {				
		
		
		/* Execute the initiallization, except initializing the ServerSocket,
		 * for this adapter.
		 */
		Adapter_Inilization(c, prefix);
		
		socketAdpter = scktAdpt;

		/* 
		 * Initialize the ServerSocket used by this adapter. If the initialization of 
		 * ServerSocket fails, a runtime exception is thrown. Only Node Adapter starts
		 * socket initialization. It not only initializes the socket in Node Adapter, 
		 * but also initialize the socket in Socket Adapter.
		 */
		if ((!AdptIndicator.equals("scktadpt")) && !InitializeServerSocket(scktAdpt, config, AddrString))
		{
			throw new HyperCastFatalRuntimeException ("Could not initialize local unicast socket!");
		}
			
	}

	/** 
	 * Constructs a TCP_UnicastAdapter object with a HyperCastConfig 
	 * and ServerSocket object.
	 * @param	config		a HyperCastConfig object.
	 * @param	ssock		a ServerSocket object.
	 * @param	prefix		property prefix.
	 */
	public TCP_UnicastAdapter(HyperCastConfig config, ServerSocket ssock, String prefix){
		
		// Execute the initiallization, except initializing the ServerSocket,
	        // for this adapter.
		 
		Adapter_Inilization(config, prefix);
		
		serverSocket = ssock;
	}

	/** 
	 * Finish the initialization for this TCP_UnicastAdapter object.
	 * 
	 * @param	c		a HyperCastConfig object.
	 * @param	prefix		property prefix.
	 */
	public void Adapter_Inilization(HyperCastConfig c, String prefix) 
	{				
		/*
		 * Assign the adapter indicator.
		 */
		if (prefix.startsWith("NodeAdpt"))
		{
			AdptIndicator = "nodeadpt";
			PROPERTY_NAME_PREFIX = "/Public/NodeAdapter/"+prefix+"/";
		} 
		else
		{
			AdptIndicator = "scktadpt";
			PROPERTY_NAME_PREFIX = "/Public/SocketAdapter/"+prefix+"/";
		} 

		config = c;		
		callback = null;

		tcpServer = null;
        
		tcpTimeProcessor = null;
        
		DeadlockMonitor = null;
        
		outgoings = new Hashtable();
		incomings = new Vector();
		receivers = new Vector();

		UReceivedPackets = 0;
		UReceivedBytes = 0;
		USentPackets = 0;
		USentBytes = 0;
    
		if (prefix.startsWith("NodeAdpt"))
		{
			AllowRandomPhyAddrPort = config.getTextAttribute
					(XmlUtil.createXPath(PROPERTY_NAME_PREFIX+"AllowRandomPhyAddrPort"));
		} 

		RECEIVE_BUFFER_SIZE = config.getPositiveIntAttribute
				(XmlUtil.createXPath(PROPERTY_NAME_PREFIX+RECEIVE_BUFFER_SIZE_PROPERTY_SUFFIX));

		TIMEOUT_TIME = config.getPositiveIntAttribute
				(XmlUtil.createXPath(PROPERTY_NAME_PREFIX+TIMEOUT_TIME_PROPERTY_SUFFIX));

		DROP_TIME = config.getPositiveIntAttribute
				(XmlUtil.createXPath(PROPERTY_NAME_PREFIX+DROP_TIME_PROPERTY_SUFFIX));

		MAX_BACKLOG = config.getPositiveIntAttribute
				(XmlUtil.createXPath(PROPERTY_NAME_PREFIX+MAX_BACKLOG_PROPERTY_SUFFIX));

		deathGuard = new Object();

		state = STOPPED;
		stateGuard = new Object();

		InitStatisticsStructure();
	}


	/**
	 * Initialize sockets running on Node Adapter and Socket Adapter.
	 * 
	 * @param	ScktAdpter	 A I_UnicastAdapter object representing Socket Adapter.
	 * @param	config	 an HyperCastConfig object.
	 * @param	AddrString	 A string carring physical address information.
	 */
	private boolean InitializeServerSocket(I_UnicastAdapter ScktAdpter, HyperCastConfig config, String AddrString)
	{
		/*
		 * Set initial values for two ports and InetAddress.
		 */
		int portNodeAdpt = 0;
		int portScktAdpt = 0;
		InetAddress ipAddr = null;

		/*
		 * If this is Socket Adapter, return. The initialization of sockets in Socket Adapter
		 * is trigged from Node Adapter.
		 */
		if (AdptIndicator.equals("scktadpt"))
		{
			return true;
		}
		
		/*
		 * If this is Node Adapter and physical address is specified (in string format) with papameter,
		 * the specified physical address string is used to create the physical address of this adapter.
		 */
		String addrStr;
		if (AddrString != null) {
			addrStr = AddrString;
		}
		/*
		 * If this is Node Adapter and no physical address is specified (in string format) with papameter,
		 * the adapter reads configuration file to get the information of physical address.
		 */
		else {
			String addrType = config.getTextAttribute
					(XmlUtil.createXPath(PROPERTY_NAME_PREFIX + "PhyAddr"));
			addrStr = config.getTextAttribute
					(XmlUtil.createXPath(PROPERTY_NAME_PREFIX + "PhyAddr/" + addrType));
		}
		
		String[] paStr = INETv4TwoPorts.getPartsFromAddrAndPortsString(addrStr);

		/*
		 * Create InetAddress.
		 */			
		if (!paStr[0].equals("")) {
			try
			{
			   ipAddr = InetAddress.getByName(paStr[0]);
			}
			catch (UnknownHostException e)
			{
			   //throw new HyperCastFatalRuntimeException (""+addrStr + " is a bad address string.**" + e);
			   config.err.println("Cannot create InetAddress from " + ipAddr + " : " + e);
			   ipAddr = null;
			}
		}

		/*
		 * Obtain the first port number.
		 */
		if (!paStr[1].equals("")) {
			try 
			{
			  portNodeAdpt = Integer.parseInt(paStr[1]);
			  if ((portNodeAdpt < 0) && (portNodeAdpt != -1))
			  {
				throw new IllegalArgumentException("" + paStr[1] + " is not a available value for port number.");
			  }
			}
			catch (NumberFormatException e) 
			{
			  throw new IllegalArgumentException("Unable to parse  \"" + paStr[1] + "\" as a port number.");
			}
		}
		
		/*
		 * Obtain the second port number.
		 */
		if (!paStr[2].equals("")) {
			try 
			{
				portScktAdpt = Integer.parseInt(paStr[2]);
				if ((portScktAdpt < 0) && (portScktAdpt != -1))
				{
				  throw new IllegalArgumentException("" + paStr[2] + " is not a available value for port number.");
				}
			}
			catch (NumberFormatException e) 
			{
			  throw new IllegalArgumentException("Unable to parse  \"" + paStr[2] + "\" as a port number.");
			}
		}


		/*
		 * Initialize the DatagramSocket with obtained IP address and port information.
		 */
		if (!ProcessingSocketInitialization(ScktAdpter, portNodeAdpt, portScktAdpt, ipAddr)) {
			/*
			 * When the initialization of the DatagramSocket fails, if AllowRandomPhyAddrPort
			 * is specified to "Yes", the adapter will try to initialize the DatagramSocket
			 * without specifying port number, i.e., asking system to pick port number.
			 */
			if ((AllowRandomPhyAddrPort.equalsIgnoreCase("Yes")) && ((portNodeAdpt > 0) || (portScktAdpt > 0))) {
				if (!ProcessingSocketInitialization(ScktAdpter, 0, 0, ipAddr)) {
					return false;
				}
				else {
					return true;
				}
			}
			else {
				return false;
			}
		}
		else {
			return true;
		}

	}


	/**
	 * Process the initialization of underlay network sockets.
	 * 
	 * @param	ScktAdpter	 	A I_UnicastAdapter object representing Socket Adapter.
	 * @param	portNodeAdpt	port number of the socket in Node Adapter.
	 * @param	portScktAdpt	port number assigned to the socket in Socket Adapter.
	 * @param	ipAddr			IP address to which the socket is bounded.
	 */
	private boolean ProcessingSocketInitialization(I_UnicastAdapter ScktAdpter, int portNodeAdpt, int portScktAdpt, InetAddress ipAddr)
	{
		/*
		 * If the port number of Node Adapter is specified as a positive number,
		 * first initialize the socket in Node Adapter and then the socket in Socket
		 * Adapter. 
		 */
		if (portNodeAdpt > 0)
		{
			/*
			 * Initialize the socket in Node Adapter. If it fails, 
			 * stop the initialization process and return.
			 */
			if (!createServerSocket(portNodeAdpt, ipAddr))
			{
				return false;
			}
			
			/*
			 * If Socket Adapter is not specified, return.
			 */
			if (ScktAdpter == null)
			{
			   return true;
			}

			/*
			 * Initialize the socket in Socket Adapter by calling Reset method. If it fails, 
			 * stop the initialization process and return.
			 */
			String addressStr;
			if (ipAddr != null) 
			{
				addressStr = new String(getIPString(ipAddr)+":"+portScktAdpt);
			}
			else
			{
				addressStr = new String(":"+portScktAdpt);
			}
			if (!ScktAdpter.Reset(addressStr))
			{
                            // TODO: release socket that was created for node?
				return false;
			}
		}
		/*
		 * If the port number of Node Adapter is not specified (-1 or 0), first the 
		 * socket in Socket Adapter and then the socket in Node Adapter.
		 */
		else
		{
			/*
			 * If Socket Adapter is not specified, skip the process of initializing the 
			 * socket for Socket Adapter. Otherwise initialize the socket in Socket Adapter.
			 */
			if (ScktAdpter != null)
			{
				/*
				 * Initialize the socket in Socket Adapter by calling Reset method. If it fails, 
				 * stop the initialization process and return.
				 */
				String addressStr;
				if (ipAddr != null) 
				{
					addressStr = new String(getIPString(ipAddr)+":"+portScktAdpt);
				}
				else
				{
					addressStr = new String(":"+portScktAdpt);
				}
				if (!ScktAdpter.Reset(addressStr))
				{
					return false;
				}
			}

			/*
			 * Initialize the socket in Node Adapter. If it fails, 
			 * stop the initialization process and return.
			 */
			if (!createServerSocket(portNodeAdpt, ipAddr))
			{
				return false;
			}
			
		}

		/*
		 * The initialization is successful. Return.
		 */
		return true;
	}


	private static Random rand = new Random();

	/**
	 * Creates a ServerSocket.
	 * 
	 * @param	portNum	 the port number for the TCP unicast socket. 
	 * 					 If portNum is set to 0, the port will be the minimal
	 * 					 available port shosen by system; If portNum is -1,
	 * 					 the port number is chosen randomly.
	 * @param	ipAddr 	 IP address to which the created socket is binded.
	 * 					 If it is null, IP address will be chosen by system.
	 */
	private boolean createServerSocket(int portNum, InetAddress ipAddr)
	{

		/*
		 * If IP address is not specified, the local ServerSocket is bound to 
		 * the default IP address chosen by system.
		 */
		if (ipAddr == null) {
			for (int retries = 0; retries < MAX_RETRIES_TO_OPEN_SOCKET; retries++)
			{
				try 
				{
						/*
						 * If parameter is 0 or positive, the parameter is used to create
						 * the port number.
						 */
					if (portNum != -1)
					{
						serverSocket = new ServerSocket(portNum);
					}
						/*
						 * If port number parameter is set to -1, a random number is used as the port number.
						 */
					else
					{
						serverSocket = new ServerSocket(Math.abs(rand.nextInt()) % (256*256));
					}
					 return true;
			   }
			   catch (IOException  se) 
			   {
					/*
					 * If portNum is -1, which means the port number is randomly chosen,
					 * it may conflict with the port number used in another adapter (Node
					 * Adapter or Socket Adapter). In this case, retry to get another port 
					 * number. Otherwise, the creation of DatagramSocket fails. Return.
					 */
					if (portNum == -1)
					{
					   continue;
					}
					else
					{
					   return false;
					}
			   }
			}
		}
		/*
		 * If IP address is specified, the local DatagramSocket is bound to this address.
		 */
		else {
			for (int retries = 0; retries < MAX_RETRIES_TO_OPEN_SOCKET; retries++)
			{
			   try 
			   {
						/*
						 * If parameter is 0 or positive, the parameter is used to create
						 * the port number.
						 */
					 if (portNum != -1)
					 {
						serverSocket = new ServerSocket(portNum, MAX_BACKLOG, ipAddr);
					 }
						/*
						 * If port number parameter is set to -1, a random number is used as the port number.
						 */
					 else {
						serverSocket = new ServerSocket(Math.abs(rand.nextInt()) % (256*256), MAX_BACKLOG, ipAddr);
					 }
					 return true;
			   }
			   catch (IOException  se) 
			   {
					/*
					 * If portNum is -1, which means the port number is randomly chosen,
					 * it may conflict with the port number used in another adapter (Node
					 * Adapter or Socket Adapter). In this case, retry to get another port 
					 * number. Otherwise, the creation of DatagramSocket fails. Return.
					 */
					if (portNum == -1)
					{
					   continue;
					}
					else
					{
					   return false;
					}
			   }
			}
		}

		/*
		 * If the the process reaches here, that means the retry times reachs 
		 * the maximum retry times, stop the initializing local DatagramSocket.
		 */
		return false;
	}


	/**
	 * Get string representation of IP address in a InetAddress object.
	 * @param	addr	An InetAddress object. 
	 */
	public String getIPString(InetAddress addr) {
		if (addr == null) return new String("");
		
		String addrstr = addr.toString();
		int slash_loc = addrstr.indexOf('/');
		return addrstr.substring(slash_loc+1, addrstr.length());
	}


	/** Get information of this adapter.
	 */ 
	public Object getInfo() {
		/*
		 * It returns a String object that carries the port number 
		 * used by the socket in this adapter.
		 */
	   int port = serverSocket.getLocalPort();
	   return new String(""+port);
	}


	/**
	 * Resets the local unicast socket with a ServerSocket object.
	 * @param	SocketObj		object containing information used to finish reset. 
	 */
	public boolean Reset(Object SocketObj) {
		/*
		 * If parameter infoObj is an instance of DatagramSocket, this 
		 * DatagramSocket object is used to replace the DatagramSocket in this
		 * adapter if they are different. This case happens when TCP_UDP_Adapter 
		 * is specified.
		 */
		if (SocketObj instanceof ServerSocket) {
			String localIPstr = getIPString(serverSocket.getInetAddress());
			String parIPstr = getIPString(((ServerSocket) SocketObj).getInetAddress());
			
			if ((localIPstr.equals(parIPstr)) && (serverSocket.getLocalPort() == ((ServerSocket) SocketObj).getLocalPort()))
			{
				return true;
			}
			
			try {
			   if (serverSocket != null) serverSocket.close();
			}
			catch (IOException  se) 
			{
				return false;
			}
			serverSocket = (ServerSocket) SocketObj;

		} 
		/*
		 * If parameter infoObj is a string, this string will bring IP address and port number
		 * information that are used to reset the DatagramSocket in this adapter. The passed
		 * string has format "IPaddrString:port".
		 */
		else if (SocketObj instanceof String) {
			String addrstring = (String) SocketObj;
			
			/*
			 * If addString is "", close the DatagramSocket and return.
			 */
			if (addrstring.equals(""))
			{
				if (serverSocket != null)
				{
					try {
					   serverSocket.close();
					}
					catch (IOException  se) 
					{
						return false;
					}
				}
				return true;
			}
			
			int cl_loc = addrstring.indexOf(':');
			String ipString = addrstring.substring(0, cl_loc);
			int portnumber = Integer.parseInt(addrstring.substring(cl_loc+1, addrstring.length()));
			
			/*
			 * If the local DatagramSocket satisfies the requirement of reset parameters, do nothing and return.
			 */			
			if ((serverSocket != null) && 
				((portnumber <= 0) ||
				((portnumber > 0) && (serverSocket.getLocalPort() == portnumber) && (ipString.equals(getIPString(serverSocket.getInetAddress()))))))
			{
				return true;
			}

			/*
			 * Otherwise the DatagramSocket is reset.
			 */
			if (serverSocket != null) {
				try {
				   serverSocket.close();
				}
				catch (IOException  se) 
				{
					return false;
				}
			}
	
			InetAddress ipaddress = null;
			/*
			 * If ipString is specified (not ""), it is used to create the local DatagramSocket,
			 * i.e. the local DatagramSocket is bound to it. Another case is that the ipString is 
			 * obtained from calling USocket.getLocalAddress(), where USocket is a DatagramSocket 
			 * instance. In this case, ipAddress is always "0.0.0.0".
			 */
			if ((!ipString.equals("")) && (!ipString.equals("0.0.0.0"))) 
			{
				try
				{
					ipaddress = InetAddress.getByName(ipString);
				}
				catch (UnknownHostException e)
				{
				   System.err.println(ipString + " is a bad address string.**" + e);
				   ipaddress = null;
				}
			}

			/*
			 * If ipString is "" or "0.0.0.0", the default IP address picked by system will be used 
			 * to create the local Datagram Socket, i.e. the local DatagramSocket is bound 
			 * to this default IP address.
			 */
			return createServerSocket(portnumber, ipaddress);
		}
		
		return true;
	}

	
	/** 
	 * Set <code>I_Callback</code> object of this adapter. <p>
	 * @param	cb	an object of type <code>I_AdapterCallback</code>
	 *
	 * @throws IllegalStateException.
	 */	
	public void setCallback(I_AdapterCallback cb) {
        if (state != STOPPED)
        {
			throw new IllegalStateException("Tried to change callback while TCP_UnicastAdapter was running!");
        }
        callback = cb;
	}			


	/** 
	 * Set a <code>Timer</code> event. <p>
	 * @param	timer_id	a time event ID.
	 * @param	time_ms		a time in ms.
	 *
	 * @throws IllegalStateException.
	 */		
	public void setTimer(Object timer_id, long time_ms) {
        if (state == STOPPED || state == STARTING)
            throw new IllegalStateException("TCP_UnicastAdapter: tried to set Timer when adapter was not running.");

        if (IP_Timer == null)
            throw new IllegalStateException("TCP_UnicastAdapter: tried to set Timer when Timer is null.");

		IP_Timer.setTimer(timer_id, time_ms);	
	}
	
	
	/** 
	 * Get a <code>Timer</code> event. <p>
	 * @param	timerid	a time event ID
	 * @return	time in ms
	 */			
	public long getTimer(Object timerid) {
        if (state == STOPPED || state == STARTING)
            throw new IllegalStateException("TCP_UnicastAdapter: tried to get Timer when adapter was not running.");

        if (IP_Timer == null)
            throw new IllegalStateException("TCP_UnicastAdapter: tried to get Timer when Timer is null.");

		return IP_Timer.getTimer(timerid);
	
	}
	
	
	/** 
	 * Clear a <code>Timer</code> event. <p>
	 * @param	timerid	a time event ID
	 */				
	public void clearTimer(Object timerid) {
        if (state == STOPPED || state == STARTING)
            throw new IllegalStateException("TCP_UnicastAdapter: tried to clear Timer when adapter was not running.");

        if (IP_Timer == null)
            throw new IllegalStateException("TCP_UnicastAdapter: tried to clear Timer when Timer is null.");

		IP_Timer.clearTimer(timerid);
		
	}


    /** 
	 * Get current system time. <p>
	 * This function is not supported by TcpAdapter	 
     */				
	public long getCurrentTime() {
        return System.currentTimeMillis();
    }
    		
  /** 
   * Creates physical address from a byte array of address - the inverse of 
   * <code>toByteArray</ode>. The first 4 bytes reprsent the IP address, 
   * and the last 8 bytes represent the Node Adapter port number and Socket
   * Adapter port number.
   *
   * @param	byteAddressAndPorts  an byte array containing a physical address
   * @param	offset  the position in byteAddressAndPort from where availabe
   * 			data is read.
   * @return	a new instance of <code>INETv4TwoPorts</code>
   * @see		INETv4TwoPorts#toByteArray
   */
   public I_PhysicalAddress createPhysicalAddress(byte[] byteAddressAndPorts, int offset) {
	   byte[] withoutShift = new byte[INETv4TwoPorts.ADDRESS_SIZE];
	   System.arraycopy(byteAddressAndPorts, offset, withoutShift, 0, INETv4TwoPorts.ADDRESS_SIZE);
	   return new INETv4TwoPorts(withoutShift);
   }

	
   /** 
    * Creates physical address object from a string.
    * @param	addrString  an string representing a physical address.
    * @return	a new instance of <code>INETv4TwoPorts</code>
    */
   public I_PhysicalAddress createPhysicalAddress(String addrString) {
       INETv4TwoPorts newINETv4 = new INETv4TwoPorts(addrString);
	   if (newINETv4.getInetAddress() == null) return null;
	   else return newINETv4;
   }


  /** 
   * Creates a new physical address.
   * @return	a new instance of <code>INETv4TwoPorts</code>
   */
  public I_PhysicalAddress createPhysicalAddress() {
	  /*
	   * Get the port number used in Socket Adapter.
	   */
	  int portScktAdpt = 0;
	  if (socketAdpter != null)
	  {
		 portScktAdpt = Integer.parseInt((String) socketAdpter.getInfo());
	  } 


	  /*
	   * Get the port number used in NodeAd Adapter.
	   */
	  int portNodeAdpt = serverSocket.getLocalPort();


	  InetAddress localAddr = serverSocket.getInetAddress();

	  /*
	   * Create a INETv4TwoPorts object with two ports. Ths IP address is determined by
	   * the system.
	   */
	  INETv4TwoPorts newINETv4 = new INETv4TwoPorts(localAddr, portNodeAdpt, portScktAdpt);
	  if (newINETv4.getInetAddress() == null)
	  {
		  throw new HyperCastFatalRuntimeException ("Could not create physical address.");
	  }
	  return newINETv4;
  }


  /** Returns the size of physical address, when carried in a packet,
   *  recognized by this adapter. 
   */
  public int getPhysicalAddressSize()
  {
	  return INETv4TwoPorts.ADDRESS_SIZE;
  }



  /** 
   * Creates underlay address from a byte array of address - the inverse of 
   * <code>toByteArray</ode>. The first 4 bytes reprsent the IP address, 
   * and the last 4 bytes represent the port number.
   *
   * @param	byteAddressAndPort  an byte array containing a physical address
   * @param	offset  the position in byteAddressAndPort from where availabe
   * 			data is read.
   * @return	a new instance of <code>INETv4AndOnePort</code>
   * @see		INETv4OnePort#toByteArray()
   */
   public I_UnderlayAddress createUnderlayAddress(byte[] byteAddressAndPort, int offset) {
	   byte[] withoutShift = new byte[INETv4OnePort.ADDRESS_SIZE];
	   System.arraycopy(byteAddressAndPort, offset, withoutShift, 0, INETv4OnePort.ADDRESS_SIZE);
	   return new INETv4OnePort(withoutShift);
   }

	
   /** 
	* Creates underlay address object from a string.
	* @param	addrString  an string representing a physical address.
    * @return	a new instance of <code>INETv4AndOnePort</code>
	*/
   public I_UnderlayAddress createUnderlayAddress(String addrString) {
	   INETv4OnePort newINETv4 = new INETv4OnePort(addrString);
	   if (newINETv4.getInetAddress() == null) return null;
	   else return newINETv4;
   }


   /** 
	* Creates a new underlay address.
    * @return	a new instance of <code>INETv4AndOnePort</code>
	*/
   public I_UnderlayAddress createUnderlayAddress() {
	   /*
		* Get the port number used in NodeAd Adapter.
		*/
	   int port = serverSocket.getLocalPort();

	   InetAddress localAddr = serverSocket.getInetAddress();

	   /*
		* Create a INETv4OnePort object with one port. Ths IP address is determined by
		* the system.
		*/
	   INETv4OnePort newINETv4 = new INETv4OnePort(localAddr, port);
	   if (newINETv4.getInetAddress() == null)
	   {
		   throw new HyperCastFatalRuntimeException ("Could not create physical address.");
	   }
	   return newINETv4;
   }


  /** Returns the size of underlay address, when carried in a packet,
   *  of an I_UnicastAdapter object. 
   */
  public int getUnderlayAddressSize()
  {
	  return INETv4OnePort.ADDRESS_SIZE;
  }




    /** 
     * Sends a message to a unicast address.
     * A TCP connection may have to be created before the packet is sent.
	 */
	public void sendUnicastMessage(I_NetworkAddress dstaddr, I_Message msg )
    {
		try{
	        OL_Message olmsg = (OL_Message) msg;
	        if (olmsg.getHopLimit() == 0)
	        {
	            System.err.println ("Attempt to send message with 0 hop limit.");
	            Thread.currentThread().dumpStack();
	            throw new IllegalArgumentException ("Attempt to send message with 0 hop limit.");
	        }
		}
		catch(ClassCastException cce){
			//Do nothing if msg is not OL_Message
		}
		catch(Exception e){
			e.printStackTrace(config.err);
		}

        if (state == STOPPED || state == STARTING)
            throw new IllegalStateException("TCP_UnicastAdapter: tried to send message when adapter was not running.");

		//System.err.println("TCP_UnicastAdapter: sendUnicastMessage: dstaddr is " + dstaddr.toString());

        int attempt;
        for (attempt = 0; attempt < 5; attempt++) 
        {
            OutgoingSocketObject oso = null;
            synchronized (outgoings) 
            {
                oso = (OutgoingSocketObject) outgoings.get(dstaddr);
                if (oso == null) 
                {
                    // create new connection and insert it into hashtable
                    try 
                    {
						oso = new OutgoingSocketObject(dstaddr, this);
			 //System.out.println("TCP_UnicastAdapter: send TCP unicast message");
                    }
                    catch (IOException ioe)
                    {
                        config.err.println("Unable to open connection to neighbor.  Dropping msg.  Exception:" + ioe.getMessage());
                        return;
                    }
                }
            }
        
            try 
            {
                oso.send(msg);
                return;
            }
            catch (IOException ioe) 
            {
                //config.err.println("Encountered Exception \"" + ioe.getMessage() + "\" while writing to outgoing socket. (Attempt #" + (attempt+1) + ")");
                oso.close();
            }
        }

        config.err.println("Giving up after " + (attempt) + " attempts to send.");
	}


	/**
	 * Start this TcpAdapter.
	 */
	public void Start() {

        synchronized (stateGuard) 
        {        
            while (true) 
            {
                if (state == STOPPED) 
                {
                    state = STARTING;
                    break;
                }
                else if (state == STARTED) 
                {
                    throw new IllegalStateException("Tried to start TCP_UnicastAdapter twice!");
                }

                try {
                    stateGuard.wait();
                }
                catch (InterruptedException ie) 
                {
                }
            }
        }

        if (callback == null)
            throw new IllegalStateException("Tried to start TCP_UnicastAdapter without setting callback!");

        tcpServer = new TcpServer(serverSocket, this);
		(new Thread(tcpServer)).start();


        DeadlockMonitor = new DeadlockMonitor(this, DROP_TIME, TIMEOUT_TIME);
        (new Thread(DeadlockMonitor)).start();

        synchronized (stateGuard) 
        {
            state = STARTED;
            stateGuard.notify();
        }
	}
		
    /** This function starts the TimerProc thread.
      */
        public  void startTimer() {
			if (IP_Timer != null) {
			   if (!tcpTimeProcessor.isDead()) {
				 System.out.println("Attempt to start timer thread twice!");
				 return;
			   }
			   else {

				tcpTimeProcessor = new Tcp_TimeEventProcessor(this);

				(new Thread(tcpTimeProcessor)).start();
				return;
			   }
		    }
            IP_Timer = new Timer();

            tcpTimeProcessor = new Tcp_TimeEventProcessor(this);


            (new Thread(tcpTimeProcessor)).start();
        }

	/**
	 * Stop this TcpAdapter.  This may take time to complete.  During that time,
        * callbacks may occur.  When the call to Stop() completes, no further callbacks
        * will be made.
	 */	
	public void Stop() {		

        synchronized (stateGuard) 
        {
            while (true) 
            {
                if (state == STARTED) 
                {
                    state = STOPPING;
                    break;
                }
                else if (state == STOPPED) 
                {
                    throw new IllegalStateException("Tried to stop TCP_UnicastAdapter twice!");
                }
                
                try {
                    stateGuard.wait();
                }
                catch (InterruptedException ie) 
                {
                }
            }
        }


        tcpServer.die();

        DeadlockMonitor.die();

        // wait for these threads to die...
        synchronized (deathGuard)  
        {
	   if (!(tcpTimeProcessor == null)) {
               while (!tcpServer.isDead() || !tcpTimeProcessor.isDead()) 
               {
                   try {
                       deathGuard.wait();
                   }
                   catch (InterruptedException ie) 
                   {
                   }
               }
        } else {
               while (!tcpServer.isDead())
               {
                   try {
                       deathGuard.wait();
                   }
                   catch (InterruptedException ie)
                   {
                   }
                }
     	   }
        }

        // close sockets and stop receiver threads.
        synchronized (incomings) 
        {
            for (int i = 0; i < incomings.size(); i++) 
            {
                IncomingSocketObject iso = (IncomingSocketObject) incomings.elementAt(i);
                iso.close();
            }
        }

        synchronized (outgoings) 
        {
            Enumeration e = outgoings.elements();
            while (e.hasMoreElements()) 
            {
                OutgoingSocketObject oso = (OutgoingSocketObject) e.nextElement();
                oso.close();
            }
        }

        synchronized (receivers) 
        {
            for (int i = 0; i < receivers.size(); i++) 
            {
				TcpReceiver tr = (TcpReceiver) receivers.elementAt(i);
                tr.die();
            }
        }

            
        // wait until no more callbacks will be called.
        synchronized (deathGuard) 
        {
            while (true) 
            {

                boolean done = true;
            
                synchronized (incomings) 
                {
                    for (int i = 0; i < incomings.size(); i++) 
                    {
                        IncomingSocketObject iso = (IncomingSocketObject) incomings.elementAt(i);
                        if (!iso.isClosed())
                            done = false;
                    }
                }

                synchronized (receivers) 
                {
                    for (int i = 0; i < receivers.size(); i++) 
                    {
						TcpReceiver tr = (TcpReceiver) receivers.elementAt(i);
                        if (!tr.isDead())
                            done = false;
                    }
                }
            
                if (done)
                    break;
                
                try 
                {
                    deathGuard.wait();
                }
                catch (InterruptedException ie) 
                {
                }
            }
        }

        synchronized (stateGuard) 
        {
            state = STOPPED;
            stateGuard.notify();
        }
    }

    /** This function stops the TimerProc thread.
      */
        public  void stopTimer() {
           if (tcpTimeProcessor != null) tcpTimeProcessor.die();
        }

		
	/** Get the number of packets recerved by this Adapter.
	 */	
	public long getUReceivedPackets() {
		return UReceivedPackets;		
	}
	
	
	/** Get the number of bytes received by this Adapter. 
	 */
	public long getUReceivedBytes() {
		return UReceivedBytes;		
	}
		
	/**	Get the number of packets sent by this Adapter. 
	 */	
	public long getUSentPackets() {
		return USentPackets;		
	}
	
	
	/** Get the number of bytes sent by this Adapter. 
	 */	
	public long getUSentBytes() {
		return USentBytes;		
	}


	/**
	 * Initialize the data structure for handling the statistics defined in this object.
	 * In this method, a StatsProcessor instance is created which contains all statistics
	 * , with each represented by a StatsElement instance, supported in this object.
	 */
	private void InitStatisticsStructure()
	{
		/*
		 * Get the statistics name defined in the configuration file for this I_Stats instance.
		 */
		statisticsName = config.getTextAttribute (XmlUtil.createXPath(PROPERTY_NAME_PREFIX+"StatName"));

		/*
		 * Initial StatsProcessor instance.
		 */
		statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, false/* writable flag */);
		
		statsPro.addStatsElement("UPacketsReceived"/* name */, new UnicastPacketsReceived()/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);

		statsPro.addStatsElement("UBytesReceived", new UnicastBytesReceived(), 1, 1);

		statsPro.addStatsElement("UPacketsSent", new UnicastPacketsSent(), 1, 1);

		statsPro.addStatsElement("UBytesSent", new UnicastBytesSent(), 1, 1);
	}

	/** @see I_Stats#getStatsName() */
	public String getStatsName ()
	{
		return statisticsName;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName (String name)
	{
		statisticsName = name;
	}


	/** 
	 * Return the result of query for the statistics specified by the given xpath.
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance represents the statistics to be queried.
	 * @see I_Stats#getStats
	 */
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//System.out.println("TCP_UnicastAdapter:getStats: xpath string is "+xpath.getPatternString());
		//System.out.println("TCP_UnicastAdapter:getStats: call statsPro.getStatsResult(doc, xpath)");
		return statsPro.getStatsResult(doc, xpath);
	}

	/** 
	 * Set the statistics specified by the given xpath. The value actually set is
	 * returned.
	 * 
	 * @param doc   	Document used for create new elements or nodes.
	 * @param xpath 	XPath instance represents the statistics to be queried.
	 * @param newValue	Element representing the value or sub-tree to be set.
	 * @see I_Stats#setStats
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	{
		//System.out.println("TCP_UnicastAdapter:setStats: xpath string is "+xpath.getPatternString());
		//System.out.println("TCP_UnicastAdapter:setStats: call statsPro.setStatsResult(doc, xpath, newValue)");
		return statsPro.setStatsResult(doc, xpath, newValue);
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in read schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//System.out.println("TCP_UnicastAdapter:getReadSchema: xpath string is "+xpath.getPatternString());
		//System.out.println("TCP_UnicastAdapter:getSchema: call statsPro.getReadSchemaResult(doc, xpath)");
		return statsPro.getReadSchemaResult(doc, xpath);
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in write schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getWriteSchema(Document, XPath)
	 */
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return statsPro.getWriteSchemaResult(doc, xpath);
	}
	

	/**
	 * This class manages simple statistic "UPacketsReceived".
	 */
	class UnicastPacketsReceived extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+UReceivedPackets);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "UBytesReceived".
	 */
	class UnicastBytesReceived extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+UReceivedBytes);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "UPacketsSent".
	 */
	class UnicastPacketsSent extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+USentPackets);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "UBytesSent".
	 */
	class UnicastBytesSent extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+USentBytes);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
	
}	


/** This class manages an outgoing socket.  It provides
 * synchronization for multiple threads trying to perform a send().
 * It will also interrupt any thread trying to perform a send() when
 * the close() method is called.  It automatically inserts itself into
 * TcpAdapter.outgoings when created.  It removes itself when
 * close()d.  
 */
class OutgoingSocketObject 
{
    /** Address that this socket connected to. */
	//private INETv4TwoPorts childPA;    
	private I_Address childPA;    

    /** The socket. */
    private Socket s;

    /** The OutputStream of the socket. */
    private OutputStream os;

    /** The adapter that created this object. */
    private TCP_UnicastAdapter parent;

    /** Ideally, the last time a message was written to this socket.
         * Actually, every time a message is written, the value is rest to
         * 0.  If its value is 0 when read, the value is reset to the
         * current time.  By this, we avoid calling
         * System.currentTimeMillis() for every write.  */
    private long lastTimeWritten;
    
    /** The thread waiting write writing this socket. */
    private Thread waitingThread;

    /** The lock for the variable "waitingThread".  Used to guarantee
         * that the interrupt sent to <code>waitingThread</code> reaches
         * it. */
    private Object waitingThreadGuard;

    /** Set to true when the socket is closed. */
    private boolean isClosed;
    


    OutgoingSocketObject(I_Address cPA, TCP_UnicastAdapter p) throws IOException
    {
        childPA = cPA;
        parent = p;
        lastTimeWritten = 0;
        
		InetAddress dstinetaddress = null;
		int portnum = 0;
		
		/*
		 * If destination is a underlay address, it contains IP
		 * address and one port.
		 */
		if (childPA instanceof I_UnderlayAddress)
		{
			INETv4OnePort dstuaaddr = (INETv4OnePort) childPA;
			dstinetaddress = dstuaaddr.getInetAddress();

			portnum = dstuaaddr.getPort();
		}
		/*
		 * If destination is a physical address, it contains IP
		 * address and two ports. Depending on that this adapter is 
		 * Socket Adapter or Node Adapter, the corresponding port 
		 * number is retrived from destination physical address.
		 */
		else if (childPA instanceof I_PhysicalAddress)
		{
			INETv4TwoPorts dstpaaddr = (INETv4TwoPorts) childPA;
			
			dstinetaddress = dstpaaddr.getInetAddress();

			if (parent.AdptIndicator.equals("nodeadpt")) 
			{
				portnum = dstpaaddr.getNodeAdptPort();
			} 
			else
			{
				portnum = dstpaaddr.getScktAdptPort();
			} 
		}
		else
		{
			throw new HyperCastWarningRuntimeException ("Destination address has incorrect format.");
		}


		s = new Socket(dstinetaddress, portnum);			
            
        os = s.getOutputStream();				

        waitingThread = null;
        waitingThreadGuard = new Object();
        isClosed = false;
        
        // if all this goes okay, then it is entered into hashtable.
        parent.outgoings.put(childPA, this);
    }
    

/** Returns true if there has not been a write to the socket in the last 
 * <code>timePeriod</code> milliseconds. */
    boolean noWriteInTime(int timePeriod, long currentTime)
    {
        if (lastTimeWritten == 0)
            lastTimeWritten = currentTime;

        return (currentTime - lastTimeWritten > timePeriod);
    }

/** Closes the socket and interrupts any thread doing a write. */
    void close() 
    {
        synchronized (waitingThreadGuard) 
        {
            if (isClosed)
                return;

            parent.outgoings.remove(childPA);

            try 
            {
                os.close();
                s.close();
            }
            catch (IOException ioe) 
            {
               // parent.config.log.println("Encountered Exception \""+ ioe.getMessage() + "\" while closing down OutgoingSocketObject in TCP_UnicastAdapter.");
            }

            //parent.config.log.println("TCP_UnicastAdapter: Closed connection to " + childPA.toString());

            if (waitingThread != null)
                waitingThread.interrupt();
            isClosed = true;
        }
    }

/** Converts the message to a byte[] and writes that to the OutputStream. 
 * If an interrupt occurs, the function returns normally.
 * @throws IOException for any socket error besides InterruptedIOException.
 */
public synchronized void send(I_Message msg) throws IOException
    {
        byte[] msgarray = msg.toByteArray();				

        //System.out.println("Tcp sending....");
        //This call will block here until this packet has been sent out!!						
        try 
        {
            try 
            {
                try 
                {
                    synchronized (waitingThreadGuard) 
                    {
                        if (isClosed) 
                        {
                            //parent.config.log.println("TCP_UnicastAdapter: Attempted write to a closed OutgoingSocketObject - dropping message.");
                            return;
                        }
                        waitingThread = Thread.currentThread();
                    }
                        
                    os.write(msgarray);
                    os.flush();						
                        
                }
                finally 
                {
                    synchronized (waitingThreadGuard) 
                    {
                        waitingThread = null;
                    }
                }   
            }
            finally
            {
                //if (Thread.interrupted()) 
                //    parent.config.seriousErr.println("SERIOUS ERROR: This println should have been interrupted before it could take place.");
            }
        }
        catch (InterruptedIOException ioe) 
        {
           // parent.config.log.println("TCP_UnicastAdapter: Write to an OutgoingSocketObject was interrupted.");
            return;
        }

        parent.USentPackets++;
        parent.USentBytes += msgarray.length;
        lastTimeWritten = 0;
    }
}


/** This object manages an incoming socket connection.  It provides
 * synchronization for multiple threads receive()ing.  It will also
 * interrupt and receive()ing thread when the object is close()d.
 * <p>
 * This object manages a buffer of TCP_UnicastAdapter.RECEIVE_BUFFER_SIZE.
 * It is a limit on the maximum size of a message.  
 *<p>
 * This thread calls the createMessage() function of the
 * I_AdapterCallback implementer.  */

class IncomingSocketObject 
{
    /** The socket. */
    Socket s;
    /** The InputStream of the socket. */
    InputStream is;
    /** The object implementing the createMessage() function. */
    I_AdapterCallback callback;
    /** The TCP_UnicastAdapter that created this object. */
    TCP_UnicastAdapter parent;
    

    /** The buffer for holding data read from the InputStream. 
    *  The portion of the buffer containing the data is in 
    * <code>receiveBuffer[startValidBytes[0]]</code> to 
    <code>receiveBuffer[endValidBytes-1]</code>*/
    byte[] receiveBuffer;    

    /** startValidBytes[0] holds the location in <code>receiveBuffer</code> 
     * of first byte read from the InputStream.
     * startValidBytes is an array of length 1.
     * @see I_AdapterCallback#restoreMessage(byte[], int[], int)
     */
    int[] startValidBytes;
    /** One beyond the index of the final byte in <code>receiveBuffer</code>
     * that was read from the InputStream. */
    int endValidBytes;

    /** The thread currently reading from the InputStream.*/
    Thread waitingThread;
    /** A lock.  Used to guarantee that interrupts do affect the <code>waitingThread</code>
     * while it is reading from the InputStream. */
    Object waitingThreadGuard;
    /** True if the socket has been closed. */
    boolean isClosed;

    /** True while the synchronized thread is inside an I_AdapterCallback function. */
    boolean inCallback;

    /** Ideally, the last time a thread performed the read() call to
     * the InputStream.  Actually, every time the call is done, the
     * value is rest to 0.  If the variable's value is read as a 0,
     * the value is reset to the current time.  By this, we avoid
     * calling System.currentTimeMillis() for every read().  */
    long lastTimeAttemptToRead;

    /** Ideally, the last time a thread returned from the read() call
     * to the InputStream.  Actually, every time the call completes,
     * the value is rest to 0.  If the variable's value is read as a
     * 0, the value is reset to the current time.  By this, we avoid
     * calling System.currentTimeMillis() for every read().  */
    long lastTimeRead;
    

    IncomingSocketObject(Socket sock, I_AdapterCallback cb, TCP_UnicastAdapter p) throws IOException
    {
        s = sock;
        callback = cb;
        parent = p;
    
        receiveBuffer = new byte[parent.RECEIVE_BUFFER_SIZE];  
        startValidBytes = new int[1];
        startValidBytes[0] = 0;
        endValidBytes = 0;
    
        waitingThread = null;
        waitingThreadGuard = new Object();
        isClosed = false;
        inCallback = false;

        lastTimeAttemptToRead = 0;
        lastTimeRead = 0;
        
        is = s.getInputStream();

      //  parent.config.log.println("TCP_UnicastAdapter: Created incoming connection.");

        // if all goes okay, entered into vector.
        parent.incomings.addElement(this);
    }


/** Returns true if there a read has not completed in the last
 * <code>timePeriod</code> milliseconds. */
    boolean noReadInTime(int timePeriod, long currentTime)
    {
        if (lastTimeRead == 0)
            lastTimeRead = currentTime;

        return (currentTime - lastTimeRead > timePeriod);
    }
    
/** Returns true if there a read call has not been made in the last
 * <code>timePeriod</code> milliseconds. */
    boolean noAttemptToReadInTime(int timePeriod, long currentTime)
    {
        if (lastTimeAttemptToRead == 0)
            lastTimeAttemptToRead = currentTime;

        return (currentTime - lastTimeAttemptToRead > timePeriod);
    }

/** Returns true if data is in buffer or InputStream has bytes available. */
    boolean dataWaiting() 
    {
        if (endValidBytes != startValidBytes[0])
            return true;
        else 
        {
            try 
            {
                return is.available() != 0;
            }
            catch (IOException ioe) 
            {
                // let someone else close it...
                return true;
            }
        }
    }
    

/** Closes the socket and interrupts any thread currently read()ing. */
    void close() 
    {
        synchronized (waitingThreadGuard) 
        {
            if (isClosed)
                return;

            parent.incomings.removeElement(this);

            try 
            {
                is.close();
                s.close();
            }
            catch (IOException ioe) 
            {
                //parent.config.log.println("Encountered Exception \""+ ioe.getMessage() + "\" while closing down IncomingSocketObject in OL_TCP_Adapter.");
            }

            //parent.config.log.println("TCP_UnicastAdapter: Closed incoming connection.");
        
            if (waitingThread != null)
                waitingThread.interrupt();
            isClosed = true;
        }
    }

/** Returns true if the thread will never call an function in I_AdapterCallback again. */
    boolean isClosed() 
    {
        return (isClosed && !inCallback);
    }
    


    /** Private function that adds bytes to the buffer from the InputStream. 
    * @throws IOException for socket error
    * @throws InterruptedIOException if socket is closed
    */
private void readBytes() throws IOException
    {
        int bytesRead = 0;
        try 
        {
            try 
            {
                synchronized (waitingThreadGuard) 
                {
                    if (isClosed)
                        throw new InterruptedIOException("closed IncomingSocketObject");
                    waitingThread = Thread.currentThread();
                    lastTimeAttemptToRead = 0;
                }
                
                bytesRead = is.read(receiveBuffer, endValidBytes, receiveBuffer.length - endValidBytes);	
                if (bytesRead < 0)
                    throw new InterruptedIOException("read from closed socket InputStream");
                endValidBytes += bytesRead;
            
            }
            finally 
            {
                synchronized (waitingThreadGuard) 
                {
                    waitingThread = null;
                    lastTimeRead = 0;
                }
            }
        }
        finally
        {
            //if (Thread.interrupted())
               // parent.config.seriousErr.println("SERIOUS ERROR: This println should have been interrupted before it could take place.");
        }

        parent.UReceivedBytes += bytesRead;
    }


/** Returns a new message as gotten from I_AdapterCallback.createMessage.
 * This function will only read from the InputStream if the buffer does not
 * contain a whole message.  
 * @throws IllegalArgumentException if bytestream gets corrupted (passed from restoreMessage())
    * @throws IOException for socket error
    * @throws InterruptedIOException if socket is closed
 */
public synchronized I_Message receive() throws IOException
    {
        I_Message msg = null;


        try 
        {
            inCallback = true;
            try 
            {
				//To ensure that non-zero length byte array is passed to restoreMessage.
                //if (!isClosed)
				if ((!isClosed) && (endValidBytes - startValidBytes[0] > 0)) {
                    msg = callback.restoreMessage(receiveBuffer, startValidBytes, endValidBytes);
				}
            }
            finally 
            {
                synchronized (parent.deathGuard) 
                {
                    inCallback = false;
                    parent.deathGuard.notify();
                }
            }
        }
        catch (IllegalArgumentException iae)
        {
            // corrupt connection.  Throw it so that socket gets closed. 
            parent.config.err.println("TCP connection became corrupted."+iae.getMessage());
            
            //should we close this socket????
            this.close();
            
            throw iae;
        }
        catch (Exception e) 
        {
            parent.config.err.println("Exception occured in I_AdapterCallback.restoreMessage()");
            e.printStackTrace(parent.config.err);
            msg = null;
        }
        
        
        // else keep reading until we get a message.
        while (msg == null) 
        {
            // if empty, start reading at beginning.
            if (startValidBytes[0] == endValidBytes)
            {
                startValidBytes[0] = 0;
                endValidBytes = 0;
            }
            
            // if reached end of buffer, copy all to beginning and then read.
            if (endValidBytes == receiveBuffer.length)
            {
                if (startValidBytes[0] == 0) 
                    throw new RuntimeException("Message size exceeded "+receiveBuffer.length+" bytes!!!");
                
                // according to documentation, this is safe.
                // (i.e. works if dst array is same as src array.)
                System.arraycopy(receiveBuffer, startValidBytes[0], receiveBuffer, 0, endValidBytes - startValidBytes[0]);
                endValidBytes = endValidBytes - startValidBytes[0];
                startValidBytes[0] = 0;
                
            }
            
            readBytes();
            
            try 
            {
                inCallback = true;
                try 
                {
					//To ensure that non-zero length byte array is passed to restoreMessage.
					//if (!isClosed)
					if ((!isClosed) && (endValidBytes - startValidBytes[0] > 0))
                        msg = callback.restoreMessage(receiveBuffer, startValidBytes, endValidBytes);
                }
                finally 
                {
                    synchronized (parent.deathGuard) 
                    {
                        inCallback = false;
                        parent.deathGuard.notify();
                    }
                }
            }
            catch (IllegalArgumentException iae)
            {
                // corrupt connection.  Throw it so that socket gets closed. 
                parent.config.err.println("TCP connection became corrupted."+iae.getMessage());
                //should we close this socket????
                this.close();
                
                throw iae;
            }
            catch (Exception e) 
            {
                parent.config.err.println("Exception occured in I_AdapterCallback.restoreMessage()");
                e.printStackTrace(parent.config.err);
                msg = null;
            }

        }

        parent.UReceivedPackets++;
        
        return msg;
    }
}


/**
 * This class accepts connections on the ServerSocket and, for each connection, 
 * creates an IncomingSocketObject, TcpReceiver, and a receiving thread. */

class TcpServer implements Runnable {	 
     /** The TCP_UnicastAdapter that created this object.*/
private TCP_UnicastAdapter Parent;		

    /** a ServerSocket */
private ServerSocket serversock;	

     /** The thread doing the accept() calls.*/
private Thread myThread;
    /** True if die() has been called. */
private boolean dying;
    /** True if the thread has completed the function run(). */
private boolean died;

	/** True if <code>myThread</code> is in the accept call. */
private boolean waiting;
    /** Lock to make sure that interrupt is only sent when <code>myThread</code>
     * is blocked on the accept call. */
private Object waitingGuard;


	TcpServer(ServerSocket serversocket, TCP_UnicastAdapter adapter) {
		Parent = adapter;	
		serversock = serversocket;

        dying = false;
        died = false;

        waiting = false;
        waitingGuard = new Object();
	}

/** Causes <code>myThread</code> to return from the run() method,
 * effectively killing it. 
 */
public void die() 
     {
         synchronized (waitingGuard) 
         {
             dying = true;
             if (waiting)
                 myThread.interrupt();  // incase it is blocking on the accept call.
         }
     }

/** Returns true if the thread will never call an function in I_AdapterCallback again. */
public boolean isDead() 
    {
        return (died || waiting);
    }
    
	
/** Performs accept() calls, and for each connection, 
* creates an IncomingSocketObject, TcpReceiver, and a receiving thread. */
	public void run() 
     {
        myThread = Thread.currentThread();
		while(true) {								
            Socket insocket = null;

			try 
            {
                try 
                {
                    try 
                    {
                        synchronized (waitingGuard) 
                        {   
                            waiting = true;
                        }
                        
                        if (dying)
                            break;
                        
                        insocket = serversock.accept(); 
                    }
                    finally 
                    {
                        synchronized (waitingGuard) 
                        {   
                            waiting = false;
                        }
                    }
                }
                finally
                {
                    // prevent possible interrupt later
                    //if (Thread.interrupted()) 
                       // Parent.config.seriousErr.println("SERIOUS ERROR: This println should have been interrupted before it could take place.");
                }
			}   
            catch (InterruptedIOException iioe) 
            {
            }
			catch (IOException e) {		
                //Parent.config.err.println("TCP_UnicastAdapter: ServerSocket accept error! Exception:" + e.getMessage());
                continue;
			}

            if (dying) 
            {
                try 
                {
                    insocket.close();
                }
                catch (Exception e)
                {
                }
                break;
            }


            IncomingSocketObject iso = null;
            try 
            {
                iso = new IncomingSocketObject(insocket, Parent.callback, Parent);
            }
            catch (IOException ioe)
            {
                Parent.config.err.println("TCP_UnicastAdapter: Error while creating IncomingSocketObject! Exception:" + ioe.getMessage());
                continue;
            }
            
			TcpReceiver tr = new TcpReceiver(iso, Parent.callback, Parent);
            (new Thread(tr)).start();
            //Parent.config.log.println("TCP_UnicastAdapter: Created IncomingSocketObject and started Receiver thread.");
		} 


        synchronized (Parent.deathGuard) 
        {
            died = true;
            Parent.deathGuard.notify();
        }
	}
}



/**
 * This class provides the run() method for each receiver thread.
 * Messages are read from the associated IncomingSocketObject and
 * then passed to the I_AdapterCallback.messageArrivedFromAdapter method.
 */
class TcpReceiver implements Runnable {	
    /** Where messages come from. */
    private IncomingSocketObject iso;
    /** Where messages go to. */
	private I_AdapterCallback callback;

    /** The TCP_UnicastAdapter that created this object. */
    private TCP_UnicastAdapter parent;
    
    /** True if the function die() has been called. */
    private boolean dying;
    /** True while the thread is in the I_AdapterCallback function <code>messageArrivedFromAdapter</code>. */
    private boolean inCallback;


	TcpReceiver(IncomingSocketObject iso_arg, I_AdapterCallback adaptercallback, TCP_UnicastAdapter p) {
        iso = iso_arg;
		callback = adaptercallback;		
        parent = p;

        dying = false;
        inCallback = false;
    }


    /** Reads from IncomingSocketObject and sends them to I_AdapterCallback.
     * If the IncomingSocketObject throws an InterruptedIOException, the thread
     * close()s the IncomingSocketObject and then dies. */
	public void run() {
        try 
        {
            while (true)
            {
                I_Message msg = iso.receive();

                inCallback = true;
                try 
                {
                    if (!dying)
                        callback.messageArrivedFromAdapter(msg); 
                }
                catch (Exception e) 
                {
                    parent.config.err.println("Encountered Exception \"" + e.getMessage() + "\" in call to I_AdapterCallback.messageArrivedFromAdapter() in TCP_UnicastAdapter.");
                    e.printStackTrace(parent.config.err);
                }

                synchronized (parent.deathGuard) 
                {
                    inCallback = false;
                    parent.deathGuard.notify();
                }
            }
        }
        catch (Exception e) 
        {
            parent.config.err.println("Encountered Exception \"" + e.getMessage() + "\" while reading from incoming socket.");
            iso.close();
        }
    }

    /** Stop the thread indirectly. */
    void die() 
    {
        dying = true;
    }

    /** True if the thread will never make another call to an I_AdapterCallback method. */
    boolean isDead() 
    {
        return (dying && !inCallback);
    }
}


/**
 * This class provides the run() method for each dropper thread.
 * A single messages is read from the associated IncomingSocketObject
 * and then dropped.
 */
class TcpDropper implements Runnable 
{
    /** Where messages come from. */
    private IncomingSocketObject iso;

    /** The TCP_UnicastAdapter that created this object. */
    private TCP_UnicastAdapter parent;
    

	TcpDropper(IncomingSocketObject iso_arg, TCP_UnicastAdapter p) {
        iso = iso_arg;
        parent = p;
    }

    /** Reads a message from IncomingSocketObject and then drops it.
     * If the IncomingSocketObject throws an InterruptedIOException, the thread
     * close()s the IncomingSocketObject and then dies. */
	public void run() {
        try 
        {
            I_Message msg = iso.receive();
        }
        catch (Exception e) 
        {
           // parent.config.err.println("Encountered Exception \"" + e.getMessage() + "\" while reading from incoming socket.");
            iso.close();
        }
    }
}

		
	
/**
 * The run() method of this object checks sockets for deadlock and
 * timeout.  The thread checks all the sockets periodically.  If an
 * incoming one is not read for more than a certain time, it will
 * creates a dropper thread to ensure progress.  If a socket is idle
 * for more than a certain time, the socket is closed.
 */
class DeadlockMonitor implements Runnable {
    /** TCP_UnicastAdapter that created this object. */
private TCP_UnicastAdapter Parent;	
    /** The time a socket must be ignored by its receiving thread
        before a message is dropped. */
private int dropTime;
    /** The time a socket must be idle before being closed. */
private int timeoutTime;

    /** The deadlock monitoring thread */
private Thread myThread;
    /** True if the function die() has been called. */
private boolean dying;

    /** True if <code>myThread</code> is sleeping. */
private boolean waiting;
    /** A lock to ensure that <code>myThread</code> is only interrupted while sleeping. */
private Object waitingGuard;
    

	DeadlockMonitor(TCP_UnicastAdapter adapter, int btime, int ttime) {		
        Parent = adapter;			
        dropTime = btime;
        timeoutTime = ttime;

        dying = false;
        waiting = false;
        waitingGuard = new Object();
	}
	
/** Causes <code>myThread</code> to return from the run() method,
 * effectively killing it. 
 */
public void die() 
     {
         synchronized (waitingGuard) 
         {
             dying = true;
             if (waiting)
                 myThread.interrupt();  // incase it is blocking on the wait call.
         }
     }

    /** Periodically checks all object in
     * <code>TCP_UnicastAdapter.incoming</code> and
     * <code>TCP_UnicastAdapter.outgoings</code> to ensure they are active
     * and being read. */
public void run() 
    {
        myThread = Thread.currentThread();
		while(true) {
            try {
                try 
                {
                    synchronized (waitingGuard) 
                    {   
                        waiting = true;
                    }   
                  
                    if (dying)
                        break;
                    
                    Thread.sleep(dropTime/4);
                }
                finally 
                {
                    synchronized (waitingGuard) 
                    {   
                        waiting = false;
                    }
                }

                // should clear interrupted bit.
                if (Thread.interrupted())
                    Thread.sleep(dropTime/4);
                
            } catch (InterruptedException e) {
                continue;
            }

            if (dying)
                break;

           
           long t = System.currentTimeMillis();
           synchronized (Parent.incomings) 
           {
               for (int i = 0; i < Parent.incomings.size(); i++) 
               {
                   IncomingSocketObject iso = (IncomingSocketObject) Parent.incomings.elementAt(i);
                   
                   synchronized (iso.waitingThreadGuard) 
                   {
                       if (iso.waitingThread != null &&
                           iso.noReadInTime(timeoutTime, t)) 
                       {
                           // close socket down.
                          // Parent.config.log.println("TCP_UnicastAdapter: Closing unused incoming socket.");
                           iso.close();
                       }
                       else if (iso.waitingThread == null &&
                                iso.noAttemptToReadInTime(dropTime, t) &&
                                iso.dataWaiting()) 
                       {
                           // start message killing thread.
                           //Parent.config.log.println("TCP_UnicastAdapter: Starting dropper on incoming socket which is not being read.");
                           Thread thread = new Thread(new TcpDropper(iso, Parent));
                           thread.start();
                       }
                   }
               }
           }
       
           synchronized (Parent.outgoings) 
           {
               Enumeration e = Parent.outgoings.elements();
               while (e.hasMoreElements()) 
               {
                   OutgoingSocketObject oso = (OutgoingSocketObject) e.nextElement();
               
                   // if no write in timeout time...
                   if (oso.noWriteInTime(timeoutTime, t)) 
                   {
                       // close socket down.
                      // Parent.config.log.println("TCP_UnicastAdapter: Closing unused outgoing socket.");
                       oso.close();
                   }
               }
           }
        }
	}			
}




/** 
 * This class starts a thread to process time event. <p>
 * <br>
 * The processor takes the time event with the smallest timestamp from 
 * the time_event queue. <br> 
 * If this timestamp > currenttime,
 * the processor sleep for (timestamp - currenttime), then processes the time event. <br>
 * If this timestamp <= currenttime,
 * the processor processes the time event at once.  */

class Tcp_TimeEventProcessor implements Runnable {

    /** The TCP_UnicastAdapter that created this object. */
private TCP_UnicastAdapter Parent;

    /** The thread performing the timer functions */
private Thread myThread;
    /** True if the function die() has been called. */
private boolean dying;
    /** True if the thread has returned from the run() method. */
private boolean died;

    /** True if <code>myThread</code> is currently sleeping. */
private boolean waiting;
    /** Lock used to ensure that <code>myThread</code> is only interrupted while sleeping. */
private Object waitingGuard;


	Tcp_TimeEventProcessor(TCP_UnicastAdapter uadapter) {
		Parent = uadapter;

        dying = false;
        died = false;

        waiting = false;
        waitingGuard = new Object();
	}


/** Causes <code>myThread</code> to return from the run() method,
 * effectively killing it. 
 */
public void die() 
     {
         synchronized (waitingGuard) 
         {
             dying = true;
             if (waiting)
                 myThread.interrupt();  // incase it is blocking on the wait call.
         }
     }

/** Returns true if the thread will never make another call to an I_AdapterCallback function. */
public boolean isDead() 
    {
        return (died || waiting);
    }


/** Takes time events from the TimeEventQueue and 
 * sends them to the protocol node (I_AdapterCallback) for processing.
 */
	public void run() {	    		
        myThread = Thread.currentThread();
		
		
		//System.out.println("IP_TimeEventProcessor started.....");		
		
		while (true)
        {
            Object id;

            try 
            {
                try 
                {
                    synchronized (waitingGuard) 
                    {   
                        waiting = true;
                    }

                    if (dying)
                        break;
                    
                    id = Parent.IP_Timer.sleepUntilNextEvent();
                }
                finally 
                {
                    synchronized (waitingGuard) 
                    {   
                        waiting = false;
                    }
                }

                // prevent possible interrupt in handler.
                if (Thread.interrupted())
                    Thread.sleep(1000); // should be interrupted & throw exception.

                if (dying) 
                {
                    // stick event back in queue.
                    Parent.IP_Timer.setTimer(id, 0);	
                    break;
                }
            
                
                try 
                {
                    Parent.callback.timerExpired(id);
                }
                catch (Exception e) 
                {
                   // Parent.config.err.println("IP_TimeEventProcessor encountered Exception in timerExpired() callback: " + e.getMessage());
                   // e.printStackTrace(Parent.config.err);
                }
            }
            catch (InterruptedException ie) 
            {
            }
		}	

        synchronized (Parent.deathGuard) 
        {
            died = true;
            Parent.deathGuard.notify();
        }
	} 
}



