/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.adapters;

import hypercast.*;
import hypercast.util.XmlUtil;

import java.io.*;
import java.util.*;
import java.net.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.apache.xpath.*;

/** 
 * This class implements all the functions related with a UDP unicast 
 * adapter. <p>
 * There are three threads: <br>
 *  1. UnicastReceiver thread receives data from a unicast 
 *     address and puts it into a buffer <br>
 *  2. UDP_MsgProcessor thread reads Message from 
 *     MessageBuffer, then pass it to protocol node (I_AdapterCallback) 
 *     for procession <br>
 *  3. UDP_TimeEventProcessor thread reads TimeEvent in this 
 *     adapter, then pass it to protocol node (I_AdapterCallback) 
 *     for procession. The Timer is started/stoped explicitly outside
 *     this class. <p>
 * 
 * There is a FIFO style MessageBuffer in charge of storing messages 
 * received from UnicastReceiver (and MulticastReceiver if this class is 
 * extended by UDP_MulticastAdapter) threads. This class supports 
 * INETv4TwoPorts type physical address.
 * 
 * @author HyperCast Team
 * @author Dongwen Wang
 * @author Michael Nahas
 * @version 2.0, February 13, 2001.
 * @author Jianping Wang
 * @version 3.0, Jul. 24, 2004 
 */

public class UDP_UnicastAdapter implements I_IPv4_UnicastAdapter
{

    	/** Prefix for all property names read in this adapter.*/
	protected static String PROPERTY_NAME_PREFIX;

		/** The property object for configuration this adapter */
	HyperCastConfig property;

		/** The protocol node that runs on this adapter */
	I_AdapterCallback Node;

		/** The timer that works with this adapter */
	Timer UDP_Timer = null;
	
		/** The unicast datagram socket of this adapter */
	DatagramSocket USock = null;

		/** The pipe like buffer embedded in this adapter */
	MessageBuffer InputChannel;

		/** The Maximum length in bytes that this adapter can receive */
	int MaximumPacketLength;
	
		/*
		 * The port number is needed by UnicastReceiver when it needs to 
		 * reset the DatagramSocket in this adapter.
		 */
		/** The port number of the UDP DatagramSocket of this adapter */
	int UPort;   

		/** The string determines if Node adapter should try random port 
		 * number when physical addresss assignment fails. If it is set to 
		 * "yes", the adapter will try random port number when the adapter 
		 * fails in assign physical addresss; Otherwise, the adapter will 
		 * not try.
		 */
	String AllowRandomPhyAddrPort;   

		/** The string which indicate this adapter is Node Adapter or
		 * Socket Adapter: 
		 * "nodeadpt" - Node Adapter, "sockadpt" - Socket Adapter.
		 */
	String AdptIndicator;   

		/** The pointer to socket adapter */
	I_UnicastAdapter socketAdpter;

		/** Total number of packets received from the unicast datagram socket */
	long UReceivedPackets;
		/** Total number of packets sent to the unicast datagram socket */
	long USentPackets;
		/** Total number of bytes received from the unicast datagram socket */	
	long UReceivedBytes;
		/** Total number of bytes sent to the unicast datagram socket */
	long USentBytes;
	
		/** The object used to synchronize unicast socket */
	Object USockLocker;
	
		/** Unicast receiver object */
	private UnicastReceiver URecv;

		/** Message processing object */
	private UDP_MsgProcessor MsgProc;  

		/** Time event processing object */
	private UDP_TimeEventProcessor TimerProc = null;  
	
     /** The object used to wait() on when waiting for other threads to die. */
	Object deathGuard;

    /** Current state of the node.  It goes from STOPPED to STARTING
         * to STARTED to STOPPING and then back to STOPPED. */
    protected int state;

    /** value for <code>state</code>. */
    final static int STOPPED = 0;
    /** value for <code>state</code>. */
    final static int STARTING = 1;
    /** value for <code>state</code>. */
    final static int STARTED = 2;
    /** value for <code>state</code>. */
    final static int STOPPING = 3;
    
    /** Used to protect <code>state</code> when its value is changed. */
	protected Object stateGuard;

	/** String indicates the state of time event processing object */
    String TimerProcDebug = "Dead";
	/** String indicates the state of message processing object */
    String MsgProcDebug;
	/** String indicates the state of unicast receiver object */
    String URecvDebug;

	/**
	 * A StatsProcessor instance which handles the statistics in this object.
	 */
	protected StatsProcessor statsPro;
	
	/**
	 * The statistics name for this object (as an I_Stats instance). Its default
	 * name is defined in the socket configuration schema file. It can be assigned
	 * by any other object through method setStattisticsName defined in the 
	 * I_Stats interface.
	 */
	protected String statisticsName;

	/** 
	 * Constructs a UDP_UnicastAdapter object.<p>
	 * 
	 * @param	config		an HyperCastConfig object.
	 * @param	scktAdpt	object representing the Socket Adapter. It is set to null when 
	 * 						creating the Socket Adapter.
	 * @param	AddrString	string containing physical address information.
	 * @param	prefix		property prefix.
	 */
	public UDP_UnicastAdapter(HyperCastConfig config, I_UnicastAdapter scktAdpt, String AddrString, String prefix){
		
		/* Execute the initiallization, except initializing the DatagramSocket,
		 * for this adapter.
		 */
		Adapter_Inilization(config, prefix);

		socketAdpter = scktAdpt;

		/* 
		 * Initialize the DatagramSocket used by this adapter. If the initialization of 
		 * DatagramSocket fails, a runtime exception is thrown. Only Node Adapter starts
		 * socket initialization. It not only initializes the socket in Node Adapter, 
		 * but also initialize the socket in Socket Adapter.
		 */
		if ((!AdptIndicator.equals("scktadpt")) && (!InitializeDatagramSocket(scktAdpt, property, AddrString)))
		{
			throw new HyperCastFatalRuntimeException ("Could not initialize local unicast socket!");
		}

	}

	/** 
	 * Constructs a UDP_UnicastAdapter object with a HyperCastConfig 
	 * and DataGramSocket object.
	 * @param	config		an HyperCastConfig object.
	 * @param	usock		a DatagramSocket object.
	 * @param	prefix		property prefix.
	 */
	public UDP_UnicastAdapter(HyperCastConfig config, DatagramSocket usock, String prefix){
		
		/* Execute the initiallization, except initializing the DatagramSocket,
		 * for this adapter.
		 */
		Adapter_Inilization(config, prefix);
		
		USock = usock;
		UPort = USock.getLocalPort();
	}

	/** 
	 * Execute the initialization, except initializing the DatagramSocket, for this 
	 * UDP_UnicastAdapter object.
	 * 
	 * @param	config		a HyperCastConfig object.
	 * @param	prefix		property prefix.
	 */
	private void Adapter_Inilization(HyperCastConfig config, String prefix){
		
		property = config;

		/*
		 * Assign the adapter indicator.
		 */
		if (prefix.startsWith("NodeAdpt"))
		{
			AdptIndicator = "nodeadpt";
			PROPERTY_NAME_PREFIX = "/Public/NodeAdapter/"+prefix+"/";
		} 
		else
		{
			AdptIndicator = "scktadpt";
			PROPERTY_NAME_PREFIX = "/Public/SocketAdapter/"+prefix+"/";
		} 
		
		MaximumPacketLength = property.getPositiveIntAttribute
				(XmlUtil.createXPath(PROPERTY_NAME_PREFIX + "MaximumPacketLength"));

		if (prefix.startsWith("NodeAdpt"))
		{
			AllowRandomPhyAddrPort = property.getTextAttribute
					(XmlUtil.createXPath(PROPERTY_NAME_PREFIX+"AllowRandomPhyAddrPort"));
		}
    
		USockLocker = new Object();
		deathGuard = new Object();
		
		UReceivedPackets = 0;
		UReceivedBytes = 0;
		USentPackets = 0;
		USentBytes = 0;
		
		/* 
		 * Initialize the MessageBuffer
		 */
		int adapterBufferSize = property.getPositiveIntAttribute
				(XmlUtil.createXPath(PROPERTY_NAME_PREFIX + "MessageBufferSize"));

		if (prefix.startsWith("NodeAdpt"))
		{
                    InputChannel = new MessageBuffer(property, adapterBufferSize, "Node Adapter UDP Adapter");
		} 
		else
		{
                    InputChannel = new MessageBuffer(property, adapterBufferSize, "Socket Adapter UDP Adapter");
		} 
		//InputChannel = new MessageBuffer(property, adapterBufferSize, "UDP Adapter");

		state = STOPPED;
		stateGuard = new Object();

		/* 
		 * Initialize the statistics name array.
		 */
		InitStatisticsStructure();
	}
	

	/**
	 * Initialize DatagramSocket used by this adapter.
	 * 
	 * @param	ScktAdpter	 A I_UnicastAdapter object representing Socket Adapter.
	 * @param	config	 an HyperCastConfig object.
	 * @param	AddrString	 A string carring physical address information.
	 */
	private boolean InitializeDatagramSocket(I_UnicastAdapter ScktAdpter, HyperCastConfig config, String AddrString)
	{
		/*
		 * Set initial values for two ports and InetAddress.
		 */
		int portNodeAdpt = 0;
		int portScktAdpt = 0;
		InetAddress ipAddr = null;

		/*
		 * If this is Socket Adapter, return. The initialization of sockets in Socket Adapter
		 * is trigged from Node Adapter.
		 */
		if (AdptIndicator.equals("scktadpt"))
		{
			return true;
		}
		
		/*
		 * If this is Node Adapter and physical address is specified (in string format) with papameter,
		 * the specified physical address string is used to create the physical address of this adapter.
		 */
		String addrStr;
		if (AddrString != null) {
			addrStr = AddrString;
		}
		/*
		 * If this is Node Adapter and no physical address is specified (in string format) with papameter,
		 * the adapter reads configuration file to get the information of physical address.
		 */
		else {
			String addrType = config.getTextAttribute
					(XmlUtil.createXPath(PROPERTY_NAME_PREFIX + "PhyAddr"));
			addrStr = config.getTextAttribute
					(XmlUtil.createXPath(PROPERTY_NAME_PREFIX+"PhyAddr/"+addrType));
		}

		String[] paStr = INETv4TwoPorts.getPartsFromAddrAndPortsString(addrStr);

		/*
		 * Create InetAddress.
		 */			
		if (!paStr[0].equals("")) {
			try
			{
			   ipAddr = InetAddress.getByName(paStr[0]);
			}
			catch (UnknownHostException e)
			{
			   //throw new HyperCastFatalRuntimeException (""+addrStr + " is a bad address string.**" + e);
			   config.err.println("Cannot create InetAddress from " + ipAddr + " : " + e);
			   ipAddr = null;
			}
		}

		/*
		 * Obtain the first port number.
		 */
		if (!paStr[1].equals("")) {
			try 
			{
			  portNodeAdpt = Integer.parseInt(paStr[1]);
			  if ((portNodeAdpt < 0) && (portNodeAdpt != -1))
			  {
				throw new IllegalArgumentException("" + paStr[1] + " is not a available value for port number.");
			  }
			}
			catch (NumberFormatException e) 
			{
			  throw new IllegalArgumentException("Unable to parse  \"" + paStr[1] + "\" as a port number.");
			}
		}
		
		/*
		 * Obtain the second port number.
		 */
		if (!paStr[2].equals("")) {
			try 
			{
				portScktAdpt = Integer.parseInt(paStr[2]);
				if ((portScktAdpt < 0) && (portScktAdpt != -1))
				{
				  throw new IllegalArgumentException("" + paStr[2] + " is not a available value for port number.");
				}
			}
			catch (NumberFormatException e) 
			{
			  throw new IllegalArgumentException("Unable to parse  \"" + paStr[2] + "\" as a port number.");
			}
		}

		/*
		 * Initialize the DatagramSocket with obtained IP address and port information.
		 */
		if (!ProcessingSocketInitialization(ScktAdpter, portNodeAdpt, portScktAdpt, ipAddr)) {
			/*
			 * When the initialization of the DatagramSocket fails, if AllowRandomPhyAddrPort
			 * is specified to "Yes", the adapter will try to initialize the DatagramSocket
			 * without specifying port number, i.e., asking system to pick port number.
			 */
			if ((AllowRandomPhyAddrPort.equalsIgnoreCase("Yes")) && ((portNodeAdpt > 0) || (portScktAdpt > 0))) 
			{
				if (!ProcessingSocketInitialization(ScktAdpter, 0, 0, ipAddr)) {
					return false;
				}
				else {
					return true;
				}
			}
			else {
				return false;
			}
		}
		else {
			return true;
		}

	}


	/**
	 * Process the initialization of underlay network sockets.
	 * 
	 * @param	ScktAdpter	 	A I_UnicastAdapter object representing Socket Adapter.
	 * @param	portNodeAdpt	port number of the socket in Node Adapter.
	 * @param	portScktAdpt	port number assigned to the socket in Socket Adapter.
	 * @param	ipAddr			IP address to which the socket is bounded.
	 */
	public boolean ProcessingSocketInitialization(I_UnicastAdapter ScktAdpter, int portNodeAdpt, int portScktAdpt, InetAddress ipAddr)
	{
		/*
		 * If the port number of Node Adapter is specified as a positive number,
		 * first initialize the socket in Node Adapter and then the socket in Socket
		 * Adapter. 
		 */
		if (portNodeAdpt > 0)
		{
			/*
			 * Initialize the socket in Node Adapter. If it fails, 
			 * stop the initialization process and return.
			 */
			if (!createDatagramSocket(portNodeAdpt, ipAddr))
			{
				return false;
			}
			
			/*
			 * If Socket Adapter is not specified, return.
			 */
			if (ScktAdpter == null)
			{
			   return true;
			}

			/*
			 * Initialize the socket in Socket Adapter by calling Reset method. If it fails, 
			 * stop the initialization process and return.
			 */
			String addressStr;
			if (ipAddr != null) 
			{
				addressStr = new String(getIPString(ipAddr)+":"+portScktAdpt);
			}
			else
			{
				addressStr = new String(":"+portScktAdpt);
			}
			if (!ScktAdpter.Reset(addressStr))
			{
				return false;
			}
		}
		/*
		 * If the port number of Node Adapter is not specified (-1 or 0), first the 
		 * socket in Socket Adapter and then the socket in Node Adapter.
		 */
		else
		{

			/*
			 * If Socket Adapter is not specified, skip the process of initializing the 
			 * socket for Socket Adapter. Otherwise initialize the socket in Socket Adapter.
			 */
			if (ScktAdpter != null)
			{
				/*
				 * Initialize the socket in Socket Adapter by calling Reset method. If it fails, 
				 * stop the initialization process and return.
				 */
				String addressStr;
				if (ipAddr != null) 
				{
					addressStr = new String(getIPString(ipAddr)+":"+portScktAdpt);
				}
				else
				{
					addressStr = new String(":"+portScktAdpt);
				}
				if (!ScktAdpter.Reset(addressStr))
				{
					return false;
				}
			}

			/*
			 * Initialize the socket in Node Adapter. If it fails, 
			 * stop the initialization process and return.
			 */
			if (!createDatagramSocket(portNodeAdpt, ipAddr))
			{
				return false;
			}
			
		}


		/*
		 * The initialization is successful. Return.
		 */
		return true;
	}



	/*
	 * Create a local random number creator.
	 */
	private static Random rand = new Random();

	/**
	 * Creates a DatagramSocket.
	 * 
	 * @param	portNum	 the port number for the UDP unicast socket. 
	 * 					 If portN is set to 0, the port will be the minimal
	 * 					 available port shosen by system; If portN is -1,
	 * 					 the port number is chosen randomly.
	 * @param	ipAddr 	 IP address to which the created socket is bounded.
	 * 					 If it is null, IP address will be chosen by system.
	 */
	public boolean createDatagramSocket(int portNum, InetAddress ipAddr)
	{
		/*
		 * If IP address is not specified, the local DatagramSocket is bound to 
		 * the default IP address chosen by system.
		 */
		if (ipAddr == null) {
			//System.out.println("UDP_UnicastAdapter: createDatagramSocket: before creating USock: "+getIPString(ipAddr)+":"+portNum);
			for (int retries = 0; retries < MAX_RETRIES_TO_OPEN_SOCKET; retries++)
			{
			   try 
			   {
				   	 /*
				   	  * If port number parameter is set to 0, system picks the minimum available port number to use.
				   	  */
				 if (portNum == 0)
				 {
					//System.out.println("UDP_UnicastAdapter: createDatagramSocket: portNum = 0");					
					USock = new DatagramSocket();
				 }
					/*
					 * If port number parameter is set to -1, a random number is used as the port number.
					 */
				 else if (portNum == -1)
				 {
					//System.out.println("UDP_UnicastAdapter: createDatagramSocket: portNum = -1");
					USock = new DatagramSocket(Math.abs(rand.nextInt()) % (256*256));
				 }
					/*
					 * If port number parameter is set to a positive number, it is used as the port number.
					 */
				 else 
				 {
					//System.out.println("UDP_UnicastAdapter: createDatagramSocket: portNum = " + portNum);
					USock = new DatagramSocket(portNum);
				 }

				 UPort = USock.getLocalPort();

				 //System.out.println("UDP_UnicastAdapter: createDatagramSocket: After creation. USock is "+getIPString(USock.getLocalAddress())+":"+USock.getLocalPort());
				 return true;
			   }
			   catch (SocketException se) 
			   {
			   	 /*
			   	  * If portNum is -1, which means the port number is randomly chosen,
			   	  * it may conflict with the port number used in another adapter (Node
			   	  * Adapter or Socket Adapter). In this case, retry to get another port 
			   	  * number. Otherwise, the creation of DatagramSocket fails. Return.
			   	  */
				 if (portNum == -1)
				 {
					continue;
				 }
				 else
				 {
					return false;
				 }
			   }
			}
		}
		/*
		 * If IP address is specified, the local DatagramSocket is bound to this address.
		 */
		else {
			int retries;
			for (retries = 0; retries < MAX_RETRIES_TO_OPEN_SOCKET; retries++)
			{
			   try 
			   {
						/*
						 * If port number parameter is set to 0, system picks the minimum available port number to use.
						 */
					 if (portNum == 0)
					 {
						USock = new DatagramSocket(portNum, ipAddr);
					 }
						/*
						 * If port number parameter is set to -1, a random number is used as the port number.
						 */
					 else if (portNum == -1)
					 {
						USock = new DatagramSocket(Math.abs(rand.nextInt()) % (256*256), ipAddr);
					 }
						/*
						 * If port number parameter is set to a positive number, it is used as the port number.
						 */
					 else 
					 {
						USock = new DatagramSocket(portNum, ipAddr);
					 }
	
					 UPort = USock.getLocalPort();
	
					 return true;
			   }
			   catch (SocketException se) 
			   {
					/*
					 * If portNum is -1, which means the port number is randomly chosen,
					 * it may conflict with the port number used in another adapter (Node
					 * Adapter or Socket Adapter). In this case, retry to get another port 
					 * number. Otherwise, the creation of DatagramSocket fails. Return.
					 */
					if (portNum == -1)
					{
					   continue;
					}
					else
					{
					   return false;
					}
				   }
			}
		}

		/*
		 * If the the process reaches here, that means the retry times reachs 
		 * the maximum retry times, stop the initializing local DatagramSocket.
		 */
		return false;
	}


	/**
	 * Get string representation of IP address in a InetAddress object.
	 * @param	addr	An InetAddress object. 
	 */
	public String getIPString(InetAddress addr) {
		if (addr == null) return new String("");
		
		String addrstr = addr.toString();
		int slash_loc = addrstr.indexOf('/');
		return addrstr.substring(slash_loc+1, addrstr.length());
	}


	/** Get information of this adapter. 
	 */ 
	public Object getInfo() {
		/*
		 * It returns a String object that carries the port number 
		 * used by the socket in this adapter.
		 */
	   int port = USock.getLocalPort();

	   return new String(""+port);
	}


	/**
	 * Resets the local unicast socket.
	 * @param	infoObj		object containing information used to finish reset. 
	 */
	public boolean Reset(Object infoObj) {
		/*
		 * If parameter infoObj is an instance of DatagramSocket, this 
		 * DatagramSocket object is used to replace the DatagramSocket in this
		 * adapter if they are different. This case happens when TCP_UDP_Adapter 
		 * is specified.
		 */
		if (infoObj instanceof DatagramSocket) {
			String localIPstr = getIPString(USock.getLocalAddress());
			String parIPstr = getIPString(((DatagramSocket) infoObj).getLocalAddress());
			
			if ((localIPstr.equals(parIPstr)) && (USock.getLocalPort() == ((DatagramSocket) infoObj).getLocalPort()))
			{
				return true;
			} 
		    
		    if (USock != null) 
		    {
		    	USock.close();
		    } 
		    USock = (DatagramSocket) infoObj;
		}
		/*
		 * If parameter infoObj is a string, this string will bring IP address and port number
		 * information that are used to reset the DatagramSocket in this adapter. The passed
		 * string has format "IPaddrString:port".
		 */
		else if (infoObj instanceof String) {
			String addrstring = (String) infoObj;
			
			/*
			 * If addString is "", close the DatagramSocket and return.
			 */
			if (addrstring.equals(""))
			{
				if (USock != null) USock.close();
				return true;
			}
			
			int cl_loc = addrstring.indexOf(':');
			String ipString = addrstring.substring(0, cl_loc);
			int portnumber = Integer.parseInt(addrstring.substring(cl_loc+1, addrstring.length()));

			/*
			 * If the local DatagramSocket satisfies the rerquirement of reset parameters, do nothing and return.
			 */			
			if ((USock != null) && 
				((portnumber <= 0) ||
				((portnumber > 0) && (USock.getLocalPort() == portnumber) && (ipString.equals(getIPString(USock.getLocalAddress()))))))
			{
				return true;
			}

			/*
			 * Otherwise the DatagramSocket is reset.
			 */
			if (USock != null) USock.close();

			InetAddress ipaddress = null;
			/*
			 * If ipString is specified (not ""), it is used to create the local DatagramSocket,
			 * i.e. the local DatagramSocket is bound to it. Another case is that the ipString is 
			 * obtained from calling USocket.getLocalAddress(), where USocket is a DatagramSocket 
			 * instance. In this case, ipAddress is always "0.0.0.0".
			 */
			if ((!ipString.equals("")) && (!ipString.equals("0.0.0.0"))) 
			{
				try
				{
					ipaddress = InetAddress.getByName(ipString);
				}
				catch (UnknownHostException e)
				{
				   System.err.println(ipString + " is a bad address string.**" + e);
				   ipaddress = null;
				}
			}
			/*
			 * If ipString is "" or "0.0.0.0", the default IP address picked by system will be used 
			 * to create the local Datagram Socket, i.e. the local DatagramSocket is bound 
			 * to this default IP address.
			 */
			return createDatagramSocket(portnumber, ipaddress);
		}
		
		return true;
	}

	/**
	 * Sets the I_AdapterCallback object for this Adapter.
	 * @param	node	an I_AdapterCallback object.
	 */
	public void setCallback(I_AdapterCallback node) {
        if (state != STOPPED)
            throw new IllegalStateException("Tried to change callback while UDP_UnicastAdapter was running!");
		Node = node;
	}


	/** 
	 * Set a <code>Timer</code> event. <p>
	 * @param	timer_id	a time event ID
	 * @param	time_ms		a time in ms
	 */		
     public void setTimer(Object timer_id, long time_ms) {
        if (state == STOPPED || state == STARTING)
            throw new IllegalStateException("UDP_UnicastAdapter: tried to set Timer when adapter was not running.");

        if (UDP_Timer == null) 
            throw new IllegalStateException("UDP_UnicastAdapter: tried to set Timer when Timer is null.");

        UDP_Timer.setTimer(timer_id, time_ms);
	}
	
	
	/** 
	 * Get a <code>Timer</code> event. <p>
	 * @param	timerid	a time event ID
	 * @return	time in ms
	 */			
	public long getTimer(Object timerid) {
        if (state == STOPPED || state == STARTING)
            throw new IllegalStateException("UDP_UnicastAdapter: tried to get Timer when adapter was not running.");

        if (UDP_Timer == null) 
            throw new IllegalStateException("UDP_UnicastAdapter: tried to get Timer when Timer is null.");

        return UDP_Timer.getTimer(timerid);
	}
	
	
	/** 
	 * Clear a <code>Timer</code> event. <p>
	 * @param	timerid	a time event ID
	 */				
	public synchronized void clearTimer(Object timerid) {
        if (state == STOPPED || state == STARTING)
            throw new IllegalStateException("UDP_UnicastAdapter: tried to clear Timer when adapter was not running.");

        if (UDP_Timer == null) 
            throw new IllegalStateException("UDP_UnicastAdapter: tried to clear Timer when Timer is null.");

        UDP_Timer.clearTimer(timerid);
	}

	
	/**
	 * Returns the time as given by System.currentTimeMillis().
	 */
	public long getCurrentTime() {
		return System.currentTimeMillis();
    }
	

	/**
	 * Starts all the threads in this Adapter
	 */
	public void Start() {		

        synchronized (stateGuard) 
        {
            while (true) 
            {
                if (state == STOPPED) 
                {
                    state = STARTING;
                    break;
                }
                else if (state == STARTED) 
                {
                    throw new IllegalStateException("Tried to start UDP_UnicastAdapter twice!");
                }

                try 
                {
                    stateGuard.wait();
                }
                catch (InterruptedException ie) 
                {
                }
            }
        }
        
        if (Node == null)
            throw new IllegalStateException("Tried to start UDP_UnicastAdapter without setting callback!");
            
        /* 
         * create receiver and message procressing objects
         */
		URecv = new UnicastReceiver(InputChannel, this);;
		MsgProc = new UDP_MsgProcessor(this);  

		/* 
		 * Creates two threads in the adapter
		 */
		Thread UThread = new Thread(URecv);
		Thread RecvThread = new Thread(MsgProc);  

		/* 
		 * Starts these threads
		 */
		UThread.start();
		RecvThread.start();

        synchronized (stateGuard) 
        {
            state = STARTED;
            stateGuard.notify();
        }
	}
	
    /** This method starts the TimerProc thread.
      */
	public	void startTimer() {		
        if (UDP_Timer != null) {
		   if (!TimerProc.isDead()) {
			 System.out.println("Attempt to start timer thread twice!");
		   	 return;
		   }
		   else {

			 TimerProc = new UDP_TimeEventProcessor(this);  

			 Thread TimerThread = new Thread(TimerProc);  
			 TimerThread.start();		
			 return;     
		   }
	   }

	   UDP_Timer = new Timer();

	   TimerProc = new UDP_TimeEventProcessor(this);  

	   Thread TimerThread = new Thread(TimerProc);  
	   TimerThread.start();
	}
	
	
	/**
	 * Stops all these threads on this Adapter.  Callbacks may still occur while this function
     * is being executed; it will only return when it can guarantee no more Callbacks will be called.
	 */
	public	void Stop() {		
        synchronized (stateGuard) 
        {
            while (true) 
            {
                if (state == STARTED) 
                {
                    state = STOPPING;
                    break;
                }
                else if (state == STOPPED) 
                {
                    throw new IllegalStateException("Tried to stop UDP_UnicastAdapter twice!");
                }

                try {
                    stateGuard.wait();
                }
                catch (InterruptedException ie) 
                {
                }
            }
        }

        
        myStop();


        synchronized (stateGuard) 
        {
            state = STOPPED;
            stateGuard.notify();
        }
	}	

    /** This function actually stops the adapter's threads.      *  It is called by Stop(). */
	protected void myStop() 
    {
		MsgProc.die();
		URecv.die();

        synchronized (deathGuard)
        {
	    if (!(TimerProc == null)) {
               while (!TimerProc.isDead() || !MsgProc.isDead() || !URecv.isDead()) 
               {
                   try 
                   {
                       deathGuard.wait();
                   }
                   catch (InterruptedException ie) 
                   {
                   }
               }
            } else {
               while (!MsgProc.isDead() || !URecv.isDead()) 
               {
                   try 
                   {
                       deathGuard.wait();
                   }
                   catch (InterruptedException ie) 
                   {
                   }
               }
	    	}
        }
    }
    
    /** This function stops the TimerProc thread.
      */
	public	void stopTimer() {		
           if (TimerProc != null) TimerProc.die();
	}
	
	
    /** 
     * Sends a message to a unicast address. <p>
	 * <br>
	 *  1. Create a byte array containing the information <br>
	 *  2. Create a DatagramPacket <br>
	 *  3. Send Packet. 
     * @param	dstaddr	destination physical address 
	 * @param	msg  the message to be sent
	 */
	public void sendUnicastMessage(I_NetworkAddress dstaddr, I_Message msg ) {
        if (state == STOPPED || state == STARTING)
            throw new IllegalStateException("UDP_UnicastAdapter: tried to send message when adapter was not running.");
		
		InetAddress dstinetaddress = null;
		int portnum = 0;
		
		/*
		 * If destination is a underlay address, it contains IP
		 * address and one port.
		 */
		if (dstaddr instanceof I_UnderlayAddress)
		{
			INETv4OnePort dstuaaddr = (INETv4OnePort) dstaddr;
			dstinetaddress = dstuaaddr.getInetAddress();

			portnum = dstuaaddr.getPort();
		}
		/*
		 * If destination is a physical address, it contains IP
		 * address and two ports. Depending on that this adapter is 
		 * Socket Adapter or Node Adapter, the corresponding port 
		 * number is retrived from destination physical address.
		 */
		else if (dstaddr instanceof I_PhysicalAddress)
		{
			INETv4TwoPorts dstpaaddr = (INETv4TwoPorts) dstaddr;
			
			dstinetaddress = dstpaaddr.getInetAddress();

			if (AdptIndicator.equals("nodeadpt")) 
			{
				portnum = dstpaaddr.getNodeAdptPort();
			} 
			else
			{
				portnum = dstpaaddr.getScktAdptPort();
			} 
		}
		else
		{
			throw new HyperCastWarningRuntimeException ("Destination address has incorrect format.");
		}
		
		byte[] Content = msg.toByteArray();   

		boolean validsend = true;
		
		/*
		 * Create datagram packet.
		 */
		DatagramPacket dpacket = 
			new DatagramPacket(Content, 
				Content.length,
				//dstipaddr.getInetAddress(),         
				//dstipaddr.getPort());         
				dstinetaddress,
				portnum);               

		/*
		 * Send the packet to the underlay network.
		 */
		synchronized (USockLocker) {
			try {
                    
				USock.send(dpacket);                 //--------> network
			} catch (IOException e) {
				validsend = false;
				property.err.println("UDP_UnicastAdapter encountered exception:" + e.toString());
				property.err.println("UDP_UnicastAdapter could not send packet to:" + dstaddr.toString());
				//e.printStackTrace(property.err);
			}
	  	}

		if (validsend) {
			USentPackets++;
			USentBytes += Content.length;
		}

	}

	
   /** 
	* Creates physical address from a byte array of address - the inverse of 
	* <code>toByteArray</ode>. The first 4 bytes reprsent the IP address, 
	* and the last 8 bytes represent the Node Adapter port number and Socket
	* Adapter port number.
	*
	* @param	byteAddressAndPorts  an byte array containing a physical address
	* @param	offset  the position in byteAddressAndPort from where availabe
	* 			data is read.
	* @return	a new instance of <code>INETv4TwoPorts</code>
	* @see		INETv4TwoPorts#toByteArray()
	*/
	public I_PhysicalAddress createPhysicalAddress(byte[] byteAddressAndPorts, int offset) {
		byte[] withoutShift = new byte[INETv4TwoPorts.ADDRESS_SIZE];
		System.arraycopy(byteAddressAndPorts, offset, withoutShift, 0, INETv4TwoPorts.ADDRESS_SIZE);

		INETv4TwoPorts newINETv4 = new INETv4TwoPorts(withoutShift);

		return newINETv4;
	}

	
	/** 
	 * Creates physical address object from a string.
	 * @param	addrString  an string representing a physical address.
	 * @return	a new instance of <code>INETv4TwoPorts</code>
	 */
	public I_PhysicalAddress createPhysicalAddress(String addrString) {
		INETv4TwoPorts newINETv4 = new INETv4TwoPorts(addrString);
		if (newINETv4.getInetAddress() == null) return null;
		else return newINETv4;
	}

	/** 
	 * Creates a new physical address.
	 * @return	a new instance of <code>INETv4TwoPorts</code>
	 */
	public I_PhysicalAddress createPhysicalAddress() {
		/*
		 * Get the port number used in Socket Adapter.
		 */
		int portScktAdpt = 0;
		if (socketAdpter != null)
		{
			portScktAdpt = Integer.parseInt((String) socketAdpter.getInfo());
		} 

		/*
		 * Get the port number used in NodeAd Adapter.
		 */
		int portNodeAdpt = USock.getLocalPort();

		InetAddress localAddr = USock.getLocalAddress();

		/*
		 * Create a INETv4TwoPorts object with two ports. Ths IP address is determined by
		 * the system.
		 */
		//INETv4TwoPorts newINETv4 = new INETv4TwoPorts(portNodeAdpt, portScktAdpt);
		INETv4TwoPorts newINETv4 = new INETv4TwoPorts(localAddr, portNodeAdpt, portScktAdpt);
		if (newINETv4.getInetAddress() == null)
		{
			throw new HyperCastFatalRuntimeException ("Could not create physical address.");
		}

		return newINETv4;
	}

	/** Returns the size of physical address, when carried in a packet,
	 *  recognized by this adapter. 
	 */
	public int getPhysicalAddressSize()
	{
		return INETv4TwoPorts.ADDRESS_SIZE;
	}


	/** 
	 * Creates underlay address from a byte array of address - the inverse of 
	 * <code>toByteArray</ode>. The first 4 bytes reprsent the IP address, 
	 * and the last 4 bytes represent the port number.
	 *
	 * @param	byteAddressAndPort  an byte array containing a physical address
	 * @param	offset  the position in byteAddressAndPort from where availabe
	 * 			data is read.
	 * @return	a new instance of <code>INETv4AndOnePort</code>
	 * @see		INETv4OnePort#toByteArray()
	 */
	 public I_UnderlayAddress createUnderlayAddress(byte[] byteAddressAndPort, int offset) {
		 byte[] withoutShift = new byte[INETv4OnePort.ADDRESS_SIZE];
		 System.arraycopy(byteAddressAndPort, offset, withoutShift, 0, INETv4OnePort.ADDRESS_SIZE);
		 return new INETv4OnePort(withoutShift);
	 }

	
	 /** 
	  * Creates underlay address object from a string.
	  * @param	addrString  an string representing a physical address.
	  * @return	a new instance of <code>INETv4AndOnePort</code>
	  */
	 public I_UnderlayAddress createUnderlayAddress(String addrString) {
		 INETv4OnePort newINETv4 = new INETv4OnePort(addrString);
		 if (newINETv4.getInetAddress() == null) return null;
		 else return newINETv4;
	 }


	 /** 
	  * Creates a new underlay address.
	  * @return	a new instance of <code>INETv4AndOnePort</code>
	  */
	 public I_UnderlayAddress createUnderlayAddress() {
		 /*
		  * Get the port number used in NodeAd Adapter.
		  */
		 int port = USock.getLocalPort();

		 InetAddress localAddr = USock.getLocalAddress();

		 /*
		  * Create a INETv4OnePort object with one port. Ths IP address is determined by
		  * the system.
		  */
		 INETv4OnePort newINETv4 = new INETv4OnePort(localAddr, port);
		 if (newINETv4.getInetAddress() == null)
		 {
			 throw new HyperCastFatalRuntimeException ("Could not create physical address.");
		 }
		 return newINETv4;
	 }


	/** Returns the size of underlay address, when carried in a packet,
	 *  recognized by this adapter. 
	 */
	public int getUnderlayAddressSize()
	{
		return INETv4OnePort.ADDRESS_SIZE;
	}


	/**
	 * Returns the number of received unicast packets from this adapter.
	 */
	public long getUReceivedPackets() {
		return UReceivedPackets;		
	}

	
	/**
	 * Returns the number of received bytes of unicast packets from this adapter.
	 */
	public long getUReceivedBytes() {
		return UReceivedBytes;		
	}
	
	
	/**
	 * Returns the number of unicast packets sent to this adapter.
	 */	
	public long getUSentPackets() {
		return USentPackets;		
	}

	
	/**
	 * Returns the number of bytes of unicast packets sent to this adapter.
	 */		
	public long getUSentBytes() {
		return USentBytes;		
	}
	
	
	/**
	 * Returns the number of unprocessed messages in BoundedBuffer.
	 */
	public int getNumInQueue() {
		return InputChannel.numInQueue();
	}




	/**
	 * Initialize the data structure for handling the statistics defined in this object.
	 * In this method, a StatsProcessor instance is created which contains all statistics
	 * , with each represented by a StatsElement instance, supported in this object.
	 */
	private void InitStatisticsStructure()
	{
		/*
		 * Get the statistics name defined in the configuration file for this I_Stats instance.
		 */
		statisticsName = property.getTextAttribute (XmlUtil.createXPath(PROPERTY_NAME_PREFIX+"StatName"));

		/*
		 * Initial StatsProcessor instance.
		 */
		statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);
		
		statsPro.addStatsElement("RecvBuf"/* name */, InputChannel/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);

		statsPro.addStatsElement("UPacketsReceived", new UnicastPacketsReceived()/* I_Stats object */, 1, 1);

		statsPro.addStatsElement("UBytesReceived", new UnicastBytesReceived()/* I_Stats object */, 1, 1);

		statsPro.addStatsElement("UPacketsSent", new UnicastPacketsSent()/* I_Stats object */, 1, 1);

		statsPro.addStatsElement("UBytesSent", new UnicastBytesSent()/* I_Stats object */, 1, 1);
	}

	/** @see I_Stats#getStatsName() */
	public String getStatsName ()
	{
		return statisticsName;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName (String name)
	{
		statisticsName = name;
	}


	/** 
	 * Return the result of query for the statistics specified by the given xpath.
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance represents the statistics to be queried.
	 * @see I_Stats#getStats
	 */
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//System.out.println("UDP_UnicastAdapter:getStats: xpath string is "+xpath.getPatternString());
		//System.out.println("UDP_UnicastAdapter:getStats: call statsPro.getStatsResult(doc, xpath)");
		return statsPro.getStatsResult(doc, xpath);
	}

	/** 
	 * Set the statistics specified by the given xpath. The value actually set is
	 * returned.
	 * 
	 * @param doc   	Document used for create new elements or nodes.
	 * @param xpath 	XPath instance represents the statistics to be queried.
	 * @param newValue	Element representing the value or sub-tree to be set.
	 * @see I_Stats#setStats
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	{
		//System.out.println("UDP_UnicastAdapter:setStats: xpath string is "+xpath.getPatternString());
		//System.out.println("UDP_UnicastAdapter:setStats: call statsPro.setStatsResult(doc, xpath, newValue)");
		return statsPro.setStatsResult(doc, xpath, newValue);
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in read schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//System.out.println("UDP_UnicastAdapter:getReadSchema: xpath string is "+xpath.getPatternString());
		//System.out.println("UDP_UnicastAdapter:getSchema: call statsPro.getReadSchemaResult(doc, xpath)");
		return statsPro.getReadSchemaResult(doc, xpath);
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in write schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getWriteSchema(Document, XPath)
	 */
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return statsPro.getWriteSchemaResult(doc, xpath);
	}
	

	/**
	 * This class manages simple statistic "UPacketsReceived".
	 */
	class UnicastPacketsReceived extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+UReceivedPackets);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "UBytesReceived".
	 */
	class UnicastBytesReceived extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+UReceivedBytes);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "UPacketsSent".
	 */
	class UnicastPacketsSent extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+USentPackets);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "UBytesSent".
	 */
	class UnicastBytesSent extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+USentBytes);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
	
}	
	




/**
 * The UnicastReceiver class starts a receiver thread to receive messages 
 * from a DatagramSocket.
 */

class UnicastReceiver implements Runnable {
    /** Buffer of the adapter. */
	MessageBuffer adapterbuffer;
    /** The adapter. */
	UDP_UnicastAdapter Parent;
    /** Offset used in calls to I_AdapterCallback.restoreMessage() */
        int[] offset;
    
    /** The receiving thread. */
	private Thread myThread;
    /** True if the function die() has been called. */
	private boolean dying;
    /** True if <code>myThread</code> has returned from run(). */
    private boolean died;

    /** True while <code>myThread</code> is waiting to receive. */
    boolean waiting;
    /** Guard used to protect <code>waiting</code> when it is changed. */
	private Object waitingGuard;	

	/**
	 * Constructs a unicast receiver thread with a pointer of a buffer and 
	 * an unicast adapter which includes a DatagramSocket.
	 * @param	bdbuffer	a pointer to the BoundedBuffer
	 * @param	uadapter	an unicast adapter 
	 */	
	UnicastReceiver(MessageBuffer bdbuffer, UDP_UnicastAdapter uadapter) {
		adapterbuffer = bdbuffer;
		Parent = uadapter;

		offset = new int[1];
		dying = false;	
        died = false;
		waiting = false;
        waitingGuard = new Object();		
	}


	/** 
	 * Starts running the unicast receiver threads and moves messages 
	 * from a Unicast Socket to the DropTailReceiveBuffer
	 */
	
	public void run() {
        Parent.URecvDebug = "JustCreated";

		myThread = Thread.currentThread();
		
		DatagramPacket dpacket;
		byte[] dbuffer;
		I_Message msg;
		
		// Allocate memory 
		dbuffer = new byte[Parent.MaximumPacketLength];

		dpacket = new DatagramPacket(dbuffer, dbuffer.length);

		while (true) {
			// Reset the DatagramPacket size!!! Otherwise, it'll reuse last msg length
			dpacket.setLength(Parent.MaximumPacketLength);			
			
			offset[0] = 0;
			
			try 
            {
                try 
                {
                    try 
                    {
                        synchronized (waitingGuard) 
                        {
                            waiting = true;
                            Parent.URecvDebug = "Waiting";
                        }				

                        // need to check this inside "waiting" guard.
                        if (dying)
                            break;
				
                        //may block here!
                        Parent.USock.receive(dpacket);			

                    }
                    finally 
                    {
                        synchronized (waitingGuard) 
                        { 
                            Parent.URecvDebug = "DoneWaiting";
                            waiting = false;

                            if (Thread.interrupted()) // doesn't appear that you can interrupt a call to receive... could send a "dummy" socket to self...
                                break;
                        }
                    }
                }
                finally 
                {
                    //if (Thread.interrupted()) 
                    //{
                    //    Parent.property.seriousErr.println("SERIOUS ERROR: This println should have been interrupted before it could take place.");
                    //}
                }
                
                // also need to check this after "waiting" guard
                if (dying)
                    break;
                
							
				Parent.UReceivedPackets++;
				Parent.UReceivedBytes += dpacket.getLength();
                try 
                {
                    Parent.URecvDebug = "InCallback";
                    msg = Parent.Node.restoreMessage(dpacket.getData(), offset, dpacket.getLength());
                    Parent.URecvDebug = "WritingToBuffer";

                    if (msg != null)	
                    {
                        adapterbuffer.nonblockingWrite(msg);
                    }
                    else 
                    {
                        Parent.property.err.println("restoreMessage() returned null! Perhaps the setting of maximum packet length is too small in your configuration.");
                        byte[] tmp_buf = new byte[dpacket.getLength()];
                        System.arraycopy(dpacket.getData(), 0, tmp_buf, 0, dpacket.getLength());
                        //Parent.property.err.println("Message (hex): " + HTTP_ServerUtility.toHexString(tmp_buf));
                        Parent.property.err.println("Message src: " + dpacket.getAddress() + ":" + dpacket.getPort());
                    }
                }
                catch (Exception e)
                {
                    Parent.property.err.println("restoreMessage() encounter exception:" + e.getMessage());
                    if (e instanceof IllegalArgumentException)
                    {
                        byte[] tmp_buf = new byte[dpacket.getLength()];
                        System.arraycopy(dpacket.getData(), 0, tmp_buf, 0, dpacket.getLength());
                        //Parent.property.err.println("Message (hex): " + HTTP_ServerUtility.toHexString(tmp_buf));
                        Parent.property.err.println("Message src: " + dpacket.getAddress() + ":" + dpacket.getPort());
                    }
                    else
                    {
                        e.printStackTrace(Parent.property.err);
                    }
                    
                }
				
			} 			
            catch (InterruptedIOException iioe) 
            {
                // thread will die when it gets back to top of loop.
            }
			catch (IOException e) {
                if (!("Socket closed".equals(e.getMessage()) || "socket closed".equals(e.getMessage())) ) {
                    Parent.property.err.println("Unicast Receiver had other error: " + e.toString());
                    Parent.property.err.println("Restarting socket.");

                    synchronized (Parent.USockLocker) {
                        Parent.USock.close();
                        try {        
							Parent.USock = new DatagramSocket(Parent.UPort);
                        } catch (Exception e2) {
                            throw new HyperCastFatalRuntimeException ("Error with socket and unable to restart it", e2);
                        }
                    }
                    Parent.property.err.println("Socket restarted.");
                }
			}			
            Parent.URecvDebug = "EndOfLoop";
		}  
        

        synchronized (Parent.deathGuard) 
        {
            died = true;
            Parent.URecvDebug = "Dead";
            Parent.deathGuard.notify();
        }
	}

	
	/** This function causes <code>myThread</code> to die eventually. */
	public void die() {
        synchronized (waitingGuard) {
            dying = true;
            if (waiting)
                myThread.interrupt();  // incase it is blocking on the wait call.
        }
    }  

    /** Returns true if <code>myThread</code> will never call an I_AdapterCallback method again. */
public boolean isDead() 
    {
        return (died || waiting);
    }
}
	
	
	
	
	
	
	

/**
 * This class starts a thread to read message from a MessageBuffer and 
 * calls I_AdapterCallback(protocol node) for procession.
 */

class UDP_MsgProcessor implements Runnable {
    /** The adapter. */
	UDP_UnicastAdapter Parent;

    /** The receiving thread. */
	private Thread myThread;

    /** True if the function die() has been called. */
	private boolean dying;
    /** True if <code>myThread</code> has returned from run(). */
    private boolean died;

    /** True while <code>myThread</code> is waiting to receive. */
	boolean waiting;
    /** Guard used to protect <code>waiting</code> when it is changed. */
	private Object waitingGuard;	
		
	/**
	 * Constructs the thread with an Adapter object.
	 */
	UDP_MsgProcessor(UDP_UnicastAdapter uadapter) {
		Parent = uadapter;
		
		dying = false;	
		died = false;	

		waiting = false;
        waitingGuard = new Object();
	}

	
	/** 
	 * Starts running the thread to read messages from 
	 * MessageBuffer and call I_AdapterCallback for procession.
	 */	
	public void run() {	    
		myThread = Thread.currentThread();
		I_Message msg = null;
				

        Parent.MsgProcDebug = "JustStarted";

		while (true) {
            try 
            {
                try 
                {
                    try 
                    {
                        synchronized (waitingGuard) 
                        {
                            Parent.MsgProcDebug = "Waiting";
                            waiting = true;
                        }
				
                        // need to check this inside "waiting" guard.

                        if (dying)
                            break;

                        // May block here
                        msg = Parent.InputChannel.read();

                    }
                    finally 
                    {						
                        synchronized (waitingGuard) 
                        { 
                            Parent.MsgProcDebug = "DoneWaiting";
                            waiting = false;
                        }
                    }
                }
                finally
                {
                    // clear interrupted bit.
                    if (Thread.interrupted()) 
                    {
                        Parent.MsgProcDebug = "ClearingInterruptBit";
                        Thread.sleep(1000);
                    }
                }
            }
            catch (InterruptedException ie) 
            {
                continue;
            }

            if (dying)
                break;

			// msg should never be null... (but Java has errors.)
			if (msg != null) {													
                try 
                {
                    Parent.MsgProcDebug = "InCallback";
                    Parent.Node.messageArrivedFromAdapter(msg);
                }
                catch (Exception e)
                {
                    Parent.property.err.println("Node.messageArrivedFromAdapter encountered exception: " + e.getMessage());
                    e.printStackTrace(Parent.property.err);
                }
			}
            Parent.MsgProcDebug = "EndOfLoop";
		} 

        synchronized (Parent.deathGuard) 
        {
            died = true;
            Parent.MsgProcDebug = "Dead";
            Parent.deathGuard.notify();
        }
	}

	
	
	/** This function causes <code>myThread</code> to die eventually. */
	public void die() {
         synchronized (waitingGuard) {
             dying = true;
             if (waiting)
                 myThread.interrupt();  // incase it is blocking on the wait call.
         }
     }
	
    /** Returns true if <code>myThread</code> will never call an I_AdapterCallback method again. */
public boolean isDead() 
    {
        return (died || waiting);
    }
}



/** 
 * This class starts a thread to process time event. <p>
 * <br>
 * The processor takes the time event with the smallest timestamp from 
 * the time_event queue. <br> 
 * If this timestamp > currenttime,
 * the processor sleep for (timestamp - currenttime), then processes the time event. <br>
 * If this timestamp <= currenttime,
 * the processor processes the time event at once.
 */

class UDP_TimeEventProcessor implements Runnable {
    /** The adapter. */
	UDP_UnicastAdapter Parent;

    /** The receiving thread. */
	private Thread myThread;

    /** True if the function die() has been called. */
	private boolean dying;
    /** True if <code>myThread</code> has returned from run(). */
    private boolean died;

    /** True while <code>myThread</code> is waiting to receive. */
    boolean waiting;
    /** Guard used to protect <code>waiting</code> when it is changed. */
	private Object waitingGuard;	
	
	
	/**
	 * Contructs the thread with a UDP_UnicastAdapter object. 
	 */
	UDP_TimeEventProcessor(UDP_UnicastAdapter uadapter) {
		Parent = uadapter;
		
		dying = false;	
		died = false;	

		waiting = false;
        	waitingGuard = new Object();
	}

	
    /*
     * Starts running the thread to take a time event from the TimeEventQueue and 
	 * send it to the I_AdapterCallback object for processing.
	 */
	public void run() 
    {	    		
			
        Parent.TimerProcDebug = "JustStarted";
		
		//System.out.println("UDP_TimeEventProcessor started.....");		
		myThread = Thread.currentThread();

		Object id;
		
		while (true) {
			try 
            {
                try 
                {
                    try 
                    {
                        synchronized (waitingGuard) 
                        {
                            waiting = true;
                            Parent.TimerProcDebug = "Waiting";
                        }
                        
                        if (dying)
                            break;
			    
                        //may block here
                        id = Parent.UDP_Timer.sleepUntilNextEvent();
	
                    }
                    finally 
                    {
                        synchronized (waitingGuard) 
                        { 
                            waiting = false;
                            Parent.TimerProcDebug = "DoneWaiting";
                        }
                    }
                }
                finally
                {
                    if (Thread.interrupted()) 
                    {
                        Parent.TimerProcDebug = "ClearingInterruptBit";
                        Thread.sleep(1000);
                    }
                }   

                if (dying) 
                {
                    Parent.TimerProcDebug = "ResettingTimer";
                    // stick timer back in queue
                    Parent.UDP_Timer.setTimer(id, 0);	
                    break;
                }
                
                try 
                {
                    Parent.TimerProcDebug = "InCallback";
                    Parent.Node.timerExpired(id);
                }
                catch (Exception e) 
                {   
                    Parent.property.err.println("Encountered exception \""+e.getMessage() + "\" in timerExpired()");
                    e.printStackTrace(Parent.property.err);
                }
            }          
            catch (InterruptedException e) 
            {
            }						
            Parent.TimerProcDebug = "EndOfLoop";
		}	    
	
        synchronized (Parent.deathGuard) 
        {
            died = true;
            Parent.TimerProcDebug = "Dead";
            Parent.deathGuard.notify();
        }
    } 
	
	
	
	
	/** This function causes <code>myThread</code> to die eventually. */
	public void die() {
         synchronized (waitingGuard) {
             dying = true;
             if (waiting)
                 myThread.interrupt();  // incase it is blocking on the wait call.
         }
     }
	
    /** Returns true if <code>myThread</code> will never call an I_AdapterCallback method again. */
public boolean isDead() 
    {
        return (died || waiting);
    }
}
