/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.events;

import hypercast.*;

import java.util.Vector;

/**
 * Class of event that is particular to the naming system. Object of
 * this class contain information about bindings that have been
 * requested by an application.
 *
 * @see Naming
 * @see MessageStoreFSM_Naming
 *
 * @author HyperCast Team
 * @author Greg Mattes
 * @version Spring 2005
 */
public class NAMING_EVENT
    extends NOTIFICATION_EVENT
{
    /** Bindings that an application requested. */
    private ImmutableNamingBinding [] bindings;

    /** The unique identifier of the query that caused this event */
    private int serialNumber;

    /**
     * Builds an event for the naming service. The bindings here are
     * copies of those actually contained in the implementation of the
     * naming service. This is done so that the application cannot
     * harm the internals of the naming service implemenation.
     */
    public NAMING_EVENT (final int serialNumber, final ImmutableNamingBinding [] bindings)
    {
        super (System.currentTimeMillis(), null);
        this.serialNumber = serialNumber;
        this.bindings = bindings;
    }

    /**
     * Returns the serial number of the query associated with this event.
     */
    public int getSerialNumber() {
        return serialNumber;
    }

    /**
     * Returns true if at least one binding in this object has the specified name.
     */
    public boolean containsName (final String name)
    {
        boolean result = false;
        for (int i = 0; i < bindings.length; ++i)
        {
            final String bindingName = bindings[i].getName();
            if (name.equals (bindingName))
            {
                result = true;
                break;
            }
        }
        return result;
    }

    /**
     * Returns true if at least one binding in this object has the specified logical address.
     */
    public boolean containsLogicalAddress (final String logicalAddressString)
    {
        boolean result = false;
        for (int i = 0; i < bindings.length; ++i)
        {
            final String bindingLogicalAddressString = bindings[i].getLogicalAddressString();
            if (logicalAddressString.equals (bindingLogicalAddressString))
            {
                result = true;
                break;
            }
        }
        return result;
    }

    /**
     * Returns the array of all bindings in this object.
     */
    public ImmutableNamingBinding [] getBindings() {
        return bindings;
    }

    /**
     * Returns an array of the subset of bindings contained by this object that were created locally.
     */
    public ImmutableNamingBinding [] getLocalBindings()
    {
        final Vector result = new Vector();

        for (int i = 0; i < bindings.length; ++i)
            if (bindings[i].isLocal())
                result.add (bindings[i]);

        return (ImmutableNamingBinding []) result.toArray (new ImmutableNamingBinding [0]);
    }

    /**
     * Returns an array of the subset of bindings contained by this object that have the specified name.
     */
    public ImmutableNamingBinding [] getBindingsWithName (final String name)
    {
        final Vector result = new Vector();
        for (int i = 0; i < bindings.length; ++i)
        {
            final String bindingName = bindings[i].getName();

            if (name.equals (bindingName))
                result.add (bindings[i]);
        }
        return (ImmutableNamingBinding []) result.toArray (new ImmutableNamingBinding [0]);
    }

    /**
     * Returns an array of the subset of bindings contained by this object that have the specified logical address.
     */
    public ImmutableNamingBinding [] getBindingsWithLogicalAddress (final String logicalAddressString)
    {
        final Vector result = new Vector();
        for (int i = 0; i < bindings.length; ++i)
        {
            final String bindingLogicalAddressString = bindings[i].getLogicalAddressString();

            if (logicalAddressString.equals (bindingLogicalAddressString))
                result.add (bindings[i]);
        }
        return (ImmutableNamingBinding []) result.toArray (new ImmutableNamingBinding [0]);
    }

    /**
     * Returns a string that contains the representation of all
     * bindings in this event. Each binding is separated with a new
     * line.
     */
    public String toString()
    {
        final StringBuffer buf = new StringBuffer();

        buf.append ("Query S/N: " + serialNumber + "\n");

        for (int i = 0; i < bindings.length; ++i)
            buf.append (bindings[i] + "\n");

        return buf.toString();
    }
}
